package edu.cmu.sei.pacc.absdom;

import java.util.*;

import edu.toronto.cs.algebra.IAlgebra;
import edu.toronto.cs.algebra.AlgebraValue;

import edu.toronto.cs.mvset.MvSet;
import edu.toronto.cs.mvset.MvSetFactory;

import edu.toronto.cs.expr.Expr;
import edu.toronto.cs.expr.ExprFactory;
import edu.toronto.cs.expr.BoolOp;

import static edu.toronto.cs.expr.BoolOp.*;
import static edu.cmu.sei.pacc.absdom.VariantOp.*;


public class MvSetAbstractDomain
{

  /**
   * An ExprFactory to create expressions
   */
  protected ExprFactory exprFactory;
  /**
   * An MvSetFactory to create mvSets for this abstract domain
   */
  protected MvSetFactory mvSetFactory;
  
  /**
   * A map from atomic expressions to information about their
   * encoding. This map should be insync with {@code idTable}.
   */
  protected Map<Expr,AtomicInfo> varTable;

  /**
   * A map from mvset variable ids to atomic information. This map
   * should be insync with {@code varTable}.
  */
  protected Map<Integer,AtomicInfo> idTable;
  
  /**
   * largest ID used so far
   */
  protected int maxID;

  /**
   * number of variants to create by default 
   */
  protected int defaultVariantsSize;
  

  /**
   * Constructs an MvSetAbstractDomain with {@code maxID} value of 1,
   * and {@code defaultVariantsSize} of 2.
   *
   * @param mvSetFactory an {@code MvSetFactory} to construct mvsets
   * @param exprFactory an {@code ExprFactory} to construct expressions
   */
  public MvSetAbstractDomain (MvSetFactory mvSetFactory, 
			      ExprFactory exprFactory)
  {
    this (mvSetFactory, exprFactory, 1, 2);
  }


  /**
   * @param mvSetFactory an {@code MvSetFactory} to construct mvsets
   * @param exprFactory an {@code ExprFactory} to construct expressions
   */
  public MvSetAbstractDomain (MvSetFactory mvSetFactory, 
			      ExprFactory exprFactory, 
			      int maxID, int defaultVariantsSize)
  {
    varTable = new IdentityHashMap<Expr,AtomicInfo> ();
    idTable = new TreeMap<Integer,AtomicInfo> ();
    
    this.mvSetFactory = mvSetFactory;
    this.exprFactory = exprFactory;
    this.maxID = maxID;
    this.defaultVariantsSize = defaultVariantsSize;
  }

  /**
   * Abstract an expression by an mv-set. Same as 
   * {@link #alpha(Expr)}.
   * 
   * @param e an expression 
   * @return an mv-set abstracting the input expression
   */
  public MvSet doAbstract (Expr e)
  {
    return alpha (e);
  }
  
  /**
   * Concretize an mv-set by an expreession
   * Same as {@link #gamma(MvSet)}.
   *
   * @param v an mv-set 
   * @return an expression that is a concretization of {@code v}.
   */
  public Expr doConcretize (MvSet v)
  {
    return gamma (v);
  }
  
  /**
   * Abstract an expression by an mv-set.
   * 
   * @param e an expression 
   * @return an mv-set abstracting the input expression
   */
  public MvSet alpha (Expr e)
  {
    MvSet res = null;

    if (VariantOp.isVariantExpr (e) && 
	! isAtomic (VariantOp.mainVariant (e)))
      throw new IllegalArgumentException ("Expressions with non-atomic" + 
					  " variant expressions are not " + 
					  "supported");
    
    if (isAtomic (e))
      res =  atomicInfo (e).getMvSet ();
//     else if (BoolOp.isIff (e))
//       res = alpha (e.arg (0)).eq (alpha (e.arg (1)));
    else if (BoolOp.isTrue (e))
      res = mvSetFactory.top ();
    else if (BoolOp.isFalse (e))
      res = mvSetFactory.bot ();
    else if (BoolOp.isNot (e))
      res = alpha (e.arg (0)).not ();
    else if (BoolOp.isAnd (e))
      {
	res = mvSetFactory.top ();
	for (Expr arg : e.args ())
	  res = res.and (alpha (arg));	
      }
    else if (BoolOp.isOr (e))
      {
	res = mvSetFactory.bot ();
	for (Expr arg : e.args ())
	  res = res.or (alpha (arg));
      }
    else if (BoolOp.isImpl (e))
      {
	res = alpha (e.arg (0)).impl (alpha (e.arg (1)));
      }
    else if (BoolOp.isLpmi (e))
      {
	res = alpha (e.arg (1)).impl (alpha (e.arg (0)));
      }
    else
      throw 
	new IllegalArgumentException ("Cannot abstract an expression: " + e);
    
    return res;
  }

  /**
   * Concretize an mv-set by an expreession
   *
   * @param v an mv-set 
   * @return an expression that is a concretization of {@code v}.
   */
  public Expr gamma (MvSet v)
  {
    // -- algebra of the mvSetFactory we are dealing with
    IAlgebra algebra = mvSetFactory.getAlgebra ();
    
    Expr res = null;
    
    if (v.equals (mvSetFactory.bot ()))
      res = exprFactory.falseExpr ();
    else if (v.equals (mvSetFactory.top ()))
      res = exprFactory.trueExpr ();
    else
      {
	for (Iterator it = v.cubeIterator (); it.hasNext ();)
	  {
	    AlgebraValue[] cube = (AlgebraValue[]) it.next ();

	    // XXX ensure that all cubes end with TOP
	    // XXX since we only deal with Boolean sets for now
	    assert cube [cube.length - 1] == algebra.top ();

	    Expr cubeExpr = null;

	    for (int i = 0; i < (cube.length - 1); i++)
	      {
		if (cube [i].isNoValue ()) continue;
		
		Expr varExpr = atomicInfoById (i).getName ();
		
		if (cube [i].isBot ())
		  varExpr = lnot (varExpr);
		
		cubeExpr = 
		  (cubeExpr == null) ? 
		  varExpr : land (cubeExpr, varExpr);
	      }
	    res = (res == null) ? cubeExpr : lor (res, cubeExpr);
	  }
      }
    
    return res;
  }
  
  
  /**
   * Returns the expression factory used by this class
   *
   * @return an {@code ExprFactory} object
   */
  public ExprFactory getExprFactory ()
  {
    return exprFactory;
  }
  
  /**
   * Retyrns the mv-set factory used by this class.
   *
   * @return an {@code MvSetFactory} object
   */
  public MvSetFactory getMvSetFactory ()
  {
    return mvSetFactory;
  }
  

  /**
   * Sets how many variants are created by default.
   *
   * @param v number of variants to be created by default
   */
  public void setDefaultVariantsSize (int v)
  {
    defaultVariantsSize = v;
  }

  /**
   * Returns the number of variants that are created by default.
   * 
   * @return number of variants to be created by default
   */
  public int getDefaultVariantsSize ()
  {
    return defaultVariantsSize;
  }
  

  /**
   * Returns true if expression {@code v} is associated with a single
   * mv-set variable 
   * 
   * @param v an expression
   * @return true if {@code v} is associated with an mv-set variable
   */
  public boolean isAllocated (Expr v)
  {
    return varTable.containsKey (v);
  }


  /**
   * Returns {@code AtomicInfo} associated with an id
   * 
   * @param id a numeric identifier
   * @return associated {@code AtomicInfo} object, or {@code null}
   */
  protected AtomicInfo atomicInfoById (int id)
  {
    return idTable.get (id);
  }
  

  /**
   * Returns {@code AtomicInfo} associated with an expression. If the
   * expression does have an associated {@code AtomicInfo}, one is
   * created for it.
   * 
   * @param an expression from which {@code AtomicInfo} is throught
   * @return {@code AtomicInfo} for {@code v}.
   */
  protected AtomicInfo atomicInfo (Expr v)
  {
    checkAtomic (v);
    AtomicInfo info = varTable.get (v);
    if (info == null)
      {
	// -- get next available id
	// -- allocate necessary information
	// -- put into enough tables
	// -- allocate as many auxilaries as needed
    
	info = new AtomicInfo (nextId (), v);

	// -- allocate mvSet for info.getId () variable 
	MvSet mvSet = mvSetFactory.var (info.getId (), 
					mvSetFactory.getAlgebra ().top (),
					mvSetFactory.getAlgebra ().top ());
	info.setMvSet (mvSet);
	mvSet = null;
	addAtomicInfo (info);

	// -- if we added a variant, register it with it's main variant
	if (info.isVariant ())
	  info.mainVariant ().addVaraint (info);

	// -- otherwise, create as many variants as needed by default
	else
	  {
	    // -- create all default variants
	    // XXX Look into allocating variants in a block to guarantee
	    // XXX a good variable ordering
	    for (int i = 1; i <= defaultVariantsSize; i++)
	      {
		AtomicInfo variantInfo = 
		  new AtomicInfo (nextId (), 
				  variantExpr (info.getName (), i),
			      null);
		// -- allocate mvset for variantInfo.getId ()
		mvSet = mvSetFactory.var (variantInfo.getId (), 
					  mvSetFactory.getAlgebra ().top (),
					  mvSetFactory.getAlgebra ().top ());

		variantInfo.setMvSet (mvSet);
		mvSet = null;
		info.addVaraint (variantInfo);
		addAtomicInfo (variantInfo);
	      }
	  }
	
      }
    return info;
    
  }
  

  /**
   * Add {@code AtomicInfo} into internal table 
   * 
   * @param info {@code AtomicInfo} to add
   */
  protected void addAtomicInfo (AtomicInfo info)
  {
    varTable.put (info.getName (), info);
    idTable.put (info.getId (), info);
  }
  


  /**
   * Returns true if an expression is considered atomic. Every atomic
   * expression is abstracted by a single mv-set variable. Note that
   * {@link #alpha} must be able to handle every non-atomic expression.
   *
   * @param v an expression whose automicity is thought
   *
   */
  protected boolean isAtomic (Expr v)
  {
    return v.op ().getClass () != BoolOp.class;
  }

  /**
   * Returns a variable set in the encoding of the abstract domain. In
   * particular, quantification operation on mv-sets require
   * quantified variables to be represened by mv-set as well.
   *
   * @param vars a collection of variables to be turned into a variable set
   * @return a veriable set 
   */
  public MvSet varSet (Collection<Expr> vars)
  {
    return varSet (vars, -1);
  }
  
  /**
   * Same as {@link #varSet(Collection)}, but creates a set of
   * variants of the varaibles in {@code vars}. The collection {@code
   * vars} should not contain any variants.
   * 
   * @param vars a set of variables to be turned into a variable set
   * @param variant a number indicating which variant of the set is needed
   * @return a variable set
   */
  public MvSet varSet (Collection<Expr> vars, int variant)
  {
    int[] varIds = new int [vars.size ()];
    
    // -- iterate over all expressions in vars and 
    // -- construct an array of all variables
    int counter = 0;
    for (Expr v : vars)
      {
	AtomicInfo info = varTable.get (v);
	
	if (variant >= 0)
	  info = info.getNthVariant (variant);
	varIds [counter++] = info.getId ();
      }

    return mvSetFactory.buildCube (varIds);
  }

  /**
   * Same as {@link #varSet(Colletion)}, but returns a set of all
   * variables of a particular variant.
   *
   * @param variant the variant of the variables that is thought
   * @return a variable set
   */
  public MvSet varSet (int variant)
  {
    assert variant >= 0;
    
    List<Expr> idList = new LinkedList<Expr> ();
    for (AtomicInfo v : varTable.values ())
      {
	if (v.variantNum () == variant)
	  idList.add (v.getName ());
      }
    return varSet (idList);
  }
  

  int[] permuteMap (Collection<Expr> vars, int variantFrom, int variantTo)
  {
    // XXX MvSetFactory should know what is the expect size of the map
    int[] map = new int [maxID];

    // -- initilize the map as an identity mapping
    for (int i = 0; i < map.length; i++)
      map [i] = i;

    for (Expr v : vars)
      {
	AtomicInfo info = varTable.get (v);

	map [info.getNthVariant (variantFrom).getId ()] =
	  info.getNthVariant (variantTo).getId ();
      }
    
    return map;
  }
  
  int[] permuteMap (int variantFrom, int variantTo)
  {
    List<Expr> vars = new LinkedList<Expr> ();
    for (AtomicInfo v : varTable.values ())
      if (!v.isVariant ()) vars.add (v.getName ());
    return permuteMap (vars, variantFrom, variantTo);
  }

  public int[] permuteMap (Map<Expr,Expr> exprMap)
  {
    // -- see permuteMap(Collection,int,int) for comments
    int[] map = new int [maxID];
    for (int i = 0; i < map.length; i++)
      map [i] = i;

    for (Map.Entry entry : exprMap.entrySet ())
      {
	AtomicInfo fromInfo = varTable.get (entry.getKey ());
	AtomicInfo toInfo = varTable.get (entry.getValue ());
	
	map [fromInfo.getId ()] = toInfo.getId ();
      }

    return map;
  }
  
  
  /***
      XXX Add "permuter" methods that return a permuter object and not
      an integer map. 
   ***/
  
  

  /**
   * A convinience method that throws an excpetion if its argument is
   * not an atomic expression.
   *
   * @param v an expression 
   * @throws IllegalArgumentException if {@code v} is not an atomic
   * expression
   */
  protected void checkAtomic (Expr v) throws IllegalArgumentException
  {
    if (!isAtomic (v)) 
      throw 
	new IllegalArgumentException ("Expression " + v + " is not atomic");

  }
  
  /**
   * Returns the next available identifier for an mv-set
   * variable. Subsequent calls return larger identifiers
   * 
   * @return next available identifier
   */
  protected int nextId ()
  {
    return maxID++;
  }
  

  /**
   * Prints debug information on the console.
   */
  protected void debugDump ()
  {
    for (Map.Entry entry : varTable.entrySet ())
      {
	System.out.println (entry.getKey () + ": " + entry.getValue ());
      }
    
  }
  


  /**
   * Bookkeeping information about an atomic expression and its
   * association to an mv-set */
  public class AtomicInfo 
  {
    
    /**
     * Identifier of the mv-set variable corresponding to this expression 
     */
    int id;

    /**
     * The expression that names the identifier. This is a variant
     * expression, if this is an info for a variant.
     */
    Expr name;
    
    /**
     * An mv-set representing the identifier 
     */
    // XX not sure if this is a good idea to keep 
    // XX a singleton mvset at all time?!
    MvSet mvSet;

    /**
     * A map of all variants of this expression. Only stored with the
     * main variant
     */
    Map<Integer,AtomicInfo> variants;
    

    /**
     * Constructs an atomic info with a {@code null} mv-set
     * 
     * @param id an identifier for this {@code AtomicInfo}
     * @param name the name of this {@code AtomicInfo}
     */
    public AtomicInfo (int id, Expr name)
    {
      this (id, name, null);
    }
    

    /**
     * Constructs an {@code AtomicInfo} with id, name, and an mvset.
     * 
     * @param id an identifier for this {@code AtomicInfo}
     * @param name the name of this {@code AtomicInfo}
     * @param mvSet an mvSet of this {@code AtomicInfo}
     */
    public AtomicInfo (int id, Expr name, MvSet mvSet)
    {
      this.id = id;
      this.name = name;
      this.mvSet = mvSet;
      
      if (! isVariant ())
	variants = new TreeMap<Integer,AtomicInfo> ();
      else
	variants = null;
    }

    /**
     * Returns true if this {@code AtomicInfo} is a variant.
     * 
     * @return true if this {@code AtomicInfo} is a variant.
     */
    public boolean isVariant ()
    {
      return isVariantExpr (name);
    }
    
    /**
     * Returns the variant number of this atomic info.
     * 
     * @return returns 0 if this {@code AtomicInfo} is not a variant,
     * and its variant number otherwise.
     */
    public int variantNum ()
    {
      return isVariant () ? VariantOp.variantNum (name) : 0;
    }

    /**
     * Returns {@code AtomicInfo} of the Nth variant corresponding to
     * an expression represented by this variant.
     *
     * @param n which variant to get
     * @return {@code AtomicInfo} of variant {@code n}
     */
    public AtomicInfo getNthVariant (int n)
    {
      assert !isVariant () : "a variant does not have another variant!";
    
	// -- a non-variant is it's own 0th variant
      if (n == 0) return this;
      return variants.get (n);
    }
    

    public int getId ()
    {
      return id;
    }
    public Expr getName ()
    {
      return name;
    }

    public AtomicInfo mainVariant ()
    {
      if (!isVariant ()) return this;
      return atomicInfo (VariantOp.mainVariant (name));
    }
    

    public void setMvSet (MvSet v)
    {
      mvSet = v;
    }
    public MvSet getMvSet ()
    {
      return mvSet;
    }
  
    
    public void addVaraint (AtomicInfo info)
    {
      assert !isVariant () : "Only main atomic epxressions can have variants";
	
      // XXX check that info is in fact a variant of this
       variants.put (info.variantNum (), info);
    }


    public String toString ()
    {
      return id + ": " + name;
    }
    



  }
  
  
}
