package edu.toronto.cs.clang;

import java.io.PrintWriter;

import java.util.*;

import edu.toronto.cs.yasm.cfg.CFGraph;
import edu.toronto.cs.yasm.cfg.CFGraph.*;
import edu.toronto.cs.expr.*;

import static edu.toronto.cs.expr.CNamedDeclOp.*;

/**
 * A C program composed of a table of control-flow graphs and symbol table.
 */
public class CProgram
{
  Map<Integer, CFGraph> cfgTable; // uid -> CFGraph
  CSymbolTable symbolTable;

  public CProgram (Map<Integer, CFGraph> cfgTable, CSymbolTable symbolTable)
  {
    this.cfgTable = cfgTable;
    this.symbolTable = symbolTable;
  }

  public Map<Integer, CFGraph> getCFGTable ()
  {
    return cfgTable;
  }

  public CSymbolTable getCSymbolTable ()
  {
    return symbolTable;
  }

  public CFGraph getCFG (Expr funcDecl)
  {
    return cfgTable.get (declUID (funcDecl));
  }


  /**
   * Returns a list of all function declarations
   *
   * @return a <code>List</code> value
   */
  public List<Expr> functions ()
  {
    List<Expr> decls = new LinkedList<Expr> ();
    
    for (int key : cfgTable.keySet ())
      decls.add (symbolTable.getDecl (key));
    return decls;
  }
  

  /**
   * Iterates over all function declarations that have function bodies.
   *
   * @return an <code>Iterable</code> value
   */
  public Iterable<Expr> allCfgFunctions ()
  {
    return new Iterable<Expr> ()
      {
	public Iterator<Expr> iterator ()
	  {
	    return new Iterator<Expr> ()
	      {
		Iterator<Integer> it = cfgTable.keySet ().iterator ();
		
		public boolean hasNext ()
		  {
		    return it.hasNext ();
		  }
		public Expr next ()
		  {
		    return symbolTable.getDecl (it.next ());
		  }
		public void remove ()
		  {
		    throw new UnsupportedOperationException ();
		  }
	      };
	  };
	
      };
  }
  

  public void print (PrintWriter out)
  {
    out.println (symbolTable.toString ());

    // -- iterate over all basic blocks
    for (Integer uid : cfgTable.keySet ())
    {
      System.out.print ("Function: ");
      System.out.println (declName (symbolTable.getDecl (uid)));

      CFGraph cfg = cfgTable.get (uid);

      System.out.println ("labels: ");
      for (String label : cfg.getLabelMap ().keySet ())
	System.out.println
	  (label + " -> " + cfg.getLabelMap ().get (label).getIndex ());

      System.out.println ("labelUids: ");
      for (Integer labelUid : cfg.getLabelUIDMap ().keySet ())
	System.out.println
	  (labelUid + " -> "
	   + cfg.getLabelUIDMap ().get (labelUid).getIndex ());

      System.out.println ();

      for (BasicBlock bb : cfg.allBlocksFwd ())
      {
	System.out.println ("Saw block: " + bb.getIndex ());
	System.out.println ("With successor edges: ");
	for (Edge e : bb.getSuccs ())
	  System.out.println ("\t" + e);
	System.out.println ("And predecessor edges: ");
	for (Edge e : bb.getPreds ())
	  System.out.println ("\t" + e);

	System.out.println ("Body:");
	for (Expr stmt : bb.getStmts ())
	  System.out.println (stmt);

	System.out.println ();
      }

      System.out.println ();
    }

  }
}
