package edu.toronto.cs.cparser;

import java.io.*;
import java.util.*;

import antlr.*;
import antlr.collections.AST;
import antlr.debug.misc.ASTFrame;

import edu.toronto.cs.cparser.*;
import edu.toronto.cs.cparser.block.*;

/**
 * Insert statements into the code in specified locations. What and where to insert should
 * be specified in the instructions file. This file should have the following format:
 * functionName ==> [at Call, at Return, at Entry, at Exit]
 * 
 * @author <a href="mailto:anton@age.cs">Anton Selyshchev</a>
 * @version 1.0
 */
class SimpleInstrumenter
{
  
  
  static HashMap map = new HashMap();
  
  private SimpleInstrumenter () {}

  // -- show a tree representation of an AST
  public static void drawFrame(String name, AST ptree)
  {
    ASTFrame frame = new ASTFrame(name, ptree);
    frame.setVisible(true);
  }

  /**
   * Reads a file into a String
   *
   * @param f a <code>String</code> specifying the path to the file to be read
   * @return a <code>String</code> with the content of the file
   * @exception Exception if the file could not be read
   */
  public static String readFileIntoString(String f) throws Exception{
    BufferedReader bufReader = new BufferedReader(new FileReader (f));
    String fileText = ""; // this variable will store unparsed instructions
    String line = bufReader.readLine();
    while(line != null){
      fileText = fileText + "\n" + line;
      line = bufReader.readLine();
    }
    return fileText;
  }
  
    
  /**
   * Parses the Instructions file and puts the instructions in the map, 
   * where keys are function names and values are the arrays of statements for each postion.
   * functionName ==> [at Call, at Return, at Entry, at Exit]
   *
   * Instruction file should have the following format:
   * functionName.position (or 'state' keyword to insert global variables 
   * {
   *    Statements;
   * }
   *  
   * @param instrFile a <code>String</code> a path to the instructions file
   * @exception Exception if the file could not be read
   */
  public static void readInstructions(String instrFile) throws Exception
  {
        
    StringTokenizer tokenizer = new StringTokenizer(readFileIntoString(instrFile));
    int statementNumber = -1; 
    String[] statements = {"","","","",""}; 
    String fName = ""; 
    String temp;
    // stack is introduced to keep track of brackets which are also used to enclose instructions.
    int stack;

    // Start reading through the fileText and parsing instructions

    while(tokenizer.hasMoreElements()){
      temp = tokenizer.nextToken(); // reads the first token which is functName.location
      
      // find out where this statement should be inserted
      if(temp.indexOf(".call") != -1){statementNumber = 0;}
      else if(temp.indexOf(".return") != -1){statementNumber = 1;}
      else if(temp.indexOf(".entry") != -1){statementNumber = 2;}
      else if(temp.indexOf(".exit") != -1){statementNumber = 3;}
      else if(temp.indexOf("state") != -1){statementNumber = 4;} 
      else{throw new Exception("invalid instructions format. token is--" + temp + "--, not functName.location");} 
      
      if(statementNumber != 4) fName = temp.substring(0,temp.indexOf(".")); //read function name if there is one 
      temp = tokenizer.nextToken(); // skips {
      
      stack = 1;
      // read the statement
      while(stack != 0){
         temp = tokenizer.nextToken();  
	 if(temp.equals("{")){stack++;}
	 else if(temp.equals("}")){stack--;}
         if(stack !=0){statements[statementNumber] = statements[statementNumber] + " " + temp;} 
      }
      if(statementNumber == 4){  // if we are now working with state
        map.put("state", statements[4]);
      }
      // if we are not then the function name must be present. Check if it is already in the map  
      else{
        String[] tempArray = {"---","---","---","---"};
	if(map.containsKey(fName)){ 
	  tempArray = (String[])map.get(fName); 
	}
        tempArray[statementNumber] = statements[statementNumber];
        map.put(fName, tempArray);
      }
      statements[statementNumber] = ""; 
    } 
  }
  
  /**
   * Runs the program taking as paramters the file to be modified and the instructions.
   *
   * @param args a <code>String[]</code> value
   * @exception Exception if an error occurs
   */
  public static void main(String[] args) throws Exception{
     if(args.length == 0 || args.length == 1){
       String helpMessage = "\nThis program takes two parameters:\n1st: path to the code to be changed\n2nd: path to the instructions file" +
	 "\n\nInstructions format must be:\nfunctionName.location{statement to be inserted}\nfunctionName." + 
	 "anotherLocation{another statement}\n\nPossible locations are:\ncall\nreturn\nenter\nexit";
       
       if(args.length != 0 && args[0].equals("--help")){
	 System.out.println(helpMessage);
	 System.exit(0);
       }
       else{
	 System.err.println(helpMessage);
	 System.exit(1);
       }
     }

      String programName = args[0];
      Reader reader = new FileReader (programName);

      // Read the instructions file if it is given
      if(args.length > 1 ){
	String instructions = args[1];
        readInstructions(instructions);
      }
      
      GnuCLexer lexer =
        new GnuCLexer ( reader );
      lexer.setTokenObjectClass("edu.toronto.cs.cparser.CToken");

      // Parse the input expression.
      GnuCParser parser = new GnuCParser ( lexer );

      // set AST node type to TNode or get nasty cast class errors
      parser.setASTNodeType (TNode.class.getName());
      TNode.setTokenVocabulary ("edu.toronto.cs.cparser.GNUCTokenTypes");

      parser.translationUnit ();

			TNode root = new TNode ();
			root.setFirstChild (parser.getAST ());
			GnuCInstrEmitter treeParser =
			  new GnuCInstrEmitter (new PreprocessorInfoChannel (), map);
			treeParser.setASTNodeClass (TNode.class.getName ());
			treeParser.translationUnit (parser.getAST ());
			
      }
      
  }
  


