package edu.toronto.cs.expr;

import java.util.*;

/**
 * Boolean operators.
 *
 * <p>This class provides common Boolean opeartors and utility
 * functions to construct and manipulate boolean expressions. </p>
 *
 * <p>Additionally, it provides functions that do shallow
 * simplification of Boolean expressions. By convention, these
 * functions are named {@code lXXX}. For example, {@code and} returns
 * an expression that is a syntactic conjunction of its arguments, but
 * {@code land} returns an expression that is only semantically
 * equivalent to the conjunction of its arguments. By convention,
 * {@code lXXX} methods are only allowed to do shallow
 * simplification. This is done in order for them to be as efficient
 * as possible.
 *
 *
 * @author <a href="mailto:arie@kanga.fm.sandbox">Arie Gurfinkel</a>
 * @version 1.0
 */
public class BoolOp extends NamedOp 
{
  public static BoolOp TRUE = new BoolOp ("true", 0);
  public static BoolOp FALSE = new BoolOp ("false", 0);
  
  public static BoolOp AND = new BoolOp ("&&", -1);
  public static BoolOp OR = new BoolOp ("||", -1);
  public static BoolOp XOR = new BoolOp ("xor", -1);
  public static BoolOp NOT = new BoolOp ("!", 1);
  public static BoolOp IMPL = new BoolOp ("->", 2);
  public static BoolOp IFF = new BoolOp ("<->", 2);
  public static BoolOp LPMI = new BoolOp ("<-", 2);
  public static BoolOp ORN = new BoolOp ("\\/\n  ", -1);
  
  private BoolOp (String _name, int _arity)
  {
    super (_name, _arity);
  }


  // -- tester methods
  /**
   * Returns true if {@code e} is a true expression.
   *
   * @param e an <code>Expr</code> value
   * @return a <code>boolean</code> value
   */
  public static final boolean isTrue (Expr e)
  {
    return e.op () == TRUE;
  }
  /**
   * Returns true if {@code e} is a false expression.
   *
   * @param e an <code>Expr</code> value
   * @return a <code>boolean</code> value
   */
  public static final boolean isFalse (Expr e)
  {
    return e.op () == FALSE;
  }

  /**
   * Returns true if e is a negation expression.
   *
   * @param e an <code>Expr</code> value
   * @return a <code>boolean</code> value
   */
  public static final boolean isNot (Expr e)
  {
    return e.op () == NOT;
  }
  /**
   * Returns true if e is a conjunction expression.
   *
   * @param e an <code>Expr</code> value
   * @return a <code>boolean</code> value
   */
  public static final boolean isAnd (Expr e)
  {
    return e.op () == AND;
  }
  /**
   * Returns true if e is a disjunction expression.
   *
   * @param e an <code>Expr</code> value
   * @return a <code>boolean</code> value
   */
  public static final boolean isOr (Expr e)
  {
    return e.op () == OR;
  }
  /**
   * Returns true if e is an implication expression.
   *
   * @param e an <code>Expr</code> value
   * @return a <code>boolean</code> value
   */
  public static final boolean isImpl (Expr e)
  {
    return e.op () == IMPL;
  }
  /**
   * Returns true if e is a reveresed implication expression.
   *
   * @param e an <code>Expr</code> value
   * @return a <code>boolean</code> value
   */
  public static final boolean isLpmi (Expr e)
  {
    return e.op () == LPMI;
  }
  
  

  // -- creation methods

  /**
   * Returns a conjunction of its arguments.
   *
   * @param e1 an <code>Expr</code> value
   * @param e2 an <code>Expr</code> value
   * @return an <code>Expr</code> value
   */
  public static final Expr and (Expr e1, Expr e2)
  {
    return e1.getFactory ().op (AND).binApply (e1, e2);
  }
  
  /**
   * Returns a conjunction of a non-empty list of expressions.
   *
   * @param el a <code>List<Expr></code> value
   * @return an <code>Expr</code> value
   */
  public static final Expr and (List<Expr> el)
  {
    assert el.size () > 0 : "Cannot create conjunction of empty list";
    if (el.size () == 1) return el.get (0);
    return el.get (0).getFactory ().op (AND).naryApply (el);
  }

  /**
   * Returns a disjunction of its arguments.
   *
   * @param e1 an <code>Expr</code> value
   * @param e2 an <code>Expr</code> value
   * @return an <code>Expr</code> value
   */
  public static final Expr or (Expr e1, Expr e2)
  {
    return e1.getFactory ().op (OR).binApply (e1, e2);
  }

  /**
   * Returns a disjunction of a non-empty list of expressions.
   *
   * @param el a <code>List<Expr></code> value
   * @return an <code>Expr</code> value
   */
  public static final Expr or (List<Expr> el)
  {
    assert el.size () > 0 : "Cannot create disjunction of empty list";
    if (el.size () == 1) return el.get (0);
    return el.get (0).getFactory ().op (OR).naryApply (el);
  }

  /**
   * Returns a negation of its argument.
   *
   * @param e an <code>Expr</code> value
   * @return an <code>Expr</code> value
   */
  public static final Expr not (Expr e)
  {
    return e.getFactory ().op (NOT).unaryApply (e);
  }

  /**
   * Returns an implication between its arguments.
   *
   * @param e1 an <code>Expr</code> value
   * @param e2 an <code>Expr</code> value
   * @return an <code>Expr</code> value
   */
  public static final Expr impl (Expr e1, Expr e2)
  {
    return e1.getFactory ().op (IMPL).binApply (e1, e2);
  }

  /**
   * Returns equivalence between its arguments.
   *
   * @param e1 an <code>Expr</code> value
   * @param e2 an <code>Expr</code> value
   * @return an <code>Expr</code> value
   */
  public static final Expr iff (Expr e1, Expr e2)
  {
    return e1.getFactory ().op (IFF).binApply (e1, e2);
  }

  /**
   * Returns reverse implication between its arguments.
   *
   * @param e1 an <code>Expr</code> value
   * @param e2 an <code>Expr</code> value
   * @return an <code>Expr</code> value
   */
  public static final Expr lpmi (Expr e1, Expr e2)
  {
    return e1.getFactory ().op (LPMI).binApply (e1, e2);
  }
  

  // -- creation with simplification methods
  
  /**
   * Returns an expression logically equivalent to negation of its argument.
   *
   * @param e an <code>Expr</code> value
   * @return an <code>Expr</code> value
   */
  public static final Expr lnot (Expr e)
  {
    if (e.op () == NOT) return e.arg (0);
    else if (isFalse (e)) return e.getFactory ().trueExpr ();
    else if (isTrue (e)) return e.getFactory ().falseExpr ();
    return not (e);
  }
  
  /**
   * Returns an expression logically equivalent to conjunction of its
   * arguments.
   *
   * @param e1 an <code>Expr</code> value
   * @param e2 an <code>Expr</code> value
   * @return an <code>Expr</code> value
   */
  public static final Expr land (Expr e1, Expr e2)
  {
    if (e1 == e2) return e1;
    else if (isTrue (e1)) return e2;
    else if (isTrue (e2)) return e1;
    else if (isFalse (e1) || isFalse (e2)) 
      return e1.getFactory ().falseExpr ();

    return and (e1, e2);
  }

  /**
   * Returns an expression logically equivalent to disjunction of its
   * arguments.
   *
   * @param e1 an <code>Expr</code> value
   * @param e2 an <code>Expr</code> value
   * @return an <code>Expr</code> value
   */  
  public static final Expr lor (Expr e1, Expr e2)
  {
    if (e1 == e2) return e1;
    else if (isFalse (e1)) return e2;
    else if (isFalse (e2)) return e1;
    else if (isTrue (e1) || isTrue (e2)) 
      return e1.getFactory ().trueExpr ();
    
    return or (e1, e2);
  }

  /**
   * Strips all negations from {@code e} and returns its absolute value.
   *
   * @param e an <code>Expr</code> value
   * @return an <code>Expr</code> value
   */
  public static final Expr labs (Expr e)
  {
    Expr r = e;
    while (r.op () == NOT) r = r.arg (0);
    return r;
  }
  
  /**
   * Collects all terms, i.e. leaves, of a boolean expression. The
   * result is stored in the provided set, and is also returned for
   * convinience.
   *
   * @param e an expression whose terms are thought.
   * @param s a set into which terms are returned
   * @return the set {@code s}
   * @throws NullPointerException if either {@code e} or {@code s} are null
   */
  public static final Set<Expr> terms (Expr e, Set<Expr> s)
  {
    // -- traverse the expression and collect all non-boolean terms
    // -- into the set s. We assume boolean operators cannot occur inside
    // -- non-boolean terms
    if (! (e.op () instanceof BoolOp))
      s.add (e);
    else
      for (int i = 0; i < e.arity (); i++)
	terms (e.arg (i), s);

    return s;
  }
}

