package edu.toronto.cs.expr;

import static edu.toronto.cs.expr.RationalOp.*;

public class ComparisonOp extends NamedOp
{
  public static ComparisonOp EQ = new ComparisonOp ("=", 2);
  public static ComparisonOp NEQ = new ComparisonOp ("!=", 2);
  public static ComparisonOp LEQ = new ComparisonOp ("<=", 2);
  public static ComparisonOp GEQ = new ComparisonOp (">=", 2);
  public static ComparisonOp GT = new ComparisonOp(">", 2);
  public static ComparisonOp LT = new ComparisonOp("<", 2);
 
  private ComparisonOp (String name, int arity)
  {
    super (name, arity);
  }

  // Creation methods

  /** Creates an equality expression {@code lhs = rhs}.
   *
   * @param lhs an <code>Expr</code> value
   * @param rhs an <code>Expr</code> value
   */
  public static final Expr eq (Expr lhs, Expr rhs)
  {
    return lhs.getFactory (). op (ComparisonOp.EQ).binApply (lhs, rhs);
  }

  /** Creates an equality expression {@code lhs = rhs}.
   *
   * @param lhs an <code>Expr</code> value
   * @param rhs an integer value
   */
  public static final Expr eq (Expr lhs, int rhs)
  {
    return eq(lhs, lhs.getFactory ().intExpr(rhs));
  }

  /** Creates an unequality expression {@code lhs != rhs}.
   *
   * @param lhs an <code>Expr</code> value
   * @param rhs an <code>Expr</code> value
   */
  public static final Expr neq (Expr lhs, Expr rhs)
  {
    return lhs.getFactory (). op (ComparisonOp.NEQ).binApply (lhs, rhs);
  }

  /** Creates an unequality expression {@code lhs != rhs}.
   *
   * @param lhs an <code>Expr</code> value
   * @param rhs an integer value
   */
  public static final Expr neq (Expr lhs, int rhs)
  {
    return neq(lhs, lhs.getFactory ().intExpr(rhs));
  }

  /** Creates an inequality expression {@code lhs <= rhs}.
   *
   * @param lhs an <code>Expr</code> value
   * @param rhs an <code>Expr</code> value
   */
  public static final Expr leq (Expr lhs, Expr rhs)
  {
    return lhs.getFactory (). op (ComparisonOp.LEQ).binApply (lhs, rhs);
  }

  /** Creates an inequality expression {@code lhs <= rhs}.
   *
   * @param lhs an <code>Expr</code> value
   * @param rhs an integer value
   */
  public static final Expr leq (Expr lhs, int rhs)
  {
    return leq(lhs, lhs.getFactory ().intExpr(rhs));
  }

  /** Creates an equality expression {@code lhs >= rhs}.
   *
   * @param lhs an <code>Expr</code> value
   * @param rhs an <code>Expr</code> value
   */
  public static final Expr geq (Expr lhs, Expr rhs)
  {
    return lhs.getFactory (). op (ComparisonOp.GEQ).binApply (lhs, rhs);
  }

  /** Creates an inequality expression {@code lhs >= rhs}.
   *
   * @param lhs an <code>Expr</code> value
   * @param rhs an integer value
   */
  public static final Expr geq (Expr lhs, int rhs)
  {
    return geq(lhs, lhs.getFactory ().intExpr(rhs));
  }

  /** Creates an inequality expression {@code lhs > rhs}.
   *
   * @param lhs an <code>Expr</code> value
   * @param rhs an <code>Expr</code> value
   */
  public static final Expr gt (Expr lhs, Expr rhs)
  {
    return lhs.getFactory (). op (ComparisonOp.GT).binApply (lhs, rhs);
  }

  /** Creates an inequality expression {@code lhs > rhs}.
   *
   * @param lhs an <code>Expr</code> value
   * @param rhs an integer value
   */
  public static final Expr gt (Expr lhs, int rhs)
  {
    return gt(lhs, lhs.getFactory ().intExpr(rhs));
  }

  /** Creates an inequality expression {@code lhs < rhs}.
   *
   * @param lhs an <code>Expr</code> value
   * @param rhs an <code>Expr</code> value
   */
  public static final Expr lt (Expr lhs, Expr rhs)
  {
    return lhs.getFactory (). op (ComparisonOp.LT).binApply (lhs, rhs);
  }

  /** Creates an inequality expression {@code lhs < rhs}.
   *
   * @param lhs an <code>Expr</code> value
   * @param rhs an integer value
   */
  public static final Expr lt (Expr lhs, int rhs)
  {
    return lt(lhs, lhs.getFactory ().intExpr(rhs));
  }

  /** 
   * Returns true iff {@code e} is a comparison expression.
   * 
   * @param e 
   * @return 
   */
  public static final boolean isComparison (Expr e)
  {
    return e.op () instanceof ComparisonOp;
  }

  /** 
   * Returns true iff {@code e} is an equality expression. 
   * 
   * @param e 
   * @return 
   */
  public static final boolean isEq (Expr e)
  {
    return e.op () == EQ;
  }

  /** 
   * Returns true iff {@code e} is a not-equals expression. 
   * 
   * @param e 
   * @return 
   */
  public static final boolean isNeq (Expr e)
  {
    return e.op () == NEQ;
  }

  /** 
   * Returns true iff {@code e} is a greater-than expression. 
   * 
   * @param e 
   * @return 
   */
  public static final boolean isGt (Expr e)
  {
    return e.op () == GT;
  }

  /** 
   * Returns true iff {@code e} is a less-than expression. 
   * 
   * @param e 
   * @return 
   */
  public static final boolean isLt (Expr e)
  {
    return e.op () == LT;
  }

  /** 
   * Returns true iff {@code e} is a greater-than-or-equal-to expression. 
   * 
   * @param e 
   * @return 
   */
  public static final boolean isGeq (Expr e)
  {
    return e.op () == GEQ;
  }

  /** 
   * Returns true iff {@code e} is a less-than-or-equal-to expression. 
   * 
   * @param e 
   * @return 
   */
  public static final boolean isLeq (Expr e)
  {
    return e.op () == LEQ;
  }


  /** 
   * Returns an expression logically equivalent to the equality of its
   * arguments. 
   * 
   * @param e1 
   * @param e2 
   * @return 
   */
  public static final Expr lequal (Expr e1, Expr e2)
  {
    ExprFactory fac = e1.getFactory ();

    if (e1 == e2)
      return fac.op (BoolOp.TRUE);

    if (isInt(e1) && isInt(e2) && getN(e1) != getN(e2))
      return fac.op (BoolOp.FALSE);

    // if they're not syntactically equivalent, they may be semantically
    // equivalent, so we can't simplify
    return eq (e1, e2);
  }

  /** 
   * Returns an expression logically equivalent to {@code e1 &gt; e2}. 
   * 
   * @param e1 
   * @param e2 
   * @return 
   */
  public static final Expr lgt (Expr e1, Expr e2)
  {
    ExprFactory fac = e1.getFactory ();

    if (e1 == e2)
      return fac.op (BoolOp.FALSE);

    if (isInt (e1) && isInt (e2))
      if (getN (e1) > getN (e2))
        return fac.op (BoolOp.TRUE);
      else
        return fac.op (BoolOp.FALSE);

    return gt (e1, e2);
  }

  /** 
   * Returns an expression logically equivalent to {@code e1 &lt; e2}. 
   * 
   * @param e2 
   * @param e2 
   * @return 
   */
  public static final Expr llt (Expr e1, Expr e2)
  {
    Expr e = lgt (e2, e1);

    if (isGt (e))
      return lt (e.arg (1), e.arg (0));

    return e;
  }

  public static void main (String[] args)
  {
    ExprFactory fac = new ExprFactoryImpl ();
    System.out.println (lequal (fac.intExpr (1), fac.intExpr (1)));
  }
}
