package edu.toronto.cs.expr;


public class CompoundTypeOp extends NamedOp
{
  /**
   * Functional type: {@literal A -> B}
   *
   */
  public static final CompoundTypeOp FUN_TYPE = 
    new CompoundTypeOp ("->", 2);
  
  /**
   * Range type: {@literal [a..b]}
   *
   */
  public static final CompoundTypeOp RNG_TYPE = 
    new CompoundTypeOp ("RNG_TYPE", 2);
  
  /**
   * Array type: {@literal ARRAY INDEX_TYPE OF VALUE_TYPE}
   *
   */
  public static final CompoundTypeOp ARRAY_TYPE =
    new CompoundTypeOp ("ARRAY_TYPE", 2);

  /**
   * A sub-type: {@literal SUBTYPE predicate}
   *
   */
  public static final CompoundTypeOp SUB_TYPE = 
    new CompoundTypeOp ("SUB_TYPE", 1);

  /**
   * Pointer type: {@literal POINTER REF_TYPE}
   */
  public static final CompoundTypeOp PTR_TYPE =
    new CompoundTypeOp ("PTR_TYPE", 1);
  
  public CompoundTypeOp (String name, int arity)
  {
    super (name, arity);
  }
  
  /**
   * Creates a functional type with domain {@code dom} and range
   * {@code rng}.
   *
   * @param dom an <code>Expr</code> value
   * @param rng an <code>Expr</code> value
   * @return an <code>Expr</code> value
   */
  public static final Expr funT (Expr dom, Expr rng)
  {
    return dom.getFactory ().op (FUN_TYPE).binApply (dom, rng);
  }

  /**
   * Creates an integer range type with minimum bound {@code min} and
   * maximum bound {@code max}.
   *
   * @param min an <code>Expr</code> value
   * @param max an <code>Expr</code> value
   * @return an <code>Expr</code> value
   */
  public static final Expr rngT (Expr min, Expr max)
  {
    return min.getFactory ().op (RNG_TYPE).binApply (min, max);
  }
  
  /**
   * Creates an array type with domain {@code dom} and range {@code rng}.
   *
   * @param dom an <code>Expr</code> value
   * @param rng an <code>Expr</code> value
   * @return an <code>Expr</code> value
   */
  public static final Expr arrayT (Expr dom, Expr rng)
  {
    return dom.getFactory ().op (ARRAY_TYPE).binApply (dom, rng);
  }

  /**
   * Creates a subtype out of a predicate {@code pred}.
   *
   * @param pred an <code>Expr</code> value
   * @return an <code>Expr</code> value
   */
  public static final Expr subT (Expr pred)
  {
    return pred.getFactory ().op (SUB_TYPE).unaryApply (pred);
  }

  /** 
   * Creates a type: pointer to {@code refType}. 
   * 
   * @param refType 
   * @return 
   */
  public static final Expr ptrT (Expr refType)
  {
    return refType.getFactory ().op (PTR_TYPE).unaryApply (refType);
  }

  /**
   * Returns true if {@code e} is a functional type.
   *
   * @param e an <code>Expr</code> value
   * @return a <code>boolean</code> value
   */
  public static final boolean isFunT (Expr e)
  {
    return e.op () == FUN_TYPE;
  }
  /**
   * Returns truen if {@code e} is a range type.
   *
   * @param e an <code>Expr</code> value
   * @return a <code>boolean</code> value
   */
  public static final boolean isRngT (Expr e)
  {
    return e.op () == RNG_TYPE;
  }
  /**
   * Returns true if {@code e} is an array type.
   *
   * @param e an <code>Expr</code> value
   * @return a <code>boolean</code> value
   */
  public static final boolean isArrayT (Expr e)
  {
    return e.op () == ARRAY_TYPE;
  }
  /**
   * Returns true if {@code e} is a subtype.
   *
   * @param e an <code>Expr</code> value
   * @return a <code>boolean</code> value
   */
  public static final boolean isSubT (Expr e)
  {
    return e.op () == SUB_TYPE;
  }

  /** 
   * Returns true iff {@code e} is a pointer type. 
   * 
   * @param e 
   * @return 
   */
  public static final boolean isPtrT (Expr e)
  {
    return e.op () == PTR_TYPE;
  }

  /**
   * Extracts the domain of a functional or an array type.
   *
   * @param e an <code>Expr</code> value
   * @return an <code>Expr</code> value
   */
  public static final Expr dom (Expr e)
  {
    return e.arg (0);
  }
  /**
   * Extracts a range of a functional or an array type.
   *
   * @param e an <code>Expr</code> value
   * @return an <code>Expr</code> value
   */
  public static final Expr rng (Expr e)
  {
    return e.arg (1);
  }
  /**
   * Extracts the minimal bound of a range type.
   *
   * @param e an <code>Expr</code> value
   * @return an <code>Expr</code> value
   */
  public static final Expr min (Expr e)
  {
    return e.arg (0);
  }
  /**
   * Extracts the maximal bound of a range type.
   *
   * @param e an <code>Expr</code> value
   * @return an <code>Expr</code> value
   */
  public static final Expr max (Expr e)
  {
    return e.arg (1);
  }

  /**
   * Extracts the predicate of a subtype.
   *
   * @param e an <code>Expr</code> value
   * @return an <code>Expr</code> value
   */
  public static final Expr pred (Expr e)
  {
    return e.arg (0);
  }
}

