package edu.toronto.cs.expr;

import edu.toronto.cs.expr.*;
import java.util.*;
import edu.toronto.cs.tp.SymbolTable;

import static edu.toronto.cs.expr.VariableOp.*;
import static edu.toronto.cs.expr.CompoundTypeOp.*;
import static edu.toronto.cs.expr.ArrayOp.*;
import static edu.toronto.cs.expr.CILIndirectionOp.*;
import static edu.toronto.cs.expr.SimpleTypeOp.*;
import static edu.toronto.cs.expr.RationalOp.*;
import static edu.toronto.cs.expr.NumericOp.*;

import static edu.toronto.cs.expr.CVarDeclOp.*;

/**
 * A type inferencer. Uses a symbol table to get the types of
 * variables, and caches subexpressions so that subsequent lookups are
 * fast.
 *
 * Supported:
 *   - variables: INT, CHAR, REAL, ARRAY
 *   - integer and real arithmetic
 *   - pointer arithmetic p +/- n
 *   - array selections and updates
 *   - dereference and address-of
 *   - integer and real literals
 *
 * Not supported but should be:
 *   - character literals
 *   - field access through "." and "-&gt;"
 *
 * @author <a href="mailto:tomhart@cs.toronto.edu">Tom Hart</a>
 * @version 1.0
 */
public class TypeInferencerCached implements TypeInferencer
{

  Map <Expr, Expr> cache;
  SymbolTable st;

  public TypeInferencerCached (SymbolTable _st)
  {
    st = _st;
    cache = new IdentityHashMap<Expr,Expr> ();
  }

  public Expr getType (Expr e)
  {
    Expr t;

    t = cache.get (e);
    /***********************************************************************
     *
     * Base case: cache hit
     *
     **********************************************************************/
    if (t != null) {
      //System.out.println ("Cache hit on " + e + "! :)");
      return t;
    }

    /***********************************************************************
     *
     * Base case: constant
     *
     **********************************************************************/
    if (e.op () instanceof RationalOp)
      if (isInt (e))
        return e.getFactory ().op (SimpleTypeOp.INT);
      else 
        return e.getFactory ().op (SimpleTypeOp.REAL);

    /***********************************************************************
     *
     * Base case: variables
     *
     **********************************************************************/
    if (isCVarDecl (e) || isVar (e))
      return st.typeOf (e);

    /***********************************************************************
     *
     * Base case: comparisons
     *
     **********************************************************************/
    if (e.op () instanceof ComparisonOp)
      // Using INT as a stand-in for boolean
      return e.getFactory ().op (SimpleTypeOp.INT);

    /***********************************************************************
     *
     * Recursive case: arithmetic (integers, reals, and pointers)
     *
     **********************************************************************/
    if (e.op () instanceof NumericOp) {

      if (e.arity () == 1) {
        t = getType (e.arg (0));
        cache.put (e, t);
        return t;
      }

      else if (e.arity () == 2) {
	// -- if at least one argument of a binary expression is a pointer
	// -- then the whole expression is a pointer.
	// -- examples with int *p: 
	// --   p + 1  and 1 + p
        Expr e1 = getType (e.arg (0));
        Expr e2 = getType (e.arg (1));

        if (isPtrT (e1)) {
          cache.put (e, e1);
          return e1;
        }
        if (isPtrT (e2)) {
          cache.put (e, e2);
          return e2;
        }

        assert (e1.op () instanceof SimpleTypeOp
                && e2.op () instanceof SimpleTypeOp) :
          "Unexpected types! \n  e1: " + e1 + "\n  e2: " + e2;
	
        // This'll work with ints and pointers. Extend - char, real, etc.
        cache.put (e, e1);
        return e1;
      } else {
        throw new RuntimeException ("Unsupported numeric operation in " + e);
      }
    }

    /***********************************************************************
     *
     * Recursive case: dereferences and addr-ofs
     *
     **********************************************************************/
    if (e.op () instanceof CILIndirectionOp) {
      if (isDeref (e)) {
        t = getType (e.arg (0));
        assert isPtrT (t) : "Expected pointer, got " + t + " from " + e;
        cache.put (e, t.arg (0));
        return t.arg (0);
      } else if (isAddrOf (e)) {
        t = getType (e.arg (0));
        cache.put (e, ptrT (t));
        return ptrT (t);
      } else if (isDot (e)) {
        throw new RuntimeException 
          ("Support for '.' not implemented yet: " + e);
      } else if (isArrow (e)) {
        throw new RuntimeException 
          ("Support for '->' not implemented yet " + e);
      } else
        throw new RuntimeException 
          ("Unsupported indirection operation in " + e);
    }

    /***********************************************************************
     *
     * Recursive case: arrays
     *
     **********************************************************************/
    if (e.op () instanceof ArrayOp) {
      if (isArraySel (e)) {
        t = getType (getArray (e));
        cache.put (e, rng (t));
        return rng (t);
      } else if (isArrayUpd (e)) {
        t = getType (getArray (e));
        cache.put (e, t);
        return t;
      } else
        throw new RuntimeException ("Unsupported array operation in " + e);
    }

    throw new RuntimeException ("Unsupported operation in " + e);
  }

  public static void main (String[] args) throws Exception
  {
    // XXX This needs to be turned into a junit test
    ExprFactory fac = new ExprFactoryImpl ();

    Expr intT = fac.op (INT);
    Expr charT = fac.op (CHAR);
    Expr ar_int_T = CompoundTypeOp.arrayT (intT, intT);
    Expr ar_char_T = CompoundTypeOp.arrayT (intT, charT);
    Expr ipT = ptrT (intT);
    Expr ippT = ptrT (ptrT (intT));

    Expr A = fac.var("A");
    Expr C = fac.var("C");
    Expr ip = fac.var ("ip"); // int* ip; (ADDRESS (INT))
    Expr ipp = fac.var ("ipp"); // int** ipp; (ADDRESS (ADDRESS (INT)))
    Expr x = fac.var ("x");

    SymbolTable st = new SymbolTable ();
    st.put (ip, ipT);
    st.put (ipp, ippT);
    st.put (x, intT);
    st.put(A, ar_int_T);
    st.put(C, ar_char_T);

    TypeInferencerCached ti = new TypeInferencerCached (st);

    Expr deref_ip = CILIndirectionOp.deref (ip);
    Expr deref_ipp = CILIndirectionOp.deref (ipp);
    Expr deref_deref_ipp =
      CILIndirectionOp.deref (CILIndirectionOp.deref (ipp));

    Expr addrx = CILIndirectionOp.addrOf (x);
    Expr deref_addrx = CILIndirectionOp.deref (addrx);

    System.out.println("type of 0 is " + ti.getType (fac.intExpr(0)));
    System.out.println("type of x is " + ti.getType (x));
    System.out.println("type of &x is " + ti.getType (addrx));
    System.out.println("type of *&x is " + ti.getType (deref_addrx));
    System.out.println("type of x + (-x) is " +
                       ti.getType (plus (x, un_minus (x))));
    System.out.println("type of ip is " + ti.getType (ip));
    System.out.println("type of *ip is " + ti.getType (deref_ip));
    System.out.println("type of *ipp is " + ti.getType (deref_ipp));     
    System.out.println("type of ip + 1 is " + ti.getType (plus (ip, 1)));
    System.out.println("type of *(ip + 1) is " + 
                       ti.getType (deref (plus (ip, 1))));
    System.out.println("type of A is " + ti.getType (A));
    System.out.println("type of A[0] is " + ti.getType (sel (A, 0)));
    System.out.println("type of A[x |-> x][*ip] is " 
                       + ti.getType (sel (upd (A, x, x), deref_ip)));
  }

}
