package edu.toronto.cs.tp;

import java.util.*;


import edu.toronto.cs.expr.*;
import edu.toronto.cs.tp.cvcl.ValidityChecker;
import edu.toronto.cs.tp.cvcl.CVectorExpr;
import edu.toronto.cs.tp.cvcl.Type;
import edu.toronto.cs.util.ArrayMarkedList;
import edu.toronto.cs.tp.cvcl.CVCLUtil;
import edu.toronto.cs.yasm.YasmStatistics;
import java.util.Stack;


/**
 * Implementation of {@link edu.toronto.cs.tp.ITheoremProver} based on
 * a binary interface to CVCLite.
 *
 * @author <a href="mailto:arie@kanga.fm.sandbox">Arie Gurfinkel</a>
 * @version 1.0
 */

public class CLCVCLTheoremProver extends AbstractTheoremProver
{
  YasmStatistics stats = new YasmStatistics ();

  // -- an instance of CVCLProcess
  CVCLProcess process;
  List<Expr> assertions;
  List<Integer> stack;
  boolean debug;
  ArrayMarkedList markedList;
  // the process' initial stack level
  private int cvclBase;

  // -- an instance of a converter between CVCL expressions and
  // -- edu.toronto.cs.expr.Expr
  CVCLExprPrinter exprConverter;

  /**
   * A class implementing the {@code TPInitializer} interface can be
   * given our {@code CVCLProcess}, and can use it to make
   * declarations, etc.
   * 
   * This lets us easily make a bunch of declarations which would be
   * awkward to make using an {@code Expr} tree, and require a lot of
   * new classes; for example, {@code NAT:TYPE = SUBTYPE(LAMBDA
   * (x:INT): x > 0)}.
   */
  public static interface TPInitializer
  {
    public void init (CVCLProcess p);
  }
 /**
   * Constructs a {@code CLCVCLTheoremProver} given an expression
   * factory.
   *
   * @param _exprFactory an <code>ExprFactory</code> value
   */
  public CLCVCLTheoremProver (ExprFactory _exprFactory) 
  {
    super (_exprFactory);
    init (false);
  }

  /**
   * Constructs a {@code CLCVCLTheoremProver} given an expression
   * factory and a flag to turn on the logging mode.
   *
   * @param _exprFactory an <code>ExprFactory</code> value
   */
  public CLCVCLTheoremProver (ExprFactory _exprFactory, boolean logging) 
  {
    super (_exprFactory);
    init (logging);
  }

  private void init (boolean logging)
  {
    process = new CVCLProcess (logging);
    cvclBase = process.where ().getStackLevel ();
    
    debug = false;
    
    exprConverter = new CVCLExprPrinter();

    assertions = new ArrayList();

    markedList = new ArrayMarkedList();
  }

  /**
   * Gives this object's {@code CVCLProcess} to an initializer which 
   * can make some declarations.
   *
   * @param i an object implementing the <code>TPInitializer</code>
   * interface
   */
  public void initialize (TPInitializer i)
  {
    i.init(process);
  }

  /** 
   * Returns the CVCL process; breaks encapsulation, oh well.
   * 
   * @return the CVCL process
   */
  public CVCLProcess getProcess ()
  {
    return process;
  }

  /**
   * This method declares a variable given its name and type as Expressions
   * @param name- a name of the variable. type- it's type 
   */
  public void declare (Expr name, Expr type)
  {
    process.declare (exprConverter.print (name), exprConverter.print (type));
    if (debug) printDebug ("declare");
    
  }

  /**
   * This method checks if a query is true or not.
   * @param expr- the expression to be evaluated 
   */
  public boolean query (Expr q)
  {
    stats.startTPQuery ();
    if (process.query (exprConverter.print (q)).isValid ())
    {
      if(debug) printDebug ("query " + q + " true");
      stats.stopTPQuery ();
      return true;
    }
    if(debug) printDebug("query " + q + " false");
    stats.stopTPQuery ();
    return false;
    
  }

  /**
   * Returns a list of the indices of the elements in other as found in base.
   */
  private static List<Integer> findCommonIndices (List base, List other)
  {

    List<Integer> commonIndices = new ArrayList<Integer> (other.size ());

    for (Iterator it = other.iterator (); it.hasNext ();)
    {
      Object o = it.next ();
      int index = base.indexOf (o);
      if (index < 0)
        throw new MismatchException (o.toString ());

      commonIndices.add (new Integer (index));
    }

    return commonIndices;
  }

  public static class MismatchException extends RuntimeException
  {
    public MismatchException (String msg)
    {
      super ("Unable to match " + msg);
    }
  }

  /**
   * Returns a list of the Exprs in exprs indexed by indices, in the order of
   * indices.
   */
  private static List<Expr> getMany (List<Expr> exprs, List<Integer> indices)
  {
    List<Expr> matchedExprs = new ArrayList<Expr> (indices.size ());
    for (Iterator it = indices.iterator (); it.hasNext ();)
      matchedExprs.add (exprs.get (((Integer) it.next ()).intValue ()));
    return matchedExprs;
  }

  /**
   * Returns a list of the Exprs corresponding to the assumptions used in the
   * last valid query.
   */
  public List<Expr> getAssumptionsUsed ()
  {
    stats.startTPGetAssm ();
    List<Expr> result = getMany (markedList, getIndicesOfAssumptionsUsed ());
    stats.stopTPGetAssm ();
    return result;
  }

  public List<Integer> getIndicesOfAssumptionsUsed ()
  {
    stats.startTPGetAssm ();
    CVCLProcess.WhereMessage whereM = process.where();
    CVCLProcess.AssertListMessage assertListM = process.dump_assumptions();
    List<Integer> result =
      findCommonIndices (whereM.getAsserts (), assertListM.getAsserts ());
    stats.stopTPGetAssm ();
    return result;
  }

  /**
   * This method assert an expression.
   * @param expr- the expression to be asserted 
   */
  public void assertFormula (Expr expr)
  {
    stats.startTPAssert ();
    process.assertFormula (exprConverter.print (expr));
    markedList.add(expr);
    //assertions.add(expr);
    if (debug) printDebug ("assertFormula");
    stats.stopTPAssert ();
  }

  /**
   * This method returns the current stack level.
   */
  public int stackLevel ()
  {
    return markedList.getNumMarks();
    //return stack.size ();
  }

  /**
   * remebers how many assertions we had at the time
   * of the push, pop will get us back here   *
   */
  public void push ()
  {
    markedList.mark();
    process.push ();
    if (debug) printDebug ("push");
  }

  public void pop ()
  {
    markedList.restore();
    process.pop();
    
    if (debug) printDebug ("pop");
            
  }

  public void popto (int i)
  {
    markedList.restore (i);
    process.popto (i + cvclBase);
    if (debug) printDebug ("popTo (" + i + ")");
              
  }

  /**
   * Removes assertions which are not in the current stack level
   *
   */
  void syncAssertions (int bound)
  {    
    // -- synchronize assertions with the stack
    if (bound == 0)
      assertions.clear ();
    else
      while (assertions.size () > bound)
        assertions.remove (assertions.size () - 1);
  }

  /**
   * This method returns all the assumptions that were made.
   */

  public List<Expr> printAssumptions()
  {
    return assertions;
  }

  public void setStats (YasmStatistics stats)
  {
    this.stats = stats;
    process.setStats (stats);
  }

  private void printDebug(String methodName)
  {
    System.out.println ("=== " + methodName + " ============================");
    System.out.println ("   numberOfMarks: " + markedList.getNumMarks  ());
    try {
      CVCLProcess.WhereMessage whereM = process.where();
      System.out.println ("   stackLevel: " + process.getStackLevel ());
      System.out.println ("   whereMessage Asserts: " + whereM.getAsserts ());
    }
    catch(Exception e){
      System.out.println("Where called in the invalid query state " + e);
    }
    System.out.println ("   Expr list:            " + markedList);
    System.out.println ("===================================================");

  }
  

}

