package edu.toronto.cs.tp;

import java.util.*;

import edu.toronto.cs.expr.*;
import edu.toronto.cs.tp.cvcl.ValidityChecker;
import edu.toronto.cs.tp.cvcl.CVectorExpr;
import edu.toronto.cs.tp.cvcl.Type;
import edu.toronto.cs.tp.cvcl.CVCLUtil;
import edu.toronto.cs.yasm.YasmStatistics;

import edu.toronto.cs.util.*;

import static edu.toronto.cs.expr.BoolOp.*;
import static edu.toronto.cs.expr.SimpleTypeOp.*;
import static edu.toronto.cs.expr.CompoundTypeOp.*;
import static edu.toronto.cs.expr.ComparisonOp.*;
import static edu.toronto.cs.expr.CILIndirectionOp.*;

/**
 * A theorem prover for C expressions.
 *
 * Background assertions are made at the initial stack level and are
 * permanent in that the stack should never return to the initial
 * level; furthermore, they must be chosen when the theorem prover is
 * created.
 *
 * @author <a href="mailto:arie@kanga.fm.sandbox">Arie Gurfinkel</a>,
 * <a href="mailto:kelvin@cs.toronto.edu">Kelvin Ku</a>, <a
 * href="mailto:tomhart@cs.toronto.edu">Tom Hart</a>
 * @version 1.0
 */

public class CTheoremProver extends AbstractTheoremProver
{
  ITheoremProver tp;
  CToAddrModel addrModel;
  SymbolTable symbolTable;

  List<Expr> assertions;
  List<Expr> backgroundAssertions;
  Stack<Integer> stack;

  YasmStatistics stats = new YasmStatistics ();
  
  void doBackgroundAssertions (List<Expr> el)
  {
    for (Expr e : el) {
      Expr rw = addrModel.rewrite (e);
      tp.assertFormula (rw);
    }
    push ();
    backgroundAssertions = el;
  }

  void init (ExprFactory fac, 
             ITheoremProver tp,
             SymbolTable symbolTable,
             List<Expr> bgAssertions) {
    this.tp = tp;

    this.symbolTable = symbolTable;

    addrModel = new CToAddrModel (fac, symbolTable);

    assert (tp instanceof CLCVCLTheoremProver);

    ((CLCVCLTheoremProver) tp).initialize (addrModel);

    assertions = new LinkedList<Expr> ();

    stack = new Stack<Integer> ();

    doBackgroundAssertions (bgAssertions);

    assert tp.stackLevel () == 1;
    assert stack.size () == 1;//0;
  }

  public CTheoremProver (ExprFactory fac, ITheoremProver tp,
      SymbolTable symbolTable)
  {
    super(fac);
    init (fac, tp, symbolTable, new LinkedList<Expr> ());
  }

  public CTheoremProver (ExprFactory fac, 
                         ITheoremProver tp,
                         SymbolTable symbolTable,
                         List<Expr> bgAssertions)
  {
    super(fac);
    init (fac, tp, symbolTable, bgAssertions);
  }

  /**
   * Declares a thing as having the specified name and type, that is, binds the
   * name to the type.
   *
   * @param name the name of the thing being declared
   * @param type the thing's type
   */
  
  public void declare (Expr name, Expr type)
  {
    // XXX doesn't look like we will actually declare CIL variables to the
    // tp since the tp will only work with addresses
    // XXX actually, it's the memory model which determines what declarations
    // need to be made; under an address model, address types and referencing
    // functions need to be declared, but not variables; under a variable
    // model, variables need to be declared
    // XXX for now assume the TP comes with the necessary type, variable, and
    // function declarations made
    // tp.declare (addrModel.rewrite (name), addrModel.rewrite (type));
  }

  /**
   * This method checks if a query is true or not.
   * @param expr- the expression to be evaluated 
   */

  public boolean query (Expr q)
  {
    Expr rwQuery = addrModel.rewrite (q);
    SystemOutWrapper.writer.println ();
    SystemOutWrapper.writer.println (">>> query " + q);
    SystemOutWrapper.writer.println (">>> context " + assertions);
//    SystemOutWrapper.writer.println (">>> rewritten " + rwQuery);
    boolean result = tp.query (rwQuery);
    SystemOutWrapper.writer.println ();
    SystemOutWrapper.writer.println (">>> result " + result);
    return result;
  }

  /**
   * Returns a list of the Exprs corresponding to the assumptions used in the
   * last valid query. Do NOT include background assumptions.
   */
  public List<Expr> getAssumptionsUsed ()
  {
    /* Convert CVCL assumptions into C assumptions so we have
	  predicates over C variables. Can do this by getting the
	  indices of the CVCL assumptions used. No "unconversion"
	  necessary. 
	  
	  We explicitly do *not* want to get the "byproduct"
	  assertions.
    */

    List<Expr> assumptionsUsed = new LinkedList<Expr> ();

    // XXX getIndicesOfAssumptionsUsed is not in ITheoremProver interface
    for (Integer index :
           ((CLCVCLTheoremProver) tp).getIndicesOfAssumptionsUsed ()) {
      // Adjust the index to account for background assertions.
      int base = index.intValue ();
      if (base < backgroundAssertions.size ()) continue;
      int adjustedIndex = base - backgroundAssertions.size ();
      assumptionsUsed.add (assertions.get (adjustedIndex));
    }

    SystemOutWrapper.writer.println ();
    SystemOutWrapper.writer.println (">>> assmUsed " + assumptionsUsed);

    return assumptionsUsed;
  }

  /**
   * This method asserts an expression.
   * @param expr- the expression to be asserted 
   */
  public void assertFormula (Expr assertion)
  {
    /* Adds a formula to CVCL's list of assumptions. Turns the expr
	  into a CVCL expression. Must work with
	  getAssumptionsUsed. 
    */
    Expr rwAssertion = addrModel.rewrite (assertion);
    SystemOutWrapper.writer.println ();
    SystemOutWrapper.writer.println (">>> assert " + assertion);
//    SystemOutWrapper.writer.println (">>> rewritten " + rwAssertion);
    tp.assertFormula (rwAssertion);
    assertions.add (assertion);
  }

  /** Might die. */
  /*public void assertBackgroundFormula (Expr expr)
  {
    throw new RuntimeException ("not handling background assertions yet");
    }*/

  /**
   * This method returns the current stack level.
   */
  public int stackLevel ()
  {
    assert stack.size () == tp.stackLevel () : 
      "stacks out of sync:\n"
      + "\tstack.size () == " + stack.size () 
      + "\ttp.stackLevel () == " + tp.stackLevel ()
      + "\tTP assertions: " + ((CLCVCLTheoremProver)tp).printAssumptions ();

    return tp.stackLevel ();
  }

  public void push ()
  {
    // -- remebers how many assertions we had at the time
    // -- of the push, pop will get us back here
    stack.push (assertions.size ());
    tp.push ();
  }

  public void pop ()
  {
    if (stack.isEmpty ()) 
      throw new IllegalStateException ("The stack is empty");

    // -- remove the top of the stack
    syncAssertions (stack.pop ());
    tp.pop ();
  }

  public void popto (int i)
  {
    if (stackLevel () == i) return ;

    // Must have i >= 1 because background assumptions live at level 0
    if (stack.size () < i || i < 1)
      throw new IllegalArgumentException ("Cannot pop to level " + i);

    int bound = stack.size ();
    while (stack.size () > i)
    {
      bound = stack.pop ();
    }

    syncAssertions (bound);
    tp.popto (i);
  }

  public void setStats (YasmStatistics stats)
  {
    this.stats = stats;
    ((CLCVCLTheoremProver) tp).setStats (stats);
    addrModel.setStats (stats);
  }

  /**
   * Removes assertions which are not in the current stack level
   *
   */
  void syncAssertions (int bound)
  {    
    // -- synchronize assertions with the stack
    if (bound == 0)
      assertions.clear ();
    else
      while (assertions.size () > bound)
        assertions.remove (assertions.size () - 1);
  } 
}

