package edu.toronto.cs.tp;

import junit.framework.*;

import edu.toronto.cs.expr.*;
import edu.toronto.cs.cparser.*;
import static edu.toronto.cs.expr.BoolOp.*;
import static edu.toronto.cs.expr.SimpleTypeOp.*;
import static edu.toronto.cs.expr.CompoundTypeOp.*;
import static edu.toronto.cs.expr.ComparisonOp.*;
import static edu.toronto.cs.expr.CILIndirectionOp.*;
import static edu.toronto.cs.expr.NumericOp.*;

/** 
 * A collection of C theorem prover tests.
 * <p>
 * JUnit checks all methods prefixed with "test".
 * 
 * @author Kelvin Ku (kelvin@cs.toronto.edu)
 * @version 
 */
public class CTheoremProverTestCase extends TestCase
{
  // some common expressions
  private Expr x;
  private Expr ax;
  private Expr y;
  private Expr ay;
  private Expr ip;
  private Expr intT;
  private Expr charT;

  ExprFactory fac;
  CTheoremProver ctp;
  SymbolTable st;

  protected void setUp ()
  {
    fac = new ExprFactoryImpl ();

    st = new SymbolTable ();

    intT = fac.op (INT);
    charT = fac.op (CHAR);

    // int x
    x = fac.var ("x");

    // &x
    ax = fac.op (ADDR_OF).unaryApply (x);

    // int y
    y = fac.var ("y");

    // &y
    ay = fac.op (ADDR_OF).unaryApply (y);

    // int* ip
    ip = fac.var ("ip");

    st.put (x, intT);
    st.put (y, intT);
    st.put (ip, ptrT (intT));

    ctp = new CTheoremProver (fac, new CLCVCLTheoremProver (fac, true), st);
  }

  protected void tearDown ()
  {
  }

  public void testScalarAddrsNotEqual ()
  {
    // &x != &y
    Expr ax_neq_ay = not (eq (ax, ay));

    assertTrue (ctp.quickQuery (ax_neq_ay));
  }

  public void testAddrNotZero ()
  {
    // taken from play/models/cprog/ptr/ptr3.c:

    // &x != 0
    Expr ax_neq_0 = not (eq (ax, fac.intExpr (0)));

    assertTrue (ctp.quickQuery (ax_neq_0));
  }

  public void testStarAddr1 ()
  {
    // taken from play/models/cprog/ptr/ptr4.c:

    // *&y
    Expr star_ay = deref (ay);

    // *&y = y
    Expr star_ay_eq_y = eq (star_ay, y);

    assertTrue (ctp.quickQuery (star_ay_eq_y));
  }

  public void testStarAddr2 ()
  {
    // y > 0
    Expr y_gt_0 = gt (y, fac.intExpr (0));

    // *&y
    Expr star_ay = deref (ay);

    // y > 0 => *&y > 0
    Expr sanity = impl (y_gt_0, gt (star_ay, fac.intExpr (0)));

    assertTrue (ctp.quickQuery (sanity));
  }

  public void testAliasDeref ()
  {
    // taken from play/models/cprog/ptr/ptr6.c:
 
    // ip = &y
    Expr ip_eq_ay = eq (ip, ay);

    // y > 0
    Expr y_gt_0 = gt (y, fac.intExpr (0));

    // ip = &y AND y > 0
    Expr p = and (ip_eq_ay, y_gt_0);

    // *ip > 0
    Expr q = gt (deref (ip), fac.intExpr (0));

    // p => q
    Expr p_impl_q = impl (p, q);

    assertTrue (ctp.quickQuery (p_impl_q));
  }


  public void testPtrArith ()
  {
    // *ip = 0 => *(ip + 1) = 0 -- should be FALSE, but if ip is treated as an
    // int pointer, the offset is thrown away and only the base object
    // (supposedly an int) is considered
    Expr ptrArith =
      impl (
        eq (deref (ip), 0),
        eq (deref (plus (ip, 1)), 0));

    assertFalse (ctp.quickQuery (ptrArith));
  }

  public void testInconsistent () throws Exception
  {
    // dest  PTR_TYPE CHAR
    // buf_ext PTR_TYPE CHAR
    // buf
    // src
    // NOT(dest > buf_ext AND NOT(dest + 1 > buf_ext))

    Expr dest = fac.var ("dest");
    Expr buf_ext = fac.var ("buf_ext");

    st.put (dest, ptrT (charT));
    st.put (buf_ext, ptrT (charT));

    // have to recreate ctp since we changed st, which kinda sucks
    ctp = new CTheoremProver (fac, new CLCVCLTheoremProver (fac), st);

    Expr q = CILTransformerUtil.getExpr (fac,
        "(dest > buf_ext && !(dest + 1 > buf_ext))");

    assertFalse (ctp.quickQuery (q));
  }

  public void testInconsistentContext () throws Exception
  {
    Expr p1 = CILTransformerUtil.getExpr (fac, "x > 0");
    Expr p2 = CILTransformerUtil.getExpr (fac, "!(x + 1 > 0)");

    Expr target = fac.falseExpr ();

    ctp.assertFormula (p1);
    ctp.assertFormula (p2);

    assertTrue (ctp.query (target));
  }
}

