package edu.toronto.cs.tp;

import java.util.*;
import edu.toronto.cs.expr.*;
import edu.toronto.cs.yasm.YasmStatistics;

import static edu.toronto.cs.expr.SimpleTypeOp.*;
import static edu.toronto.cs.expr.CompoundTypeOp.*;
import static edu.toronto.cs.expr.CILIndirectionOp.*;
import static edu.toronto.cs.expr.NumericOp.*;
import static edu.toronto.cs.expr.ComparisonOp.*;
import static edu.toronto.cs.expr.BoolOp.*;
import static edu.toronto.cs.expr.VariableOp.*;
import static edu.toronto.cs.expr.ArrayOp.*;

import static edu.toronto.cs.expr.CVarDeclOp.*;

/** 
 * Translates CIL Exprs to address model Exprs, where each variable
 * <code>x</code> is associated with a unique non-zero arbitrary address
 * <code>a</code> so that "x" is value-at-<code>a</code>, "&amp;x" is
 * <code>a</code>, and "&#42;x" is value-at-value-at-<code>a</code>; the
 * mapping between variables and addresses is maintained over the lifetime of
 * the converter (that is, not just per conversion).
 *
 * When translating to CVCL, an address becomes a tuple <OBJECT,
 * OFFSET>.  The OBJECT is the unique non-zero value mentioned above,
 * while the OFFSET, for a scalar variable, is 0. Pointers into arrays
 * are represented in the same way, but the OFFSET can be non-zero,
 * and represents the offset into the array. This way, we only have to
 * deal with a single address type for both scalars and arrays.
 *
 * <p>
 * XXX Should generate declarations, e.g., NAT, ADDR, INT_REF, ADDR_REF in Expr
 * form for an Expr-to-TP-language converter to pass on to the TP
 * ^^^ Why? That seems like over-engineering. - Tom
 * <p>
 * XXX Should generate by-product assertions (doesn't seem like there are any
 * under this memory model)
 *
 * @author Kelvin Ku (kelvin@cs.toronto.edu), Tom Hart (tomhart@cs.toronto.edu)
 * @version 
 */
public class CToAddrModel
  implements CLCVCLTheoremProver.TPInitializer, ExprRewriter
{
  YasmStatistics stats = new YasmStatistics ();

  // variable -> unique address
  Map<Expr, Integer> varAddressMap;

  RefFunctionMap refFunctions;

  // maximum address appearing in varAddressMap
  int maxAddress;

  ExprFactory fac;

  SymbolTable symbolTable;

  ////////////////////////////////////////////////////////////////////////////
  // Reference Functions:
  //
  // To reduce communication overhead, we're using very short acronyms 
  // for names and types. In addition, we're inlining functions for 
  // pointer comparison and arithmetic which "logically" are separate 
  // functions; this is done to avoid lambda-expressions.
  //
  // IAA is "INT_ARRAY_AT"
  // IA is "INT_ARRAY_REF"
  // I is "INT_REF"
  // Similarly for ADDR and CHAR.
  //
  // Hence,
  //   IAR((3,0)) = INT_ARRAY_REF((3,0))
  //   It returns (*3)[0], interpreted as an integer.
  //
  // If we have,
  //   int *x; // Address (1,0)
  //   int y;  // Address (2,0)
  // These would be written as,
  //   A((1,0))  --> Read the contents of (1,0), interpret as an address
  //   I((2,0))  --> Read the contents of (2,0), interpret as an int
  // *x would then be written as,
  //   I(A((1,0)))
  ////////////////////////////////////////////////////////////////////////////

  private static class RefFunctionMap
  {
    private Expr intT;
    private Expr charT;
    private Expr addrT;
    private Expr intArrayT;
    private Expr charArrayT;
    private Expr addrArrayT;

    // range type -> referencing function declaration
    Map<Expr, Expr> refFunctionMap;

    ExprFactory fac;

    public RefFunctionMap (ExprFactory fac)
    {
      refFunctionMap = new IdentityHashMap<Expr, Expr> ();
      this.fac = fac;
      declareRefFunctions ();
    }

    public Set<Expr> keySet()
    {
      return refFunctionMap.keySet();
    }

    /**
     * Creates a single declaration (name : domain -> range) and adds
     * it to the reference functions map.
     */
    private void declareRefFunction (String name, Expr type)
    {
      Expr funcName = fac.var (name);
      Expr funcType = CompoundTypeOp.funT (addrT, type);
      refFunctionMap.put (type, BindOp.bind (funcName, funcType));
    }

    /** 
     * Creates declarations (name : domain -> range) for referencing and adds
     * them to the map with key
     */
    private void declareRefFunctions ()
    {
      // primitive types which the Ops can't make themselves
      intT = fac.op (INT);
      charT = fac.op (CHAR);
      addrT = fac.op (ADDR);
      intArrayT = CompoundTypeOp.arrayT(intT, intT);
      charArrayT = CompoundTypeOp.arrayT(intT, charT);
      addrArrayT = CompoundTypeOp.arrayT(intT, addrT);

      // Mappings TYPE_REF : ADDR -> TYPE
      declareRefFunction("I", intT);
      declareRefFunction("C", charT);
      declareRefFunction("A", addrT);

      // Mappings TYPE_ARRAY_REF : ADDR -> ARRAY OF TYPE
      declareRefFunction("IA", intArrayT);
      declareRefFunction("CA", charArrayT);
      declareRefFunction("AA", addrArrayT);
    }

    /** 
     * Returns the referencing function with type {@code ADDRESS -> rangeType}.
     *
     * @param rangeType a type expression specifying the range of the
     * referencing function
     * @return 
     */
    public Expr get (Expr rangeType) throws RefFunctionMapException
    {
      // XXX treat pointer values as plain address
      if (isPtrT (rangeType))
        rangeType = rangeType.getFactory ().op (ADDR);

      if (isArrayT (rangeType)) {
        rangeType = CompoundTypeOp.arrayT (intT, 
            CompoundTypeOp.rng (rangeType));
      }

      Expr refFunc = refFunctionMap.get (rangeType);

      if (refFunc == null)
        throw new RefFunctionMapException
          ("no ref function for range type " + rangeType);

      return BindOp.name (refFunc);
    }

    public List<Expr> allRefFunctions ()
    {
      return null;
    }

    public class RefFunctionMapException extends Exception
    {
      public RefFunctionMapException (String msg)
      {
        super (msg);
      }
    }
  }

  /** 
   * Constructs a converter with the specified expression factory and symbol
   * table which associates each variable in the table with a unique arbitrary
   * address and translates reference and dereference expressions according to
   * the type information in the symbol table (<code>*x</code> to <code>INT_REF
   * (ADDR_REF (&amp;x)</code>, where <code>x</code> is of type
   * <code>int*</code>, for example).
   *
   * FIXME clean up comment and merge with class comment
   * 
   * @param fac 
   * @param symbolTable 
   */
  public CToAddrModel (ExprFactory fac, SymbolTable symbolTable)
  {
    this.fac = fac;
    this.symbolTable = symbolTable;
    varAddressMap = new IdentityHashMap<Expr, Integer> ();
    refFunctions = new RefFunctionMap (fac);
    maxAddress = 0;
    addSymbolTableVarsToMap ();
    // System.out.println("varAddressMap: " + varAddressMap);
  }

  ////////////////////////////////////////////////////////////////////////////
  //
  // Address Model Initialization
  //
  // Names have been shortened, and functions have been inlined.
  // 
  // The type "a" is an address.
  //
  // Address comparison and arithmetic is defined as follows:
  //   Comparison:
  //    (x,y) > (a,b) iff (x = a) and (y > b)
  //    Similarly for <, >=, <=. We use the built-in = and <>.
  //   Arithmetic:
  //    (x,y) + (a,b) = (x, y+b)
  //    Similarly for -.
  //    This assumes that only pointers with the same base object will
  //    be added/subtracted.
  // Address comparison and arithmetic is inlined.
  //
  ////////////////////////////////////////////////////////////////////////////
  public void init (CVCLProcess p)
  {
    // Types.
    // Make ALL addresses pairs of the form <OBJECT, OFFSET> The
    // OBJECT part gets passed to a referencing function.  The OFFSET
    // part gets used for array indexing, if applicable.
    //
    // Since an address is a tuple, it is represented by the using 
    // the TupleOp operator.
    p.declare("a", "TYPE = [INT, INT]"); // ADDR
    
    // We used to declare CHAR to be an INT; the reference functions 
    // have simply been redefined so that CHAR_REF (abbreviated to "C") 
    // returns an INT.

    // Functions used by the reference functions. They map the OBJECT
    // part of an address to an array.
    p.declare("IAA", "INT -> ARRAY INT OF INT"); // INT_ARRAY_AT
    p.declare("CAA", "INT -> ARRAY INT OF INT"); // CHAR_ARRAY_AT
    p.declare("AAA", "INT -> ARRAY INT OF a");   // ADDR_ARRAY_AT

    // Reference functions. The symbol table tells us what reference 
    // function to use for each object address.
    // 
    // Primitive reference functions operate on addresses. They use 
    // the array reference functions.
    // 
    // I, C, A, stand for INT_REF, CHAR_REF, and ADDR_REF, respectively.
    p.declare("I", "a -> INT = LAMBDA (x: a) : IAA(x.0)[x.1]");
    p.declare("C", "a -> INT = LAMBDA (x: a) : CAA(x.0)[x.1]");
    p.declare("A", "a -> a = LAMBDA (x: a) : AAA(x.0)[x.1]");

    // Array reference functions. They map an ADDR to the array that
    // starts at the OBJECT part. The OFFSET part is used separately
    // for indexing.
    //
    // IA, CA, AA, stand for INT_ARRAY_REF, CHAR_ARRAY_REF, and ADDR_REF,
    // respectively.
    p.declare("IA", "a -> ARRAY INT OF INT " + "= LAMBDA (x: a) : IAA(x.0)");
    p.declare("CA", "a -> ARRAY INT OF INT " + "= LAMBDA (x: a) : CAA(x.0)");
    p.declare("AA", "a -> ARRAY INT OF a " + "= LAMBDA (x: a) : AAA(x.0)");

    // integer abs function
    // XXX cvcl includes a library of rational functions including abs, but i
    // can't figure out how to call them from the interactive interface
    // XXX if this is the only way to define abs in cvcl, is this the right
    // place to do it? this is not a C-specific function
    p.declare ("ABS",
      "INT->INT = LAMBDA(x:INT): (IF x < 0 THEN -x ELSE x ENDIF)");
  }

  private void addSymbolTableVarsToMap ()
  {
    for (Expr e : symbolTable.varSet ())
      addVarToMap (e);
  }

  private void addVarToMap (Expr e)
  {
    assert !varAddressMap.containsKey (e) : "trying to add var twice: " + e;
    // assert e.op () instanceof VariableOp : "not a variable: " + e;

    // map e to the next available address, set maxAddress to that address
    varAddressMap.put (e, ++maxAddress);
  }

  private Expr ptrAdd (Expr e1, Expr e2)
  {
    // e1 a pointer, e2 an int... do an assert

    return TupleOp.tuple (dot (e1, 0), plus (dot (e1, 1), e2));
  }

  private Expr ptrSub (Expr e1, Expr e2)
  {
    return TupleOp.tuple (dot (e1, 0), minus (dot (e1, 1), e2));
  }

  /**
   * This should ONLY be used by rewite(). Before returning, it sets
   * this to be the type of the expression it re-wrote, so that the
   * caller can infer how to interpret its operator. For example, if
   * "p" is a pointer, and we're rewriting "p + 1", rewrite(p) will
   * set returnType to "pointer", so that rewrite(p + 1) will know to
   * interpret "+" as pointer addition, instead of integer addition.
   */
  private Expr returnType;

  /** 
   * Recursively converts a CIL Expr to an address model Expr, replacing
   * referencing (x) and dereferencing (*x) expressions with REF function
   * applications.
   *
   * Note that this method is only called by {@code CTheoremProver} when 
   * adding assertions and making queries. 
   *
   * Some operators have different interpretations depending on the
   * type of their operands. For example, addition of an integer and a
   * pointer is interpreted differently than addition of two
   * integers. To deal with this problem, rewrite() does type
   * inference by placing the type of the most recently rewritten
   * subexpression in the {@code returnType} variable, above.
   *
   * {@code rewrite()} can handle an expression {@code e} involving:
   *   - integer and rational literals
   *   - variables whose types are in the symbol table
   *   - the "address of" operator, applied to variables, arrays, and 
   *     array subscripts
   *   - dereferences of pointer expressions
   *   - array selections and updates
   *   - (p+n), (n+p), (p-n), where p is a pointer and n is an integer
   *   - arbitrary integer arithmetic
   *   - equalities and inequalities of scalars and pointers
   *
   * Other expressions may work, but would be handled by the catch-all case 
   * at the end of the method. Your milage may vary.
   * 
   * XXX For an array A, rewrite (*A) will trigger an assertion violation; 
   * XXX without assertions turned on, it'll generate gibberish.
   *
   * XXX For an array A, rewrite (A) will yield a reference to the array 
   * XXX object corresponding to A. To get a pointer to the start of A, 
   * XXX we need to do &A. Probably fixable, but translation is easier 
   * XXX this way.
   *
   * XXX We don't deal with typecasting at all.
   *
   * @param e the CIL Expr to be converted
   * @return the address model representation of {@code e}
   */
  public Expr rewrite (Expr e)
  {
    try
    {
      stats.startAddrModel ();

      // in the following comments, let e be the variable "x" and addr be &x

      /************************************************************************
       *
       * Literal
       *
       * XXX Assumes all literals are ints or reals
       *
       ***********************************************************************/
      if (e.op () instanceof RationalOp) {
        if (RationalOp.isInt (e))
          returnType = fac.op (INT);
        else
          returnType = fac.op (REAL);
        return e;
      }

      /************************************************************************
       *
       * Variable
       *
       * convert "x" to value-at-addr
       *
       ***********************************************************************/
      if (isVar (e) || isCVarDecl (e)) {
        returnType = typeOf (e);
        return FunctionApplOp.apply (getRefFunction (typeOf (e)), addressOf (e));
      }

      /************************************************************************
       *
       * Address-of
       *
       * convert "&x" to addr
       * XXX in general, have to handle &e for any lval e
       *
       ***********************************************************************/
      if (e.op () == CILIndirectionOp.ADDR_OF)
      {
        assert (e.arg (0).op () instanceof VariableOp
            || ArrayOp.isArraySel (e.arg (0))) : 
          "Cannot take address of non-lvalue " + e;

        if (ArrayOp.isArraySel (e.arg (0))) {
          Expr ind = rewrite (ArrayOp.getIndex (e.arg (0)));
          returnType = ptrT (rng (typeOf (ArrayOp.getArray (e.arg (0)))));
          return ptrAdd (addressOf (e.arg (0).arg (0)), ind);
        }
        returnType = CompoundTypeOp.ptrT(typeOf (e.arg (0)));
        return addressOf (e.arg (0));
      }

      /************************************************************************
       *
       * Dereference
       *
       * convert "*x" to value-at-(value-at-addr)
       *
       ***********************************************************************/
      if (e.op () == CILIndirectionOp.DEREF) {
        Expr df = rewrite (e.arg (0));
        assert (isPtrT (returnType)) : 
          returnType + " is not a pointer" +
            "\n  e : " + e;

        returnType = returnType.arg (0);
        Expr ref = getRefFunction (returnType);
        return FunctionApplOp.apply (ref, df);
      }

      /************************************************************************
       *
       * Array Selection
       *
       * convert A[x] to INT_ARRAY_REF(OBJ(ADDR_REF(ADDR)))[rewrite(x)]
       *
       ***********************************************************************/
      if (isArraySel (e))
      {
        /* Interpreting the name of an array as referring to the array itself,
         * not a pointer to the array. Change this if we add a level of 
         * indirection to bring us closer to C.
         *
         * Actually, for stack-allocated arrays, A = &A = &(A[0]), so we're
         * not so far from C. :-)
         */
        Expr ar = rewrite (getArray (e));
        assert (CompoundTypeOp.isArrayT (returnType)) : 
          "Expected array, got " + returnType + 
            " from rewriting " + ar + " in expression " + e;

        return ArrayOp.sel (ar, rewrite (getIndex (e)));
      }

      /************************************************************************
       *
       * Arithmetic
       *
       * Check if it's pointer arithmetic.
       *
       ***********************************************************************/
      if (e.op () == NumericOp.PLUS && e.arity () == 2) {
        Expr e1, e2;
        e1 = rewrite (e.arg (0));
        if (isPtrT (returnType)) {
          // New return type is pointer
          Expr rtold = returnType;
          e2 = rewrite (e.arg (1));
          returnType = rtold;
          return ptrAdd (e1, e2);
        }
        e2 = rewrite (e.arg (1));
        if (isPtrT (returnType)) {
          // New return type is pointer; do nothing to returnType
          return ptrAdd (e2, e1);
        }

        return fac.op (e.op ()).binApply (e1, e2);
      } else if (e.op () == NumericOp.MINUS && e.arity () == 2) {
        Expr e1, e2;
        e1 = rewrite (e.arg (0));
        if (isPtrT (returnType)) {
          // New return type is pointer
          Expr rtold = returnType;
          e2 = rewrite (e.arg (1));
          returnType = rtold;
          return ptrSub (e1, e2);
        }

        e2 = rewrite (e.arg (1));
        return fac.op (e.op ()).binApply (e1, e2);
      }

      /************************************************************************
       *
       * Comparison
       *
       * Check if it's pointer comparison
       *
       ***********************************************************************/
      if (e.op () instanceof ComparisonOp) {
        // All comparison ops have arity 2
        Expr e1 = rewrite (e.arg (0));
        Expr r1 = returnType;
        Expr e2 = rewrite (e.arg (1));
        Expr r2 = returnType;

        // Special case: check if we're comparing a pointer to an
        // array. If so, turn the array into a pointer to its first
        // element. Should deal with CIL's insistence on turning &(A[0])
        // into just "A".
        if (isPtrT (r1) && isArrayT (returnType)) {
          e2 = rewrite (CILIndirectionOp.addrOf (e.arg (1)));
          r2 = returnType;
        }

        returnType = fac.op (INT); // XXX stand-in for BOOL

        if (isPtrT (r1)) {
          // Special case: check if we're comparing a pointer to 0 (NULL)
          // Assume that no programmer will ever "if (NULL == p)"
          if ((e.op () == ComparisonOp.EQ || e.op () == ComparisonOp.NEQ)
              && e2 == fac.intExpr (0))
            return fac.op (e.op ()).binApply 
              (e1, TupleOp.tuple (fac.intExpr (0) , fac.intExpr (0)));

          assert (isPtrT (r2)) : 
            "Don't know how to compare " + e1 + " to " + e2 + 
              "\n LHS type: " + r1 + "\n RHS type: " + r2 +
              "\n e       : " + e;

          if (e.op () == ComparisonOp.EQ || e.op () == ComparisonOp.NEQ)
            // Nothing special, "=", "/=" works for tuples in CVCL
            return fac.op (e.op ()).binApply (e1, e2);

          // Use the appropriate comparison function; as explained above, 
          // they're all inlined.
          if (e.op () == ComparisonOp.LT)
            return and (eq (dot (e1, 0), dot (e2, 0)), 
                        lt (dot (e1, 1), dot (e2, 1)));
          else if (e.op () == ComparisonOp.GT)
            return and (eq (dot (e1, 0), dot (e2, 0)), 
                        gt (dot (e1, 1), dot (e2, 1)));
          else if (e.op () == ComparisonOp.LEQ)
            return and (eq (dot (e1, 0), dot (e2, 0)), 
                        leq (dot (e1, 1), dot (e2, 1)));
          else if (e.op () == ComparisonOp.GEQ)
            return and (eq (dot (e1, 0), dot (e2, 0)), 
                        geq (dot (e1, 1), dot (e2, 1)));
          else 
            throw new RuntimeException("Unanticipated operation " + e.op ());
        }

        return fac.op (e.op ()).binApply (e1, e2);
      }

      /************************************************************************
       *
       * Everything else.
       *
       * recursively handle n-ary expressions, including array updates
       *
       ***********************************************************************/  
      List<Expr> convertedArgs = new LinkedList<Expr> ();
      for (Object arg : e.args ())
        convertedArgs.add (rewrite ((Expr) arg));

      // Check if it's an array update; this can change return type
      if (isArrayUpd (e)) returnType = typeOf (e.arg (0));

      // Predicate expressions change return type to boolean.
      // XXX Call booleans ints for now. Should work.
      if (e.op () instanceof ComparisonOp || e.op () instanceof BoolOp)
        returnType = fac.op (INT);

      // Assume return type otherwise doesn't change here.
      return fac.op (e.op ()).naryApply (convertedArgs);
    }
    finally
    {
      stats.stopAddrModel ();
    }
  }

  public void setStats (YasmStatistics stats)
  {
    this.stats = stats;
  }

  /** 
   * Returns the type of the specified dereferencing expression, that is, the
   * type of the value obtained by applying all the dereferencing operators in
   * the expression.
   *
   * For example, where the declared type of a variable {@code x} is
   *
   * <code>
   *   (PTR (PTR (INT))) (which represents int** x)
   * </code>
   *
   * the dereferencing expressions over x have the following types:
   *
   * <code>
   *   x    : ADDRESS
   *   *x   : ADDRESS
   *   **x  : INT
   *   *&x  : ADDRESS
   *   **&x : ADDRESS
   * </code>
   *
   * @param d
   * @return 
   */
  private Expr derefType (Expr d)
  {
    // depth to descend in the type chain
    int depth = 0;

    // descend down the dereferencing chain to get to the variable at the end
    // also move a pointer into the corresponding type chain down and up
    // according to the * (down) and & (up) operators encountered
    while (CILIndirectionOp.isDeref (d) || CILIndirectionOp.isAddrOf (d))
    {
      if (CILIndirectionOp.isDeref (d))
        depth++;

      if (CILIndirectionOp.isAddrOf (d))
        depth--;

      d = d.arg (0);
    }

    assert VariableOp.isVar (d) : "not a variable " + d;

    return typeAt (typeOf (d), depth);
  }

  /** 
   * Returns the subexpression (representing a type) at the specified depth
   * along the leftmost branch in the specified type expression.
   *
   * Intended to be used to traverse address chains; for example, letting t be
   * the type of x declared as int** x, then
   *
   *   t = ADDRESS (ADDRESS (ADDRESS (INT)))
   *
   * and
   *
   *   typeAt (t, 0) = ADDRESS (ADDRESS (ADDRESS (INT)))
   *   typeAt (t, 1) = ADDRESS (ADDRESS (INT))
   *   typeAt (t, 2) = ADDRESS (INT)
   * 
   * @param type a type expression
   * @param depth a non-negative integer specifying the depth to traverse to
   * @return 
   */
  private Expr typeAt (Expr type, int depth)
  {
    while (depth > 0)
    {
      type = type.arg (0);
      depth--;
    }

    return type;
  }

  /** 
   * Returns the name of the referencing function: {@code ADDRESS ->
   * RANGETYPE}.
   *
   * @param rangeType a type expression specifying the range of the referencing
   * function
   * @return 
   */
  private Expr getRefFunction (Expr rangeType)
  {
    try
    {
      return refFunctions.get (rangeType);
    }
    catch (Exception ex)
    {
      System.out.println ("Range type failure on: " + rangeType);
      System.out.println ("Map is: ");
      for (Expr e : refFunctions.keySet())
        System.out.println (e + " -> " + getRefFunction(e));
      throw new RuntimeException (ex);
    }
  }

  /** 
   * Returns the type of the specified expression.
   * 
   * It only gets the type of a variable from the symbol table so it doesn't do
   * any inference like: (INT + INT) : INT.
   *
   * @param e
   * @return
   */
  private Expr typeOf (Expr e)
  {
    // Strip away array updates
    while (isArrayUpd (e)) e = e.arg (0);

    try
    {
      return symbolTable.typeOf (e);
    }
    catch (Exception ex)
    {
      throw new RuntimeException (ex);
    }
  }

  /** 
   * Returns the address of variable <code>e</code> as an address 
   * expression <OBJECT, OFFSET>.
   * 
   * @param e a variable Expr
   * @return the address of <code>e</code>
   */
  private Expr addressOf (Expr e)
  {
    return TupleOp.tuple (fac.intExpr (varAddressMap.get (e).intValue ()), 
                          fac.intExpr (0));
  }

  /**
   * Test case which accesses private methos, and therefore can't
   * be put into JUnit test case.
   */
  private static void testTypeAt ()
    throws Exception
    {
      ExprFactory fac = new ExprFactoryImpl ();
      Expr intT = fac.op (INT);
      Expr charT = fac.op (CHAR);

      Expr x = fac.var ("x"); // int* x; (ADDRESS (INT))
      Expr xT = ptrT (ptrT (intT));
      SymbolTable st = new SymbolTable ();
      st.put (x, xT);
      CToAddrModel converter = new CToAddrModel (fac, st);

      System.out.println (converter.typeAt (converter.typeOf (x), 0));
      System.out.println (converter.typeAt (converter.typeOf (x), 1));
      System.out.println (converter.typeAt (converter.typeOf (x), 2));
    }

  public static void main (String[] args) throws Exception
  {
    System.out.println("\ntestTypeAt=======================================");
    testTypeAt ();
  }
}
