package edu.toronto.cs.tp;

import junit.framework.*;

import edu.toronto.cs.expr.*;
import static edu.toronto.cs.expr.BoolOp.*;
import static edu.toronto.cs.expr.SimpleTypeOp.*;
import static edu.toronto.cs.expr.CompoundTypeOp.*;
import static edu.toronto.cs.expr.ComparisonOp.*;
import static edu.toronto.cs.expr.CILIndirectionOp.*;

/** 
 * A collection of C to address model rewriter tests.
 * <p>
 * JUnit checks all methods prefixed with "test".
 *
 * Run this with:
 *  $ java -ea junit.textui.TestRunner edu.toronto.cs.tp.CToAddrModelTestCase
 *
 * XXX There really *should* be some assertTrue()'s in here, and there used to 
 * XXX be, but I removed them. New builds tend to want to initialize the 
 * XXX varAddressMap with different variables getting different addresses, 
 * XXX which makes things a big pain. Also, the address model translation 
 * XXX is still somewhat prone to change.
 *
 * @author <a href="mailto:tomhart@cs.toronto.edu">Tom Hart</a>
 * @version 
 */
public class CToAddrModelTestCase extends TestCase
{
  public static void testSymbolTableAndAddressMapping ()
    throws Exception
  {
    System.out.println("\n\n\ntestSymbolTableAndAddressMapping ():\n");
    
    ExprFactory fac = new ExprFactoryImpl ();
    Expr intT = fac.op (INT);
    Expr charT = fac.op (CHAR);
    
    Expr x = fac.var ("x"); // int x
    Expr addr_x = fac.op (CILIndirectionOp.ADDR_OF).unaryApply (x);
    Expr y = fac.var ("y"); // int y
    Expr addr_y = fac.op (CILIndirectionOp.ADDR_OF).unaryApply (y);
    Expr ip = fac.var ("ip"); // int* ip
    Expr addr_ip = fac.op (CILIndirectionOp.ADDR_OF).unaryApply (ip);
    SymbolTable st = new SymbolTable ();
    st.put (x, intT);
    st.put (y, intT);
    st.put (ip, ptrT (intT));
    CToAddrModel converter = new CToAddrModel (fac, st);
    
    System.out.println ("x: " + st.typeOf (x) + ", &x: "
			+ converter.rewrite (addr_x));
    System.out.println ("y: " + st.typeOf (y) + ", &y: "
			+ converter.rewrite (addr_y));
    System.out.println ("ip: " + st.typeOf (ip) + ", &ip: "
			+ converter.rewrite (addr_ip));
  }
  
  public static void testRef ()
    throws Exception
  {
    System.out.println("\n\n\ntestRef ():\n");
    
    ExprFactory fac = new ExprFactoryImpl ();
    Expr intT = fac.op (INT);
    Expr charT = fac.op (CHAR);
    
    Expr x = fac.var ("x"); // int x
    Expr addr_x = fac.op (CILIndirectionOp.ADDR_OF).unaryApply (x);
    Expr y = fac.var ("y"); // int y
    Expr addr_y = fac.op (CILIndirectionOp.ADDR_OF).unaryApply (y);
    Expr x_plus_y = fac.op (NumericOp.PLUS).binApply (x, y);
    Expr neg_x = fac.op (NumericOp.UN_MINUS).unaryApply (x);
    SymbolTable st = new SymbolTable ();
    st.put (x, intT);
    st.put (y, intT);
    CToAddrModel converter = new CToAddrModel (fac, st);
    CVCLExprPrinter p = new CVCLExprPrinter();

    System.out.println ("x has addr " + converter.rewrite (addr_x));
    System.out.println ("y has addr " + converter.rewrite (addr_y));
    System.out.println ("'x' is " + converter.rewrite (x));
    System.out.println ("'y' is " + converter.rewrite (y));
    System.out.println ("'x + y' is " + converter.rewrite (x_plus_y));
    //assertTrue(p.print(converter.rewrite (x_plus_y)).equals
    //           ("(INT_REF((2,0)) + INT_REF((1,0)))"));
    System.out.println ("'-x' is " + converter.rewrite (neg_x));
  }

  public static void testDeref ()
    throws Exception
  {
    System.out.println("\n\n\ntestDeref ():\n");

    ExprFactory fac = new ExprFactoryImpl ();

    Expr intT = fac.op (INT);
    Expr charT = fac.op (CHAR);

    Expr ip = fac.var ("ip"); // int* ip; (ADDRESS (INT))

    Expr ipT = ptrT (intT);

    Expr ipp = fac.var ("ipp"); // int** ipp; (ADDRESS (ADDRESS (INT)))
    Expr ippT = ptrT (ptrT (intT));

    Expr x = fac.var ("x");

    SymbolTable st = new SymbolTable ();
    st.put (ip, ipT);
    st.put (ipp, ippT);
    st.put (x, intT);

    Expr deref_ip = CILIndirectionOp.deref (ip);
    Expr deref_ipp = CILIndirectionOp.deref (ipp);
    Expr deref_deref_ipp =
      CILIndirectionOp.deref (CILIndirectionOp.deref (ipp));

    Expr addrx = CILIndirectionOp.addrOf (x);
    Expr deref_addrx = CILIndirectionOp.deref (addrx);

    CToAddrModel converter = new CToAddrModel (fac, st);
    CVCLExprPrinter p = new CVCLExprPrinter();

    System.out.println ("ip is " + converter.rewrite (ip));
    System.out.println ("*ip is " + converter.rewrite (deref_ip));
    System.out.println ("*ipp is " + converter.rewrite (deref_ipp));
    System.out.println ("**ipp is " + converter.rewrite (deref_deref_ipp));
    System.out.println ();
    System.out.println ("x is " + converter.rewrite (x));
    System.out.println ("&x is " + converter.rewrite (addrx));
    System.out.println ("*(&x) is " + converter.rewrite (deref_addrx));
  }

  public static void testRefFunctions ()
  {
    System.out.println("\n\n\ntestRefFunctions ():\n");

    ExprFactory fac = new ExprFactoryImpl ();
    SymbolTable st = new SymbolTable ();
    CToAddrModel converter = new 
      CToAddrModel (fac, st);
    System.out.println (converter.refFunctions);
  }

  public static void testArrays ()
    throws Exception
  {
    System.out.println("\n\n\ntestArrays ():\n");

    ExprFactory fac = new ExprFactoryImpl ();
    SymbolTable st = new SymbolTable ();

    Expr intT = fac.op (INT);
    Expr charT = fac.op (CHAR);
    Expr a = fac.var("A");
    Expr c = fac.var("C");
    Expr x = fac.var("x");
    // Type int
    st.put(x, intT);

    // Type int []
    Expr ar_int_T = CompoundTypeOp.arrayT (intT, intT);
    st.put(a, ar_int_T);

    // Type char []
    Expr ar_char_T = CompoundTypeOp.arrayT (intT, charT);
    st.put(c, ar_char_T);
	 	 
    CToAddrModel converter = new CToAddrModel (fac, st);
    CVCLExprPrinter p = new CVCLExprPrinter();

    System.out.println("A is " + p.print (converter.rewrite (a)));

    Expr as = ArrayOp.sel(a, x);
    System.out.println("A[x] is " + p.print (converter.rewrite (as)));

    Expr as2 = ArrayOp.sel(a, as);
    System.out.println("A[A[x]] is " + 
                       p.print (converter.rewrite (as2)));

    Expr aupd = ArrayOp.upd(a, x, x);
    System.out.println("A WITH [x] := x is " + 
                       p.print (converter.rewrite (aupd)));

    String s = p.print (converter.rewrite(ArrayOp.sel(aupd, x)));
    System.out.println("(A WITH [x] := x)[x] is " + s);

    System.out.println("C is " + p.print (converter.rewrite (c)));
    System.out.println("C[x] is " + 
                       p.print(converter.rewrite (ArrayOp.sel (c, x))));

  }

  public static void testPtrArith ()
    throws Exception
  {
    System.out.println("\n\n\ntestPtrArith ():\n");

    ExprFactory fac = new ExprFactoryImpl ();
    SymbolTable st = new SymbolTable ();

    Expr intT = fac.op (INT);
    Expr charT = fac.op (CHAR);
    Expr a = fac.var("A");
    Expr c = fac.var("C");
    Expr x = fac.var("x");
    st.put (x, intT);
	 
    Expr seqP = fac.var("p");
    Expr seqQ = fac.var("q");
    st.put (seqP, CompoundTypeOp.ptrT(intT));
    st.put (seqQ, CompoundTypeOp.ptrT(intT));
	 
    CToAddrModel converter = new CToAddrModel (fac, st);
    CVCLExprPrinter p = new CVCLExprPrinter();
    System.out.println("p is " + p.print (converter.rewrite (seqP)));

    Expr pp1 = NumericOp.plus (seqP, fac.intExpr(1));
    Expr pm1 = NumericOp.minus (seqP, fac.intExpr(1));
    Expr _1pp = NumericOp.plus (fac.intExpr(1), seqP);
	 
    System.out.println("p+1 is " + 
		       p.print(converter.rewrite (pp1)));
    
    System.out.println("1+p is " + 
		       p.print(converter.rewrite (_1pp)));

    System.out.println("p-1 is " + 
		       p.print(converter.rewrite (pm1)));

    System.out.println("(p + 1) - 3 is " +
		       p.print(converter.rewrite 
			       (NumericOp.minus(pp1, fac.intExpr (3)))));

    System.out.println("p + 1 = q is " +
		       p.print 
		       (converter.rewrite 
			(ComparisonOp.eq (pp1, seqQ))));

    System.out.println("*p is " + 
		       p.print (converter.rewrite 
				(CILIndirectionOp.deref (seqP))));

    String s = p.print (converter.rewrite ((CILIndirectionOp.deref 
                                            (NumericOp.plus (seqP, 1)))));
    System.out.println("*(p+1) is " + s);
  }

  public static void testPtrComp ()
    throws Exception
  {
    System.out.println("\n\n\ntestPtrComp ():\n");

    ExprFactory fac = new ExprFactoryImpl ();
    SymbolTable st = new SymbolTable ();

    Expr intT = fac.op (INT);
    Expr charT = fac.op (CHAR);
    Expr C = fac.var("C");
    Expr x = fac.var("x");
    // Type int
    st.put(x, intT);

    // Type char []
    Expr ar_char_T = CompoundTypeOp.arrayT (intT, charT);
    st.put(C, ar_char_T);

    Expr seqP = fac.var("p");
    Expr seqPT = CompoundTypeOp.ptrT(charT);
    st.put(seqP, seqPT);

    Expr seqQ = fac.var("q");
    st.put (seqQ, CompoundTypeOp.ptrT(charT));
	 
    CToAddrModel converter = new CToAddrModel (fac, st);
    CVCLExprPrinter p = new CVCLExprPrinter();

    Expr spC = CILIndirectionOp.addrOf(ArrayOp.sel(C,0));
    System.out.println("C is " + p.print (converter.rewrite (C)));
    System.out.println("&C is " + p.print 
                       (converter.rewrite (CILIndirectionOp.addrOf (C))));
    System.out.println("&(C[0]) is " + p.print (converter.rewrite(spC)));

    Expr p_eq_spC = ComparisonOp.eq (seqP, spC);
    System.out.println("p == &(C[0]) is " + 
                       p.print (converter.rewrite (p_eq_spC)));

    Expr plq = ComparisonOp.lt (seqP, seqQ);
    System.out.println("(p < q) is " + p.print (converter.rewrite (plq)));

    Expr pleq = ComparisonOp.leq (seqP, seqQ);
    System.out.println("(p <= q) is " + p.print (converter.rewrite (pleq)));

    Expr pgeqC = ComparisonOp.lt (seqP, spC);
    System.out.println("(p >= &(C[0])) is " + 
                       p.print (converter.rewrite (pgeqC)));

    Expr peqz = ComparisonOp.eq (seqP, 0);
    System.out.println("(p == 0) is " + p.print (converter.rewrite (peqz)));

    Expr peqC = ComparisonOp.eq (seqP, C);
    System.out.println("(p == C) is " + p.print (converter.rewrite (peqC)));
  }

}

