package edu.toronto.cs.yasm;

import java.util.*;
import java.io.*;

import java.util.prefs.*;

import edu.toronto.cs.util.*;
import edu.toronto.cs.util.Clapi.*;

import edu.toronto.cs.yasm.pcfa.*;
import edu.toronto.cs.yasm.pgenerator.*;
import edu.toronto.cs.ctl.*;
import edu.toronto.cs.ctl.antlr.*;
import edu.toronto.cs.ctl.antlr.CTLNodeParser.*;
import edu.toronto.cs.mvset.*;
import edu.toronto.cs.modelchecker.*;
import edu.toronto.cs.cfa.CFAMvSetFactory.*;
import edu.toronto.cs.boolpg.BoolProgramCompiler;
import edu.toronto.cs.proof2.*;
import edu.toronto.cs.proof2.CTLProver.*;
import edu.toronto.cs.algebra.*;
import edu.toronto.cs.yasm.pprogram.*;
import edu.toronto.cs.yasm.abstractor.*;
import edu.toronto.cs.expr.*;
import edu.toronto.cs.yasm.util.*;
import edu.toronto.cs.yasm.wp.WPComputer;
import edu.toronto.cs.yasm.reachability.*;
import edu.toronto.cs.yasm.nontermination.*;
import edu.toronto.cs.yasm.abspatterns.*;

import edu.toronto.cs.cparser.*;
import edu.toronto.cs.tp.*;


/**
 * YasrApp.java
 *
 * Yet Another Software Model Checker
 *
 * Created: Tue Jun 22 13:38:03 2004
 *
 * @author <a href="mailto:maxin@epoch.cs">Xin Ma</a>
 * @version 1.0
 */
public class YasrApp 
{
  public static boolean DEBUG = true;  

  /**
   * Standard Output
   *
   */
  public static PrintWriter out = 
    new PrintWriter (new OutputStreamWriter (System.out), true);

  /**
   * Standad Error
   *
   */
  public static PrintWriter err = 
    new PrintWriter (new OutputStreamWriter (System.err), true);

  /**
   * Preferences of the application
   *
   */
  public static final Preferences PREFS 
    = Preferences.userRoot ().node ("edu/toronto/cs/yasm/YasrApp");


  PProgram pProgram = null;

  PProgramReachabilityCompiler compiler = null;  

  Refiner refiner = null;

  File cFile;

  File initPredFile;

  String absPatternFilename;
  AbstractionPatternDB absPatternDB = null;
  AbsPatternBacktracker apBacktracker = null;
  
  /**
   * CTL Property being checked
   *
   */
  CTLNode origProp = null;
  CTLNode prop;

  /**
   * Model used for reachability analysis
   *
   */
  ReachabilityStructure model;
  

  /**
   * Our model-checker
   *
   */
  ReachabilityChecker yasr;

  NonTerminationChecker yasnt;
  
  /**
   * If true then a proof is produced,  false by default
   *
   */
  boolean doProof = true;

  /**
   * true if functions should be inlined
   *
   */
  boolean inline = false;
  
  /**
   * our algebra
   *
   */
  BelnapAlgebra algebra;


  String refinerType = "new"; // <-- override this
  String pGeneratorType = "bend:wpd:wps:wpa";
  int selectorType = -1;
  boolean stmtBlocking = false;


  // found is set to true iff, during the current CEGAR iteration, one of 
  // the following is true:
  //  (1) Yasr found at least MAX_PREDICATES_PER_REFINEMENT new predicates
  //  (2) the back-end refiner successfully tightened up the abstraction
  //  (3) the abstraction pattern backtracker made us backtrack
  //  (4) at the END of refinement, Yasr found at least one new predicate 
  //      during the search for predicates
  boolean found = false;
  int newPredFound = 0;
  static int MAX_PREDICATES_PER_REFINEMENT = 1;

  // to give different edges a chance to be refined
  List<Edge> refinedEdges;

  ExprFactory fac;
  
  boolean useHyperEdges;
  boolean unknownInit;

  
  StopWatch overall = new StopWatch ();


  // -- for EG
  List pathForEG;
  Map  statesOnPathForEG;
  

  // Count of the number of the iterations of the CEGAR loop
  int refinementIterations;

  // Used by canUseAbstractionPattern to make sure we use at most 
  // one abstraction pattern per iteration
  boolean usedAbstractionPatternThisIteration;

  // canUseAbstractionPattern will set this if all the predicates we 
  // found unroll loops which are already being pattern-abstracted
  boolean mustBacktrackOne = false;

  boolean makeConsistent = false;


  boolean checkingPrecond = false;

  List<String> precondStubLabels;
  Map<CTLNode, List<Pred>> predicateHistory;

  private String whackOf (char c)
  {
    int whackAmount = 500;
    StringBuffer sb = new StringBuffer (whackAmount);
    for (int i = 0; i < whackAmount; i++)
      sb.append (c);
    return new String (sb);
  }
  
  public YasrApp () 
  {
    fac = new ExprFactoryImpl ();
    // setting up some defaults
    // redundant, but safe
    doProof = true;

    useHyperEdges = true;
    unknownInit = true;
    pGeneratorType = "bend:wpd:wps:wpa";
    refinerType = "cbj";
    refinementIterations = 0;

    precondStubLabels = new LinkedList <String> ();
    predicateHistory = new HashMap <CTLNode, List<Pred>> ();
  }

  public void setCFile (File f)
  {
    cFile = f;
  }
  

  public void setInitPredFile (File f)
  {
    initPredFile = f;
  }

  public void setAbsPatternFilename (String s)
  {
    absPatternFilename = s;
  }

  public boolean usingAbstractionPatterns ()
  {
    return absPatternDB != null;
  }

  public void setProp (CTLNode v)
  {
    YasrApp.out.println (">>> prop is now " + v);
    prop = v;
  }
  
  public CTLNode getProp ()
  {
    return prop;
  }

  public void setRefinerType (String v)
  {
    refinerType = v;
  }

  public void setSelectorType (String s)
  {
    if (s.equals ("dd"))
      selectorType = PProgram.DDSelector;
    else if (s.equals ("bin"))
      selectorType = PProgram.BinSelector;
    else if (s.equals ("lin"))
      selectorType = PProgram.LinSelector;
    else
      throw new RuntimeException ("Unknown selector type: " + s);
  }

  public void setStmtBlocking (boolean v)
  {
    stmtBlocking = v;
  }

  public void setMaxPredsPerRefinement (int i)
  {
    MAX_PREDICATES_PER_REFINEMENT = i;
  }

  public void setPGeneratorType (String v)
  {
    pGeneratorType = v;
  }
  
  public void setHyper (boolean v)
  {
    useHyperEdges = v;
  }
  public void setUnknownInit (boolean v)
  {
    unknownInit = v;
  }

  


  /**
   * doProof setter
   *
   * @param v a <code>boolean</code> value
   */
  public void setDoProof (boolean v)
  {
    doProof = v;
  }

  public boolean isDoProof ()
  {
    return doProof;
  }
  

  public void setInline (boolean v)
  {
    inline = v;
  }
  
  public void setModel (ReachabilityStructure v)
  {
    model = v;
    algebra = model.getAlgebra ();
  }
  
 
  public void setReachabilityChecker (ReachabilityChecker v)
  {
    yasr = v;
  }


  


  static Refiner getRefiner (String name, ExprFactory fac, PProgram p)
  {
    if (name.equals ("local-cbj-vs"))
      return new LocalCBJPredicateRefinerVisitor (fac, p, true, true);
    else if (name.equals ("local-cbj-vs-i"))
      return new LocalCBJPredicateRefinerVisitor (fac, p, false, true);
    else if (name.equals ("local-cbj-vs-s"))
      return new LocalCBJPredicateRefinerVisitor (fac, p, true, false);
    else if (name.equals ("local-cbj-vs-i-s"))
      return new LocalCBJPredicateRefinerVisitor (fac, p, false, false);

    
    throw new RuntimeException ("Unknown refiner: " + name);
  }

  /** 
   * call with resetting set to true if you want to forget all 
   * our predicates, and go back to a null abstraction
   *
   */
  ReachabilityStructure compileCProgram (boolean resetting)
    throws antlr.ANTLRException, FileNotFoundException
  {

    GnuCParser gnuCParser = null;

    if (pProgram == null || resetting)
      {
	//stats.startParse ();
	gnuCParser = GnuCParserUtil.getParser
	  (new FileReader (cFile));

	Expr programTree =
	  CILTransformerUtil.getProgram (fac, gnuCParser, stmtBlocking);
	//stats.stopParse ();
	  
	
	pProgram = (new NullExprAbstractor (fac)).
	  doProgramAbstract (programTree, selectorType);	  

        PProgramPatcher.patch (pProgram);

	// initialize symbol table
	SymbolTable st =
	  SymbolTableInitializer.init (pProgram, new SymbolTable (), true);

	pProgram.setSymbolTable (st);
      }

    if (!resetting)
      for (String label : pProgram.getLabelledStatementMap ().keySet ())
	if (label.contains ("PRECONDITION"))
	  precondStubLabels.add (label);

    if (refiner == null || resetting)
      {
        refiner = getRefiner (refinerType, fac, pProgram);

	// XXX Old way of communicating the symbol table to 
	// XXX the theorem prover
	refiner.doDeclsRefine (pProgram);


        if (gnuCParser != null && initPredFile != null && !resetting) {
          // parse initial predicates using the declarations of
          // the target C program
          
          gnuCParser =
            GnuCParserUtil.getParser
            (new FileReader (initPredFile), gnuCParser.symbolTable);
          
          Expr predTree =
            CILTransformerUtil.getProgram (fac, gnuCParser);
          
          List <Expr> predNames = PredicateExtractor.extract (predTree);
          for (Expr e : predNames)
            refiner.addToNewPredicates 
              (new Pred (e, PredicateSource.HINT));
          
        }

        // Initialize the abstraction pattern database, if we're 
        // using one
        if (absPatternFilename != null && !resetting) {
          absPatternDB = new AbstractionPatternDB 
            (new AbstractionPatternFileParser (absPatternFilename), fac);
          apBacktracker = new AbsPatternBacktracker ();
        }

      }

    //stats.startRefine ();
    pProgram = refiner.doProgramRefine (pProgram);
    //    stats.stopRefine ();


    System.err.println ();
    System.err.println ("Refined Boolean Program");
    System.err.println ();
    //       pProgram.print (YasrApp.err);
    System.err.println ();
    System.err.println ("Refined Boolean Program END");
    System.err.println ();

    //Ou Wei - test 
    //System.out.println ("Test Printing Begin ....");

    //       PrintWriter myout = new PrintWriter(System.out);
    //       pProgram.print(myout);
    //       myout.flush ();

    pProgram.accept 
      (new PrintingVisitor (new IndentPrintWriter (System.out)), null);
    //System.out.println ("Test Printing End ....");


    if (compiler == null && !resetting)
      {
	if (inline)
	  compiler = new InliningReachabilityCompiler ();
	else
	  compiler = new SummarizingReachabilityCompiler ();
      }
      

    compiler.setPProgram (pProgram);
    // XXX just for now
    //compiler.setMaxDDVars (8);
    compiler.setMaxDDVars (100);
    //compiler.setMaxDDVars (20);
      

    setModel (compiler.compile ());

    return model;
  }

  ReachabilityStructure compileCProgram ()
    throws antlr.ANTLRException, FileNotFoundException
  {
    return compileCProgram (false);
  }

  ReachabilityStructure resetAbstraction ()
  {
    try {
      return compileCProgram (true);
    } catch (antlr.ANTLRException ex) {
      YasrApp.out.println ("Could not parse C program: " + ex);
      ex.printStackTrace (YasrApp.err);
    } catch (FileNotFoundException ex) {
      YasrApp.out.println ("Could not find C file: " + ex);
      ex.printStackTrace (YasrApp.err);
    }
    
    throw new RuntimeException ("shouln't get here");
  }

  void forceFullAbstractionRecalculation (Collection<Pred> pl)
  {
    resetAbstraction ();
    for (Pred p : pl) {
      if (p.getSource () instanceof PredicateSource.WPPredicateSource) {
        ((PredicateSource.WPPredicateSource) p.getSource ()).updateStmt 
          (pProgram);
      }
      refiner.addToNewPredicates (p);
    }
  }

  /**
   * Prepare our property to be checked
   *
   */
  CTLNode prepareProperty ()
  {
    if (! (prop instanceof CTLEFNode))
      {
	YasrApp.out.println ("Property is not supported");
	System.exit (1);
      }
    
    CTLNode right = prop.getRight ();
    
    if (! (right instanceof CTLEqualsNode))
      {
	YasrApp.out.println ("Only 'EF pc = XXX' is supported");
	System.exit (1);
      }
    
    String pcValue = right.getRight ().toString ();
    
    YasrApp.err.println ("Looking for reachability of " + pcValue);
    PStmt stmt = pProgram.getStmt (pcValue);
    if (stmt == null) 
      {
	YasrApp.err.println ("No such statement!");
	System.exit (1);
      }
    
    NodeSet nodes = model.getNodeSet ().createNext ();
    nodes.add (new Node (new Node.NodeId (stmt.getId ()),
			 model.getMvSetFactory ().top ()));
    model.setNodeSet (nodes);
    return null;
  }



  /**
   * Prepare our property to be checked
   *
   */
  CTLNode prepareEGProperty ()
  {

    // -- set "END" node with "FALSE", other nodes willl be set "TRUE"
    // -- during the "initialize ()" of in NonTerminationChecker.java

    String pcValue = "END";
    
    YasrApp.err.println ("Check EG true");
    PStmt stmt = pProgram.getStmt (pcValue);
    if (stmt == null) 
      {
	YasrApp.err.println ("No such statement!");
	System.exit (1);
      }
    
    NodeSet nodes = model.getNodeSet ().createNext ();
    nodes.add (new Node (new Node.NodeId (stmt.getId ()),
			 model.getMvSetFactory ().bot ()));
    model.setNodeSet (nodes);
    return null;
  }



  AlgebraValue extractValue ()
  {
    int initId = pProgram.getStmt ("init").getId ();
    Node initNode = model.getNodeSet ().get (new Node.NodeId (initId));
    
    if (initNode == null) return model.getAlgebra ().bot ();

    
//     YasrApp.err.println ("init node is labeled by: ");
//     YasrApp.err.println (initNode);
    
//     YasrApp.err.println ("looking at another node");
//     YasrApp.err.println (model.getNodeSet ().get (new Node.NodeId(12)));

//     YasrApp.err.println ("looking at next to another node");
//     YasrApp.err.println (model.getNodeSet ().get (new Node.NodeId(13)));

//     YasrApp.err.println ("looking at the before error node");
//     YasrApp.err.println (model.getNodeSet ().get (new Node.NodeId(14)));

//     YasrApp.err.println ("looking at the before error node");
//     YasrApp.err.println (model.getNodeSet ().get (new Node.NodeId(15)));

//     YasrApp.err.println ("looking at the before error node");
//     YasrApp.err.println (model.getNodeSet ().get (new Node.NodeId(16)));
    

//     YasrApp.err.println ("looking at init node before extractValue");
//     YasrApp.err.println (initNode);
    
//     YasrApp.err.println ("fooEntry");
//     YasrApp.err.println (model.getNodeSet ().get (new Node.NodeId (3)));
    
//     YasrApp.err.println ("five is");
//     YasrApp.err.println (model.getNodeSet ().get (new Node.NodeId (5)));


//     YasrApp.err.println ("and 4 is");
//     YasrApp.err.println (model.getNodeSet ().get (new Node.NodeId (4)));
      


    MvSet initValue = initNode.getValue ();
    
    initValue = initValue.
      existAbstract (model.getPreCube ());
    
    
    

    //     YasrApp.err.println ("PreCube is: ");
    //     YasrApp.err.println (new Node (new Node.NodeId (0),
    // 				   model.getPreCube ()));
    //     YasrApp.err.println ("Extracted: ");
    //     YasrApp.err.println ("Dumping");
    //     YasrApp.err.println (new Node (new Node.NodeId (0), initValue));


    return initValue.getValue ();
  }
  


  // cribbed from BoundsCheckExtractor
  private void printAsAssertion (CTLNode _prop)
  {
    CTLNode right = _prop.getRight ();
    
    if (! (right instanceof CTLEqualsNode)) {
      throw new RuntimeException ("Only 'EF pc = XXX' is supported");
    }
    
    String pcValue = right.getRight ().toString ();
    
    PStmt stmt = pProgram.getStmt (pcValue);
    if (stmt != null) {
      PStmt errorStmt = stmt;
      for (PStmt pred : stmt.getPreds ()) {
        if (pred instanceof IfPStmt) {
          IfPStmt errorPred = (IfPStmt)pred;
          Expr negAssert = ((IfPStmt)errorPred).getCond ().getCond ();
          YasrApp.out.println ("\t assert" + BoolOp.labs (BoolOp.not (negAssert)));
          return;
        }
      }
    }
    YasrApp.out.println ("\t Can't represent property as an assertion, sorry. :-/");
  }

  
  public AlgebraValue run () throws CTLNodeParserException, 
				    antlr.ANTLRException,
				    FileNotFoundException
  {
  
  
    AlgebraValue value = null;
    NodeSet result = null;
  
    do
      {
	YasrApp.out.println ("Compiling C program: " + cFile);
	if (refiner != null)
	  {
	    YasrApp.out.println ("YasrApp: Old predicates: " + 
				 refiner.getPredicates ());
	    YasrApp.out.println ("YasrApp: New predicates: " + 
				 refiner.getNewPredicates ());
	  }
	
	compileCProgram ();
      

	// -- force garbage collection 
	System.gc ();

	// -- convert the property to checkable form
	if (prop instanceof CTLEGNode)
	  {
	    prepareEGProperty ();  

	    // -- construct the model-checker
	    yasnt = new NonTerminationChecker (model);
	  }
	else
	  {
	    prepareProperty ();	      

	    // -- construct the model-checker
	    yasr = new ReachabilityChecker (model);
	  }
    

	

	// -- do the actual model-checking
	YasrApp.out.println ();
	YasrApp.out.println ();
	YasrApp.out.println ("***** Model-Checking ***** ");

	StopWatch sw = new StopWatch ();



	// -- convert the property to checkable form
	if (prop instanceof CTLEGNode)
	  {
	    yasnt.doNonTermination ();
	  }
	else
	  {
	    yasr.doReachablity ();
	  }
    



	YasrApp.out.println ("Done in: " + sw);


	// -- finally, print what we think is the result in the 
	// -- initial state
	
	// -- Print  " \forall {s \in States} INIT(s) -> result(s) "
	// --
	value = extractValue ();	

	// -- old print-out of result 

	// 	YasrApp.out.println ();
	// 	YasrApp.out.println ("******* SEE RESULT BELOW ***********");
	// 	YasrApp.out.println ("result is: " + value);
	// 	YasrApp.out.println ("********** LOOK UP  ^ **********");
	
	// 	//YasrApp.err.println (model.getNodeSet ().dump ());
	
	// 	YasrApp.out.println ("Predicates used: ");
	// 	YasrApp.out.println (refiner.getPredicates ());

	// -- new print-out of result

	YasrApp.out.println ();
	YasrApp.out.println ("MODEL-CHECKING");
	YasrApp.out.println ("\tPROPERTY: " + prop);
        printAsAssertion (prop);
	YasrApp.out.println ("\tRESULT: " + value);

	YasrApp.out.println ("Predicates used: ");
		
	YasrApp.out.println (refiner.getPredicates ());
	YasrApp.out.println ();


	
	// -- stop if asked to run only once
	if (!isDoProof ())
	  break;
	  
	found = false;
        newPredFound = 0;
        mustBacktrackOne = false;
	if (value.equals (algebra.infoBot ()))
	  {
	    YasrApp.out.println ("Unknown result, starting refinement");
	    if (prop instanceof CTLEGNode)
	      doNTRefine ();
	    else
	      doRefine ();

	    if (!found && checkingPrecond) {
	      checkingPrecond = false;
	      backtrackAP ();
	    }
	  }

	if (value.equals (algebra.bot ())) {

	  predicateHistory.put (prop, new ArrayList (refiner.getPredicates ()));

	  if (origProp == null) {
	    // ERROR unreachable, now check preconds (if any) without abspatts 
	    origProp = prop;
	    checkingPrecond = true;
	  }
	  if (usingAbstractionPatterns ()
	      && apBacktracker.canBacktrack ()
	      && apBacktracker.currentPrecondsIterator ().hasNext ()) {

	    String precondStubLabel =
	      apBacktracker.currentPrecondsIterator ().next ();
	    setProp (CTLNodeParser.parse ("EF pc = " + precondStubLabel));
	    forceFullAbstractionRecalculation (
		apBacktracker.currentPreconds ().values ());
	    substPreconds (apBacktracker.currentPreconds ());
	    found = true; // force next iteration of main loop
	    continue;
	  } else {
	    // all preconds hold, done
	    value = algebra.bot ();
	    prop = origProp;
	    break; // XXX move on to next loop?
	  }
	} else if (value.equals (algebra.top ()) && checkingPrecond) {
	  checkingPrecond = false;
	  backtrackAP ();
	}

        // Don't give up unless we didn't manage to find any predicates
        // whatsoever.
        found |= newPredFound > 0;

        refinementIterations++;

        if (refinementIterations > 100) assert (false) : "infinite loop";
      } while (found);

    return value;
  
  } // end of run

  void backtrackAP ()
  {
      prop = origProp;
      List<Pred> backtrackPreds = apBacktracker.backtrackOne ();
      assert (backtrackPreds != null);
      forceFullAbstractionRecalculation (backtrackPreds);
      substPreconds (apBacktracker.currentPreconds ());
      found = true;
  }

  // -- non-termination refine
  void doNTRefine ()
  {
    ReachabilityProver prover = initNTProver (algebra, model.getEdgeSet (),
					      model.getPreCube ());
    int initId = pProgram.getStmt ("init").getId ();
    Node initNode = model.getNodeSet ().get (new Node.NodeId (initId));
    ProofGoal rootGoal = new EGPathGoal (null, 
					 model.getNodeSet (),
					 algebra.infoBot (),
					 initNode);
    
    pathForEG = new ArrayList ();
    statesOnPathForEG = new HashMap ();
    
    findCycle (prover, rootGoal);
    
  }
  
  boolean timedOutAndBacktracked ()
  {
    List<Pred> backtrackPreds = apBacktracker.tick ();
    if (backtrackPreds != null) {
      forceFullAbstractionRecalculation (backtrackPreds);
      return true;
    }
    return false;
  }

  List<Pred> getPreconditionPreds ()
  {
    List<Pred> allPreds = refiner.getPreds ();
    List<Pred> ret = new LinkedList<Pred> ();
    
    for (Pred p : allPreds)
      if (p.isPrecondition ()) ret.add (p);

    return ret;
  }

  void doRefine ()
  {
    // If we're using abstraction patterns, we may have hit a threshold,
    // in which case we have to backtrack to an earlier abstraction.
    if (usingAbstractionPatterns () && timedOutAndBacktracked ()) {
      //assert false : "I timed out and backtracked";
      found = true;
      return;
    }

    System.out.println ("BATMAN:");
    System.out.println ("model nodes == " + model.getNodeSet ().size());
    System.out.println ("model edges == " + model.getEdgeSet ().size());

    // We're allowed to use an abstraction pattern this iteration, 
    // if we're using them
    usedAbstractionPatternThisIteration = false;

    refinedEdges = new ArrayList<Edge> ();

    ReachabilityProver prover = initProver (algebra, 
					    model.getEdgeSet (),
					    model.getPreCube ());
    
    int initId = pProgram.getStmt ("init").getId ();
    Node initNode = model.getNodeSet ().get (new Node.NodeId (initId));

    ProofGoal rootGoal = new PathProofGoal (null, 
					    model.getNodeSet (),
					    algebra.infoBot (),
					    initNode);

    List<EdgeProofGoal> refinableEdges = new ArrayList<EdgeProofGoal> ();
    getRefinableEdges (prover, rootGoal, refinableEdges);
    System.out.println ("TOM: " + refinableEdges.size () + 
                        " edges available for refinement ");
    for (EdgeProofGoal epg : refinableEdges) {
      System.out.println (epg);
      PStmt src = pProgram.getStmt (epg.getSrc ().getId ().intValue ());
      PStmt dst = pProgram.getStmt (epg.getDst ().getId ().intValue ());
      System.out.println ("\tsrc  : " + src.getLabel ());
      System.out.println ("\tdest : " + dst.getLabel ());
    }

    /*
    List<EdgeProofGoal> usableEdges;
    boolean havePreconditions = false;
    for (Pred p : refiner.getPreds ()) {
      Expr e = p.getName ();
      if (ComparisonOp.isLeq (e)
          && e.arg (0).op () == NumericOp.MINUS) {
        havePreconditions = true;
        break;
      }
    }
    if (havePreconditions) {
      // we have preconditions; restrict the edges we consider to those 
      // which are before the loop that they're preconditions for
      usableEdges = new ArrayList<EdgeProofGoal> ();
      for (EdgeProofGoal epg : refinableEdges) {
        if (epg.getSrc ().getId ().intValue () <= 20)
          usableEdges.add (epg);
      }
      } else {
      usableEdges = refinableEdges;
      }

    for (EdgeProofGoal epg : usableEdges) {
      refineEdge (epg);
      if (found) break;
    }
    */
    findPath (prover, rootGoal);

    // XXX Why not just change the "mustBacktrackOne" to !found?
    // XXX Right now, we can terminate with a "maybe", which is wrong.
    // 
    // XXX And right now, mustBacktrackOne is never set to true!!!!
    if (usingAbstractionPatterns () 
        && mustBacktrackOne
        && apBacktracker.canBacktrack ()) {
      //assert false : "I had to backtrack";
      List<Pred> backtrackPreds = apBacktracker.backtrackOne ();
      assert (backtrackPreds != null);
      forceFullAbstractionRecalculation (backtrackPreds);
      substPreconds (apBacktracker.currentPreconds ());
      found = true;
    }

  }
  

  
  void findCycle (ReachabilityProver prover, ProofGoal goal)
  {
    // -- the bound can be removed, currently set it a very 
    // -- big number
    findCycle (prover, goal, 10000);
  }
  
  void findCycle (ReachabilityProver prover, ProofGoal goal, 
		  int bound)
  {

    // -- save information for EG path
    if(goal instanceof StateProofGoal)
      {	
	pathForEG.add (goal);


	// -- check if the same state has been visited before
	// -- if so, we have found a cycle
	// -- otherwise, add this states to the state set

	Node.NodeId id =  ( (StateProofGoal)goal).getState ().getId ();
	MvSet state = ( (StateProofGoal)goal).getState ().getValue ();

	if (statesOnPathForEG.containsKey (id))
	  {
	    Set stateSet = (HashSet) statesOnPathForEG.get (id);

	    if (!stateSet.add (state))
	      {
		System.err.println ("\n >>>>> A Cycle is detected !!!\n");
		
		printEGPath ();
		return;
	      }
	    else
	      {
		statesOnPathForEG.put (id,stateSet);
	      }
	  }
	else
	  {
	    Set stateSet = new HashSet ();
	    stateSet.add (state);
	    statesOnPathForEG.put (id,stateSet);
	  }


	// -- check if we have found a node whose EG value is true
	if (! ((StateProofGoal)goal).getValue ().equals (algebra.infoBot ())) 
	  {

	    System.err.println ("\n >>>>> Found a EG(true)=true node !!!\n");

	    printEGPath ();
	    return;
	  }
	else
	  {
	    if (bound == 0) 
	      {

		System.err.println 
		  ("\n >>>>> Reach the bound for detecting EG path !!!\n");

		printEGPath ();
		return;
	      }
	  }


      }

    YasrApp.out.println ("\n*************************************" +
			 "**********************");

    if (goal instanceof StateProofGoal)
      {
	YasrApp.out.println ("Looking at StateProofGaol " + goal);
	YasrApp.out.println ("\t" + 
			     pProgram.getStmt (((StateProofGoal)goal).
					       getState ().getId ().
					       intValue ()));
      }
    else
      {
	if (goal instanceof EdgeProofGoal)
	  YasrApp.out.println ("Looking at EdgeProofGoal " + goal);
	else
	  {
	    throw new RuntimeException ("Wrong type of proof goal" + goal);
	  }
      }
    
    
    
    prover.expand (goal);
    ProofGoal[] subGoals = goal.getSubGoals ();
    if (subGoals == null) 
      {
	// -- hit a point when no proof rules are applicable
	// -- see if we can refine
	if (goal instanceof EdgeProofGoal)
	  {
	    if (((EdgeProofGoal) goal).getValue ().equals (algebra.infoBot ()))
	      {
		YasrApp.out.println ("Edge to refine: " + goal);
		refineEdge ((EdgeProofGoal) goal);
		return;
	      }

	    // -- else : true edge; no refinement is needed 
	  }
      }
    else
      {
	for (int i = 0; i < subGoals.length && !found; i++)
	  findCycle (prover, subGoals [i], bound - 1);
      }


  }
  
  // -- ugly way to print EG path 
  void printEGPath ()
  {
    
    if (pathForEG.size () == 0)
      return;

    System.err.println ("\n EG path <<<<<<<<<<<<<<<<<<<");
    
    for (Iterator it = pathForEG.iterator (); it.hasNext ();)
      {
	StateProofGoal goal = (StateProofGoal) it.next ();
	
	System.err.println ("\nNode  : " + goal.getState ().getId () +
			    "    (EG true) Value : " + goal.getValue ());
	

	CTLNode[] nodeState = 
	  model.getStatePresenter ().toCTL (goal.getState ().getValue ());

	System.err.println ("Predicate Assignment  : " + 
			    Arrays.asList (ctlToExpr (nodeState)));

	ProofGoal[] subGoals = goal.getSubGoals ();
	if (subGoals != null) 
	  {
	    // -- print the edge information, assume the EdgeProofGoal
	    // -- is always the second subgoal
	    if (subGoals[1] instanceof EdgeProofGoal)
	      {
		System.err.println ("\nby edge : " + 
				    (EdgeProofGoal) subGoals[1]);
	      }
	    else
	      {		
		throw new RuntimeException ("Woring subgoal type!");
	      }
	  }    
      }

    System.err.println ("\n >>>>>>>>>>>>>>>>>>>");
                          
    pathForEG = new ArrayList ();
    statesOnPathForEG = new HashMap ();
    return;
  }
  
  
  String[]  ctlToExpr (CTLNode [] ctl)
  {
    String [] result = new String [ctl.length];
    
    for (int i=0; i < ctl.length; i++)
      result[i] = ctlToExpr (ctl[i]);
    
    return result;
  }

  String ctlToExpr (CTLNode ctl)
  {
    String ctlName = ctl.getLeft ().toString ();
    String value = ctl.getRight ().toString ();

    String result = 
      pProgram.getPredTable ().
      getByCtlName (ctlName).getName ().toString ();
    
    if (value.equals ("false"))
      result = "!" + result;
    if (value.equals ("maybe"))
      result = "?" + result + "?";
    return result;
  }
  
  boolean edgeAlreadyRefined (ProofGoal goal)
  {
    EdgeProofGoal epg = (EdgeProofGoal) goal;
    int srcID = epg.getEdge ().getSrc ().intValue ();
    int destID = epg.getEdge ().getDst ().intValue ();

    for (Edge e : refinedEdges) {
      if (e.getSrc ().intValue () == srcID &&
          e.getDst ().intValue () == destID)
        return true;
    }

    refinedEdges.add (epg.getEdge ());
    return false;
  }
  
  void getRefinableEdges (ReachabilityProver prover, 
                          ProofGoal goal,
                          List<EdgeProofGoal> acc)
  {
    if (goal instanceof StateProofGoal &&
	(! ((StateProofGoal)goal).getValue ().equals (algebra.infoBot ())))
      {
	return;
      }
    prover.expand (goal);
    ProofGoal[] subGoals = goal.getSubGoals ();
    if (subGoals == null) 
      {
	if (goal instanceof EdgeProofGoal &&
	    ((EdgeProofGoal) goal).getValue ().equals (algebra.infoBot ()))
	  {
            if (! acc.contains ((EdgeProofGoal)goal))
              acc.add ((EdgeProofGoal)goal);
	    return;
	  }
      }
    else
      {
	for (int i = 0; i < subGoals.length && !found; i++)
	  getRefinableEdges (prover, subGoals [i], acc);
      }
  }

  void findPath (ReachabilityProver prover, ProofGoal goal)
  {
    if (goal instanceof StateProofGoal &&
	(! ((StateProofGoal)goal).getValue ().equals (algebra.infoBot ())))
      {
	return;
      }
//     else if (goal instanceof EdgeProofGoal)
//       {
// 	YasrApp.out.println ("Edge to refine: " + goal);
// 	if (((EdgeProofGoal)goal).getValue ().equals (algebra.infoBot ()))
// 	  refineEdge ((EdgeProofGoal) goal);
// 	//found = true;
// 	return;
//       }
    
    //YasrApp.out.println ("Looking at " + goal);
    /*if (goal instanceof StateProofGoal)
      {
	YasrApp.out.println ("\t" + 
			     pProgram.getStmt (((StateProofGoal)goal).
					       getState ().getId ().
					       intValue ()));
                                               }*/
      
    prover.expand (goal);
    ProofGoal[] subGoals = goal.getSubGoals ();
    if (subGoals == null) 
      {
	// -- hit a point when no proof rules are applicable
	// -- see if we can refine
	if (goal instanceof EdgeProofGoal &&
	    ((EdgeProofGoal) goal).getValue ().equals (algebra.infoBot ()))
	  {
	    YasrApp.out.println ("Edge to refine: " + goal);
	    refineEdge ((EdgeProofGoal) goal);
	    return;
	  }
      }
    else
      {
	for (int i = 0; i < subGoals.length && !found; i++)
	  findPath (prover, subGoals [i]);
      }
  }
  
  boolean isStateConsistent (CTLNode[] state, PFunctionDef pfd)
  {
    PredicateTable pTable = pProgram.getPredTable ();
	  
    // XXX big ugly hack for now
    Map<String,Expr> predMap = new HashMap<String,Expr> ();
    for (Iterator it = pTable.getVariables ().iterator (); 
	 it.hasNext ();)
      {
	PredicateTable.Predicate pred = 
	  (PredicateTable.Predicate) it.next ();
	predMap.put (pred.getCtlName (), pred.getName ());
      }

    // -- here we need pProgram, refiner, and CVectorExpr of childState

    PredicateGenerator 	pg = 
      new InconsistentDestinationGenerator (pProgram,
					    refiner,
					    getStatePred (predMap, 
							  state),
                                            pfd);    
    boolean ret = pg.find ();
    System.out.println ("InconsistentDestinationGenerator returned " + ret);
    return ret;
  }


  void refineEdge (EdgeProofGoal edgeGoal)
  {
    // -- get a statement corresponding to this edge
    PStmt stmt = pProgram.getStmt (edgeGoal.getSrc ().getId ().intValue ());
    System.out.println ("TOM >>> refining " + stmt.getSourceExpr ());

    if (makeConsistent) {
      // XXX Ported from YasmApp.java
      // XXX
      // XXX This probably won't work right when we have multiple 
      // XXX functions, which is probably why it wasn't here 
      // XXX originally.
      // XXX
      // XXX What should this be for Yasr?
      
      found = isStateConsistent 
        (model.getStatePresenter ().toCTL 
         (edgeGoal.getDst ().getValue ()), stmt.getParent ());
      if (found) return;
    }

    if (stmt instanceof IfPStmt)
      refineIfStmt ((IfPStmt) stmt);
    else if (stmt instanceof PrllAsmtPStmt)
      refinePrllAsmt ((PrllAsmtPStmt) stmt, edgeGoal.getSrc (), 
		      edgeGoal.getDst ());
    else if (edgeGoal.getEdge () instanceof SummaryEdge)
      {
	YasrApp.out.println ("Refining summary edge: " + edgeGoal);
	
	SummaryEdge edge = (SummaryEdge) edgeGoal.getEdge ();
	//MvSet sumValue = edge.getRelation ().toMvSet ();
	MvSet sumValue = edgeGoal.getSrc ().getValue ().and 
	  (edgeGoal.getDst ().getValue ().
	   renameArgs (model.getPreToAuxMap ()));
	
	//sumValue = sumValue.renameArgs (model.getPostToAuxMap ());
	
	ProofGoal rootGoal = 
	  new PathProofGoal (edgeGoal, model.getSumNodeSet (),
			     edgeGoal.getValue (),
			     new Node (edgeGoal.getSrc ().getId (),
				       sumValue));
	ReachabilityProver prover = initProver (algebra, 
						model.getEdgeSet (),
						model.getPreCube ().and 
						(model.getAuxCube ()));
	
	findPath (prover, rootGoal);	
      }
    else if (stmt instanceof SkipPStmt 
             && stmt.getLabel ().equals ("init"))
      {
        // XXX As above, probably not right for multi-function programs, 
        // XXX but without this, Yasr says "maybe" for some examples 
        // XXX which should be "false."
        found = isStateConsistent 
          (model.getStatePresenter ().toCTL 
           (edgeGoal.getDst ().getValue ()), stmt.getParent ());	  
      }
    else
      {
	System.err.println ("Not IfPStmt/PrllAsmtPStmt/SummaryEdge:" +
			    "Do nothing");
	
      }
    
    
  }

  void refineIfStmt (IfPStmt ifStmt)
  {
    Expr realCond = ifStmt.getCond ().getOrigCond ();
    ifStmt.getCond ().setCond (realCond);
    
    while (realCond.op () == BoolOp.NOT)
      realCond = realCond.arg (0);

    if (BoolOp.isTrue (realCond) || BoolOp.isFalse (realCond)) {
      YasrApp.out.println ("Oops, constant expression " + realCond + 
                           " cannot be a predicate");
      return;
    }
    
    YasrApp.out.println ("Trying to add IF predicate");
    YasrApp.out.println (realCond);
    if (!refiner.getPredicates ().contains (realCond))
      {
        newPredFound++;
	found = newPredFound >= MAX_PREDICATES_PER_REFINEMENT;
        mustBacktrackOne = false;
	//refiner.addPredicate (realCond);
        refiner.addToNewPredicates 
          (new Pred (realCond, 
                     new PredicateSource.ConditionalPredicateSource (ifStmt)));
      }
    else
      YasrApp.out.println ("Ooops, have it already");
    
  }
  
  void refinePrllAsmt (PrllAsmtPStmt asmt, Node src, Node dst)
  {
    
    CTLNode[] srcState = model.getStatePresenter ().toCTL (src.getValue ());
    CTLNode[] dstState = model.getStatePresenter ().toCTL (dst.getValue ());
    
    
    System.out.println ("refining between states:");
    System.out.println ("src state: " + Arrays.asList (srcState));
    System.out.println ("dst state: " + Arrays.asList (dstState));
      
    PredicateGenerator pg = createPredicateGenerator (asmt, srcState, 
						      dstState);
    
    List<Pred> newPred = null;
    if (pg.find  ()) {
	newPred = pg.getNewPreds ();
        newPredFound += newPred.size () + 1;
        found = newPredFound >= MAX_PREDICATES_PER_REFINEMENT;
    }
    if (newPred != null 
        && !newPred.isEmpty () 
        && !canUseAbstractionPattern (newPred)) {

      System.out.println ("about to add predicates " 
                          + Pred.getNames (newPred));
      System.out.println ("newPredFound == " + newPredFound);
      System.out.println ("found        == " + found);

      for (Pred p : newPred) {
        mustBacktrackOne = false;
        refiner.addToNewPredicates (p);
      }

    }
  }

  /*
   * Replaces the cond of each precond stub in pProgram: if the stub is in
   * preconds, replaces the cond with the associated not(precond), otherwise
   * the cond is replaced with false. Assumes a precond stub looks like this:
   * if (cond)
   *   loopHeaderLabel_PRECONDITIONi: while (1) ;
   *
   * XXX this doesn't belong here ...
   */
  void substPreconds (Map<String,Pred> preconds)
  {
    for (String precondStubLabel : precondStubLabels)
    {
      Pred precond;
      if ((precond = preconds.get (precondStubLabel)) != null)
	substCond (precondStubLabel, BoolOp.not (precond.getName ()));
      else
	substCond (precondStubLabel, fac.falseExpr ());
    }
  }

  void substCond (String precondStubLabel, Expr cond)
  {
    PStmt precondStub = pProgram.getStmt (precondStubLabel);
    for (PStmt p : precondStub.getPreds ()) // predecessors
      if (p instanceof IfPStmt)
	((IfPStmt) p).setCond (new PCond (cond, cond));
  }

  /**
   * Given a list of new predicates generated by propagating
   * assignments, see if we're propagating an assignment over the same
   * predicates multiple times (and thus unrolling a loop containing
   * said assignment). If so, figure out which loop is being unrolled,
   * and see if there's an abstraction pattern that can help us verify
   * the program without loop unrolling.
   *
   */
  boolean canUseAbstractionPattern (List<Pred> pl)
  {
    if (usedAbstractionPatternThisIteration
        || !usingAbstractionPatterns ()) return false;

    LoopUnrollingDetector l = new LoopUnrollingDetector (pProgram, pl);
    if (l.loopUnrollingDetected ()) {
      System.err.println ("Loop unrolling detected! Potentially unrolled loops: " + l.getUnrolledLoops ());
      for (IfPStmt loop : l.getUnrolledLoops ()) {

        // Get the parameters of the abstraction pattern
        String label = loop.getLabel ();
        BoundsCheckExtractor b = 
          new BoundsCheckExtractor (pProgram, prop,
                                    absPatternDB.getBoundedVars (label));

        // Bounds check extraction can fail if, for example, none of
        // the vars in the bounds check are bounded by this loop.
        //
        // XXX This API is gross. Clean it up at some point in the future.
        if (! b.extractionWorked ()) {
          System.out.println ("couldn't extract a bounds check");
          continue;
        }

        Expr var = b.getVar ();
        Expr bound = b.getBound ();
        IfPStmt assertion = b.getAssertion ();
        int tries = apBacktracker.getTries (label);
        System.out.println ("var = " + var);
        System.out.println ("bound = " + bound);
        System.out.println ("tries = " + tries);

        // If we're already using an abstraction pattern for this
        // loop, don't introduce a second one
        if (apBacktracker.alreadyPatternAbstracting (label)) continue;

        List<Pred> patternPreds = 
          absPatternDB.getAbstraction (pProgram, label, 
                                       var, bound, assertion,
                                       tries);

        // No more abstraction patterns to use
        if (patternPreds == null) continue;

        apBacktracker.registerChoice (label, refiner.getPreds (), patternPreds);

	substPreconds (apBacktracker.currentPreconds ());

        usedAbstractionPatternThisIteration = true;

        // Add the new preds. Make sure we don't end up with multiple
        // copies of the same predicate, and that any predicates in
        // the pattern that we already have are made new, so that
        // this'll work with incremental abstraction. (not well tested)
        List <Expr> oldPreds = Pred.getNames (refiner.getPreds ());
        assert (! patternPreds.isEmpty ());
        for (Pred p : patternPreds) {
          if (! oldPreds.contains (p.getName ())) {
            refiner.addToNewPredicates (p);
          } else {
            ((LocalCBJPredicateRefinerVisitor) refiner).makeNew 
              (p.getName ());
          }
        }
        found = true; // regardless of maxpreds, return true if we get 
                      // to use an abstraction pattern
        return true;

      }
    }

    // We didn't manage to use an abstraction pattern. Prune the
    // predicates which unroll only pattern-abstracted loops from our
    // list.
    List<Pred> uselessPreds = new ArrayList<Pred> ();
    for (Pred p : pl) {
      List <IfPStmt> pUnrolls = l.getUnrolledLoops (p);
      if (! pUnrolls.isEmpty ()) {
        boolean useful = false;
        for (IfPStmt loop : pUnrolls) {
          if ( !apBacktracker.alreadyPatternAbstracting (loop.getLabel ())) {
            useful = true;
            break;
          }
        }
        if (!useful) uselessPreds.add (p);
      }
    }
    newPredFound -= uselessPreds.size () + 1;
    found = newPredFound >= MAX_PREDICATES_PER_REFINEMENT;
    pl.removeAll(uselessPreds);
    System.out.println ("preds pruned to : " + Pred.getNames (pl));
    System.out.println ("removed         : " + Pred.getNames (uselessPreds));
    return false;
  }

  public ReachabilityProver initNTProver (BelnapAlgebra algebra, 
					  EdgeSet edges, 
					  MvSet stateCube)
  {
    ReachabilityProver prover = new ReachabilityProver ();
    prover.addRule (new EGStepRule (algebra, edges, stateCube));
    prover.addRule (new FunctionSummaryRule (algebra, stateCube));
    return prover;
    
  }
  

  public ReachabilityProver initProver (BelnapAlgebra algebra, 
					EdgeSet edges,
					MvSet stateCube)
  {
    ReachabilityProver prover = new ReachabilityProver ();
    prover.addRule (new PathRule (algebra, stateCube));
    prover.addRule (new SimpleMinPathRule (algebra));
    prover.addRule (new StepRule (algebra, edges, stateCube));
    prover.addRule (new FunctionSummaryRule (algebra, stateCube));
    return prover;
  }


  PredicateGenerator createPredicateGenerator (Object stmt, 
					       CTLNode[] parentState, 
					       CTLNode[] childState)
  {
    PredicateTable pTable = pProgram.getPredTable ();

    Map<String,Expr> predMap = new HashMap<String,Expr> ();
    for (Iterator it = pTable.getVariables ().iterator (); it.hasNext ();)
      {
	PredicateTable.Predicate pred = (PredicateTable.Predicate) it.next ();
	predMap.put (pred.getCtlName (), pred.getName ());
      }
    // Map symbolicMap = refiner.getMemoryModel ((PrllAsmtPStmt) stmt);    
    // WPComputer wp = refiner.getWPComputer ((PrllAsmtPStmt) stmt);
    WPComputer wp = ((PrllAsmtPStmt) stmt).getWPComputer ();

    // XXX This is dead --- why is it here?
    List<Expr> preds = refiner.getPredicates ();
    
    // pState, cState --- valuations of all predicates in parent and
    // child states, respectively; all CTLNodes in parentState and
    // childState are CTLEqualsNodes
    List<Expr> pState = getStatePred (predMap, parentState);
    List<Expr> cState = getStatePred (predMap, childState); 
    System.out.println ("Refining between states: ");
    System.out.println ("pState = " + pState);
    System.out.println ("cState = " + cState);
    
    return findPredicateGenerator (pGeneratorType,  pProgram, 
				   (PrllAsmtPStmt) stmt,
				   refiner, wp, pState, cState);
    
  }

  static PredicateGenerator findPredicateGenerator (String name,
						    PProgram pProgram,
						    PrllAsmtPStmt stmt,
						    Refiner refiner,
						    WPComputer wp,
						    List<Expr> pState,
						    List<Expr> cState)
  {
    ChainPredicateGenerator pg = new ChainPredicateGenerator ();
    
    String[] result = name.split (":");
    for (int i = 0; i < result.length; i++ )
      {
	String generatorName = result[i];
	if (generatorName.equals ("wpd"))
	  {
	    pg.add (new WPPredicateGenerator.WPDiff 
		    (pProgram,
		     (PrllAsmtPStmt)stmt,
		     refiner, wp, pState, cState));
	  }
	else if (generatorName.equals ("wps"))
	  {
	    pg.add (new WPPredicateGenerator.WPSame 
		    (pProgram, 
		     (PrllAsmtPStmt)stmt,
		     refiner, wp, pState, cState));
	  }
	else if (generatorName.equals ("wpa"))
	  {
	    pg.add (new WPPredicateGenerator.WPAll
		    (pProgram, 
		     (PrllAsmtPStmt)stmt,
		     refiner, wp, pState, cState));
	  }
	else if (generatorName.equals ("bend"))
	  {
	    pg.add (new BackEndGenerator 
		    (pProgram, 
		     (PrllAsmtPStmt) stmt,
		     refiner, wp, pState, cState));
	  }
	else 
	  throw new NoSuchElementException ("Unknow Predicate Generator in "
					    + name + " of " + generatorName);
      }
    return pg;
  }

  private List<Expr> getStatePred (Map<String,Expr> predMap, CTLNode[] state)
  {
    List<Expr> vExpr = new ArrayList<Expr>();
    for (int i = 0; i < state.length; i++)
      {
	Expr leftChild = predMap.get (state [i].getLeft ().toString ());

        // if a predicate p is true, encode it as the Expr p, otherwise, 
        // encode it as the Expr ~p
	if (state [i].getRight ().toString ().equals ("false"))
	  vExpr.add 
	    (refiner.exprFactory ().op (BoolOp.NOT).unaryApply (leftChild));
	else
	  vExpr.add (leftChild);
      }
    return vExpr;
  }
  

  public void parseCmdLine (String[] args)
    throws CTLNodeParserException
  {
  OptParser optParser = new OptParser ().startClass (YasrApp.class);

  BooleanOpt runOnceOpt = optParser.opt ().longName ("once").
    description ("does not go into the refinement loop").argumentArity (0).
    asBoolean ().defaultValue (false);
  StringOpt ctlPropOpt = 
    optParser.opt ().longName ("property").shortName ('p').
    description ("property to be checked").required ().
    asString ();

//   BooleanOpt cFile = optParser.opt ().longName ("pfile").
//     shortName ('c').
//     description ("input file is assumed to be a predicate program").
//     asBoolean ().defaultValue (true);

  StringOpt refinerOpt = optParser.opt ().longName ("refiner").
    description ("Refiner to be used. Choices are:\n" + 
		 //"\told -- complete slow search\n" + 
		 "new -- incremental full search\n" + 
		 "csp1 -- first try at CSP based search\n" + 
		 "cbj -- conflict-driven back jumping with prunning\n" + 
		 "local-cbj-vs -- LocalCBJPredicateRefinerVisitor\n" +
		 "cbj-i -- as cbj but don't check inconsistencies\n" + 
		 "cbj-s -- as cbj but don't abstract statements\n" + 
		 "cbj-s-i -- as cbj but only use cheap heuristics\n" + 
		 "cffc -- forward check with back jumping with prunning\n"+ 
		 "cffc-i -- as cffc but don't check inconsistencies\n" + 
		 "cffc-s -- as cffc but don't abstract statements\n" + 
		 "cffc-s-i -- as cffc but only use cheap heuristics\n" + 
		 "cffc-d -- as cffc but no use all predicates\n" +
		 "cffc-s-i-d -- take a guess\n" +
		   "\tDEFAULT IS local-cbj-vs\n")
    .asString ().defaultValue ("local-cbj-vs");

  StringOpt pGeneratorOpt = optParser.opt ().longName ("pg").
    description ("The order in which predicate generators are used\n " + 
		   "  separated by colon (':'). The choices are:\n" + 
		   "wpd -- weakest precondition based on difference\n" + 
		   "wps -- weakest precondition based on similarities\n" + 
		   "bend -- backend generator that solves\n" + 
		   "\tincompleteness of the refiner\n" + 
		   "\tDEFAULT IS bend:wpd:wps\n").asString ().
    defaultValue ("bend:wpd:wps:wpa");

  BooleanOpt hyperOpt = optParser.opt ().longName ("nohyper").
    description ("disables hyper edges").argumentArity (0).
    asBoolean ().defaultValue (false);

  BooleanOpt nonDetInitOpt = optParser.opt ().longName ("ndinit").
    description ("if present initial values of predicates are " + 
		 "assumed to be non-deterministic").
    argumentArity (0).asBoolean ().defaultValue (false);

  BooleanOpt inlineOpt = optParser.opt ().longName ("inline").
    description ("Inline function calls").
    argumentArity (0).asBoolean ().defaultValue (false);
    
  FileNameOpt initPredOpt = optParser.opt ().longName ("init-pred").
    description ("A file containing intial predicates").asFileName ();
  FileNameOpt inputFileOpt = 
    optParser.opt ().nameless ().required ().
    description ("boolean program").asFileName ();

  StringOpt selectorOpt = optParser.opt ().longName ("sel").
    description ("Return selector type. The choices are:\n" +
      "dd -- Decision diagram selector\n" +
      "bin -- Binary search PProgram selector\n" +
      "lin -- Linear search PProgram selector\n" +
      "\tDEFAULT IS dd\n").
    asString ().defaultValue ("dd");  

  BooleanOpt stmtBlockingOpt = optParser.opt ().longName ("blocking").
    description ("if present, consecutive simple assignments are merged " +
                 "into a basic block\n" +
                 "\tDEFAULT IS true\n").
    argumentArity (0).asBoolean ().defaultValue (true);

  IntOpt maxPredsPerRefinementOpt = optParser.opt ().longName ("maxpreds").
    description ("Max. number of predicates to discover per " +
                 "predicate-generation pass\n" +
                 "\tDEFAULT is 1\n").
    asInt ().defaultValue (1);             

  FileNameOpt absPatternOpt = optParser.opt ().longName ("prooftemplates").
    description ("A file containing a database of proof template instances").
    asFileName ();
  
  // -- actually parse the command line
  OptResult result = optParser.run (args);

  // -- initialize our class


  setCFile (new File (inputFileOpt.getString (result)));
  setProp (CTLNodeParser.parse (ctlPropOpt.getString (result)));
  setDoProof (!runOnceOpt.getBoolean (result));
  setInline (inlineOpt.getBoolean (result));
  setHyper (!hyperOpt.getBoolean (result));
  setUnknownInit (!nonDetInitOpt.getBoolean (result));


  setRefinerType (refinerOpt.getString (result));
  setPGeneratorType (pGeneratorOpt.getString (result));

  setSelectorType (selectorOpt.getString (result));
  setStmtBlocking (stmtBlockingOpt.getBoolean (result));

  setMaxPredsPerRefinement (maxPredsPerRefinementOpt.getInt (result));
  
  String initPredFileName = initPredOpt.getString (result);
  if (initPredFileName != null)
    setInitPredFile (new File (initPredOpt.getString (result)));
  else
    setInitPredFile (null);

  setAbsPatternFilename (absPatternOpt.getString (result));

} // end of parseCmdLine

 
  public static void main (String[] args)
{
  try
  {

    // -- create our application
    YasrApp yasmApp = new YasrApp ();

    // -- parse command line option
    yasmApp.parseCmdLine (args);


    System.err.println ("max preds: " + MAX_PREDICATES_PER_REFINEMENT);

    AlgebraValue value = yasmApp.run ();

    YasrApp.out.println ("Done in " + yasmApp.overall);

//     YasrApp.out.println ("QUERY: " + yasmApp.refiner.getVC ().sw);


    YasrApp.out.println ();
    YasrApp.out.println ();
    YasrApp.out.println ();
    YasrApp.out.println ("FINAL RESULT");
    YasrApp.out.println ("\tPROPERTY: " + yasmApp.getProp ());
    YasrApp.out.println ("\tRESULT: " + value);
    
    {
      int totalPredicates = 0;
      Set<Pred> uniquePreds = new HashSet<Pred> ();
      YasrApp.out.println ();
      for (CTLNode cn : yasmApp.predicateHistory.keySet ()) {
	List<Pred> preds = yasmApp.predicateHistory.get (cn);
	YasrApp.out.println ("Predicates to prove " + cn + ": " 
	    + preds.size () + "\n" + preds + "\n");
	totalPredicates += preds.size ();
	uniquePreds.addAll (preds);
      }

      YasrApp.out.println ("Total number of predicates: " + totalPredicates
	  + " (" + uniquePreds.size () + " unique)");
    }

  }
  catch (CTLNodeParserException ex)
  {
    YasrApp.out.println ("Could not parse CTL property: " + ex);
    ex.printStackTrace (YasrApp.err);
  }
  catch (antlr.ANTLRException ex)
  {
    YasrApp.out.println ("Could not parse C program: " + ex);
    ex.printStackTrace (YasrApp.err);
  }
  catch (FileNotFoundException ex)
  {
    YasrApp.out.println ("Could not find C file: " + ex);
    ex.printStackTrace (YasrApp.err);
  }
  catch (Throwable ex)
  {
    YasrApp.out.println ("Unhandled Exception: " + ex);
    ex.printStackTrace ();
  }

  // -- flush all our streams

  YasrApp.out.flush ();
  YasrApp.err.flush ();
} // end of main

  
  
} // YasrApp
