package edu.toronto.cs.yasm.abspatterns;

import java.util.Map;
import java.util.HashMap;
import java.util.Set;
import java.util.List;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.Stack;
import java.util.Iterator;

import edu.toronto.cs.yasm.pprogram.Pred;

/**
 * A class to handle backtracking when using abstraction patterns. An
 * absPatternBacktracker is only intended for use with *one*
 * assertion. When we start doing verification of inter-dependent
 * loops, use a separate backtracker.
 *
 * Since an AbsPatternBacktracker is tied to a single assertion, we
 * don't have to use a variable name as a key, since we'll ALWAYS be
 * bounding the LHS of the same bounds check assertion.
 *
 * Assume that at most one loop can be abstracted using patterns per
 * iteration. (This was important in the older design, but may not be
 * essential now.)
 * 
 */
public class AbsPatternBacktracker {

  /* number of iterations we can go for without timing out on an
   * abstraction pattern. we should probably make this a command-line
   * option eventually
   *
   * This is arbitrary; should be a lot different based on whether
   * we're using el-cheapo abstraction or not.
   */
  public static final int RETRY_THRESHOLD = 20;

  /* for parameterizing the "explicit" abstraction; see the paper */
  public static final int Cmax = 5;

  abstract class AP_StackFrame {
    /* loop label -> number of times we've tried to abstract that loop */
    Map <String, Integer> loopTriesMap;

    Map<String,Pred> preconds;
    Iterator<String> precondsIter;
	// XXX this is pretty awful I think? the iterator should be in
	// YasrApp.run

    public Map<String,Pred> getPreconds ()
    {
      return preconds;
    }

    public Iterator<String> getPrecondsIterator ()
    {
      return precondsIter;
    }

    public int getTries (String loopLabel)
    {
      Integer r = loopTriesMap.get (loopLabel);
      return (r != null) ? r.intValue () : 0;
    }

    public void incTries (String loopLabel)
    {
      loopTriesMap.put (loopLabel, new Integer (getTries (loopLabel) + 1));
    }

    public Set<String> getMappedLoops ()
    {
      return loopTriesMap.keySet ();
    }

    public abstract boolean tick ();

    public abstract List<Pred> getBacktrackPreds ();

    public abstract String getLoop ();
  }

  class AP_StackFrame_Child extends AP_StackFrame {
    List<Pred> backtrackPreds;
    String myLoop;
    int timeToLive;

    public AP_StackFrame_Child (AP_StackFrame parent, 
                                String _loop,
                                List<Pred> _backtrackPreds,
				List<Pred> patternPreds)
    {
      // Deep copy of parent's loopTriesMap
      loopTriesMap = new HashMap<String, Integer> ();
      for (String curKey : parent.getMappedLoops ())
        loopTriesMap.put (curKey, new Integer (parent.getTries (curKey)));

      // Deep copy of backtrack preds given to us
      backtrackPreds = new ArrayList<Pred> ();
      backtrackPreds.addAll (_backtrackPreds);

      myLoop = _loop;

      timeToLive = RETRY_THRESHOLD;

      preconds = new HashMap <String,Pred> ();

      int precondIdx = 1;
      for (Pred p : patternPreds) 
	if (p.isPrecondition ())
	  preconds.put (p.preconditionLabels ().get (0) + "_PRECONDITION" 
	      + precondIdx++, p); // XXX this is not the place to be doing this

      precondsIter = preconds.keySet ().iterator ();
    }

    public boolean tick ()
    {
      timeToLive--;
      return (timeToLive == 0);
    }

    public List<Pred> getBacktrackPreds ()
    {
      return backtrackPreds;
    }

    public String getLoop ()
    {
      return myLoop;
    }

  }

  class AP_StackFrame_Head extends AP_StackFrame {
    public AP_StackFrame_Head ()
    {
      loopTriesMap = new HashMap<String, Integer> ();
      preconds = new HashMap <String, Pred> ();
      precondsIter = preconds.keySet ().iterator ();
    }

    public boolean tick ()
    {
      return false;
    }

    public List<Pred> getBacktrackPreds ()
    {
      assert (false) : 
        "getBacktrackPreds called on stack head --- something is wrong";
      return null;
    }

    public String getLoop ()
    {
      return "DUMMYLOOP";
    }
  }


  /* *************** And now, the main class itself.... ****************** */


  Stack <AP_StackFrame> backtrackStack;

  public AbsPatternBacktracker ()
  {
    backtrackStack = new Stack <AP_StackFrame> ();
    backtrackStack.add (new AP_StackFrame_Head ());
  }

  /**
   * Return the number of tries we've made at pattern-abstracting the 
   * loop with the given label in the current backtracking context.
   */
  public int getTries (String loopLabel)
  {
    AP_StackFrame curContext = backtrackStack.peek ();

    assert (curContext != null) : "Invalid state: backtrack stack is empty";

    return curContext.getTries (loopLabel);
  }

  /**
   * Increment the number of tries in the current context for the loop 
   * with the given label, and create a new backtracking context, saving 
   * the given set of backtracking predicates.
   */
  public void registerChoice (String loopLabel, 
                              List<Pred> _backtrackPreds,
			      List<Pred> patternPreds)
  {
    AP_StackFrame curContext = backtrackStack.peek ();

    assert (curContext != null) : "Invalid state: backtrack stack is empty";

    curContext.incTries (loopLabel);
    backtrackStack.push (new AP_StackFrame_Child (curContext,
                                                  loopLabel,
                                                  _backtrackPreds,
						  patternPreds));
  }

  /**
   * Return true iff we are already using an abstraction pattern for 
   * the loop with the given label.
   */
  public boolean alreadyPatternAbstracting (String loopLabel)
  {
    for (AP_StackFrame frame : backtrackStack)
      if (frame.getLoop ().equals (loopLabel))
        return true;
    return false;
  }

  /**
   * Go through our backtracking stack, and see if any choice we made 
   * timed out. If so, restore the stack to the context before the 
   * timeout, and return the saved backtracking predicates of the 
   * choice that timed out.
   */
  public List<Pred> tick ()
  {
    AP_StackFrame timeoutFrame = null;
    for (AP_StackFrame frame : backtrackStack)
      if (frame.tick ()) {
        timeoutFrame = frame;
        break;
      }


    if (timeoutFrame != null) {
      while (backtrackStack.pop () != timeoutFrame) {}
      return timeoutFrame.getBacktrackPreds ();
    } else 
      return null;
  }

  /**
   * Backtrack, undoing the most recent decision. 
   */
  public List<Pred> backtrackOne ()
  {
    AP_StackFrame timeoutFrame = backtrackStack.peek ();
    assert (! (timeoutFrame instanceof AP_StackFrame_Head)) : 
      ("Error: attempting to backtrack past a point where no " +
       "decisions have been made");

    backtrackStack.pop ();

    return timeoutFrame.getBacktrackPreds ();
  }

  /** 
   * Tell if there's anything to backtrack to.
   */
  public boolean canBacktrack ()
  {
    AP_StackFrame timeoutFrame = backtrackStack.peek ();
    return (! (timeoutFrame instanceof AP_StackFrame_Head));
  }

  public Map<String,Pred> currentPreconds ()
  {
    return backtrackStack.peek ().getPreconds ();
  }

  public Iterator<String> currentPrecondsIterator ()
  {
    return backtrackStack.peek ().getPrecondsIterator ();
  }
}
