package edu.toronto.cs.yasm.abspatterns;

import java.util.Set;
import java.util.HashSet;
import java.util.Map;
import java.util.HashMap;
import java.util.List;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.Iterator;

import edu.toronto.cs.expr.*;
import edu.toronto.cs.cparser.CILTransformerUtil;
import edu.toronto.cs.yasm.pprogram.*;

import static edu.toronto.cs.expr.ComparisonOp.*;
import static edu.toronto.cs.expr.ArrayOp.*;
import static edu.toronto.cs.expr.NumericOp.*;
import static edu.toronto.cs.expr.BoolOp.*;
import static edu.toronto.cs.expr.RationalOp.*;
import static edu.toronto.cs.expr.SimplifyingRewriter.*;

public class AbstractionPatternDB {

  /* loop -> var -> list of abstraction patterns */
  Map <String, Map <Expr, List<AbstractionPatternEntry> > > abstractionMap;

  ExprFactory fac;

  private static final boolean USE_INV = true;

  private void dumpDB ()
  {
    for (String s :  abstractionMap.keySet ()) {
      System.err.println ("loop = " + s);
      Map <Expr, List<AbstractionPatternEntry>> cur = abstractionMap.get (s);
      for (Expr e : cur.keySet ()) {
        System.err.println ("\tvar = " + e);
        for (AbstractionPatternEntry ape : cur.get (e)) {
          System.err.println ("\t\t" + ape);
        }
      }
    }
  }

  public AbstractionPatternDB (AbstractionPatternFileParser apfp,
                               ExprFactory _fac)
  {
    abstractionMap = 
      new HashMap 
      <String, Map <Expr, List<AbstractionPatternEntry> > > ();
    fac = _fac;

    initializeDB (apfp);

    {
      int patterns = 0;
      for (Map m : abstractionMap.values ())
	patterns += m.size ();
      System.err.println (">>> AbstractionPatternDB #patterns: " + patterns);
    }
  }

  private Expr getExpr (String exprString)
  {
    try {
      return CILTransformerUtil.getExpr (fac, exprString);
    } catch (Exception ex) {
      throw new RuntimeException (ex);
    }
  }

  /* wrapper for lookup which creates an entry if it doesn't exist */
  private Map <Expr, List<AbstractionPatternEntry> > 
    getMapForLoop (String loop) 
  {
    Map<Expr, List<AbstractionPatternEntry> > candidate 
      = abstractionMap.get (loop);
    if (candidate == null) {
      HashMap<Expr, List<AbstractionPatternEntry> > newMap 
        = new HashMap <Expr, List<AbstractionPatternEntry> > ();
      abstractionMap.put (loop, newMap);
      return newMap;
    } else {
      return candidate;
    }
  }

  /* wrapper for lookup which creates an entry if it doesn't exist */
  private List<AbstractionPatternEntry>
    getAbstractionListForVar (Map <Expr, List<AbstractionPatternEntry> > map,
                              Expr var)
  {
    List<AbstractionPatternEntry> candidate = map.get (var);
    if (candidate == null) {
      List<AbstractionPatternEntry> newList 
        = new ArrayList <AbstractionPatternEntry> ();
      map.put (var, newList);
      return newList;
    } else {
      return candidate;
    }
  }

  public List<Expr> getBoundedVars (String loopLabel)
  {
    List<Expr> ret = new ArrayList<Expr> ();

    Map <Expr, List<AbstractionPatternEntry> > map =
      getMapForLoop (loopLabel);

    ret.addAll (map.keySet ());

    return ret;
  }

  /** return a new Expr which is e, with every variable replaced by 
   *  scope::var
   */
  private Expr mangle (String scope, String exp)
  {
    class ManglingRewriter extends IdentityRewriter
    {
      String scope;
      public ManglingRewriter (String s) {
        scope = s;
      }

      public Expr rewrite (Expr e)
      {
        e = super.rewrite (e);
        if (VariableOp.isVar (e))
          return e.getFactory ().var (scope + "::" + e);
        else 
          return e;
      }
    }

    return (new ManglingRewriter (scope)).rewrite (getExpr (exp));
  }

  private void initializeDB (AbstractionPatternFileParser apfp)
  {
    while (apfp.readLine () != -1) {
      String loop = apfp.getLoop ();
      String scope = apfp.getFunction ();
      Expr var = mangle (scope, apfp.getVar ());
      List<AbstractionPatternEntry> absList 
        = getAbstractionListForVar (getMapForLoop (loop), var);
      int numVars = apfp.getNumVars ();
      int type = apfp.getType ();
      if (numVars == 1 && type == 0) {
        /* 1 ex */
        absList.add 
          (new AbstractionPatternEntry (numVars,
                                        type,
                                        mangle (scope, apfp.getBound ())));
      } else if (numVars == 1 && type == 1) {
        /* 1 str */
        String array = apfp.getArray ();
        if (array.indexOf ("base(") == -1) {
          absList.add (new AbstractionPatternEntry 
                       (numVars, type,
                        mangle (scope, array)));
        } else {
          System.err.println ("no support for base pointers yet, skipping\n");
        }
      } else if (numVars == 2 && type == 0) {
        /* 2 ex */
        if (apfp.getVarDirection () == 2) {
          /* Don't know what direction to use. Add two apes. */
          for (int i = 0; i < 2; i++) {
            absList.add (new AbstractionPatternEntry 
                         (numVars, type,
                          mangle (scope, apfp.getBound ()),
                          mangle (scope, apfp.getLeader ()),
                          i,
                          apfp.getLeaderDirection ()));
          }
        } else {
          absList.add (new AbstractionPatternEntry 
                       (numVars, type,
                        mangle (scope, apfp.getBound ()),
                        mangle (scope, apfp.getLeader ()),
                        apfp.getVarDirection (),
                        apfp.getLeaderDirection ()));
        }
      } else if (numVars == 2 && type == 1) {
        /* 2 str */
        String array = apfp.getArray ();
        if (array.indexOf ("base(") == -1) {
          absList.add (new AbstractionPatternEntry 
                       (numVars, type,
                        mangle (scope, array),
                        mangle (scope, apfp.getLeader ())));
        } else {
          System.err.println ("no support for base pointers yet, skipping\n");
        }
      } else {
        System.err.println ("loop = " + loop);
        System.err.println ("var = " + var);
        System.err.println ("numVars = " + numVars);
        System.err.println ("type = " + type);
        throw new RuntimeException ("unexpected entry");
      }
    }
  }

  private String getBaseName (Expr e)
  {
    assert (VariableOp.isVar (e)) : "not a variable " + e;

    String [] parts = e.toString ().split ("::");
    return parts[1];
  }

  private String getScope (Expr e)
  {
    assert (VariableOp.isVar (e)) : "not a variable " + e;

    String [] parts = e.toString ().split ("::");
    return parts[0];
  }

  private Pred hintPred (Expr e)
  {
    return new Pred (e, PredicateSource.HINT);
  }

  private Pred hintPred (Expr e, String precLabel)
  {
    return new Pred (e, PredicateSource.HINT, precLabel);
  }

  private AbstractionPatternEntry getAPE
    (String label, Expr var, Expr bound, int tryNum) {

    List<AbstractionPatternEntry> absList 
      = getAbstractionListForVar (getMapForLoop (label), var);
    // XXX later considre CMax
    if (tryNum >= absList.size ()) return null;
    return absList.get (tryNum);

  }

  public boolean havePattern
    (String label, Expr var, Expr bound, int tryNum) {
    return (getAPE (label, var, bound, tryNum) != null);
  }

  /* getAbstraction : String -> Expr -> Expr -> int -> List<Pred>
   *
   * return it in list form so we have some control over the order 
   * of predicates... stupid peephole optimizations..... grr....
   *
   * XXX side effect: modifies precondition if-statements
   */
  public List<Pred> getAbstraction 
    (PProgram prog, String label, 
     Expr var, Expr bound, IfPStmt assertion,
     int tryNum) {

    System.err.println ("looking up abstraction for " + var + " <= " 
                        + bound + " in loop with label " + label);

    AbstractionPatternEntry ape = getAPE (label, var, bound, tryNum);
    if (ape == null) return null;

    List<Pred> abs = new ArrayList <Pred> ();
    int numVars = ape.getNumVars ();
    int type = ape.getType ();
    Expr N = bound;

    Pred precond;
    if (numVars == 1 && type == 0) {
      /* 1 ex */
      System.out.println (">>> abs: 1ex");
      Expr M = ape.getBound ();

      // XXX twiddle M
      abs.add (hintPred (leq (var,M), label));
      abs.add (hintPred (leq (M,var)));
      try {
        ExplicitSymbolicExecutor ese = 
          new ExplicitSymbolicExecutor ((IfPStmt)prog.getStmt (label), M, var);
        int c = ese.getC (assertion);
        Expr M_prime = 
          c >= 0 ? simplify (plus (M, c)) : simplify (minus (M, -1*c));

        //System.out.println ("M = " + M + "\nc = " + c);

        abs.add (hintPred (leq (M_prime,N), label));
      } catch (SymexException e) {
        e.printStackTrace ();
        assert false;
      }
    } else if (numVars == 1 && type == 1) {
      /* 1 str */
      System.out.println (">>> abs: 1str");
      Expr A = ape.getArray ();
      Expr A_nullpos = mangle (getScope (A), getBaseName (A) + "_nullpos");

      abs.add (hintPred (leq (var,A_nullpos), label));
      abs.add (hintPred (leq (A_nullpos,var)));
      abs.add (hintPred (leq (A_nullpos,N), label));
      abs.add (hintPred (eq (sel (A,var), 0)));
      /*if (USE_INV == false) {
        abs.add (eq (sel (A,A_nullpos), 0));
        abs.add (geq (A_nullpos,0));
        }*/
    } else if (numVars == 2 && type == 0) {
      /* 2 ex */
      System.out.println (">>> abs: 2ex");
      Expr leader = ape.getLeader ();
      Expr leader_s = 
        mangle (getScope (ape.getLeader ()), 
                getBaseName (ape.getLeader ()) + "_" + label + "_s");
      Expr var_s = 
        mangle (getScope (var),
                getBaseName (var) + "_" + label + "_s");
      Expr M = ape.getBound ();
      int leaderDirection = ape.getLeaderDirection ();
      abs.add (hintPred (leq (leader,M)));
      abs.add (hintPred (leq (M,leader)));
      abs.add (hintPred (leq (M,N), label));

      /* just assume varDirection is "up" */
      if (leaderDirection == 0) { /* up */
        abs.add (hintPred (leq (minus (var,var_s), minus (leader,leader_s)), label));
        abs.add (hintPred (leq (minus (M,leader_s), minus (N,var_s)), label));
      } else { /* down */
        abs.add (hintPred (leq (minus (var,var_s), minus (leader_s,leader)), label));
        abs.add (hintPred (leq (minus (leader_s,M), minus (N,var_s)), label));
      }
    } else if (numVars == 2 && type == 1) {
      /* 2 str */
      System.out.println (">>> abs: 2str");
      Expr leader = ape.getLeader ();
      Expr leader_s = 
        mangle (getScope (ape.getLeader ()),
                getBaseName (ape.getLeader ()) + "_" + label + "_s");
      Expr var_s = 
        mangle (getScope (var),
                getBaseName (var) + "_" + label + "_s");
      Expr A = ape.getArray ();
      Expr A_nullpos = mangle (getScope (A),
                               getBaseName (A) + "_nullpos");
      Expr M = A_nullpos;

      abs.add (hintPred (leq (leader,A_nullpos), label));
      abs.add (hintPred (leq (A_nullpos,leader)));
      abs.add (hintPred (leq (A_nullpos,N), label));
      abs.add (hintPred (eq (sel (A,leader),0)));
      /*if (USE_INV == false) {
        abs.add (eq (sel (A,A_nullpos), 0));
        abs.add (geq (A_nullpos,0));
        }*/
      /* leaderDirection is up for str case */
      abs.add (hintPred (leq (minus (var,var_s), minus (leader,leader_s)), label));
      abs.add (hintPred (leq (minus (M,leader_s), minus (N,var_s)), label));
    } else {
      throw new RuntimeException ("unexpected entry");
    }

    return abs;
  }

  /**
   *
   * Throw this if the backwards symbolic execution doesn't work ---
   * ie. we somehow get out of the loop, or our iterator has something
   * that isn't a constant added to it.
   *
   */
  class SymexException extends Exception
  {
    public SymexException (String msg)
    {
      super (msg, new Exception ());
    }
  }

  /**
   * While using the single-variable explicit pattern, we have two 
   * sources of complication:
   *   (1) the check in the exit conditional could be <=, ~, or <
   *   (2) iterator could be incremented between exit conditional and 
   *       assertion
   */
  class ExplicitSymbolicExecutor {
    IfPStmt loop;
    Expr boundFromAPE;
    Expr iterator;
    Set<PStmt> visited;
    boolean found;
    int c;

    public ExplicitSymbolicExecutor (IfPStmt _loop,
                                     Expr _boundFromAPE,
                                     Expr _iterator)
    {
      loop = _loop;
      boundFromAPE = _boundFromAPE;
      iterator = _iterator;
      visited = new HashSet<PStmt> ();
      found = false;
    }

    private void foundExitConditional (IfPStmt exitCond,
                                       PStmt prev,
                                       int cur_offset)
      throws SymexException
    {
      System.out.println ("exit conditional is " + exitCond + "\nprev is " + prev);

      // Adjust the valuation of the conditional based on what path we
      // took to reach this statement
      Expr cond = exitCond.getCond ().getCond ();
      if (prev == exitCond.getElseStmt ()) { // XXX should this be elseGoto?
        System.out.println ("REACHED THROUGH ELSE BRANCH");
        if (isNot (cond))
          cond = cond.arg (0);
        else
          cond = not (cond);
      }

      System.out.println ("before stripping nots, cond == " + cond);

      // Get the number of negations, and the base condition, and see
      // what the effective condition is
      int notsStripped = 0;
      while (isNot (cond)) {
        cond = cond.arg (0);
        notsStripped++;
      }

      System.out.println (notsStripped + " nots stripped");

      if (notsStripped % 2 == 1) {
        Expr lhs = cond.arg (0);
        Expr rhs = cond.arg (1);
        if (isEq (cond)) cond = neq (lhs, rhs);
        else if (isNeq (cond)) cond = eq (lhs, rhs);
        /*
          XXX what the hell was I thinking here?
        else if (isLeq (cond)) cond = geq (lhs, rhs);
        else if (isLt (cond)) cond = gt (lhs, rhs);
        */
        else if (isGeq (cond)) cond = lt (lhs, rhs);
        else if (isGt (cond)) cond = leq (lhs, rhs);
        else 
          throw new SymexException ("don't know what to do with cond " + cond);
      }

      //System.out.println ("cur_offset = " + cur_offset);
      //System.out.println ("cond = " + cond);

      // If the valuation of the conditional implies that iterator !=
      // bound, then, assuming that iterator <= bound before the
      // conditional, iterator <= bound - 1 after the conditional.
      if (isNeq (cond) || isLt (cond))
        c = cur_offset - 1;
      else
        c = cur_offset;

      found = true;
    }

    /**
     * Tell if this statement is the exit conditional we
     * want. Unfortunately, we don't store the exact exit
     * conditional. Fake it with a ghetto implementation.
     */
    private boolean isRightExitConditional (Expr cond)
    {
      if (! isComparison (cond)) return false;
      if (! boundFromAPE.contains (cond.arg (1))) return false;
      
      Set<Expr> condVars = VariableOp.vars (cond, new HashSet<Expr> ());
      Set<Expr> boundVars = 
        VariableOp.vars (boundFromAPE, new HashSet<Expr> ());
      if (! condVars.containsAll (boundVars)) return false;
      
      return true;
    }

    /**
     * cur -- the current statement
     * prev -- the statement whose predecessor field we followed to get 
     *         here; used for interpreting ifs
     */
    private void inner (PStmt cur, PStmt prev, int cur_offset) 
      throws SymexException
    {
      /* make sure we don't reach the head of a while (1) {} loop
       * through its else branch */
      if (cur instanceof IfPStmt) {
        IfPStmt pp = (IfPStmt) cur;
        Expr cond = pp.getCond ().getCond ();
        if ((isTrue (cond) || (isInt (cond) && getN (cond) == 1))
            && prev == pp.getElseGoto ().getDest ()) {
          return;
        }
      }

      // only visit a statement once
      if (visited.contains (cur)) return;
      visited.add (cur);
      //System.out.println ("visiting " + cur);

      if (cur instanceof IfPStmt) {
        // check if conditional is the one we want
        Expr cond = labs (((IfPStmt)cur).getCond ().getCond ());
        if (isRightExitConditional (cond)) {
          foundExitConditional ((IfPStmt)cur, prev, cur_offset);
        } else if (cur == loop) {
          // recurse, but stay in the loop
          for (PStmt pred : cur.getPreds ()) {
            if (HasStmtVisitor.hasStmt (loop, pred))
              inner (pred, cur, cur_offset);
          }
        } else { // recurse
          for (PStmt pred : cur.getPreds ()) {
            inner (pred, cur, cur_offset);
          }
        }
      } else if (cur instanceof PrllAsmtPStmt) {
        // check if asmt increments or decrements iterator
        Expr sourceExpr = ((PrllAsmtPStmt)cur).getSourceExpr ();

        assert CILListOp.isStatementList (sourceExpr);

        int change = 0;
        for (Expr instr : sourceExpr.args ()) {
          assert CILAssignOp.isAssign(instr) : 
            "not an assignment: " + instr;
          
          Expr lhs = instr.arg (0);
          Expr rhs = instr.arg (1);
          if (lhs == iterator) {
            if (isPlus (rhs) 
                && rhs.arg (0) == iterator 
                && isInt (rhs.arg (1)))
              change += getN (rhs.arg (1));
            else if (isMinus (rhs) 
                     && rhs.arg (0) == iterator 
                     && isInt (rhs.arg (11)))
              change -= getN (rhs.arg (1));
            else
              throw new SymexException 
                ("don't know what to do with assignment" + cur);
          }
          /*System.out.println ("cur = " + sourceExpr + 
                              "\n\trecursing with change = " + change +
                              "\n\t cur_offset = " + (cur_offset+change));*/
        }
        

        // recurse
        for (PStmt pred : cur.getPreds ())
          inner (pred, cur, cur_offset + change);

      } else if (cur instanceof SkipPStmt || cur instanceof GotoPStmt) {
        // recurse
        for (PStmt pred : cur.getPreds ())
          inner (pred, cur, cur_offset);
      } else {
        // XXX do we need to handle GOTOs?
        throw new SymexException ("unsupported operation " + cur);
      }
    }

    /** Facade to more complicated function */
    public int getC (IfPStmt assertion)
      throws SymexException
    {
      // assertion is the place from which to start getting predecessors
      for (PStmt p : assertion.getPreds ()) {
        if (found) break;
        inner (p, assertion, 0);
      }
      
      //assert (found) : "failed to find exit conditional";
      
      return c;
    }
  }

}
