package edu.toronto.cs.yasm.abspatterns;

import java.util.List;
import java.util.Set;
import java.util.HashSet;
import java.util.LinkedList;

import edu.toronto.cs.expr.*;
import edu.toronto.cs.yasm.pprogram.*;
import edu.toronto.cs.ctl.*;

import static edu.toronto.cs.expr.NumericOp.*;
import static edu.toronto.cs.expr.ComparisonOp.*;
import static edu.toronto.cs.expr.VariableOp.*;
import static edu.toronto.cs.expr.RationalOp.*;

/**
 * Given a CTLNode which represents an "EF pc = ERROR" node, which, by
 * assumption, is the failure path of a bounds check assertion, get
 * the bounds check.
 *
 * Works assuming that the bounds check is of the form:
 *   if (! (blah ~ blah))
 *      ERROR: goto ERROR;
 * where ~ is one of <= or <.
 *
 */
public class BoundsCheckExtractor
{
  PProgram pProgram;
  PStmt errorStmt;

  Expr lhs;
  Expr var;
  Expr bound;
  IfPStmt assertion;

  List<Expr> boundedVars;

  boolean worked;

  public Expr getVar ()
  {
    assert (var != null) : "failed to extract var from bounds check";
    return var;
  }

  public Expr getBound ()
  {
    assert (bound != null) : "failed to extract bound from bounds check";
    return bound;
  }

  public IfPStmt getAssertion ()
  {
    assert (assertion != null) : 
      "failed to extract assertion from bounds check";
    return assertion;
  }

  private void fillInErrorStmt (CTLNode _prop)
  {
    CTLNode right = _prop.getRight ();
    
    if (! (right instanceof CTLEqualsNode)) {
      throw new RuntimeException ("Only 'EF pc = XXX' is supported");
    }
    
    String pcValue = right.getRight ().toString ();
    
    PStmt stmt = pProgram.getStmt (pcValue);
    if (stmt == null) {
      throw new RuntimeException ("No such statement!");
    } else {
      errorStmt = stmt;
    }

    List<PStmt> predecessors = new LinkedList<PStmt> ();
    predecessors.addAll (errorStmt.getPreds ());
    if (predecessors.contains (errorStmt)) predecessors.remove (errorStmt);
    assert predecessors.size () == 1 :
      errorStmt + " does not have a unique predecessor; CFG predecessors: " 
      + predecessors;

    PStmt errorPred = predecessors.get (0);
    assert (errorPred instanceof IfPStmt) : "not an IfPStmt " + errorPred;
    assertion = (IfPStmt)errorPred;
    fillInVarAndBound (assertion.getCond().getCond());
  }

  public BoundsCheckExtractor (PProgram _pProgram, 
                               CTLNode _prop,
                               List<Expr> _boundedVars)
  {
    pProgram = _pProgram;
    var = null;
    bound = null;
    boundedVars = _boundedVars;
    fillInErrorStmt (_prop);
  }

  public boolean extractionWorked ()
  {
    return worked;
  }

  private boolean found ()
  {
    return var != null;
  }

  private boolean isWellFormed (Expr e)
  {
    if (! (isVariable (e) || isInt (e) || isPlus (e) || isMinus (e)))
      return false;

    for (Expr child : e.args ())
      if (! (isWellFormed (child)))
        return false;

    return true;
  }

  private void fillInVarAndBound (Expr cond)
  {
    Expr e = cond;
    
    // Strip !'s
    while (BoolOp.isNot (e)) e = e.arg (0);

    assert (e.op () instanceof ComparisonOp) : "not a comparison op: " + e;

    // Abstraction patterns expect a bounds check of the form i <= N
    // But we can have bounds checks of the form i < N
    // Transform the latter using the fact that (i < N+1) == (i <= N)
    assert (isLeq (e) || isLt (e)) : "neither < nor <=: " + e;

    // LHS of bounds check must be a mixed sum/difference of variables 
    // and constants
    lhs = e.arg (0);
    assert (isWellFormed (lhs)) : 
      "lhs " + lhs + " of bounds check is not well-formed!";

    if (isLeq (e)) {
      bound = e.arg (1);
    } else { // by above assertion, must be less-than
      bound = (new SimplifyingRewriter ()).rewrite (minus (e.arg (1), 1));
    } 

    // We have the "base" bound and the lhs. Now, extract the var by
    // intersecting it with the vars the loop bounds (according to our
    // abstraction pattern database), and rewrite the bounds check.

    Set<Expr> lhsVars = new HashSet<Expr> ();
    lhsVars = vars (lhs, lhsVars);

    lhsVars.retainAll (boundedVars);

    if (lhsVars.size () != 1) {
      // Either none of the bounded variables appear in this bounds
      // check, or multiple bounded variables do. Give up.
      //
      // XXX In general, consider derived variables. Put this off
      // until either (1) we really really need it, or (2) we replace
      // the current implementation with something that isn't strung
      // together with chicken wire.
      worked = false;
      return;
    }

    var = (Expr)((lhsVars.toArray ())[0]);

    // Rewrite the bounds check. ie. if i is our iterator, and the bounds 
    // check is 
    //    blah + i - blah2 <= blah3
    // rewrite this as
    //    i <= blah3 - blah + blah2
    while (lhs != var) {
      if (isMinus (lhs)) {
        if (lhs.arg (0).contains (var)) {
          bound = plus (bound, lhs.arg (1));
          lhs = lhs.arg (0);
        } else {
          assert false : "iterator cannot be negative in bounds check";
        }
      } else if (isPlus (lhs)) {
        if (lhs.arg (0).contains (var)) {
          bound = minus (bound, lhs.arg (1));
          lhs = lhs.arg (0);
        } else {
          bound = minus (bound, lhs.arg (0));
          lhs = lhs.arg (1);
        }
      } else {
        assert false : "shouldn't get here";
      }
    }

    worked = true;
  }


}
