package edu.toronto.cs.yasm.abspatterns;

import edu.toronto.cs.yasm.pprogram.*;
import java.util.List;
import java.util.ArrayList;
import java.util.Set;
import java.util.HashSet;

import static edu.toronto.cs.yasm.pprogram.PredicateSource.*;
import static edu.toronto.cs.yasm.abspatterns.HasStmtVisitor.*;

public class LoopUnrollingDetector {
  
  PProgram prog;
  List<Pred> pl;
  Pred tail;
  PrllAsmtPStmt asmt;

  public LoopUnrollingDetector (PProgram _p, List<Pred> _pl)
  {
    prog = _p;
    pl = _pl;
    tail = null;
    asmt = null;
  }

  /**
   * Return true iff some predicate in our list causes some loop to be
   * unrolled. As a side-effect, set {@code tail}, which is necessary
   * for {@code getUnrolledLoops} to work.
   *
   */
  public boolean loopUnrollingDetected ()
  {
    for (Pred cur : pl) {
      if (predCreatesCycle (cur)) {
        tail = getTail (cur);
        assert (cur.getSource () instanceof WPPredicateSource);
        asmt = ((WPPredicateSource) cur.getSource ()).getAssignment ();
        return true;
      }
    }
    return false;
  }

  private void inner
    (PStmt start, Set<PStmt> visited, List<IfPStmt> acc, PStmt cur)
  {
    if (visited.contains (cur)) return;
    visited.add (cur);

    if (cur instanceof WhilePStmt) {
      if (hasStmt ((WhilePStmt)cur, start)) {
        acc.add ((WhilePStmt)cur);
      }
    }

    for (PStmt p : cur.getPreds ())
      inner (start, visited, acc, p);
  }


  /**
   * XXX This is a hack --- we should really examine the
   * counterexample for CFG locations which are visited repeatedly,
   * and infer which loops are being unrolled that way.
   *
   */
  private List<IfPStmt> getUnrolledLoopsFromPred (Pred p)
  {
    if (p.getSource () instanceof ConditionalPredicateSource) {
      List ret = 
        UnrolledLoopCollectorVisitor.getLoops (tail.getName (), asmt);
      if (ret.size () == 0) {
        ret = new ArrayList<IfPStmt> ();
        Set acc = new HashSet<PStmt> ();
        PStmt start = 
          ((ConditionalPredicateSource)
           (p.getSource ())).getConditional ();
        inner (start, acc, ret, start);
        return ret;
      } else {
        return ret;
      }
    }
    return new ArrayList<IfPStmt> ();
  }

  /**
   * Return the loops which may be being unrolled.
   *
   * @return a {@code List<PStmt>} value
   */
  public List<IfPStmt> getUnrolledLoops ()
  {
    return getUnrolledLoopsFromPred (tail);
  }


  /** 
   * Return a list of loops which the predicate {@code p} may be unrolling.
   */
  public List<IfPStmt> getUnrolledLoops (Pred p)
  {
    if (predCreatesCycle (p)) {
      assert (p.getSource () instanceof WPPredicateSource);
      return getUnrolledLoopsFromPred (getTail (p));
    }
    return new ArrayList<IfPStmt> ();
  }

  /**
   * Tell if adding a given {@code Pred} causes a loop to be unrolled.
   *
   */
  protected boolean predCreatesCycle (Pred p)
  {
    PredicateSource ps = p.getSource ();
    if (ps instanceof WPPredicateSource) {
      WPPredicateSource wps = (WPPredicateSource) ps;
      PrllAsmtPStmt asmt = wps.getAssignment ();
      
      // Walk the linked list of predicates, looking for a Pred whose source
      // is a WPPredicateSource whose assignment is asmt
      Pred cur = wps.getParent ();
      while (true) {
        if (cur.getSource () instanceof WPPredicateSource) {
          WPPredicateSource wpcs = (WPPredicateSource) cur.getSource ();
          if (wpcs.getAssignment () == asmt) {
            System.err.println 
              ("predicate " + 
               p.getName () + 
               "creates a cycle due to propagation of " + 
               p.getSource ());
            return true;
          }
          else cur = wpcs.getParent ();
        } else {
          return false;
        }
      }
    } else {
      return false;
    }
  }

  /** 
   * Find the tail of a list of {@code Pred}s --- ie. the first
   * {@code Pred} whose source is not a {@code WPPredicateSource}.
   *
   */
  protected Pred getTail (Pred p)
  {
    Pred cur = p;
    while (true) {
      if (cur.getSource () instanceof WPPredicateSource)
        cur = ((WPPredicateSource) cur.getSource ()).getParent ();
      else
        return cur;
    }
  }
}
