package edu.toronto.cs.yasm.abstractor;

import edu.toronto.cs.clang.*;
import edu.toronto.cs.expr.*;
import edu.toronto.cs.yasm.pprogram.*;
import edu.toronto.cs.yasm.cfg.*;
import edu.toronto.cs.yasm.cfg.CFGraph.*;

import java.util.*;
import java.io.*;


import static edu.toronto.cs.expr.CDeclOp.*;
import static edu.toronto.cs.expr.CNamedDeclOp.*;
import static edu.toronto.cs.expr.CStmtOp.*;
import static edu.toronto.cs.expr.CLabelDeclOp.labelUID;
import static edu.toronto.cs.expr.CSwitchExprOp.*;

import static edu.toronto.cs.expr.CCallExprOp.isCCallExpr;
import static edu.toronto.cs.expr.CCallExprOp.getCallee;

/**
 * Creates an intial (no-predicates) PProgram abstraction of a CProgram.
 */
public class CNullExprAbstractor
{
  
  /**
   * Expression factory
   *
   */
  protected ExprFactory efac;
  

  /**
   * A {@code CProgram} being abstractred
   */
  protected CProgram cProgram;


  /*******************************************************************/
  /* INTERNALS */
  /* Only available within the lifetime of absProgram method */
  /*******************************************************************/


  /**
   * List where statements of the final abstract program are put
   * during construction.
   *
   */
  private List<PStmt> statementList;
  
  /**
   * A map from labels to statements
   */
  private Map<String, PStmt> labelMap;


  /**
   * A list of all abstract functions, indexed by UID
   */
  private Map<Integer,PFunctionDef> functionDefs;


  /**
   * Maps BB id to a PStmt representing entry into that block
   *
   */
  private PStmt[] bbEntry;
  
  /**
   * Maps BB id to PStmt representing exit from that block
   *
   */
  private PStmt[] bbExit;

  private CFGraph cfg;
  private PFunctionDef def;


  
  /**
   * Protects construction of this object. Use {@method absProgram}
   * instead.
   *
   * @param efac an <code>ExprFactory</code> value
   */
  protected CNullExprAbstractor (ExprFactory efac)
  {
    this.efac = efac;
  }
  

  /**
   * Construct a {@code PProgram} out of a C program.
   *
   * @param efac a factory to create new expressions
   * @param cProg a C program
   * @return an abstract version of {@code cProg}
   */
  public static PProgram absProgram (ExprFactory efac, CProgram cProg)
  {
    return new CNullExprAbstractor (efac).absProgram (cProg);
  }
  

  /**********************************************************************/
  /* Abstraction routines. 
   * Name Convention: 
   *    absXXX  -- a function to abstract XXX
   */
  /**********************************************************************/


  /**
   * Abstracts the program by abstracting all of its functions
   *
   * @param _cProgram a <code>CProgram</code> value
   * @return a <code>PProgram</code> value
   */
  PProgram absProgram (CProgram _cProgram)
  {
    PFunctionDef mainDef;

    // -- set the current program
    cProgram = _cProgram;

    // -- initialize 
    statementList = new LinkedList<PStmt> ();
    labelMap = new HashMap<String, PStmt> ();
    mainDef = null;
    functionDefs = new TreeMap<Integer, PFunctionDef> ();

    // -- create designated init and END statements.  They are created
    // -- first to be the first two elements in the list of
    // -- statements. The parent field of this statements is set to
    // -- null at first, but is reset to be the main function later.
    SkipPStmt init = new SkipPStmt (null, (Expr) null);
    // this could also be a Skip since nowdays Skips and gotos
    // virtually indistinguishable. (But it really should be a goto
    // since it affects the direction of the control flow)
    GotoPStmt end = new GotoPStmt (null, (Expr) null); 
    addStmt (init);
    addLabel (init, "init");
    init.setNext (end);
    addStmt (end);
    addLabel (end, "END");
    end.setDest (end);


    // -- Iterate over all functions
    for (Expr funcDecl : cProgram.allCfgFunctions ())
      {
	PFunctionDef def = absFunction (funcDecl);
	if (CNamedDeclOp.declName (funcDecl).equals ("main"))
	  mainDef = def;
      }

    // -- patch the parents of init and END. For now, they belong to
    // -- the main function for historical reasons.
    init.setParent (mainDef);
    init.setDest (mainDef.getEntry ());
    end.setParent (mainDef);

    // patch main's return selector to goto END
    ((ReturnSelectorPStmt) mainDef.getReturnSelector ()).addDest (end);


    PProgram pp = new PProgram (efac, cProgram, functionDefs);
    pp.setStatementList (statementList);
    pp.setLabelMap (labelMap);

    // -- set parents to all functions we have already created
    for (PFunctionDef def : functionDefs.values ())
      def.setParent (pp);


    // free temporaries
    cProgram = null;

    // -- initialize 
    statementList = null;
    labelMap = null;
    mainDef = null;
    functionDefs = null;

    return pp;
  }




  /**
   * Creates a PFunctionDef abstraction of a function described by the
   * expression {@code e}
   *
   * @param e an <code>Expr</code> value
   * @return a <code>PFunctionDef</code> value
   */
  PFunctionDef absFunction (Expr funcDecl)
  {

    // -- find the control flow graph of funcDecl
    cfg = cProgram.getCFG (funcDecl);

    // -- we are only called on function declarations that have
    // -- function bodies. 
    assert cfg != null;

    // -- possibly create a blank PFunctionDef
    def = getPFunctionDef (funcDecl);

    bbEntry = new PStmt [cfg.size ()];
    bbExit = new PStmt [cfg.size ()];


    // --------------------------------------------------
    // -- FIRST PASS
    // -- create PStmts for BBs and link their syntactic next fields
    // --------------------------------------------------
    

    // -- for each BB, convert it into a sequence of PStmts, and link
    // -- to the last basic block converted. Intially, everything is
    // -- linked to a dummy entry node of the PFunctionDef.
    PStmt prev = def.getEntry ();

    assert prev != null;
    
    for (BasicBlock bb : cfg.allBlocksFwd ())
      {
	PStmt curr = absBB (bb, prev);
	// -- fill bbEntry/bbExit arrays
	bbEntry [bb.getIndex ()] = prev.getNext ();
	bbExit [bb.getIndex ()] = curr;
	prev = curr;
      }
    
    

    // --------------------------------------------------
    // -- SECOND PASS
    // -- link Dest fields (only applies to PStmts 
    // -- that are at an exit of some BB)
    // --------------------------------------------------
    for (int i = 0; i < bbExit.length; i++)
      {
	// -- some BB ids maybe unused
	if (bbExit [i] == null) continue;

	// -- proceed based on the type of bbExit [i] node
	// -- for conditionals -- resolve then/else gotos 
	// -- for switches -- resolve switch gotos
	// -- for returns  -- point them to return selector
	// -- otherwise, use the edge of the appropritate BB
	
	if (bbExit [i] instanceof ConditionalBranchPStmt)
	  {
	    resolveGoto 
	      (((ConditionalBranchPStmt)bbExit [i]).getTrueGoto ());
	    resolveGoto 
	      (((ConditionalBranchPStmt)bbExit [i]).getFalseGoto ());
	  }
	else if (bbExit [i] instanceof SwitchBranchPStmt)
	  for (PSwitchCase scase : ((SwitchBranchPStmt)bbExit [i]).cases ())
	    resolveGoto (scase.getGoto ());
	else if (bbExit [i] instanceof ReturnPStmt)
	  bbExit [i].setDest (def.getReturnSelector ());
	else 
	  {
	    List<Edge> succs = cfg.getBB (i).getSuccs ();
	    
	    // -- a simple statement can only have one successor
	    // -- or none if it is at a dead-end
	    
	    assert succs.size () <= 1 : "Got statement: " + 
	      bbExit [i] + " with " + succs.size () + " succs";

	    if (succs.isEmpty ()) continue;
	    
	    
	    // -- extract the edge and find PStmt corresponding to the
	    // -- BB at the destination
	    bbExit[i].setDest (bbEntry [succs.get (0).getDst ().getIndex ()]);
	  }
	
      }
    

    // -- deallocate temporaries
    bbEntry = null;
    bbExit = null;
    cfg = null;
    
    // -- clear def (but store it temporarily so that we can return it 
    PFunctionDef retDef = def;
    def = null;
    return retDef;
  }
  


  /**
   * Returns the PFunctionDef for the function with uid {@code uid};
   * creates one and adds it to the function table if one does not exist.
   */
  private PFunctionDef getPFunctionDef (Expr funcDecl)
  {
    int uid = declUID (funcDecl);
    PFunctionDef def = functionDefs.get (uid);

    if (def != null)
      return def;

    // -- get CFG of the function. Only functions with a CFG have a
    // -- PFunction equivalents, others are treated as skips
    CFGraph cfg = cProgram.getCFG (funcDecl);
    if (cfg == null) return null;
    

    // function not in table, create one
    def = new PFunctionDef (cfg, funcDecl);
    functionDefs.put (uid, def);

    // create selector and entry point
    // XXX no special cases for 0 or 1 call sites; we don't know the number of
    // call sites here; cumbersome to do a second pass and replace
    // ReturnSelectorPStmt with a Goto or Skip since we don't have prev links
    // for next and dest
    PStmt sel = new ReturnSelectorPStmt (def, false);
    PStmt funcEntry = new SkipPStmt (def, null);


    // XXX Do scary mangling to PFunctionDef. None of this is probably
    // needed with the GCC-XML parse

    // add selector and entry point to function's head
    // no dests for sel yet; these have to be added at each call site
    def.setHead (sel);
    def.setReturnSelector (sel);
    sel.setNext (funcEntry);

    addStmt (sel);
    addLabel (sel, def.returnSelectorLabel ());
    
    addStmt (funcEntry);
    addLabel (funcEntry, def.entryLabel ());

    def.setEntry (funcEntry);

    return def;
  }
  

  /**
   * Abstract the statements in {@code bb} by a list of PStmts. The
   * first Pstms in the list is linked to the next field of {@code
   * prev}. The last is returned.
   *
   * @param bb a <code>BasicBlock</code> value
   * @param prev a <code>PStmt</code> value
   * @return a <code>PStmt</code> value
   */
  private PStmt absBB (BasicBlock bb, PStmt head)
  {
    // -- assume that prev is not connected to anything yet
    assert head.getNext () == null;
    
    List<Expr> labels = new LinkedList<Expr> ();

    PStmt prev = head;
    
    for (Expr e : bb.getStmts ())
      {
	// -- skip all labels, but remember what they are
	if (CStmtOp.isLabelExpr (e))
	  {
	    labels.add (CStmtOp.labelExprDecl (e));
	    continue;
	  }

	// -- if we are here, `e' is not a label
	prev = absStmt (e, prev);
      }

    // -- at this point
    // -- head.getNext () is the first statement in the list
    // -- prev is the last one

    // -- if prev is head, no statements where created
    // -- create a Skip
    if (prev == head)
      {
	prev = new SkipPStmt (def);
	addStmt (prev);
	head.setNext (prev);
      }


    // -- we should always add something for a basic block
    assert head.getNext () != null;

    // -- attach labels
    for (Expr labelDecl : labels)
      addLabel (head.getNext (), labelDecl);
    
    return prev;
  }
  
  

  /**
   * Abstracts a single C statement by a list of PStmts. The list is
   * attached to the Next field of `head'. The return value is the
   * last element of the list.
   *
   * @param cStmt an <code>Expr</code> value
   * @param head a <code>PStmt</code> value
   * @return a <code>PStmt</code> value
   */
  private PStmt absStmt (Expr cStmt, PStmt head)
  {
    if (isModifyExpr (cStmt))
      return absModifyExpr (cStmt, head);
    else if (isReturn (cStmt))
      return absReturn (cStmt, head);
    else if (isCondExpr (cStmt))
      return absCondExpr (cStmt, head);
    else if (isSwitchExpr (cStmt))
      return absSwitchExpr (cStmt, head);
    else if (isCCallExpr (cStmt))
      return absCallExpr (cStmt, head);
    
    throw new RuntimeException ("Could not abstract: " + cStmt);
  }
  

  private PStmt absModifyExpr (Expr e, PStmt head)
  {
    if (isCCallExpr (e.arg (1)))
      return absCallExpr (e.arg (1), e.arg (0), head);

    PStmt v = new PrllAsmtPStmt (def, e);
    addStmt (v);
    if (head != null) head.setNext (v);
    return v;
  }
  


  private ReturnPStmt absReturn (Expr e, PStmt head)
  {
    ReturnPStmt v = new ReturnPStmt (def, e);
    addStmt (v);
    if (head != null) head.setNext (v);
    return v;
  }

  private GotoPStmt absGoto (Expr e, PStmt head)
  {
    GotoPStmt v = new GotoPStmt (def, e);
    addStmt (v);
    if (head != null) head.setNext (v);
    return v;

  }

  private ConditionalBranchPStmt absCondExpr (Expr e, PStmt head)
  {
    ConditionalBranchPStmt v = 
      new ConditionalBranchPStmt (def, e, 
				  absCond (condExprCond (e)),
				  absGoto (condExprThen (e), null), 
				  absGoto (condExprElse (e), null));

    addStmt (v);
    if (head != null) head.setNext (v);
    return v;
  }

  private PCond absCond (Expr e)
  {
    // XXX CXMLParser doesn't create these yet
    if (e != null && NondetOp.isNondet (e))
      return new PCond (e, e);

    return new PCond (e, efac.op (BiLatticeOp.IBOT));
  }


  private SwitchBranchPStmt absSwitchExpr (Expr e, PStmt head)
  {
    List<PSwitchCase> cases = new ArrayList (numCases (e));
    
    // -- abstract each case as a PSwitchCase, build the default case label
    Expr defaultCaseLabel = efac.op (BoolOp.FALSE);
    int defaultCaseIndex = -1;
    for (int c = 0; c < numCases (e); c++)
    {
      if (isDefaultCase (caseLow (e, c)))
      {
	defaultCaseIndex = c;
	continue;
      }
      Expr caseLabel = explicitCaseLabel (
	    switchCond (e), caseLow (e, c), caseHigh (e, c));
      defaultCaseLabel =
	efac.op (BoolOp.OR).binApply (defaultCaseLabel, caseLabel);
      cases.add (
	new PSwitchCase (absCond (caseLabel),
	  absGoto (caseGoto (e, c), null)));
    }

    // -- create a default PSwitchCase
    defaultCaseLabel = efac.op (BoolOp.NOT).unaryApply (defaultCaseLabel);
    cases.add (new PSwitchCase (absCond (defaultCaseLabel),
	  absGoto (caseGoto (e, defaultCaseIndex), null)));

    SwitchBranchPStmt v = 
      new SwitchBranchPStmt (def, e, cases);
    
    addStmt (v);
    if (head != null) head.setNext (v);
    return v;
  }

  private Expr explicitCaseLabel (Expr switchCond, Expr low, Expr high)
  {
    ExprFactory fac = switchCond.getFactory ();
    Expr switchVar = switchCond.arg (0);

    if (high == null)
      return fac.op (ComparisonOp.EQ).binApply (switchVar, low);

    return fac.op (BoolOp.AND).binApply (
	fac.op (ComparisonOp.LEQ).binApply (low, switchVar),
	fac.op (ComparisonOp.LEQ).binApply (switchVar, high));
  }

  private PStmt absCallExpr (Expr e, PStmt head)
  {
    return absCallExpr (e, null, head);
  }
  
  /**
   * Generates an abstraction of a function call {@code e}, which
   * potentially returns a result into variable {@code retVar}. A call
   * to an undefined function is abstracted by a skip. 
   *
   * @param e a function call expression
   * @param retVar a variable expresion for the return. null if there
   * is no return variable
   * @param head a <code>PStmt</code> value
   * @return a <code>PStmt</code> value
   */
  private PStmt absCallExpr (Expr e, Expr retVar, PStmt head)
  {

    PFunctionDef callee = getPFunctionDef (getCallee (e));

    // -- if the function does not have a body, replace with a skip
    if (callee == null)
      {
	PStmt v = new SkipPStmt (def);
	// XXX this is duplication of code (see end of this function)
	// but it is the easiest solution for now :(
	if (head != null) head.setNext (v);
	return v;
      }


    FunctionCallPStmt call = new FunctionCallPStmt (def, e);

    if (retVar != null)
      call.setReturnVar (retVar);

    FunctionCallPrologue prologue = new FunctionCallPrologue (call);
    FunctionCallEpilogue epilogue = new FunctionCallEpilogue (call);

    call.setLogues (prologue);
    call.setCallee (callee);
    prologue.setDest (callee.getEntry ());
    prologue.setNext (epilogue);

    int callIndex = callee.addReturnSite (epilogue);

    call.setCallIndex (callIndex);
    addStmt (prologue);
    addLabel (prologue, callee.callSiteLabel (callIndex));
    addStmt (epilogue);
    addLabel (epilogue, callee.returnSiteLabel (callIndex));


    if (head != null) head.setNext (prologue);
    
    return epilogue;

  }


  /**********************************************************************/
  /* SUPPORT FUNCTIONS */
  /**********************************************************************/

  /**
   * Sets the Dest of {@code pGoto} based on the C goto it represents.
   *
   * @param pGoto a <code>GotoPStmt</code> value
   * @return a <code>GotoPStmt</code> value
   */
  private GotoPStmt resolveGoto (GotoPStmt pGoto)
  {
    // -- extract the C goto expression corresponding to pGoto
    Expr cGoto = pGoto.getSourceExpr ();
    
    // -- assume there are no synthetic gotos
    assert CStmtOp.isGoto (pGoto.getSourceExpr ());

    // -- use the CFG to resolve the label to a BB, and then look the
    // -- corresponding PStmt in bbEntry
    pGoto.setDest 
      (bbEntry [cfg.bbForLabel (CStmtOp.gotoDest (cGoto)).getIndex ()]);

    return pGoto;
  }
    

  /**
   * Provides a label for a UID. Not really needed for anything.
   *
   * @param uid an <code>int</code> value
   * @return a <code>String</code> value
   */
  private String labelFor (int uid)
  {
    // XXX using uppercase L to avoid clashes with PProgram.getStmt's lookup
    // procedure
    return "LabelUID_" + uid;
  }


  /**
   * Adds statement to {@code statementList} and associates ID with
   * the statement
   *
   * @param p a <code>PStmt</code> with an unset id
   */
  private void addStmt (PStmt p)
  {
    // -- p must not have an id already
    assert p.getId () == -1;
    p.setId (statementList.size ());
    statementList.add (p);
  }

  /**
   * Attaches a label to a statement. A statement is directly
   * associated with the first label assigned to it. A label is always
   * added to labelMap as well.
   *
   * @param p a <code>PStmt</code> value
   * @param label a <code>String</code> value
   */
  private void addLabel (PStmt p, String label)
  {
    
    assert p != null;
    if (label == null) return;
    
    
    labelMap.put (label, p);
    if (!p.hasLabel ()) p.setLabel (label);
  }

  /**
   * Attaches a label described by a label decl to a statement
   *
   * @param p a <code>PStmt</code> value
   * @param labelDecl an <code>Expr</code> value
   */
  private void addLabel (PStmt p, Expr labelDecl)
  {
    // -- first,  add the name 
    addLabel (p, declName (labelDecl));
    // -- second, add label based on UID 
    addLabel (p, labelFor (labelUID (labelDecl)));
  }
}
