package edu.toronto.cs.yasm.abstractor;


import edu.toronto.cs.yasm.refiner.*;
import edu.toronto.cs.yasm.pprogram.*;
import edu.toronto.cs.expr.*;
import edu.toronto.cs.yasm.YasmStatistics;
import edu.toronto.cs.yasm.wp.*;

import edu.cmu.sei.pacc.absdom.*;
import edu.toronto.cs.algebra.TwoValAlgebra;
import edu.toronto.cs.mvset.*;

import java.util.*;

import static edu.toronto.cs.expr.BoolOp.lnot;


/**
 * Describe class LocalCBJPredicateRefinerVisitor here.
 *
 *
 * Created: Sun Oct 30 11:39:44 2005
 *
 * @author <a href="mailto:owei@age.cs">Ou Wei</a>
 * @version 1.0
 */
public class LocalCBJPredicateRefinerVisitor 
  extends LocalPredicateRefinerVisitor
{

  public void makeNew (Expr e)
  {
    // Move e to the set of new predicates so that we re-abstract it, 
    // and remove all assignments to e
    pm.makeNew (e);
    pProgram.accept (new AsmtRemoverVisitor (e), null);
  }

  // -- true if refiner should find which combinations of predicates
  // -- are inconsistent
  boolean doInconsitent;
  // -- true if refiner should use the theorem prover to approximate
  // -- statements
  boolean doSearch;



  
  // -- cache for refinement
  Map<Expr,Expr[]>  cache;
  // -- all predicates for refining pre-conditions in the currently
  // -- visited function
  List<Expr> allPrePredsList;
  // -- all predicates for refining prost-conditions in the currently
  // -- visited function
  List<Expr> allPostPredsList;
  
  MvSetAbstractDomain simplifier;

  
  /**
   * Creates a new <code>LocalCBJPredicateRefinerVisitor</code> instance.
   *
   * @see {@link LocalCBJPredicateRefinerVisitor(ExprFactory, PProgram, boolean, boolean}
   */
  public LocalCBJPredicateRefinerVisitor (ExprFactory _fac, 
					  PProgram _pProgram)
  {
    super (_fac, _pProgram);

    simplifier = 
      new MvSetAbstractDomain 
      (new MDDMvSetFactory (new TwoValAlgebra (), 10), _fac);

    doInconsitent = true;
    doSearch = true;
  }

  /**
   * doInconsitent -- if true compute inconsistent cubes
   * doSearch -- if true do abstraction search
   *
   * @param _fac an <code>ExprFactory</code> value
   * @param _pProgram a <code>PProgram</code> value
   * @param _doInconsitent a <code>boolean</code> value
   * @param _doSearch a <code>boolean</code> value
   */
  public LocalCBJPredicateRefinerVisitor (ExprFactory _fac, 
					  PProgram _pProgram,
					  boolean _doInconsitent,
					  boolean _doSearch)
  {
    super (_fac,_pProgram);

    simplifier = 
      new MvSetAbstractDomain 
      (new MDDMvSetFactory (new TwoValAlgebra (), 10), _fac);

    doInconsitent = _doInconsitent;
    doSearch = _doSearch;
  }


  /**
   * refine a predicate program <code>p</code>
   * <p>
   * compute inconsistent (invariant for modelchecking
   * process); will also consider locality later.
   * <p>
   * @param p a <code>PProgram</code> value
   * @param o an <code>Object</code> value
   * @return an <code>Object</code> value
   */
  public Object visitPProgram (PProgram p, Object o)
  {

    assert p.isDeclsRefined () == true
      : "Decls not refined; cannot refine PProgram";

    // -- this condition is true iff there are no new predicates
    // -- in this round of refinement, but this is not the first
    // -- time the pprogram is refined.
    if (pm.getNewPredicates ().isEmpty () && 
	!pm.getPredicates ().isEmpty ()) return p;

    // -- predicates is empty when this is the first run of this refiner
    if (pm.getPredicates ().isEmpty ())
      p.setInconsistent (fac.falseExpr ());

    // XXX This is wrong since it makes use of all predicates, 
    // XXX even the ones that are out of scope. 
    // XXX PProgram's inconsistent set should be computed only
    // XXX over all the global predicates!

    // ALL predicates 
    List<Expr> allPreds = pm.getAllPredicates ();

    //System.out.println("TOM Spider-");
    //System.out.println ("allPreds.isEmpty () == " + allPreds.isEmpty ());
    //System.out.println ("doInconsitent == " + doInconsitent);
  // -- computer invariant 
    if (!allPreds.isEmpty () && doInconsitent)
    {
      //System.out.println ("\tTOM Man");
      p.setInconsistent (cbjCondAbs (allPreds, 
				     tp.exprFactory ().falseExpr ())[1]);
      System.err.println ("INCONSISTENT: " + 
                          Fmt.disjuncts (p.getInconsistent ()));
    }

    // -- do refinement of the program
    return super.visitPProgram (p, o);
  }
 

  protected void setupPreds (PFunctionDef preF, PFunctionDef postF)
  {
    super.setupPreds (preF, postF);
    
    // -- extra stuff needed by this refiner

    
    // -- the list of all pre-condition predicates.  The order is
    // -- important, currently predicates come first, new predicates
    // -- come after
    allPrePredsList = new ArrayList<Expr>();
    allPrePredsList.addAll (prePreds);
    allPrePredsList.addAll (newPrePreds);

    // -- if pre- and post- functions are the same we don't need to
    // -- create a post-predicates list, but we do need to create the
    // -- cache for refinemnt. If pre- and post- functions are
    // -- different, no need for cache, but need the lsit of
    // -- post-predicates
    if (preF == postF)
      {
	allPostPredsList = allPrePredsList;
	// -- create a refinement cache for this function
	cache = new IdentityHashMap<Expr,Expr[]>();
      }
    else
      {
	allPostPredsList = new ArrayList<Expr>();
	allPostPredsList.addAll (postPreds);
	allPostPredsList.addAll (newPostPreds);
      }
  }
  
  protected void clearPreds ()
  {
    super.clearPreds ();
    
    // -- remove extra things we've created
    cache = null;
    allPrePredsList = null;
    allPostPredsList = null;
  }
  

  protected PFunctionDef refineFunctionDef (PFunctionDef fDef)
  {
    // -- need to compute inconsistent sets

    // -- compute syntactical representation of inconsistent states in
    // -- boolean programs w.r.t. to the predicates in current
    // -- functions


    // On first run of refiner, set inconsistent to false
    if (pm.getPredicates ().isEmpty ()) {
      fDef.setInconsistent (fac.falseExpr ());
    }

    if (doInconsitent)
    {
      fDef.setInconsistent (cbjCondAbs (allPrePredsList, 
					tp.exprFactory ().falseExpr ())[1]);
    }

    return fDef;
  }
  


  public Object visitPrllAsmt (PrllAsmtPStmt asmt, Object o)
  {

    // XXX This code is very similar to visitFunctionCallPrologue
    // XXX the two functions should be merged together.


    //List asmts = new ArrayList ();
    List<AsmtPStmt> asmts;
    int  start;
    int  startOfNewPredicates;

    // -- true if we found a new non-deterministically affected predicate
    boolean newNonDet = false;

    List nondetPredicates = new ArrayList<Expr> ();
    

    // -- compute the starting postion of new post-predicates
    startOfNewPredicates = postPreds.size ();

    /**
     **
     ** if doSearch is true, we go over all new and old predicates
     ** and compute the new abstraction of each statement. Therefore, 
     ** we start at the first predicate and an empty list

     ** If we do not do search, then we only need to compute new information
     ** for new predicates we have. Thus we start with previous computation
     ** which is stored in 'astm.getAsmts ()', and only process new predicates
     ** that start at the index 'startOfNewPredicates'
     **/
    if (doSearch)
      {
	start = 0;
	asmts = new ArrayList<AsmtPStmt> ();
      }
    else
      {
	start = startOfNewPredicates;
	asmts = (ArrayList<AsmtPStmt>) asmt.getAsmts ();
      }

    for (int i = start; i < allPostPredsList.size (); i++)
      {
	Expr curPost = allPostPredsList.get (i);
	Expr wpExpr = asmt.getWPComputer ().computeWP (curPost);

	assert wpExpr != null : "null wp for: " + asmt;

	Expr[] result = null;

	// -- if processing a new post-predicate, and it 
	// -- is non-deterministically affected by the assignment
	if (NondetOp.hasNondet (wpExpr))
	  {
	    // -- check if curPost is completelly non-deterministic at
	    // -- the end
	    if (wpExpr.varSet (new HashSet ()).isEmpty ())
	      {
		System.err.println ("Got predicate: " + curPost);
		System.out.println (" with wp: " + wpExpr);
		System.out.println ("it should change non-deterministically");


		if (i >= startOfNewPredicates)
		  {
		    // -- store all the predicates that are
		    // -- non-deterministically assigned
		    asmt.addNondetPred (curPost);
		    // -- set the flag to indicate that non-deterministic
		    // -- predicates set has changed
		    newNonDet = true;
		  }

		/**
		   we know that current statement changes curPost
		   non-deterministically
		   
		   add 
		    curPost := {(t,f), (f,t)}
                   to current parallel assignment
		   		   
		 */
		NDAsmtPStmt ndupdate = new NDAsmtPStmt (curPost);
		ndupdate.addUpdate (fac.trueExpr (), fac.falseExpr ());
		ndupdate.addUpdate (fac.falseExpr (), fac.trueExpr ());
		
		// -- now we need to add this update into the parallel 
		// -- assignment we are building
		// -- and go to the next predicate
		asmts.add (ndupdate);
		continue;
	      }
	    else
	      {
		// -- this predicate is afffected by a 
		// -- non-deterministic assignment
		
		// -- at the moment, just say we know nothing about it
		result = new Expr[] { fac.falseExpr (), fac.falseExpr () };
	      }
	  }

	// -- try heuristics for quick abstraction
	if (result == null)
	  result = quickCondAbs (wpExpr);

	// -- check if there is something in the cache we can use
	if (result == null)
	  result = cache.get (wpExpr);
	

	if (result == null && doSearch)
	  { 
            System.out.println ("TOM @@@@@@ approximating " + wpExpr);
	    result = cbjCondAbs (allPrePredsList, wpExpr);
	    cache.put (wpExpr, result);
	  }
	else if (result == null && !doSearch)
	  result = new Expr[] { fac.falseExpr (), fac.falseExpr () };
	else
	  ; //System.err.println ("CACHE HIT");

	asmts.add (new AsmtPStmt (curPost, result [1], result [0]));
      }

    asmt.setAsmts (asmts);

    // -- check inconsistency on non-deterministic predicates
    if (newNonDet)
      asmt.setInconsistent (cbjCondAbs (asmt.getNondetPreds (),
					tp.exprFactory ().falseExpr ())[1]);

    return asmt;
  }


  /**
   * This is similar to <code>visitPrllAsmt</code>, but we need consider
   * the different predicates from the call funcation and the function being
   * called.
   *
   * @param fpro a <code>FunctionCallPrologue</code> value
   * @param o an <code>Object</code> value
   * @return an <code>Object</code> value
   */
  public Object visitFunctionCallPrologue (FunctionCallPrologue fPro, 
					   Object o)
  {

    List<AsmtPStmt> asmts;
    int start;

    /**
     ** The next block of code does the following:
     ** if doSearch is true, we go over all new and old predicates
     ** and compute the new abstraction of each statement. Therefore, 
     ** we start at the first predicate and an empty list.

     ** If we do not do search, then we only need to compute new information
     ** for new predicates we have. Thus we start with previous computation
     ** which is stored in 'astm.getAsmts ()', and only process new predicates
     ** 
     **/
    if (doSearch)
      {
	start = 0;
	asmts = new ArrayList<AsmtPStmt> ();
      }
    else
      {
	// -- new predicates start right after the current ones
	start = postPreds.size ();
	asmts = (ArrayList<AsmtPStmt>) fPro.getAsmts ();
      }

    for (int i = start; i < allPostPredsList.size (); i++)
      {


	// -- the post-predicate we are abstracting for in this iterator
	Expr curPost = allPostPredsList.get (i);
	// -- the result of the abstraction
	Expr[] result = null;


	// -- check if this is a self-call statement	
	if (fPro.getParent () == fPro.getCall ().getFunctionDef ())
	  {
	    // -- if the prologue corresponds to a self-call statement,
	    // -- then we need deal with it specially suppose the function
	    // -- is as following
	    // 
	    //    int f (int fp_1, int fp_2, .. int fp_n) 
	    //    {
	    //        int l_1;
	    //        int l_2;
	    //        ...
	    //        int l_m;
	    //        ...
	    //        ...
	    //    }
	    //
	    // -- for each predicate p, we consider the following three cases
	    // -- (1) p defined over global variables "only", then
	    // --     p is not changed, i.e. p = (p, !p)
	    // -- (2) the variables in p contains local variables 
	    //        in {l_i}, then we know nothing about its value, 
	    //        i.e., p = {f, f}.
	    // -- (3) p defined over formal parameters {fp_i} "only", then
	    // --     p is abstraced as usual
	    //


	    // -- case (1)
	    if (pm.getPredicateScope (curPost).isEmpty ())
	      result = new Expr[] {lnot (curPost), curPost};
	    else
	      {
		// -- case (2)
		if (!fPro.getCall ().getFunctionDef ().getFormalParams ().
		    containsAll (curPost.varSet (new HashSet ())))
		  result = new Expr[] { fac.falseExpr (), fac.falseExpr () };
		else
		  ; // -- case (3)
	      }
	  }
	

	// -- compute pre-condition to approximate
	Expr wpExpr =  fPro.getWPComputer ().computeWP (curPost);
	  
        assert wpExpr != null : "null wp for: " +  fPro + ", "  + curPost;

	// -- try and do a quick abstraction first
	result = quickCondAbs (wpExpr);

	// -- if failed to abstract so far, and have to do expensive
	// -- abstraction, and there are predicates to abstract with
	if (result == null && doSearch && !allPrePredsMap.isEmpty ())
	  result = cbjCondAbs (allPrePredsList, wpExpr);
	
	// -- if we still did not find a suitable abstraction, 
	// -- use the most general one
	if (result == null)
	  result = new Expr[] { fac.falseExpr (), fac.falseExpr () };

	// -- by now, result is definitelly been set
	assert result != null;
	
	asmts.add (new AsmtPStmt (curPost, result [1], result [0]));
      }

    fPro.setAsmts (asmts);
    
    return fPro;
  }


  /**
   * A quick abstraction of {@code expr}
   *
   * @param expr an <code>Expr</code> value
   * @return an <code>Expr[]</code> value
   */
  protected Expr[] quickCondAbs (Expr expr)
  {
    // XXX For now just extract the true part out
    // XXX this may be unsound in general
    expr = ChoiceOp.lgetT (expr);
    
    // -- check if we can approximate the weakest pre-condition
    // -- exactly using one of our pre-predicates
    if (allPrePredsMap.containsKey (expr))
      return new Expr[] {lnot (expr), expr};
    if (tp.quickQuery (expr))
      //System.err.println ("QUICKTEST2");
      return new Expr[] { fac.falseExpr (), fac.trueExpr () };
    if (tp.quickQuery (ChoiceOp.lgetF (expr)))
      //System.err.println ("QUICKTEST3");
      return new Expr[] { fac.trueExpr (), fac.falseExpr () };
    
    return null;
  }

  /* Queries that trigger bugs in CVCL */
  /*String [] badQueries = {
    "((0 - main::j_VERISEC_ij2_line14_0_s) <= (0 - main::i_VERISEC_ij2_line14_0_s))"
    };*/

  /**
   * Abstracts {@code expr} using {@code preds} via CBJ strategy
   *
   * @param expr an <code>Expr</code> value
   * @return an <code>Expr[]</code> value
   */
  protected Expr[] cbjCondAbs (List<Expr> preds, Expr expr)
  {
    // XXX This is a hack. CVCL has a bug (looks like a memory leak),
    // which causes it to crash on some queries. Uncomment this, and
    // the badQueries array above, to work around this. Add bad
    // queries as you come across them. If a query is bad, return
    // (false, false). This is sound, but imprecise.
    /*String queryString = expr.toString ();
    for (int i = 0; i < badQueries.length; i++) {
      if (queryString.equals (badQueries[i])) {
        return new Expr[] {expr.getFactory ().falseExpr (),
                           expr.getFactory ().falseExpr ()};
      }
      }*/

    TPCBJPredicateAbstractor cbj = 
      new NumericTPCBJPredicateAbstractor (preds, expr, tp);
    
    if (stats != null)
      stats.startPredicateAbstraction ();
    cbj.recSearch (1);
    if (stats != null)
      stats.stopPredicateAbstraction ();

    //System.err.println("TExpr before simplification: \n\t" + cbj.getTExpr ());
    //System.err.println("TExpr after simplification: \n\t" + simplifier.gamma (simplifier.alpha (cbj.getTExpr () )));    

    return new Expr[] {simplifier.gamma (simplifier.alpha (cbj.getFExpr ())), 
                       simplifier.gamma (simplifier.alpha (cbj.getTExpr ()))};
    
  }
  
  


  /**
   * check if <code>setA</code> is subset of <code>setB</code>
   *
   * @param setA a <code>Set</code> value
   * @param setB a <code>Set</code> value
   * @return a <code>boolean</code> value
   */
  private static final boolean isSubsetOf (Set setA, Set setB)
  {
    return setB.containsAll (setA);    
  }

  // -- methods for CBJ refiner 

  /**
   * @depricated Not used
   */
  private CBJRefinerInfo getRefinerInfo (PrllAsmtPStmt asmt)
  {
    CBJRefinerInfo info = (CBJRefinerInfo) asmt.getRefinerInfo ();
    if (info == null)
    {
      info = new CBJRefinerInfo ();
      asmt.setRefinerInfo (info);
    }
    return info;
  }

  /**
   * @depricated Not used
   */
  public MemoryModel getRegularMemoryModel (PrllAsmtPStmt asmt)
  {
    return getRefinerInfo (asmt).memoryModel;
  }

  /**
   * @depricated Not used
   */
  public CVCLMemoryModel getMemoryModel (PrllAsmtPStmt asmt)
  {
    return getRefinerInfo (asmt).cvclMemoryModel;
  }

  /**
   * @depricated Not used
   */
  public WPComputer getWPComputer (PrllAsmtPStmt astmt)
  {
    return getRefinerInfo (astmt).wp;
  }

  /**
   * @depricated Not used
   */
  public class CBJRefinerInfo
  {
    MemoryModel memoryModel;
    CVCLMemoryModel cvclMemoryModel;
    WPComputer wp;

    // -- more stuff
  }  

}
