package edu.toronto.cs.yasm.abstractor;

import edu.toronto.cs.cparser.block.*;
import edu.toronto.cs.cparser.*;
import edu.toronto.cs.expr.*;
import edu.toronto.cs.yasm.*;
import edu.toronto.cs.yasm.pprogram.*;

import edu.toronto.cs.tp.*;
import edu.toronto.cs.tp.cvcl.ValidityChecker;

import edu.toronto.cs.boolpg.abstraction.StmtAbstraction;

import edu.toronto.cs.yasm.wp.*;

import antlr.*;

import static edu.toronto.cs.expr.CVarDeclOp.*;

import java.util.*;
import java.io.*;

import edu.toronto.cs.yasm.abspatterns.NullposCollectorVisitor;
import static edu.toronto.cs.expr.ComparisonOp.*;
import static edu.toronto.cs.expr.ArrayOp.*;

/**
 * Implement LocalPredicateRefinerVisitor based on 
 * PProgramPredicateManager
 *
 *
 * Created: Thu Oct 27 21:39:52 2005
 *
 * @author <a href="mailto:owei@age.cs">Ou Wei</a>
 * @version 1.0
 */
public class LocalPredicateRefinerVisitor extends BasePProgramVisitor  
  implements Refiner
{
  public edu.toronto.cs.yasm.YasmStatistics stats;


  ExprFactory fac;
  // -- our theorem prover
  ITheoremProver tp;

  // -- a condition that a theorem prover can never prove or disprove
  // -- it is used to refine conditions that we cannot represent in our
  // -- internal representation
  Expr unknownCond;


  // -- a program being refined
  PProgram pProgram;
  
  
  // -- a predicate manager that keeps track of our predicates
  PProgramPredicateManager pm;

  // -- the following variables are used during refinement
  // -- and are updated in visitFunctionDef

  // -- new predicates available for abstracting pre-conditions
  List<Expr> newPrePreds;
  // -- predicates currently used to abstract pre-condtions
  List<Expr> prePreds;
  // -- a set of all predicates available for abstracting pre-conditions
  Map<Expr,Expr> allPrePredsMap;
  // -- new predicates available for abstracting post-conditions
  List<Expr> postPreds;
  // -- predicates curerently used to abstract post-conditions
  List<Expr> newPostPreds;


  public LocalPredicateRefinerVisitor (ExprFactory _fac, PProgram _pProgram)
  {    

    this (_fac, _pProgram, new ArrayList ());

    if (edu.toronto.cs.yasm.YasrApp.DEBUG)
      {
	YasmApp.out.println ("--- Start LocalPredicateRefinerVisitor ---");
      }

  }

  private List<Expr> getInvariants ()
  {
    // For now, the only invariants we have to worry about are the 
    // nullpos invariants.
    NullposCollectorVisitor ncv = new NullposCollectorVisitor ();
    ncv.visitPProgram (pProgram, null);
    List<Expr> nullposes = ncv.getNullposes ();
    List<Expr> invariants = new ArrayList<Expr> ();
    for (Expr A_nullpos : nullposes) {
      invariants.add (eq (sel (ncv.getArray (A_nullpos), A_nullpos), 0));
      invariants.add (geq (A_nullpos, 0));
    }
    return invariants;
  }

  public LocalPredicateRefinerVisitor (ExprFactory _fac, 
				       PProgram _pProgram,
				       List _newPredicates)
  {
   
    fac = _fac;

    pProgram = _pProgram;
    pm = pProgram.getPProgramPredicateManager ();

    addToNewPredicates (_newPredicates);

    List<Expr> invariants = getInvariants ();

    //System.out.println ("invariants are:" + invariants);
    //assert (false);

    //tp = new JCVCLTheoremProver (fac);
    //tp = new CLCVCLTheoremProver (fac, true);
    tp = new CTheoremProver (fac, 
                             new CLCVCLTheoremProver (fac, true), 
                             _pProgram.getSymbolTable (),
                             invariants);

  }

  // -- ***** methods for implementing Refiner interface *****

  /**
   * Refines a predicate program with new predicates
   *
   * @param p an input predicate program
   * @return a refined version of <code>p</code>
   */
  public PProgram doProgramRefine (PProgram p)
  {
    
    assert p == pProgram 
      : "Refiner : pprogram to be refined is not consistent!";
    
    return (PProgram) p.accept (this, null);
  }
  

  // XXX should be moved out of all refiners
  public void doDeclsRefine (PProgram p)
  {

    assert p == pProgram 
      : "Refiner : pprogram to be refined is not consistent!";


    // -- process global declarations
    for (Iterator it = p.getGlobalDecls ().values ().iterator ();
        it.hasNext ();)
      doDeclRefine ((PDecl) it.next ());

    // -- process local declarations
    for (Iterator it = p.getFunctionDefs ().values ().iterator ();
        it.hasNext ();)
      doFunctionDeclsRefine ((PFunctionDef) it.next ());

    // -- crreate an unknown variable not used anywhere else
    Expr unknownVar = fac.var ("__UNKNOWN__");

    // -- tell our theorem prover that this is an integer variable
    tp.declare (unknownVar, fac.op (SimpleTypeOp.INT));

    // -- make unknown if-condition -- a condition that can never
    // -- be shown to be neither true nor false
    unknownCond = fac.op (ComparisonOp.EQ).binApply
      (unknownVar, fac.intExpr (0));

    p.setDeclsRefined (true);  
  }

  // XXX should be moved out of all refiners
  public void doFunctionDeclsRefine (PFunctionDef f)
  {
    // -- process local declarations
    for (Iterator it = f.getLocalDecls ().values ().iterator ();
        it.hasNext ();)
      doDeclRefine ((PDecl) it.next ());

    // -- declare return index
    tp.declare 
      (f.getReturnIndexVar (), tp.exprFactory ().op (SimpleTypeOp.INT));

    Expr returnType;

    // -- create return value Expr
    try
    {
      returnType = f.getReturnType (tp);
    }
    catch (RecognitionException ex)
    {
      ex.printStackTrace ();
      throw new RuntimeException
        ("Error parsing return type for function " + f.getFunctionName ());
    }

    // -- declare return value
    tp.declare (f.getReturnValueVar (),
	tp.exprFactory ().op (SimpleTypeOp.INT));
  }


  // XXX should be moved out of all refiners
  /** 
   * Registers a variable v corresponding to <code>pDecl</code> with the vc
   * and also stores a reference to v in pDecl (to prevent CVCL from garbage
   * collecting v).
   * 
   * @param pDecl the declaration to process.
   */
  public void doDeclRefine (PDecl pDecl)
  {
    pDecl.declare (tp);
  }

  /**
   * Sets a stats object to collect statistics
   *
   * @param stats an <code>YasmStatistics</code> value
   */
  public void setStats (YasmStatistics _stats)
  {    
    stats = _stats;    
  }


  /**
   * return predicates used for refinement
   *
   * @return a <code>List</code> value
   */
  public List getPredicates ()
  {
    return pm.getPredicates ();
  }

  /**
   * return predicates used for refinement, in @code{Pred}
   *
   * @return a <code>List</code> value
   */
  public List getPreds ()
  {
    return pm.getPreds ();
  }


  /**
   * retunre new predicates used for refinment
   *
   * @return a <code>List</code> value
   */
  public List<Expr> getNewPredicates ()
  {
    return pm.getNewPredicates ();
  }

  /**
   * retunre new predicates used for refinment, as @code{Pred}s
   *
   * @return a <code>List</code> value
   */
  public List<Pred> getNewPreds ()
  {
    return pm.getNewPreds ();
  }


  public ExprFactory exprFactory ()
  {
    return tp.exprFactory ();
  }
  

  /**
   * Returns a copy of {@code ITheoremProver} used by this refiner
   *
   * @return an instance of <code>ITheoremProver</code> 
   */
  public ITheoremProver getTP ()
  {
    return tp;
  }
  
  public ValidityChecker getVC ()
  {
    //return tp.getValidityChecker ();
    assert false;
    return null;
  }

  public CVCLExprConverter getCVCLConverter ()
  {
    //return tp.getExprConverter ();
    assert false;
    return null;
  }

  /**
   * add new predicates to predicate list
   *
   * @param l a <code>List</code> value
   */
  public void addToNewPredicates (List l)
  {
    for (Iterator it = l.iterator (); it.hasNext ();) {
      addToNewPredicates ((Pred) it.next());
    }
  }


  /**
   * add new predicates to predicate list - add new predicates
   * to pProgramPredicateManager
   *
   * @param p a <code>Pred</code> value : a new predicate
   * 
   */
  public void addToNewPredicates (Pred p)
  {
    pm.newPred (p);	  
  }


  /** 
   * @deprecated Use {@link #addToNewPredicates(List)}
   */
  public void addPredicates (List l)
  {
    addToNewPredicates (l);
  }

  /** 
   * @deprecated Use {@link #addToNewPredicates(Expr)}
   */
  public void addPredicate (Pred p)
  {
    addToNewPredicates (p);
  }


  /**
   * check if there are new predicates for refinement
   *
   * @return a <code>boolean</code> value
   */
  public boolean hasNewPredicates ()
  {
    return !(getNewPredicates ().isEmpty ());
  }



  // XXX should be moved out of all refiners
  /** 
   * Parses stream read from Reader; each predicate parsed is added to
   * the list of predicates.
   * 
   * @param in Source of predicates.
   */
  public void readPredicates (Reader in) 
    throws RecognitionException, TokenStreamException 
    {
      CILLexer lexer = new CILLexer (in);
      lexer.setTokenObjectClass (CToken.class.getName ());

      PredicateParser parser = new PredicateParser (lexer);
      parser.setASTNodeClass (Block.class.getName ());
      Block.setTokenVocabulary (CILTokenTypes.class.getName ());

      parser.parse();

      Block cPred = (Block) parser.getAST ();
      while (cPred != null)
      {
        /*
           System.err.println ("attempting to parse predicate:");
           TNode.setTokenVocabulary (CILTokenTypes.class.getName ());
           TNode.printTree (cPred);
         */

        Expr pred = SymbolicExecutorUtil.newSymbolicExecutor (fac).pureExpr
                    (cPred, new MemoryModel ());
        System.err.println ("Adding predicate: " + pred);
        // XXX just assume it's from a hint
        addToNewPredicates (new Pred (pred, PredicateSource.HINT));

        cPred = (Block) cPred.getNextSibling ();
      }
    }


  /**
   * return CVCL Memory Model
   *
   * @param asmt a <code>PrllAsmtPStmt</code> value
   * @return a <code>CVCLMemoryModel</code> value
   * @depricated Not used anymore.
   */
  public CVCLMemoryModel getMemoryModel (PrllAsmtPStmt asmt)
  {
    return (CVCLMemoryModel) asmt.getRefinerInfo ();
  }


  /**
   * retrun WP computer used for refinement
   *
   * @param asmt a <code>PrllAsmtPStmt</code> value
   * @return a <code>WPComputer</code> value
   * @depricated Not used anymore.
   */
  public WPComputer getWPComputer (PrllAsmtPStmt asmt)
  {
    return getPredRefinerInfo (asmt).wp;
  }

  /**
   * @depricated Not used anymore.
   */
  private PredicateRefinerInfo getPredRefinerInfo (PrllAsmtPStmt asmt)
  {
    PredicateRefinerInfo info = (PredicateRefinerInfo) asmt.getRefinerInfo ();
    if (info == null)
    {
      info = new PredicateRefinerInfo ();
      asmt.setRefinerInfo (info);
    }
    return info;
  }

  /**
   * create Memory Model
   *
   * @param s a <code>PrllAsmtPStmt</code> value
   * @return a <code>MemoryModel</code> value
   * @depricated Not used anymore.
   */
  public MemoryModel computeRegularMemoryModel (PrllAsmtPStmt s)
  {
    assert false : "Use getWPComputer!";

    return null;

  }

  /**
   * create CVCL memory model
   *
   * @param s a <code>PrllAsmtPStmt</code> value
   * @return a <code>CVCLMemoryModel</code> value
   * @depricated Not used anymore.   
   */
  public CVCLMemoryModel computeMemoryModel (PrllAsmtPStmt s)
  {
    assert false : "Use getWPComputer!";
    return null;
  }


  /**
   * @depricated Not used anymore.
   */
  class PredicateRefinerInfo 
  {
    CVCLMemoryModel mmModel;
    WPComputer wp;
  }



  // -- *****  methods from BasePProgramVisitor *****



  /**
   * refine a predicate program <code>p</code>
   * <p>
   * the declarations of the program should have been 
   * "refined" by calling <code>doDeclsRefine</code>, 
   * i.e., all the variables have been registered in
   * the Theorem Prover (validity checker)
   * <p>
   * when refinement is finished, we set new predicates
   * in PProgramPredicateManager to old.
   * <p>
   * @param p a <code>PProgram</code> value : a predicate program
   * @param o an <code>Object</code> value
   * @return an <code>Object</code> value
   */
  public Object visitPProgram (PProgram p, Object o)
  {

    assert p.isDeclsRefined () == true
      : "Decls not refined; cannot refine PProgram";

    if (edu.toronto.cs.yasm.YasrApp.DEBUG)
      dumpState (YasmApp.err);

    // -- refine functions
    for (Object v : p.getFunctionDefs ().values ())
      ((PFunctionDef) v).accept (this, o);

    // -- add new predicates to PProgram
    for (Expr e : getNewPredicates ()) p.getPredTable ().newPredicate (e);

    // -- add new predicates to Refiner (this)
    pm.merge ();

    return p;

  }
 
 
  /**
   * refine the function refered to by <code>fdef</code>
   * <p>
   * we pass the name of the function to refiner. so when
   * we refine a statement of the function, we only cosider
   * the predicates accessible by this function
   * <p>
   * @param fdef a <code>PFunctionDef</code> value : the 
   *             function to be refined
   * @param o an <code>Object</code> value 
   * @return an <code>Object</code> value
   */
  public Object visitFunctionDef (PFunctionDef fdef, Object o)
  {
    // -- first refine function calls, 
    // -- note that we have to do this even if 
    // -- there where no new predicates for this function
    refineCalls (fdef);
    
    // -- if there are no new predicates here, nothing to refine
    if (pm.getFuncNewPredicates (fdef.getFunctionName ()).isEmpty ()) return fdef;

    // -- setup predicates for refinement
    setupPreds (fdef);    
    
    // -- do any function-def level refinement. 
    fdef = refineFunctionDef (fdef);
    
    // -- refine the body of the function
    refinePStmtBlock(fdef.getHead (), fdef.getFunctionName ());

    // -- get rid of what we have created
    clearPreds ();
    
    return fdef;
  }


  /**
   * refine <code>IfPStmt</code>
   * <p>
   * considering predicate locality w.r.t. the function
   * referred to by <code>o</code>
   * <p>
   * @param s an <code>IfPStmt</code> value
   * @param funcName an <code>Object</code> value : the name of the function
   *          <code>e</code> belongs to
   * @return an <code>Object</code> value
   */
  public Object visitIf (IfPStmt s, Object funcName)
  {
 
    // System.err.println (">>> doIfRefine " + s.getLabel ());
    
    // - refine the condtion of the IfPStmt, passing the function name
    refinePCond(s.getCond (), funcName);

    // -- refine the ThenStmt, passing the function name 
    refinePStmtBlock (s.getThenStmt (), funcName);

    // -- refine the ElseStmt, passing the function name 
    refinePStmtBlock (s.getElseStmt (), funcName);
    
    return s;
    
  }


  // ----------------------------------------------------------------
  // -------- Helper refiner methods --------------------------------
  // ----------------------------------------------------------------

  // -- returns true if every operator/operand in expr is known
  private boolean knownCond (Expr expr)
  {
    if (expr == null)
      return false;

    if (isCVarDecl (expr))
      return true;

    if (expr.op () instanceof VariableOp)
      return true;

    if (expr.op () instanceof RationalOp)
      return true;

    if (expr.op () instanceof JavaObjectOp)
      return true;

    if (expr.op () instanceof ArrayOp)
	 return true;

    if (expr.op () instanceof BoolOp ||  
        (expr.op () instanceof NumericOp &&
	 expr.op () != NumericOp.MOD) ||
        expr.op () instanceof ComparisonOp ||
	expr.op () instanceof CILIndirectionOp)
      {
	for (int i = 0; i < expr.arity (); i++)
	  if (!knownCond (expr.arg (i)))
	    return false;
	return true;
      }
    
    return false;  
  }
  

  /**
   * refine a condition expression refered to by <code>e</code> and
   * return the new, refined condtion
   * <p>
   * considering locality. only check predicates accessible 
   * by the current function <code>funcName</code>
   * <p>
   * @param e a <code>PCond</code> value
   * @param funcName an <code>Object</code> value : the name of the function
   *          <code>e</code> belongs to
   * @return a <code>PCond</code> value
   */
  private PCond refinePCond (PCond e, Object funcName)
  {


    // System.err.println ("doCondRefine: " + e);

    // -- check that condition is as refined as can be already
    // -- XXX currently, each condition can be refined at most once
    // -- XXX at which point it's value changed from IBOT to something 
    // -- XXX else. 
    if (e.getCond ().op () != BiLatticeOp.IBOT)
      return e;

    // -- NOTE: YasmApp.preRefine now sets cond's abstracted condition to the
    // if-condition itself

    try
      {
	Expr sourceExpr = e.getOrigCond ();

	// -- unable to convert to cvcl, e.g. contains an unknown operator
	if (!knownCond (e.getOrigCond ())) 
	  {
	    e.setOrigCond (unknownCond);
	    return e;
	  }
	
	sourceExpr = e.getOrigCond ();  

	// -- check if the condition is a constant
	Boolean result = tp.quickCheckTruth (sourceExpr);
		

	// -- result != null iff original condition is a constant
	if (result != null)
	  {
	    e.setCond 
	      (result == Boolean.TRUE ? fac.trueExpr () : fac.falseExpr ());
	    return e;
	  }

	// -- check if condition is negated and strip the negation
	boolean negated = sourceExpr.op () == BoolOp.NOT;
	Expr condExpr = negated ? sourceExpr.arg (0) : sourceExpr;

	// -- see if the condition is equivalent to one of our 
	// -- new  predicates whose scopes contains funcName	
	for (Expr pred : newPrePreds)
	  {
	    if (condExpr == pred /* || 
				    CVCLUtil.quickQuery (vc, 
				    vc.iffExpr (cvcl.toCVCL (condExpr),
				    cvcl.toCVCL (pred))) */)
	      {
		// -- got a new predicate equivalent to the condition
		e.setCond (negated ? 
			   fac.op (BoolOp.NOT).unaryApply (pred) : pred);
		return e;
	      }
	  }
      }
    catch (Exception ex)
      {
	ex.printStackTrace ();
	throw new RuntimeException (ex);
      }
    return e;

  }

  /**
   * refine the block of predicate statements starting from 
   * the statement <code>s</code>.
   * <p>
   * <code>funcName</code> is the name of the function which the statement
   * <code>s</code> belongs to; here we do refinement by considering
   * locality.
   * <p>
   * @param s a <code>PStmt</code> value : the statement to be refined
   * @param funcName an <code>Object</code> value : the name of the function
   *          <code>s</code> belongs to
   * @return an <code>Object</code> value
   */
  private Object refinePStmtBlock(PStmt s, Object funcName)
  {

    PStmt head = s;

    // -- refine all statements in the block, except for
    // -- function-call-prologue which are handled by refineCalls
    // -- method
    for (; s != null; s = s.getNext ())
      if (! (s instanceof FunctionCallPrologue))
	  s.accept (this, funcName);
    
    
    return head;
  }



  /**
   * Does any function-def level refinement. Typically overridden in
   * subclasses. Must return a refined version of the function definition.
   *
   * @param fdef a <code>PFunctionDef</code> value
   * @return a <code>PFunctionDef</code> value
   */
  protected PFunctionDef refineFunctionDef (PFunctionDef fdef)
  {
    return fdef;
  }
  
  /**
   * Refines all functions calls of {@code fdef}
   */
  protected void refineCalls (PFunctionDef fdef)
  {
    /*
      -- we need deal with the call statments for this function
      specially, because the call statement (actually,
      FunctionCallPrologue statement) is special in terms that its
      pre-variables and post-variables are different.
      
      Our strategy is only to refine the functions for which new
      predicates are discovered.
      
      Now suppose we have a FunctionCallPorlogue statement AB: A ->
      B, which corresponds to a call edge from A to B.  Let Pred_AB
      be the predicates accessible by both A and B. Let Pred_A and
      Pred_B the predicates accessible only by A and B respectively.
      
      If a new predicate pB \in Pred_B is discovered, we will only
      refine the function B, because AB is treated as a statement of
      function A, the prologue statement AB is missed!
    
      Similarly, when refining a FunctionCallPrologue statement, we need
      to deal with the pre-variables and post-variables separately.
    */

    // -- only need to do this for functions that are actually called
    // -- from somewhere
    if (fdef.getNumCallSites () <= 0) return ;
    

    // XXX we may want to sort call-sites based on where they 
    // XXX are located to avoid creating and destroying the predicate
    // XXX sets for refinement
    for (FunctionCallPrologue p : allCallSites (fdef))
      {
	setupPreds (p.getParent (), fdef);
	// -- check that there are some new predicates 
	// -- available for refining this call
	if (newPrePreds.isEmpty () && newPostPreds.isEmpty ()) continue;
	// -- refine
	p.accept (this, fdef.getFunctionName ());
	// -- get rid of the predicates we've created
	clearPreds ();
      }
    
  }
  

  private List<FunctionCallPrologue> allCallSites (PFunctionDef fDef)
  {
    PStmt fHead = fDef.getHead ();
    
    assert ( (fHead instanceof GotoPStmt) || 
	     (fHead instanceof ReturnSelectorPStmt))
      : "Refiner : type of returnSelector for function " +
      fDef.getFunctionName () + " is not correct!";

    
    // -- we are first going to gather all function 
    // -- calls (prologues) into this list
    List<FunctionCallPrologue> l = new LinkedList<FunctionCallPrologue>();

	  
    // -- key thing to remember: a prologue is a statement right 
    // -- before an epilogue, and an epilogue is where the 
    // -- return selector is pointing

    if (fHead instanceof GotoPStmt)    
      l.add ((FunctionCallPrologue)
	     ((FunctionCallEpilogue) fHead.getDest ()).
	     getCall ().getLogues ());
    else if (fHead instanceof ReturnSelectorPStmt)
      for (Object o : ((ReturnSelectorPStmt) fHead).getDests ())
	l.add ((FunctionCallPrologue)
	       ((FunctionCallEpilogue) o).getCall ().getLogues ());

    return l;
  }
  

  protected void setupPreds (PFunctionDef fDef)
  {
    setupPreds (fDef, fDef);
  }
  
  protected void setupPreds (PFunctionDef preF, PFunctionDef postF)
  {
    newPrePreds = pm.getFuncNewPredicates (preF.getFunctionName ());
    prePreds = pm.getFuncPredicates (preF.getFunctionName ());

    /*for (Expr e : newPrePreds)
      System.out.println ("@@@@@@@@@@@@@@@@@@@@@@@@@ new predicate: " + e);*/
    
    allPrePredsMap = new IdentityHashMap<Expr,Expr>();
    for (Expr e : newPrePreds) allPrePredsMap.put (e, e);
    for (Expr e : prePreds) allPrePredsMap.put (e, e);
    
    if (preF == postF || postF == null)
      {
	newPostPreds = newPrePreds;
	postPreds = prePreds;
      }
    else
      {
	newPostPreds = pm.getFuncNewPredicates (postF.getFunctionName ());
	postPreds = pm.getFuncPredicates (postF.getFunctionName ());
      }
  }
  protected void clearPreds ()
  {
    newPrePreds = null;
    prePreds = null;
    allPrePredsMap = null;
    newPostPreds = null;
    postPreds = null;
  }
  
  

  /**
   * Dumps current state of this class for debugging.
   *
   * @param out a <code>PrintWriter</code> value
   */
  protected void dumpState (PrintWriter out)
  {
    // -- print out predicates information

    out.println ("******  Predicates ***********************");

    out.println ("LPRV: Old predicates: \n" + 
		 getPredicates ());
    out.println ("LPRV: New predicates: \n" + 
		 getNewPredicates ());


    Map funcDefs = pProgram.getFunctionDefs ();
    for (Iterator it = funcDefs.values ().iterator (); it.hasNext (); )
      {
	PFunctionDef funcDef = (PFunctionDef) it.next ();
	String funcName = funcDef.getFunctionName ();

	out.println (funcName);

	out.println ("Old Predicates (" + funcName + ") :\n" +
		     pm.
		     getFuncPredicates (funcName));

	out.println ("New Predicates (" + funcName + ") :\n" +
		     pm.
		     getFuncNewPredicates (funcName));
      }
	
    out.println ("******************************************");
    out.flush ();
  }  
  
}

