package edu.toronto.cs.yasm.abstractor;

import java.util.*;

import edu.toronto.cs.yasm.pprogram.*;
import edu.toronto.cs.tp.*;
import edu.toronto.cs.expr.*;

import static edu.toronto.cs.expr.CILDeclarationOp.*;
import static edu.toronto.cs.expr.CILFunctionDefOp.*;

import static edu.toronto.cs.expr.CSimpleTypeOp.*;

/** 
 * Initializes a symbol table with the (global and local) declarations in a
 * {@link edu.toronto.cs.yasm.pprogram.PProgram} and special variables (such as
 * function return value); converts between CIL types and Expr (Simple and
 * Compound) types.
 * <p>
 * Helper class (all methods are static).
 *
 * @author Kelvin Ku (kelvin@cs.toronto.edu)
 * @version 
 */
public class SymbolTableInitializer
{

  /**
   * Used by {@code cilToExprType} to map type specifiers to simple types.
   */
  private final static IdentityHashMap<CILTypeSpecifierOp,SimpleTypeOp> 
    typeMap = new IdentityHashMap<CILTypeSpecifierOp,SimpleTypeOp> ();
  static {
    typeMap.put (CILTypeSpecifierOp.INT, SimpleTypeOp.INT);
    typeMap.put (CILTypeSpecifierOp.CHAR, SimpleTypeOp.CHAR);
  }

  /** 
   * Initializes {@code st} with the declarations in {@code p} and special
   * variables.
   * 
   * @param p 
   * @param st 
   * @param unknownAsInt if true, convert unknown types to int
   * @return 
   */
  public static SymbolTable init (PProgram p, SymbolTable st, boolean unknownAsInt)
  {
    ExprFactory fac = p.getExprFactory ();

    if (p.getCProgram () != null)
      for (Expr decl : p.getCProgram ().getCSymbolTable ().decls ())
	if (CTypedDeclOp.isCTypedDecl (decl)
	    && !CTypeDeclOp.isCTypeDecl (decl))
	  addCDecl (decl, st, unknownAsInt);

    // -- process global declarations
    if (p.getGlobalDecls () != null)
      for (Iterator it = p.getGlobalDecls ().values ().iterator ();
	  it.hasNext ();)
	addPDecl ((PDecl) it.next (), st, unknownAsInt);

    // -- process local declarations
    for (Iterator it = p.getFunctionDefs ().values ().iterator ();
	it.hasNext ();)
      addFunctionDecls ((PFunctionDef) it.next (), st, unknownAsInt);

    // XXX used by PredicateRefiner -- var name is magic string!!!
    addDecl (fac.var ("__UNKNOWN__"), fac.op (SimpleTypeOp.INT), st);

    // XXX is this hint needed?
    p.setDeclsRefined (true);  

    return st;
  }

  private static void addPDecl (PDecl d, SymbolTable st, boolean unknownAsInt)
  {
    Expr sourceExpr = d.getSourceExpr ();

    assert isDecl (sourceExpr);

    Expr name = getDeclNameExpr (sourceExpr);
    Expr stType = cilToExprType (sourceExpr, unknownAsInt);

    addDecl (name, stType, st);
  }

  private static void addCDecl (Expr decl, SymbolTable st,
      boolean unknownAsInt)
  {
    addDecl (decl, cToExprType (CTypedDeclOp.declType (decl), unknownAsInt),
	st);
  }

  private static void addDecl (Expr decl, Expr type, SymbolTable st)
  {
    st.put (decl, type);
  }

  private static void addFunctionDecls (PFunctionDef f, SymbolTable st,
      boolean unknownAsInt)
  {
    ExprFactory fac = f.getExprFactory ();

    // -- process local declarations
    if (f.getLocalDecls () != null)
      for (Iterator it = f.getLocalDecls ().values ().iterator ();
	  it.hasNext ();)
	addPDecl ((PDecl) it.next (), st, unknownAsInt);

    // -- add decl for return index
    addDecl (f.getReturnIndexVar (), fac.op (SimpleTypeOp.INT), st);

    // -- add decl for return value
    // only do this for CIL-based functions since resultDecl is in declTable;
    // ugly way to check for this :(
    if (f.getParent ().getCProgram () == null)
      addDecl (f.getReturnValueVar (),
	  cilToExprType (f.getSourceExpr (), unknownAsInt),
	  st);
  }

  /** 
   * Converts a CIL type ({@link edu.toronto.cs.expr.CILDeclarationOp}) to an
   * Expr type.
   * 
   * @param cilType 
   * @param unknownAsInt if true, convert unknown types to int
   * @return 
   */
  private static Expr cilToExprType (Expr cilType, boolean unknownAsInt)
  {
    ExprFactory fac = cilType.getFactory ();

    Expr baseType = getBaseType (cilType);

    Expr stType;

    SimpleTypeOp sto = typeMap.get (baseType.op ());
    if (sto == null)
	 if (unknownAsInt)
	   sto = SimpleTypeOp.INT;
	 else
	   throw new RuntimeException ("unknown base type in " + baseType);
    
    // XXX extend to deal with multi-dimensional arrays
    if (isDecl(cilType) 
	   && isArrayDecl (cilType))
	 stType = CompoundTypeOp.arrayT(fac.op (SimpleTypeOp.INT),
							  fac.op (sto));
    else // Primitive
	 stType = fac.op (sto);

    // apply pointers, if any
    return ptrChain (getNumPtrs (cilType), stType);
  }

  private static Expr cToExprType (Expr cType, boolean unknownAsInt)
  {
    ExprFactory fac = cType.getFactory ();

    if (isCSimpleType (cType))
    {
      String typeName = simpleTypeName (cType);
      if (typeName.equals ("int"))
	return fac.op (SimpleTypeOp.INT);
    }
    
    if (unknownAsInt)
      return fac.op (SimpleTypeOp.INT);
    else
      throw new RuntimeException ("unknown type in " + cType);
  }

  private static Expr ptrChain (int depth, Expr baseType)
  {
    if (depth == 0)
      return baseType;

    return CompoundTypeOp.ptrT (ptrChain (depth - 1, baseType));
  }
}

