package edu.toronto.cs.yasm.cfg;

import java.util.*;
import edu.toronto.cs.expr.*;

/**
 * Describe class CFGraph here.
 *
 *
 * Created: Fri Feb 16 10:42:14 2007
 *
 * @author <a href="mailto:arie@pcawf.sei.cmu.edu">Arie Gurfinkel</a>
 * @version 1.0
 */
public class CFGraph {


  /**
   * an array of basic blocks
   */
  BasicBlock[] blocks;

  /**
   * Maps a label name to a BB it is defined in.
   */
  Map<String,BasicBlock> labelMap;

  /**
   * Maps a labelUID to a BB it is defined in.
   */
  Map<Integer,BasicBlock> labelUIDMap;


  /** 
   * an index of the last free entry in the blocks array
   * if lastFreeBlock == blocks.length then there are no free entries
   */
  int lastFreeBlock;
  
  
  
  /**
   * Creates a new <code>CFGraph</code> instance.
   *
   */
  public CFGraph () 
  {
    lastFreeBlock = 0;
    blocks = new BasicBlock [2];
    // -- allocate entry block
    BasicBlock entry = getBBSafe (0);
    // -- allocate exit block
    entry.linkAfter (getBBSafe (1));
    labelMap = new HashMap<String,BasicBlock> ();
    labelUIDMap = new TreeMap<Integer,BasicBlock> ();
  }


  /**
   * Associate basic block {@code bb} with a label given by 
   * {@code CLabelDec} expression {@code e}.
   *
   * @param e an <code>Expr</code> value
   * @param bb a <code>BasicBlock</code> value
   */
  public void addLabel (Expr e, BasicBlock bb)
  {
    addLabel (CNamedDeclOp.declName (e), CLabelDeclOp.labelUID (e), bb);
  }
  

  public void addLabel (String label, int labelUID, BasicBlock bb)
  {

    labelUIDMap.put (labelUID, bb);
    if (label != null) labelMap.put (label, bb);
  }

  public BasicBlock bbForLabelName (String label)
  {
    return labelMap.get (label);
  }

  public BasicBlock bbForLabelUID (int labelUID)
  {
    return labelUIDMap.get (labelUID);
  }

  public BasicBlock bbForLabel (Expr labelDecl)
  {
    return bbForLabelUID (CLabelDeclOp.labelUID (labelDecl));
  }
  

  
  public BasicBlock newBB ()
  {
    return getBBSafe (lastFreeBlock);
  }
  
  /**
   * Returns a basic block {@code index}, or creates a new one if it
   * does not exists
   *
   * @param index an <code>int</code> value
   * @return a <code>BasicBlock</code> value
   */
  public BasicBlock getBBSafe (int index)
  {
    ensureCapacity (index + 1);
    
    // if the block does not exist, create it
    if (blocks [index] == null)
      {
	blocks [index] = new BasicBlock (index);
	// update the pointer to the last free block
	if (index + 1 > lastFreeBlock)
	  lastFreeBlock = index + 1;
      }
    
    return blocks [index];
  }
  
  public BasicBlock getBB (int index)
  {
    if (index < lastFreeBlock) return blocks [index];

    throw new NoSuchElementException ("No block with index " + index);
  }

  public boolean hasBB (int index)
  {
    return index < lastFreeBlock && getBB (index) != null;
  }
  
  /**
   * Adds an edge between {@code src} and {@code dst}. If the edge 
   * already exists, returns an existing one.
   *
   * @param src a <code>BasicBlock</code> value
   * @param dst a <code>BasicBlock</code> value
   * @return an <code>Edge</code> value
   */
  public Edge addEdge (BasicBlock src, BasicBlock dst, boolean trueValue,
      boolean falseValue)
  {
    // -- create a new edge
    Edge newEdge = new Edge (src, dst, trueValue, falseValue);
    // -- link it
    Edge edge = newEdge.linkSrc ();
    // -- if there was no such edge already, link dest as well
    if (edge == newEdge)
      edge.linkDst ();
    return edge;
  }
  
  public Edge addEdge (BasicBlock src, BasicBlock dst)
  {
    return addEdge (src, dst, false, false);
  }

  /**
   * ensure that blocks array can contain at least {@code v} number of
   * elements
   */
  protected void ensureCapacity (int v) 
  { 
    if (v < blocks.length) return;
    
    int newSize = (v << 1);
    
    BasicBlock[] newBlocks = new BasicBlock [newSize];
    System.arraycopy (blocks, 0, newBlocks, 0, blocks.length);
    blocks = newBlocks;
  }
  
  /**
   * Trims blocks array to contain just enough space to store all
   * current blocks */

  protected void trimBlocksToSize ()
  {
    BasicBlock[] newBlocks = new BasicBlock [lastFreeBlock];
    System.arraycopy (blocks, 0, newBlocks, 0, lastFreeBlock);
    blocks = newBlocks;
  }
  
  
  public void compact ()
  {
    trimBlocksToSize ();
  }
  

  /**
   * Returns the entry block of this graph
   *
   * @return a <code>BasicBlock</code> value
   */
  public BasicBlock entry ()
  {
    return blocks [0];
  }

  /**
   * Returns the exit block of this graph
   *
   * @return a <code>BasicBlock</code> value
   */
  public BasicBlock exit ()
  {
    return blocks [1];
  }

  public Map<String, BasicBlock> getLabelMap ()
  {
    return labelMap;
  }

  public Map<Integer, BasicBlock> getLabelUIDMap ()
  {
    return labelUIDMap;
  }


  public Iterable<BasicBlock> allBlocksFwd ()
  {
    return new Iterable<BasicBlock> ()
      {
	public Iterator<BasicBlock> iterator ()
	  {
	    return new Iterator<BasicBlock> ()
	      {
		BasicBlock cur = null;
		
		public boolean hasNext ()
		  {
		    return nextOrNull () != null;
		  }

		private BasicBlock nextOrNull ()
		  {
		    if (cur == null) return entry ();
		    else return cur.getNext ();
		  }
		
		
		public BasicBlock next ()
		  {
		    if (!hasNext ())
		      throw new NoSuchElementException ();
		    
		    cur = nextOrNull ();
		    return cur;
		  }

		public void remove ()
		  {
		    throw new UnsupportedOperationException ();
		  }
	      };
	  }
      };
  }  
  
  /**
   * Returns total number of basic blocks.
   *
   * @return an <code>int</code> value
   */
  public int size ()
  {
    return lastFreeBlock;
  }
  
  public int numBlocks ()
  {
    return lastFreeBlock;
  }

  public static void main (String[] args)
  {
    CFGraph cfg = new CFGraph ();

    BasicBlock prevBB = null;

    for (int i = 0; i < 10; i++)
      {
	// -- create basic block i
	BasicBlock bb = cfg.getBBSafe (i);

	// -- link it right after the last basic block created so far
	// -- but only if the block is not already linked
	if (prevBB != null && !bb.isLinked ())
	  prevBB.linkAfter (bb);
	
	// -- update prevBB for the next iteration of the loop
	prevBB = bb;
	  
	
	// -- modify the block by setting its statments and what not
	// -- don't forget to update the label map if block statements
	// -- contain any labels in them.

	// -- now ready to add edges

	// -- add a backward edge if possible
	if (i - 2 > 0)
	  cfg.addEdge (bb, cfg.getBBSafe (i - 2));

	// -- add a fwd edge if possible
	if (i + 2 < 10)
	  cfg.addEdge (bb, cfg.getBBSafe (i + 2));
      }
    

    // -- iterate over all basic blocks
    for (BasicBlock bb : cfg.allBlocksFwd ())
      {
	System.out.println ("Saw block: " + bb.getIndex ());
	System.out.println ("With successor edges: ");
	for (Edge e : bb.getSuccs ())
	  System.out.println ("\t" + e);
	System.out.println ("And predecessor edges: ");
	for (Edge e : bb.getPreds ())
	  System.out.println ("\t" + e);

      }
    
  }
  
  

  /**
   * Describe class BasicBlock here.
   *
   *
   * Created: Fri Feb 16 10:22:28 2007
   *
   * @author <a href="mailto:arie@pcawf.sei.cmu.edu">Arie Gurfinkel</a>
   * @version 1.0
   */
  public class BasicBlock 
  {

    /**
     * Describe variable <code>index</code> here.
     *
     */
    int index;
  
    /**
     * List of expressions in this basic block 
     */
    List<Expr> stmts;

    List<Expr> phiNodes;

    /**
     * Edges to predecessors of this block 
     */
    List<Edge> preds;

    /** 
     * Edges to successors of this block 
     */
    List<Edge> succs;

    BasicBlock nextBB;
    BasicBlock prevBB;

    Object aux;
  
  
    /**
     * Creates a new <code>BasicBlock</code> instance.
     *
     */
    public BasicBlock (int _index) 
    {
      index = _index;
      preds = new LinkedList ();
      succs = new LinkedList ();
      stmts = new LinkedList<Expr> ();
      nextBB = null;
      prevBB = null;
    }


  
    public void setStmts (List<Expr> v)
    {
      stmts.clear ();
      stmts.addAll (v);
    }
    public void addStmt (Expr v)
    {
      stmts.add (v);
    }

    public int getIndex ()
    {
      return index;
    }
    
  
    public List<Expr> getStmts ()
    {
      return stmts;
    }

    public boolean isLinked ()
    {
      return nextBB != null || prevBB != null;
    }

    public BasicBlock splitAtStmt (Expr stmt)
    {
      if (CStmtOp.isLabelExpr (stmt))
	throw new IllegalArgumentException ("Cannot split at a label");

      List<Expr> prefix = new LinkedList<Expr> ();
      List<Expr> suffix = new LinkedList<Expr> (stmts);
      boolean found = false;
      
      for (Iterator<Expr> it = suffix.iterator (); it.hasNext () && !found ; )
	{
	  Expr e = it.next ();
	  if (e != stmt)
	    {
	      prefix.add (e);
	      it.remove ();
	    }
	  
	  else
	    found = true;
	}

      // -- if nothing was found, replace stmts with prefix and bail
      // -- out or found the split point, but there is nothing before
      // -- it to split
      if (!found || prefix.isEmpty ()) return this;
      
      // -- normal split case.  prefix is what goes into the current
      // -- block, and stmts is what goes into the new one
      

      // -- create a new basic block
      BasicBlock newBB = newBB ();      

      setStmts (prefix);
      newBB.setStmts (suffix);

      

      // -- add new block after the old one 
      this.linkAfter (newBB);
      
      // -- move all edge from this BB to originate from the old one
      // -- since Edge.relinkSrc modifies successor list, it is copied first
      for (Edge e : new LinkedList<Edge> (this.getSuccs ()))
	e.relinkSrc (newBB);

      // -- add cfg edge between current block and the new one
      addEdge (this, newBB);

      
      return newBB;
    }
    
    
    public BasicBlock getNext ()
    {
      return nextBB;
    }

    public void setNext (BasicBlock v)
    {
      nextBB = v;
    }
    

    public BasicBlock getPrev ()
    {
      return prevBB;
    }

    public void setPrev (BasicBlock v)
    {
      prevBB = v;
    }
    
    public List<Edge> getSuccs ()
    {
      return succs;
    }
    public List<Edge> getPreds ()
    {
      return preds;
    }
    
    
    /**
     * Link {@code bb} after this basic block.
     *
     * @param bb a <code>BasicBlock</code> value
     */
    public void linkAfter (BasicBlock bb)
    {
      bb.setPrev (this);
      bb.setNext (getNext ());
      
      if (getNext () != null)
	getNext ().setPrev (bb);

      setNext (bb);
    }
  }

  public class Edge 
  {
    boolean trueValue;
    boolean falseValue;
    BasicBlock src;
    BasicBlock dst;
    
    public Edge (BasicBlock _src, BasicBlock _dst, boolean _trueValue,
	boolean _falseValue)
    {
      src = _src;
      dst = _dst;
      trueValue = _trueValue;
      falseValue = _falseValue;
    }

    public Edge (BasicBlock _src, BasicBlock _dst)
    {
      this (_src, _dst, false, false);
    }


    public Edge relinkSrc (BasicBlock newSrc)
    {
      unlinkSrc ();
      src = newSrc;
      Edge res = linkSrc ();
      assert res == this : "Edge consistency failed!";
      return res;
    }
    

    protected void unlinkSrc ()
    {
      if (src.getSuccs ().remove (this))
	src = null;
      else
	throw new RuntimeException ("Unexpected behaviour: edge " + this + 
				    " was not registered at the source");
    }
    

    public Edge linkSrc ()
    {
      List<Edge> succs = src.getSuccs ();
      
      // -- see if this edge is already in the src list, if so nothing 
      // -- needs doing
      for (Edge e : succs)
	if (e.equals (this)) return e;
      
      // -- add the edge to the list of successors
      succs.add (this);
      return this;
    }

    public Edge linkDst ()
    {
      List<Edge> preds = dst.getPreds ();
      for (Edge e : dst.getPreds ())
	if (e.equals (this)) return e;
      
      preds.add (this);
      return this;
    }
    
    public BasicBlock getDst ()
    {
      return dst;
    }
    public BasicBlock getSrc ()
    {
      return src;
    }

    public boolean isTrueValue ()
    {
      return trueValue;
    }

    public boolean isFalseValue ()
    {
      return falseValue;
    }

    public boolean equals (Object o)
    {
      if (o instanceof Edge)
	{
	  Edge e = (Edge)o;
	  return e.getSrc () == getSrc () && e.getDst () == getDst ();
	}
      return false;
      
    }

    public String toString ()
    {
      return src.getIndex () + " --> " + dst.getIndex ();
    }
    
    
  }
  

}
