package edu.toronto.cs.yasm.eqsolver;

import java.util.*;

import edu.toronto.cs.yasm.eqsolver.EqSystem.FixPtModality;

/**
 * Describe class EqSolver here.
 *
 *
 * Created: Thu May 31 16:30:18 2007
 *
 * @author <a href="mailto:arie@pcawf.sei.cmu.edu">Arie Gurfinkel</a>
 * @version 1.0
 */
public class EqSolver<V>
{

  /**
   * The work queue that contains all variables whose values need to
   * be recomputed.
   */
  protected Queue<Integer> wq;

  /**
   * Current environment.
   */
  protected Map<Integer,V> env;
  

  protected EqSystem<V> eqSystem;
  
  
  protected List<List<VariableChangedEventListener<V>>> listeners;
  
  protected VariableChangedEventListener<V> historyObserver 
    = new SolverHistoryObserver<V> ();

  /**
   * Creates a new <code>EqSolver</code> instance for a given
   * equational system.
   *
   * @param eqSystem an equational system for the solver to solve
   */
  public EqSolver (EqSystem<V> eqSystem) 
  {
    wq = new LinkedList<Integer> ();
    initSolver (eqSystem);
  }

  
  /**
   * Initializes the solver for an equational system.
   *
   * @param eqSystem an equational system to solve
   */
  protected void initSolver (EqSystem<V> eqSystem)
  {
    this.eqSystem = eqSystem;
    
    // -- create listeners for every variable
    listeners = 
      new ArrayList<List<VariableChangedEventListener<V>>>
      (eqSystem.getFreshVar ());

    for (int i = 0; i < eqSystem.getFreshVar (); i++)
      {
	LinkedList<VariableChangedEventListener<V>> iListener = 
	  new LinkedList<VariableChangedEventListener<V>> ();
	
	if (historyObserver != null)
	  iListener.add (historyObserver);

	listeners.add (iListener);
      }
    

    // -- go over all variables
    for (int i = 0; i < listeners.size (); i++)
      {
	
	// -- defining equation for current variable
	Equation<V> eq = eqSystem.eqForVar (i);
	if (eq == null) continue;
	
	// -- register a listener with every variable that current
	// -- equation depends on
	VariableChangedEventListener<V> eqListner 
	  = new DefaultVariableChangedEventListener<V> (eq);
	
	for (int var : eq.getUseVars ())
	  listeners.get (var).add (eqListner);
      }
  }
  

  

  /**
   * Computes a fixpoint of an equational system given to this solver,
   * starting with a particular starting point
   *
   * @param env initial environment
   * @return an enviroment that is the solution to the system.
   */
  public Map<Integer,V> solve (Map<Integer,V> env)
  {
    this.env = env;

    initWorkQueue (env);
    
    while (!wq.isEmpty ())
      {
	// -- dequue a variable from the work list
	int var = wq.remove ();
	// -- compute its new value
	V newV = eqSystem.eqForVar (var).eval (env);
	
	// -- get old value, and compare
	V oldV = env.get (var);
	if (!newV.equals (oldV))
	  {
	    // -- if anything has changed, update current environment
	    // -- and notify all interested parties
	    env.put (var, newV);
	    sendVariableChangedEvent (var, oldV, newV);
	  }
      }
    
    return env;
  }

  /**
   * Returns current environment
   *
   * @return current environment
   */
  protected Map<Integer,V> getEnv ()
  {
    return env;
  }
  
  

  public void addVariableChangedEventListener 
    (int var, VariableChangedEventListener<V> l)
  {
    listeners.get (var).add (l);
  }
  
  protected void sendVariableChangedEvent (int var, V oldV, V newV)
  {
    List<VariableChangedEventListener<V>> vListeners = 
      listeners.get (var);
    
    if (vListeners == null) return;
    
    for (VariableChangedEventListener<V> l : vListeners)
      l.valueChanged (this, var, oldV, newV);
  }


  protected void initWorkQueue (Map<Integer,V> env)
  { 
    for (int var : eqSystem.boundVars ())
      if (eqSystem.eqForVar (var).maybeUnstable (env))
	addVarToWorkQueue (var);
  }
  
  protected void addVarToWorkQueue (int var)
  {
    // -- may want to check if var is already there
    wq.add (new Integer (var));
  }

  /**********************************************************************/
  /**** Event Listeners ***/
  /**********************************************************************/


  protected class SolverHistoryObserver<V> 
    implements VariableChangedEventListener<V>
  {
    public void valueChanged (EqSolver<V> solver, int var, V oldV, V newV)
    {
      System.out.println ("Variable " + var + " has changed");
    }
  }


  protected class DefaultVariableChangedEventListener<V>
    implements VariableChangedEventListener<V>
  {
    Equation<V> eq;
    
    protected DefaultVariableChangedEventListener (Equation<V> eq)
    {
      this.eq = eq;
    }


    public void valueChanged (EqSolver<V> solver, int var, V oldV, V newV)
    {
      if (eq.mayChange (solver.getEnv (), var, oldV, newV))
	addVarToWorkQueue (eq.getDefVar ());
    }
  }
  
}
