package edu.toronto.cs.yasm.pcfa;

import edu.toronto.cs.mvset.MvRelation;
import edu.toronto.cs.mvset.SeqMvRelation;
import edu.toronto.cs.mvset.MvSet;
import edu.toronto.cs.mvset.MvSetMvRelation;

/**
 * Describe class CallEdge here.
 *
 *
 * Created: Wed Jun  8 13:18:15 2005
 *
 * @author <a href="mailto:arie@eon.cs">Arie Gurfinkel</a>
 * @version 1.0
 */
public class CallEdge extends MvRelnEdge 
{

  /**
   * Summary of the function including call->body->return
   *
   */
  FunctionSummaryEdge funSumEdge;
  
  /**
   * Summary of the function body
   *
   */
  SummaryEdge bodySumEdge;

  /**
   * The return edge corresponding to this call
   *
   */
  ReturnEdge returnEdge;

  /**
   * The return edge used in FunctionSummary
   */
  ReturnEdge sumReturnEdge;

  /**
   * The call relation used in FunctionSummary
   */
  MvRelation sumCallReln;

  /**
   * Creates a new <code>CallEdge</code> instance.
   *
   * @param _src a source node of this edge, has to be a function call
   * @param _dst a destination node of this edge, has to be an entry
   * of a function
   * @param _reln the assignment of values to function arguments
   * @param _funSumEdge an edge corresponding to the completion of
   * this call, which will be updated by this {@code CallEdge}
   * @param _bodySumEdge a summary edge of the function from entry to
   * exit
   * @param _returnEdge an edge from the exit of the function to the
   * end of this call
   * @param _sumReturnEdge similar to {@code _returnEdge}, but with
   * different pre- and post-variables.
   * @param _sumCallReln similar to {@code _reln}, but with different
   * pre- and post-variables.
   */
  public CallEdge (Node.NodeId _src, Node.NodeId _dst, 
		   MvRelation _reln, FunctionSummaryEdge _funSumEdge, 
		   SummaryEdge _bodySumEdge, ReturnEdge _returnEdge,
		   ReturnEdge _sumReturnEdge,
		   MvRelation _sumCallReln) 
  {
    super (_src, _dst, _reln);
    setFunSummary (_funSumEdge);
    bodySumEdge = _bodySumEdge;
    setReturnEdge (_returnEdge);

    sumReturnEdge = _sumReturnEdge;
    sumCallReln = _sumCallReln;
  }

  
  /**
   * Creates a new <code>CallEdge</code> instance without _sumReturnEdge,
   * and _sumCallReln; this is used for being compatiable with "old"
   * way, i.e. all predicates variables are treated with "global" varaibles.
   * 
   *
   */
  public CallEdge (Node.NodeId _src, Node.NodeId _dst, 
		   MvRelation _reln, FunctionSummaryEdge _funSumEdge, 
		   SummaryEdge _bodySumEdge, ReturnEdge _returnEdge)
  {
    super (_src, _dst, _reln);
    setFunSummary (_funSumEdge);
    bodySumEdge = _bodySumEdge;
    setReturnEdge (_returnEdge);

    // -- if _sumReturnEdge and _sumCallReln are not provided, then
    // -- all varaibles are treated with global varaibles (old way)
    // -- in updateBodySummary, relation on fumSumEdge computed by
    //
    //     funSumEdge.setRelation 
    //       (new SeqMvRelation (sumCallReln,
    //     			  new SeqMvRelation (bodySumEdge.getRelation (),
    //     					     sumReturnEdge.getRelation ())));
    // should be as the same as before, i.e.
    //
    //     funSumEdge.setRelation 
    //       (new SeqMvRelation (getRelation (),
    //     			  new SeqMvRelation (bodySumEdge.getRelation (),
    //     					     returnEdge.getRelation ())));

    sumReturnEdge = _returnEdge;
    sumCallReln = _reln;
  }



  protected void setFunSummary (FunctionSummaryEdge v)
  {
    funSumEdge = v;
    funSumEdge.setCallEdge (this);
  }
  
  public FunctionSummaryEdge getFunSummary ()
  {
    return funSumEdge;
  }

  public SummaryEdge getBodySummary ()
  {
    return bodySumEdge;
  }


  protected void setReturnEdge (ReturnEdge v)
  {
    returnEdge = v;
  }

  public ReturnEdge getReturnEdge ()
  {
    return returnEdge;
  }

  public ReturnEdge getSumReturnEdge ()
  {
    return sumReturnEdge;
  }

  public MvRelation getSumCallReln ()
  {
    return sumCallReln;
  }
  
  

  /**
   * Updates body summary and returns new function summary 
   * that was affected by this change
   *
   * @param mvSet a <code>MvSet</code> value
   * @param _PostToAuxMap a <code>int []</code> value
   * @param _AuxToPostMap a <code>int []</code> value
   * @return a <code>SummaryEdge</code> value
   */
  public SummaryEdge updateBodySummary (MvSet mvSet,
					int[] _PostToAuxMap,
					int[] _AuxToPostMap)
  {
    bodySumEdge.update (mvSet);

    // XXX need extra handling of local variables 
    // XXX depending on what is the type of the returnEdge

    // -- recompute function summary


    // -- check if funSumEdge corresponds to a self-call statement
    if (funSumEdge.isSelfCall ())
      {
	// -- for self-call statement, we compose a "real" tranisition
	// relation for funcSumEdge : G X L_caller -> G X L_caller
	//
	// existAbsFwdCube : G, existAbsBwdCube : G
	//
	// Note : L_callee = L_caller for self-call
	//
	// see the comments for doSumReturnEdge () in 
	// LocalPProgramReachabilityCompilerVisitor.java
	

	// -- sumReturnEdge.getRelation () : G X L_callee -> G
	//    existAbsFwdCube : G X L_callee 
	//    existAbsBwdCube : G

	MvSet tempReln  = sumReturnEdge.getRelation ().toMvSet ();

	// -- now: tempReln : G X L_callee --> G

	// -- rename post variables G' to G''
	tempReln = tempReln.renameArgs (_PostToAuxMap);

	// -- now: tempReln : G X L_callee --> G''

	// -- bodySumEdge.getRelation : G X L_callee -> G X L_callee
	//    existAbsFwdCube : G X L_callee 
	//    existAbsBwdCube : G X L_callee

	tempReln = bodySumEdge.getRelation ().bwdImage (tempReln);


	// -- now: tempReln : G X L_callee --> G''

	// -- sumCallReln : G X L_caller -> G X L_callee
	//    existAbsFwdCube : G 
	//    existAbsBwdCube : G X L_callee

	tempReln = sumCallReln.bwdImage (tempReln);
	

	// -- now: tempReln : G X L_callee --> G''	

	// -- rename aux variables G'' back to G'
	tempReln = tempReln.renameArgs (_AuxToPostMap);


	// -- now: tempReln : G X L_callee --> G'

	// -- we want preCube be G, and postCube be G
	//    existAbsFwbCube : G, existAbsBwdCube : G

	// -- note that
	//    sumCallReln.getPreVariablesCube () : G
	//    sumReturnEdge.getRelation (). getPostVariablesCube () : G
	//
	//    sumCallReln.getPreToPostMap () : 
	//        map G to G'
	//                                     
	//    sumCallReln.getPostToPreMap () : 
	//        map (G', L'_caller) to  (G, L_caller)
	//        but actually, only (G' -> G) is used 

	MvSetMvRelation.doAssert = false;

	MvRelation tempMvReln = 
	  tempReln.
	  toMvRelation (null,
			sumCallReln.getPreVariablesCube (), 
			sumReturnEdge.getRelation ().getPostVariablesCube (),
			sumReturnEdge.getRelation ().getPreToPostMap (),
			sumCallReln.getPostToPreMap ());

	MvSetMvRelation.doAssert = true;

	funSumEdge.setRelation (tempMvReln);

      }
    else
      {
	// -- for non-self-call statement, we ensure the values of
	// -- caller's local predicates are not changed by using
	// -- special pre and post cubes for "sumCallReln" and
	// -- "sumReturnEdge" relations respectively, i.e., we do
	// -- not touch those predicates by removing the correponding
	// -- cubes.
	
	// -- the pre/post cubes in following composition is 
	// 
	// -- G --> G X L2; G X L2 --> G X L2; L2 --> empty
	//
	
	funSumEdge.setRelation 
	  (new SeqMvRelation (sumCallReln,
			      new SeqMvRelation 
			      (bodySumEdge.getRelation (),
			       sumReturnEdge.getRelation ())));
      }

    

    return funSumEdge;		       
  }


  // ----------------------------------------------------------------------
  // -- SOME THOUGHTS ABOUT A DIFFERENT IMPLEMENTATION OF THIS
  // ----------------------------------------------------------------------

  /*
    Let r be an MvRelation, and Q and mv-set representing relation
    we want to compute Q;r. This can be done as follows
    
    r: pre <-> post
    Q: aux <-> pre

    r.fwdImg (Q): aux <-> pre
     is defined by 
       (\E{pre} Q /\ r)[post->pre]

    and gives the composition. After that, rename variables:
      r.fwdImage (Q) [pre->post][aux->pre]

    if we also have t : pre <-> post, then 
    Q;r;t is
       t.fwdImage (r.fwdImage (Q))[pre->post][aux->pre]


    // -- get our Q : pre<->post
    MvSet reln = sumCallReln.getRelation ().toMvSet ();
    // -- change it into aux<->pre
    reln = reln.renameArgs (preToAuxMap);
    reln = reln.renameArgs (postToPreMap);
	   
	   
    // -- do the composition
    reln = returnEdge.getRelation ().fwdImage 
      (bodySumEdge.getRelation ().fwdImage (reln));

    // -- rename reln from aux<->pre into pre<->post
    reln = reln.renameArgs (preToPostMap);
    reln = reln.renameArgs (auxToPreMap);
	   
    // -- reln: pre <-> post
    // -- pre: G x L 
    // -- post: G
    // -- need to extend it to (GxL) <-> (GxL)
    // -- to do so need the following parts
    // -- inScope (global).pre-cube == G -- callSum has that
    // -- inScope(call[er/ee]).post-cube = GxL -- many have that
    // -- inScope(global).pre-post == G -- this is the only thing we need
    // -- inScope(call[er/ee].post-pre == GxL -- many have that

    With this implementation, we no longer will need special return-summary
    edge. One should also check for which case cex generation is easier.
   */
  
  
}
