package edu.toronto.cs.yasm.pprogram;

import edu.toronto.cs.expr.*;
import edu.toronto.cs.tp.SymbolTable;

import java.util.*;

import java.io.PrintWriter;

/**
 * PProgramPatcher.java
 *
 * Given a PProgram patches the PProgram's CFG edges: sets dest fields, adds
 * init and end statements to the entire program, and adds function return
 * selectors and logues.
 *
 * Preserves labels of CILLabelledStmts.
 *
 * @author <a href="mailto:kelvin@cs.toronto.edu">Kelvin Ku</a>
 * @version 1.0
 */
public class PProgramPatcher
{
  /**
   * Global reference to the PProgram being patched.
   */
  static PProgram pp;

  /**
   * Main method: calls patching methods on p. 
   */
  public static final void patch (PProgram p)
  {
    pp = p;
    for (PFunctionDef def : p.getFunctionDefs ().values ())
      def.setParent (p);

    // doesn't add init and end to statementList!
    addProgramInitAndEnd (p);

    // sets ids and labels of all statements
    // adds all statements, init, and end to statementList
    // XXX eccch, creates selectors and logues
    labelAllStmts (p);

    // patches dests
    computeCFGForEachFunction (p);

    // patches predecessors, given dests
    fillInPredecessors (p);
  }

  /**
   * Adds init and end statements to program.
   */
  static final void addProgramInitAndEnd (PProgram p)
  {
    SkipPStmt init = new SkipPStmt (p.getMainFunctionDef (), (Expr) null);
    init.setDest (p.getMainFunctionDef ().getHead ());
    init.setLabel ("init"); // XXX Magic string
    // getMainFunctionDef ().setHead (init);

    // -- add a self-loop at the end
    GotoPStmt g = new GotoPStmt (p.getMainFunctionDef (), (Expr) null);
    g.setDest (g);
    g.setLabel ("END"); // XXX Magic string
    // getMainFunctionDef ().getHead ().setTail (g);

    init.setNext (g);
    p.setProgramStmts (init);
  }

  /** 
   * Calls labelStmts on PProgram statements, then on each function head.
   */
  static final void labelAllStmts (PProgram p)
  {
    // has to be done first so init gets id 0
    labelStmts (p.getProgramStmts ()); 
    for (PFunctionDef def : p.getFunctionDefs ().values ())
      labelFunctionStmts (def);
  }

  /**
   * Adds return selector to <code>f</code> and calls labelStmts on its list
   * of PStmts.
   */
  static final void labelFunctionStmts (PFunctionDef f)
  {
    // System.err.println ("labelling " + f.getFunctionName ()); -- add
    // returnSelector into the list of statements -- need to recurse on this to
    // add all of the statements -- and give them an ID

    addReturnSelector (f);

    // labelStmts (f.getReturnSelector ()); -- then process the rest of
    // function statements
    labelStmts (f.getHead ());
  }

  /** 
   * Assigns a label to every unlabelled statement in the list headed by
   * <CODE>p</CODE>; descends into the children of an IfPStmt. XXX Also adds
   * each statement to statementList in the order that they are labelled ---
   * only the statements in statementList are processed by PProgramCompiler.
   */
  static final void labelStmts (PStmt p)
  {
    /*
       if (p != null)	  
       {
       System.out.println ("Labelling a " + p.getSourceExpr ());	  
       if (p.getLabel () != null)
       {
       System.out.println ("\tIt came with a label: " + p.getLabel());
       }
       else
       {
       System.out.println ("\tIt didn't come with a label");
       }
       }
       */

    if (p instanceof FunctionCallPStmt)
    {
      // XXX 2005-06-21: labels of function calls weren't added to
      // map previously?
      /*
	 if (p.getLabel () != null)
	 labelledStatementMap.put (p.getLabel (), p);
	 */

      // adds prologue and epilogue to statementList instead of
      // FunctionCallPStmt

      // XXX also sets functionDef field

      // System.err.println ("PProgram: FunctionCall: " + p.getSourceBlock ());
      FunctionCallPStmt call = (FunctionCallPStmt) p;
      /*
      System.err.println ("PProgram: labelling function call: " +
                          call.getFunctionName () + ", " +
                          new Integer (call.getCallIndex ()));
*/
      PFunctionDef def =
	pp.getFunctionDefs ().get (call.getFunctionName ());

      PStmt madeCall;  

      if (def == null)
      {
	System.err.println ("PProgram: Missing function def for: " + call.getFunctionName ());
	madeCall = new SkipPStmt (call.getParent (), call.getSourceExpr ());
	// XXX probably not the correct way to do this
	if (p.getLabel () != null)
	{
	  pp.getLabelledStatementMap ().put (p.getLabel (), madeCall);
	  madeCall.setLabel (p.getLabel ());
	}
      }
      else  
	madeCall = makeFunctionCall (def, call);

      call.setFunctionDef (def);
      call.setLogues (madeCall);
      // System.err.println ("made call");
      // madeCall.print ();
      // madeCall.setLabel (p.getSourceBlock ().getLabel ());
      if (call.getSourceExpr ().op () == CILLabelledStmtOp.LSTMT)
	madeCall.setLabel (CILLabelledStmtOp.getLabel (call.getSourceExpr ()));
      labelStmts (madeCall);

      // XXX had to put this here since if next statement is a
      // FunctionCallPStmt, it will get ignored below
      labelStmts (call.getNext ()); // will label successors as usual
      return;
    }

    if (p == null)
      return;

    int labelIndex = pp.getStatementList ().size ();  

    // NB: id is one less than labelIndex below
    p.setId (labelIndex);

    pp.getStatementList ().add (p);

    labelIndex++;

    /*
       if (p.getSourceBlock () != null)
       {  
       p.setLabel (p.getSourceBlock ().getLabel ());
       }
       */

    if (p.getSourceExpr () != null &&
	p.getSourceExpr ().op () == CILLabelledStmtOp.LSTMT)
    {
      p.setLabel (CILLabelledStmtOp.getLabel (p.getSourceExpr ()));
    }

    if (p.getLabel () == null)
      // statement has no label; generate one and tag it
      p.setLabel ("l" + labelIndex);
    else
      // statement originally had a label, add it to map
      pp.getLabelledStatementMap ().put (p.getLabel (), p);

    if (p instanceof IfPStmt)
    {
      IfPStmt ip = (IfPStmt) p;
      labelStmts (ip.getThenStmt ());
      labelStmts (ip.getElseStmt ());
    }

    labelStmts (p.getNext ());
  }

  static final void computeCFGForEachFunction (PProgram p)
  {
    for (PFunctionDef def : p.getFunctionDefs ().values ())
    {
      computeCFG (def.getHead (), def.getReturnSelector (),
	  def.getReturnSelector (), def.getReturnSelector ());
    }
  }

  /**
   * For each {@code PStmt} s, set s's predecessors to be all {@code
   * PStmt}s s' such that s is a dest of s'. Assumes that dests have
   * already been computed.
   */
  static final void fillInPredecessors (PProgram p)
  {
    for (Object o : p.getStatementList ()) {
      PStmt cur = (PStmt) o;
      List<PStmt> curDests = PStmt.getDests (cur);

      for (PStmt child : curDests)
        child.addPredecessor (cur);
    }
  }

  /**
   * Recursively patches next and dest fields of given PStmt and exit
   * statements.
   */
  static final void computeCFG
  (PStmt p, PStmt ifExitStmt, PStmt whileElseStmt, PStmt returnSelector)
  {
    PStmt nextStmt = p.getNext ();

    // XXX if a function call is the first statement of a function body, it
    // does get expanded into its logues but only because it is actually the
    // nextStmt of the function body's entry point (which gets added
    // automatically when a PFunctionDef is created)

    // -- expand function call into logues
    if (nextStmt instanceof FunctionCallPStmt)
    {
      FunctionCallPStmt call = (FunctionCallPStmt) nextStmt;
      p.setNext (call.getLogues ());
      call.getLogues ().setTail (nextStmt.getNext ());
      nextStmt = p.getNext (); // nextStmt is prologue
    }

    if (p instanceof IfPStmt)
    {

      IfPStmt ip = (IfPStmt) p;

      // -- if ip.getNext () is a FunctionCallPStmt, the logues are
      // -- patched in above, so we don't do it again here

      // -- if ip.getThenStmt () is a FunctionCallPStmt, the logues are
      // -- patched in here

      if (ip.getThenStmt () instanceof FunctionCallPStmt)
      {
	FunctionCallPStmt call = (FunctionCallPStmt) ip.getThenStmt ();
	call.getLogues ().setTail (ip.getThenStmt ().getNext ());
	ip.setThenStmt (call.getLogues ());
	ip.getThenGoto ().setDest (ip.getThenStmt ());
      }

      if (ip.getElseStmt () instanceof FunctionCallPStmt)
      {
	FunctionCallPStmt call = (FunctionCallPStmt) ip.getElseStmt ();
	call.getLogues ().setTail (ip.getElseStmt ().getNext ());
	ip.setElseStmt (call.getLogues ());
	ip.getElseGoto ().setDest (ip.getElseStmt ());
      }

      if (ip instanceof WhilePStmt)
      {
	WhilePStmt wp = (WhilePStmt) ip;

	// XXX make sure that wp.getNext () == null is handled
	PStmt wpNextDest = wp.getNext ();
	if (wpNextDest == null)
	{
	  // -- if this while is the last statement of 
	  // -- an if statement, once it is done we exit the 
	  // -- if statement as well
	  if (ifExitStmt != null) 
	    wpNextDest = ifExitStmt;
	  // -- we know that 
	  // -- whileElseStmt != null -> wp.getNext () != null
	  // 	      else if (whileElseStmt != null)
	  // 		wpNextDest = whileElseStmt;
	  // -- this while is the last statement of a function
	  else
	    wpNextDest = returnSelector;
	}

	wp.getElseGoto ().setDest (wpNextDest);

	computeCFG (wp.getThenStmt (), 
	    ifExitStmt, wp.getElseGoto ().getDest (), 
	    returnSelector);
      }
      else // p instanceof IfPStmt
      {
	PStmt ipNextDest = ip.getNext ();

	if (ipNextDest == null)
	{
	  if (ifExitStmt != null)
	    ipNextDest = ifExitStmt;
	  else
	    ipNextDest = returnSelector;
	}

	if (ip.getThenStmt () != null)
	  computeCFG
	    (ip.getThenStmt (), ipNextDest, whileElseStmt, returnSelector);
	else
	  throw new RuntimeException ("If without a body");

	if (ip.getElseStmt () == null)
	  ip.getElseGoto ().setDest (ipNextDest);
	else
	{
	  computeCFG (ip.getElseStmt (), ipNextDest, 
	      whileElseStmt, returnSelector);
	  ip.getExitGoto ().setDest (ipNextDest);
	}
      }
    }
    else if (p instanceof GotoPStmt)
    {
      if (p instanceof BreakPStmt)
	((BreakPStmt)p).setDest (whileElseStmt);

      else
      {
	// p instanceof GotoPStmt

	GotoPStmt gp = (GotoPStmt) p;

	// The goto dest is already set if it was generated in abstracting a
	// WHILE Block; so we only set it if the GotoPStmt was generated
	// for something else, e.g. an actual GOTO
	if (gp.getDest () == null)
	{
	  /*
	     System.err.println (">>> '" + gp.getTargetLabel () + "'");
	     System.err.println (">>> " + gp.getSourceExpr ());
	     System.err.println (">>> " + gp.getParent ());
	     */
	  gp.setDest (pp.getStmt (gp.getTargetLabel ()));
	}
      }
    }
    else if (p instanceof ReturnSelectorPStmt)
    {
      for (Iterator it =
	  ((ReturnSelectorPStmt) p).getTargetLabels ().iterator ();
	  it.hasNext ();)
	((ReturnSelectorPStmt) p).addDest
	  ((pp.getStmt ((String) it.next ())));
    }
    else if (p instanceof NDGotoPStmt)
    {
      for (Iterator it = ((NDGotoPStmt) p).getTargetLabels ().iterator ();
	  it.hasNext ();)
	((NDGotoPStmt) p).getDests ().add (pp.getStmt ((String) it.next ()));
    }
    else if (p instanceof PrllAsmtPStmt)
    {
      if (p instanceof ReturnPStmt)
	p.setDest (returnSelector);
    }
    else if (p instanceof SkipPStmt)
    {
      // do nothing
    }
    else if (p instanceof FunctionCallPStmt)
    {
      // XXX Do nothing for now, but this must be removed eventually; we
      // should never handle a FunctionCallPStmt directly
    }
    else
    {
      throw new RuntimeException ("Don't know how to computeCFG for: " +
	  p.getClass ());
    }

    if (!(p instanceof IfPStmt) && !(p instanceof NDGotoPStmt) &&
	p.getDest () == null)
    {
      if (ifExitStmt != null)
	p.setDest (ifExitStmt);
      // 	else if (whileElseStmt != null)
      // 	  p.setDest (whileElseStmt);
      else if (returnSelector != null)
	p.setDest (returnSelector);

      assert p.getDest () != null : "Missed a case for destination of " + p;
    }

    if (nextStmt != null)
      computeCFG (nextStmt, ifExitStmt, whileElseStmt, returnSelector);
  }

  /************************************************************************
   * PFunctionDef and FunctionCallPStmt building methods
   ************************************************************************/

  static public void addEntryPoint (PFunctionDef func)
  {
    PStmt entry = new SkipPStmt (func, null);
    entry.setLabel (func.entryLabel ());
    entry.setNext (func.getHead ());
    func.setHead (entry);
    func.setEntry (entry);
  }

  /**
   * Creates a return selector for this function if one has not been created
   * yet; returns the selector.
   */
  static public void addReturnSelector (PFunctionDef func)
  {
    PStmt returnSelector;

    int selectorType = func.getParent ().getSelectorType ();

    if (func.getNumCallSites () == 0)
      returnSelector = new GotoPStmt ("END");
    else if (func.getNumCallSites () == 1)
      returnSelector = new GotoPStmt (func.returnSiteLabel (0));
    else  
    {
      if (selectorType == PProgram.DDSelector)
	returnSelector = new ReturnSelectorPStmt (func);
      else
	throw new RuntimeException
	  ("No longer supporting non-DD return selector");
    }

    returnSelector.setLabel (func.returnSelectorLabel ());
    returnSelector.setNext (func.getHead ());
    func.setHead (returnSelector);
    func.setReturnSelector (returnSelector);
  }

  /**
   * Returns a prologue and epilogue for {@code call}.
   */
  static public PStmt makeFunctionCall
  (PFunctionDef def, FunctionCallPStmt call)
  {
    /*
    System.err.println ("Making function call for: " +
                        call.getFunctionName () + ", " +
                        call.getCallIndex ());
    */
    assert def != null : "call to null PFunctionDef!";
    FunctionCallPrologue callAsmt = new FunctionCallPrologue (call);
    callAsmt.setDest (def.getEntry ());
    callAsmt.setLabel (call.getLabel ());
    // -- not originally labelled, so we can make our own label
    if (call.getLabel () == null)
      callAsmt.setLabel (def.callSiteLabel (call.getCallIndex ()));
    FunctionCallEpilogue returnAsmt = new FunctionCallEpilogue (call);
    returnAsmt.setLabel (def.returnSiteLabel (call.getCallIndex ()));
    callAsmt.setNext (returnAsmt);
    return callAsmt;
  }
}

