package edu.toronto.cs.yasm.pprogram;

/** 
 * The source of a predicate. In OCaml, this would be
 *
 * type PredicateSource =
 *    WPPredicateSource of PrllAsmtPStmt * Pred
 *  | ConditionalPredicateSource of IfPstmt
 *  | HintPredicateSource
 *
 * This tracks whether a predicate was obtained from a branch in the
 * program, by propagating some assignment over a branch, or by a hint
 * (initial predicates or abstraction patterns).
 *
 * XXX a ConditionalPredicateSource should *really* have a *list* of 
 * XXX associated IfPStmts; OR an Pred should be able 
 * XXX to have multiple sources
 */
public abstract class PredicateSource {

  /**
   * A predicate based on the WP of a parallel assignment over another
   * predicate.
   */
  public static class WPPredicateSource extends PredicateSource 
  {
    PrllAsmtPStmt stmt;
    Pred parent;
    
    public WPPredicateSource (PrllAsmtPStmt s, Pred p)
    {
      stmt = s;
      parent = p;
    }

    public PrllAsmtPStmt getAssignment ()
    {
      return stmt;
    }

    public Pred getParent ()
    {
      return parent;
    }
    
    public String toString ()
    {
      return " ---(" + stmt + ")---> " + parent.getName ();
    }

    public void updateStmt (PProgram p)
    {
      PStmt reincarnation = p.getStmt (stmt.getLabel ());
      assert (reincarnation instanceof PrllAsmtPStmt) : "reincarnation error";
      stmt = (PrllAsmtPStmt) reincarnation;
    }
  }

  /** 
   * A predicate based on the text of a conditional.
   */
  public static class ConditionalPredicateSource extends PredicateSource 
  {
    /* Could be an IfPStmt, a ConditionalBranchPStmt, or a SwitchBranchPStmt */
    PStmt stmt;
    
    public ConditionalPredicateSource (PStmt s)
    {
      stmt = s;
    }
    
    public PStmt getConditional ()
    {
      return stmt;
    }
    
    public String toString ()
    {
      return " ----> " + stmt;
    }
  }

  public static class HintPredicateSource extends PredicateSource
  {
    
    public String toString ()
    {
      return " ----> HINT";
    }

  }
  public static final HintPredicateSource HINT = new HintPredicateSource ();
}
