package edu.toronto.cs.yasm.pprogram;

import edu.toronto.cs.util.*;
import edu.toronto.cs.expr.*;
import edu.toronto.cs.yasm.pprogram.*;

import java.io.*;
import java.util.*;

/**
 * 
 * A {@code PredicateManager} is a class that manages a set of
 * predicates during predicate abstraction. The predicates are
 * partitioned into <b>old</b> and <b>new</b>. New predicates are
 * predicates that have not yet been used during any refinement. Old
 * predicates are predicates that have been used in some refinement
 * already. The class allows to add predicates to the <b>new</b>
 * predicate list, and merge new and old sets.
 *
 * @author <a href="mailto:arie@kanga.fm.sandbox">Arie Gurfinkel</a>
 * @version 1.0
 */
public class SimplePredicateManager
{

  // -- list of predicates we know about
  List<Pred> predicateList;
  // -- list of new predicates being added to us
  List<Pred> newPredicateList;
  
  /**
   * Creates a new <code>SimplePredicateManager</code> instance.
   *
   */
  public SimplePredicateManager ()
  {
    predicateList = new ArrayList<Pred> ();
    newPredicateList = new ArrayList<Pred> ();
  }

  /**
   * Tell if we already have a predicate with the same name as {@code predicate}
   *
   * @param predicate a {@code Pred} value
   */
  protected boolean hasPredicate (Pred predicate)
  {
    return hasPredicate (predicate.getName ());
  }

  protected boolean hasPredicate (Expr predicate)
  {
    for (Pred p : newPredicateList) {
      if (p.getName () == predicate) return true;
    }
    return false;
  }


  /**
   * Adds a new predicate. Returns true if {@code predicate} has been
   * added, and false if it is already contained in the new set.
   *
   * @param predicate a predicate to be added
   */
  public boolean newPred (Pred predicate)
  {
    if (hasPredicate (predicate))
      {
	System.err.println ("Already have predicate: " + predicate);
	return false;
      }

    newPredicateList.add (predicate);
    return true;
  }

  public void makeNew (Expr predicate)
  {
    Pred p = null;
    for (Pred q : predicateList) {
      if (q.getName () == predicate) {
        p = q;
        break;
      }
    }
    assert (p != null) : "makeNew called on nonexistant predicate " 
      + predicate;

    for (Expr e : Pred.getNames (predicateList))
      System.out.println ("@@@@@@@@@@@ predicateList:" + e);
    for (Expr e : Pred.getNames (newPredicateList))
      System.out.println ("@@@@@@@@ newPredicateList:" + e);

    predicateList.remove (p);
    newPredicateList.add (p);

    for (Expr e : Pred.getNames (predicateList))
      System.out.println ("@@@@@@@@@@@ predicateList:" + e);
    for (Expr e : Pred.getNames (newPredicateList))
      System.out.println ("@@@@@@@@ newPredicateList:" + e);

  }

  /**
   * Adds a list of new predicates using {@link #newPred(Pred)}.
   *
   * @param predList a list of predicates to be added
   */
  public void newPred (List<Pred> predList)
  {
    for (Pred p : predList) newPred (p);
  }

  /**
   * Indicates that no new predicates will be added in this round, and
   * that new predicates can be merged with old.
   *
   */
  public void merge ()
  {
    predicateList.addAll (newPredicateList);
    newPredicateList = new ArrayList<Pred> ();
  }

  /**
   * Returns the list of old predicates contained in this manager,
   * as @code{Pred}s.
   *
   * @return a <code>List</code> value
   */
  public List<Pred> getPreds ()
  {
    return predicateList;
  }
  
  /**
   * Returns the list of all new predicates contained in this manager,
   * as @code{Pred}s.
   *
   * @return a <code>List</code> value
   */
  public List<Pred> getNewPreds ()
  {
    return newPredicateList;
  }


  /**
   * Returns the list of old predicates contained in this manager,
   * as @code{Expr}s.
   *
   * @return a <code>List</code> value
   */
  public List<Expr> getPredicates ()
  {
    return Pred.getNames (predicateList);
  }
  
  /**
   * Returns the list of all new predicates contained in this manager,
   * as @code{Expr}s.
   *
   * @return a <code>List</code> value
   */
  public List<Expr> getNewPredicates ()
  {
    return Pred.getNames (newPredicateList);
  }
    
}
