package edu.toronto.cs.yasm.pprogram;

import edu.toronto.cs.yasm.pcfa.*;

import edu.toronto.cs.mvset.MvSet;
import edu.toronto.cs.mvset.MvRelation;
import edu.toronto.cs.mvset.MvSetMvRelation;
import edu.toronto.cs.mvset.SeqMvRelation;

import java.util.*;


/**
 * An implementation {@code PProgramReachabilityCompiler} that
 * compiles a program into an intraprocedural CFG representation. It
 * currently inlines functions with a single call site. 
 *
 *
 * Created: Mon Nov 23 21:25:28 2005
 *
 * @author <a href="mailto:owei@age.cs">Ou Wei</a>, <a
 * href="mailto:arie@cs.toronto.edu>Arie Gurfinkel</a>
 * @version 1.1
 */
public class SummarizingReachabilityCompiler
  extends InliningReachabilityCompiler
{

  /**
   * Creates a new
   * <code>SummarizingReachabilityCompiler</code> instance.
   *
   */
  public SummarizingReachabilityCompiler () 
  {
    super ();
  }



  // ----------------------------------------------------------------------
  // -- VISITOR METHODS
  // ----------------------------------------------------------------------



  /**
   * Compiles return-selector statement.
   *
   * @param sel a <code>ReturnSelectorPStmt</code> value
   * @param o an <code>Object</code> value
   * @return an <code>Object</code> value
   */
  public Object visitReturnSelector (ReturnSelectorPStmt sel, Object o)
  {
    // -- return selector is added while we compile the call statement
    return o;
  }

  

  public Object visitFunctionCallPrologue (FunctionCallPrologue fPro, 
					   Object o)
  {
    // -- create a fun-summary edge corresponding to this prologue
    // -- and attach it to the corresponding call-edge
    
    // XXX for now, we handle functions that are only called ones
    // XXX by inlining, so such functions do not need a fun-summary edge.
    // XXX We can tell what kind of function we are dealing with
    // XXX by looking at its return selector
    if (!(fPro.getCall ().getFunctionDef ().getReturnSelector ()
	  instanceof ReturnSelectorPStmt))
      // -- use Inlining compiler to handle this 
      return super.visitFunctionCallPrologue (fPro, o);
    

    // ----------------------------------------------------------------------
    // -- REAL FUNCTOIN BODY BEGINS HERE 
    // ----------------------------------------------------------------------
    handleCall (fPro);
    return o;
  }

  


  // ----------------------------------------------------------------------
  // -- handleXXX FUNCTIONS THAT DO THE ACTUALL COMPILATION WORK. THESE
  // -- FUNCTIONS ARE EXPECTED TO STORE THEY RESULT IN THE edges FIELD.
  // ----------------------------------------------------------------------  


  protected void handleCall (FunctionCallPrologue fPro)
  {

    /**
       when processing a call, the following nodes are involved in the
       caller and callee.
       
       In the caller:
         - callSite -- where the call originates
	 - callRet  -- where the call terminates
       In the callee
         - fEntry   -- first statement of the calee
	 - fExit    -- the last statement of the callee

       We need to create the following edges
         Call (callSite, fEntry) -- caller-callee edge
         Return  (fExit, callRet)   -- callee-caller edge
         BodySum (fEntry, fExit)    -- callee edge
         FunSum  (callSite, callRet) -- calller edge
       In addition, we need 
         CallSum (callSite, fEntry) -- special version of Call edge
         ReturnSum (fExit, callRet) -- special version of Return edge

	 Note that the only requirement of post-pre map is that it application
         restores all post-variables to their pre-versions. 
         
         Dually, the only requirement for the post-cube is that it contain
         all post-variables and quantifiaction with respect to it removes
         all such variables.

         We can safelly assume that inScope(caller,callee) = inScope(global)

	 For each edge we need the following:
          Call -- inScope(caller).pre-cube, 
                  inScope(callee).post-cube,pre-post,post-pre
	  
	  CallSum -- inScope(caller,callee).pre-cube 
                     inScope(callee).post-cube,pre-post,post-pre
          
          BodySum -- inScope(callee).pre-cube,post-cube,pre-post,post-pre
	  
          FunSum  -- inScope(caller).pre-cube,post-cube,pre-post,post-pre

          Return  -- inScope(callee).vars
	             local(callee).pre-cube
                     local(caller).pre-cube
                note that our local is strange since predicates can 
                be shared between scopes even when they are not global
                (i.e., shared between all scopes)

         ReturnSum -- 
               SelfRecursion (note: caller==callee)
	         inScope(global).vars,pre-post,post-cube
		 inScope (callee).pre-cube,post-pre
               NrmlRecursion                  
                 local(callee).pre-cube
          
        
	  
     */

    // -- find caller and callee
    PFunctionDef caller = fPro.getParent ();
    PFunctionDef callee = fPro.getCall ().getFunctionDef ();
    
    // -- setup caller/callee contexts
    CompilationContext callerCtx = getContext (caller);
    CompilationContext calleeCtx = getContext (callee);

    // -- establish useful edge points
    PStmt funEntry = callee.getEntryPoint ();
    PStmt funExit = callee.getReturnSelector ();
    PStmt callRet = fPro.getNext ();
    
    // -- body summary edge
    SummaryEdge bodyEdge = 
      new SummaryEdge (toNodeId (funEntry),
		       toNodeId (funExit),
		       mvF.toMvRelation (null,
					 calleeCtx.preCube,
					 calleeCtx.postCube,
					 calleeCtx.preToPost,
					 calleeCtx.postToPre));

    // -- relation for call edge
    MvRelation callReln = doPrllAssign (fPro, callerCtx, calleeCtx);

    assert ! (callReln instanceof SkipMvRelation);

    // -- call-sum relation
    // XXX conceptually this extends callReln from a GxL1 <-> GxL2 relation
    // XXX to a GxL1 <-> GxL1xL2 relation
    // XXX doAssert is for debugging only
    MvSetMvRelation.doAssert = false;
    MvSet globalPreCube = mvSetFactory.buildCube
      (pProgram.getPredTable ().getVariableIds (inScope (caller, callee), 0));
    MvRelation callSumReln = 
      callReln.toMvSet ().toMvRelation (null, // invar
					globalPreCube, 
					calleeCtx.postCube,
					calleeCtx.preToPost,
					calleeCtx.postToPre);
    MvSetMvRelation.doAssert = true;



    // -- function summary edge
    // XXX last argument in constructor is not used, will clean later
    FunctionSummaryEdge funSumEdge
      = new FunctionSummaryEdge (toNodeId (fPro),
				 toNodeId (fPro.getNext ()),
				 mvF.toMvRelation (null,
						   callerCtx.preCube,
						   callerCtx.postCube,
						   callerCtx.preToPost,
						   callerCtx.postToPre),
				 isSelfCall (fPro), null);

    
    // -- return edge
    ReturnEdge retEdge = doReturnEdge (funExit, callRet);
    
    // -- build a call and related edges, now will add them 
    edges.add (new CallEdge (toNodeId (fPro),
			     toNodeId (funEntry),
			     callReln,
			     funSumEdge,
			     bodyEdge,
			     retEdge,
			     doSumReturnEdge (funExit, callRet),
			     callSumReln));
    edges.add (funSumEdge);
    edges.add (retEdge);
  }

  // ----------------------------------------------------------------------
  // -- doXXX FUNCTIONS THAT ARE USED BY handleXXX LEVEL TO ACTUALLY 
  // -- MANIPULATE EXPRESSIONS AND MVSETS
  // ----------------------------------------------------------------------

  protected ReturnEdge doReturnEdge (PStmt src, PStmt dst)
  {
    PFunctionDef caller = dst.getParent ();
    PFunctionDef callee = src.getParent ();
    
    CompilationContext callerCtx = getContext (caller);
    CompilationContext calleeCtx = getContext (callee);

    // -- in a recursive call, caller and callee are the same
    // -- which we'll indicate by setting caller to null.
    if (caller == callee) caller = null;

    MvSet calleeLocalPreCube = mvSetFactory.buildCube 
      (pProgram.getPredTable ().
       getVariableIds (excludeScope (calleeCtx.preds, caller), 0));
    MvSet callerLocalPreCube = caller == null ? calleeLocalPreCube :
      mvSetFactory.buildCube (pProgram.getPredTable ().getVariableIds 
			      (excludeScope (callerCtx.preds, callee), 0));
    

    return new ReturnEdge (toNodeId (src),
			   toNodeId (dst),
			   new SkipMvRelation (null,
					       calleeLocalPreCube,
					       callerLocalPreCube),
			   doSkip (calleeCtx.vars));
    
  }
  

  protected ReturnEdge doSumReturnEdge (PStmt src, PStmt dst)
  {
    PFunctionDef caller = dst.getParent ();
    PFunctionDef callee = src.getParent ();
    if (caller == callee)
      return doSumRecReturnEdge (caller, callee, src, dst);

    MvSet calleeLocalCube = mvSetFactory.buildCube 
      (pProgram.getPredTable ().getVariableIds (locals (callee), 0));
    
    return new ReturnEdge (toNodeId (src),
			   toNodeId (dst),
			   new SkipMvRelation (null, calleeLocalCube, null), 
			   null);
					       
  }

  protected ReturnEdge doSumRecReturnEdge (PFunctionDef caller, 
					   PFunctionDef callee,
					   PStmt src, PStmt dst)
  {
    CompilationContext calleeCtx = getContext (callee);
    CompilationContext callerCtx = getContext (caller);

    MvRelation reln = doSkip (globalCtx.vars).toMvRelation
      (null, calleeCtx.preCube, globalCtx.postCube, globalCtx.preToPost,
       calleeCtx.postToPre);

    return new ReturnEdge (toNodeId (src), toNodeId (dst), reln, null);
  }
  
  





  



  // ----------------------------------------------------------------------
  // -- DEBUGGING AND TESTING HELPERS
  // ----------------------------------------------------------------------
  /**
   * Runs a testcase.
   *
   * @param args a <code>String[]</code> value
   * @exception Exception if an error occurs
   */
  public static void main (String[] args) throws Exception
  {
    String fileName = args [0];

    PProgram pProgram = PProgram.parse 
      (new edu.toronto.cs.yasm.abstractor.NullExprAbstractor 
       (new edu.toronto.cs.expr.ExprFactoryImpl ()), fileName);


    SummarizingReachabilityCompiler compiler = 
      new SummarizingReachabilityCompiler ();

    compiler.setPProgram (pProgram);
    compiler.setMaxDDVars (1);
    System.out.println ("Compiling to a reachability structure structure");
    compiler.compile ();
    System.out.println ("Done");
    
  }




  // ----------------------------------------------------------------------
  // -- LEFT OVERS
  // ----------------------------------------------------------------------
  /**
   * Creates a return edge from <code>src</code> to <code>dst</code>
   * <code>forCall</code> is true if this return edge is to be used by the call
   *
   * @param src a <code>PStmt</code> value
   * @param dst a <code>PStmt</code> value
   * @param forCall a <code>boolean</code> value
   * @return a <code>ReturnEdge</code> value
   */
  /*  protected ReturnEdge __doReturnEdge (PStmt src, PStmt dst)
  {
    // -- src : the exit node of callee function (function called)
    // -- dst : the call_return in caller function (function call)


    PFunctionDef caller = dst.getParent ();
    PFunctionDef callee = src.getParent ();

    String calleeName = src.getParent ().getFunctionName ();
    String callerName = dst.getParent ().getFunctionName ();


    List callerPredicates = pm.getFuncPreds (callerName);
    List calleePredicates = pm.getFuncPreds (calleeName);

    List tmpPredicates;    
    MvSet calleLocalCube;
    MvSet callerLocalCube;

    // -- construct cubes of variables local to callee and caller
    if (caller == callee)
      {	

	// -- if this is a self-call statement, we find all the local
	// -- predicates by checking the predicates in caller (or
	// -- callee) one by one.
	
	// -- note that currently for each predicate, only two cases
	// -- are possible
	//
	// -- 1. p is defined over global variables only
	// -- 2. p is defined over local variables only

	List<Expr> locals = excludeScope (calleePredicates, 
						   null);

	MvSet localsCube = 
	  mvSetFactory.buildCube (pProgram.getPredTable ().
				  getVariableIds (locals,0));

	// -- since in self-recursion caller and callee are the same,
	// -- they have the same locals and same local cubes.
	calleeLocalCube = localsCube;
	callerLocalCube = localsCube;	
      }
    else
      {

	// -- for non-self-call statement, checking the difference
	// -- between caller's and callee's predicates
	
	// -- local callee's predicates
	tmpPredicates = excludeScope (calleePredicates, caller);
	calleeLocalCube =  
	  mvSetFactory.buildCube 
	  (pProgram.getPredTable ().getVariableIds (tmpPredicates,
						    0));

	// -- local caller's predicates
	tmpPredicates = excludeScope (callerPredicates, callee);
	callerLocalCube = 
	  mvSetFactory.buildCube 
	  (pProgram.getPredTable ().getVariableIds (tmpPredicates,
						    0));
      }

    // -- variables in the scope of the callee: callee's locals + globals
    List calleeVars = 
      pProgram.getPredTable ().getVariables (calleePredicates);


    // -- for ReturnEdge, the SkipMvRelation need to be done differently,
    //    i.e., considering that the pre-varaibles and post-variables
    //    are different.
    //    suppose A call B, and var(B) = { x, y}, and var(A) = {x, z},
    //    let S_B = (x>3 \land y>4), S_A = (x>5 \land z>6)
    //    for the returnEdge B-->A,
    //    when computing the fwdImage (postImage), we need abstract
    //    away the varible "y" which is only in the scope of B
    //    therefore, fwdImage(B-->A) S_B = (x>3)
    //
    //    when computing the bwdImage (preImage), we need abstract
    //    away the varible "z" which is only in the scope of A
    //    therefore, bwdImage(B-->A) S_A = (x>5)
    //
    //    existsAbsFwdCube : L_callee
    //    existsAbsBwdCube : L_caller
    //
    //    see class SkipMvRelation
    
    return new ReturnEdge (toNodeId (src), 
			   toNodeId (dst),
			   new SkipMvRelation (invar, 
					       calleeLocalCube,
					       callerLocalCube),
			   doSkip (calleeVars));
  }
*/


  /**
   * Creates a return edge from <code>src</code> to <code>dst</code>
   * for building function summary edge. It is almost the same
   * as <code>doReturnEdge</code>. The differences is that here
   * 
   * (1) for non-self-recursive call, "existAbstractBwdCube = null";
   *     the funcSumEdge corresponding to the call is a SeqMvRelation
   *     defined by the  "callSumReln" in call-edge, body-edge and 
   *     the "SkipMvRelation" in the (sum)return-edge i.e., 
   *     G -> G X L_callee; G X L_callee -> G X L_callee; 
   *     L_callee -> \emptyset
   *
   * (2) for self-recursive call, the "reln" (returned by
   *     getRelation()) is not the simplied "SkipMvRelation", but 
   *     the "real" identity relation for global variables with
   *     the existrAbsFwdCube and existAbsBwdCube be "G X L" and "G" 
   *     respectively. This is because we will compose return-edge
   *     with the corresponding "sumCallReln" call-edge and body-edge 
   *     to get the transition relation for funcSumEdgeg (i.e., 
   *     G  -> G X L_callee;   G X L_callee -> G X L_callee;  
   *     G X L_callee -> G, note that here L_caller = L_callee - the 
   *     same set of local predicates in caller and callee). 
   *     Note that to preserve the value of L1 during the pre
   *     and post computation of the funcSumEdge, we need set the
   *     the existAbsFwdCube as "G". 
   *
   *     See updateBodySummary () in CallEdge.java 
   *
   *  
   * Note that for in sum return-edge, only the "reln" (returned by
   *     getRelation()) is used. the "idSrcReln" (returned by
   *     getIdSrcReln()) is not used.
   *
   *
   * @param src a <code>PStmt</code> value
   * @param dst a <code>PStmt</code> value
   * @return a <code>ReturnEdge</code> value
   */
  /*protected ReturnEdge __doSumReturnEdge (PStmt src, PStmt dst)
  {
          
    // -- src : the exit node of callee function (function called)
    // -- dst : the call_return in caller function (function call)


    PFunctionDef caller = dst.getParent ();
    PFunctionDef callee = src.getParent ();

    String calleeName = src.getParent ().getFunctionName ();
    String callerName = dst.getParent ().getFunctionName ();

    // -- A calls B, on the return edge, the type is 
    //    var(B) --> var(A), therefore, we use

    List calleePredicates = pm.getFuncPreds (calleeName);

    int mapIndex;
    int[] calleePostToPreMap;

    mapIndex = 
      ( (Integer) mapFuncToPrePostMapIndex.get (calleeName)).intValue ();

    calleePostToPreMap = arrayFuncPostToPreMap[mapIndex];


    if (caller == callee)
      {
	// -- self-recursive call
	List<Expr> globalPredicates = reduceToScope (calleePredicates, null);

	List globals = 
	  pProgram.getPredTable ().getVariables (globalPredicates);

	int[] globalPreToPostMap = 
	  pProgram.getPredTable ().variableMap (globals, 0, 1);


 	MvSet calleeCube = 	  
 	  mvSetFactory.buildCube (pProgram.getPredTable ().
 				  getVariableIds (calleePredicates,0));

	MvSet globalCube = 	  
	  mvSetFactory.buildCube (pProgram.getPredTable ().
				  getVariableIds (globalPredicates,1));

	MvRelation reln = 
	  doSkip(globals).
	  toMvRelation (invar,	
			calleeCube,
			globalCube,
			globalPreToPostMap,
			calleePostToPreMap);

	
	return new ReturnEdge (toNodeId (src), 
			       toNodeId (dst),
			       reln,
			       doSkip (new ArrayList ()));
      }
    else
      {
	// -- existsAbsFwdCube : L_callee  (callee: called, caller:call)
	// -- existsAbsBwdCube : \emptyset
	

	// -- for non-self-call statement, checking the difference
	// -- between caller's and callee's predicates
	List tmpPredicates;


	// -- local to callee
	tmpPredicates = excludeScope (calleePredicates, caller);
	MvSet calleeLocalCube =
	  mvSetFactory.buildCube 
	  (pProgram.getPredTable ().getVariableIds (tmpPredicates,
						    0));

	// -- if we are creating a return edge for a call
	// -- we need to not abstract any of the variables in the 
	// -- bwdImage
	MvSet existAbstractBwdCube = null;

	return new ReturnEdge (toNodeId (src), 
			       toNodeId (dst),
			       new SkipMvRelation (invar, 
						   calleeLocalCube,
						   existAbstractBwdCube),
			       doSkip (new ArrayList ()));
      }

  }
*/

  /**
   * compile <code>FunctionCallPrologue</code> statement <code>fpro</code> 
   * <p>
   * through this method, we create the callEdge, functionSummaryEdge,
   * summaryEdge (bodyEdge), and returnEdge
   * <p> 
   * Example
   * <p>
   * prologue@@@ -- inc(5) (call edge) --> @@@ entry node
   *   |                                    |
   *   |                                  inc(x)
   *                            body    return x+1
   *   funSum (y: inc(5))                  exit
   *   edge  (y=6)                          |
   *   |                                    |
   *   |                                 summary edge
   *   |                                    |
   *   |                                    |
   *   v                                    v
   * epilogue@@@ --y=ret_in (return edge)-> @@@ exit node
   * <p>
   * @param fpro a <code>FunctionCallPrologue</code> value
   * @param o an <code>Object</code> value
   * @return an <code>Object</code> value
   */
  /* public Object old__visitFunctionCallPrologue (FunctionCallPrologue fpro, Object o)
  {

    MvSet tempInvar = invar;
    
    // -- disable current invariant for handling call-return edges.
    invar = null;

    handleFunctionCallPrologue (edges, fpro);
    
    // -- restore invar value
    invar = tempInvar;

    return o;
    
  } */

  /**
   * called by <code>visitFunctionCallPrologue</code> to compile
   * a <code>FunctionCallPrologue</code> prologue
   * 
   * @param edges an <code>EdgeSet</code> value
   * @param fpro a <code>FunctionCallPrologue</code> value
   */
  /* protected void handleFunctionCallPrologue (EdgeSet edges, 
					     FunctionCallPrologue fpro)
  {


    String callerName = fpro.getParent ().getFunctionName ();
    String calleeName = 
      fpro.getCall ().getFunctionDef ().getFunctionName ();    


    // -- A calls B, on the prologue, the type is 
    //    var(A) --> var(B), therefore, we use
    //    A.preVarCube as the relation's preVarCube, 
    //    B.postVarCube as the relation's postVarCube, and 
    //    set the variable maps using B.preToPostMap, and B.postToPreMap

    MvSet callerPreVarCube;
    MvSet callerPostVarCube;
    MvSet calleePreVarCube;
    MvSet calleePostVarCube;

    int[] callerPreToPostMap;
    int[] callerPostToPreMap;    
    int[] calleePreToPostMap;
    int[] calleePostToPreMap;

    callerPreVarCube = (MvSet) mapFuncToPreVarCube.get (callerName);
    callerPostVarCube = (MvSet) mapFuncToPostVarCube.get (callerName);
    calleePreVarCube = (MvSet) mapFuncToPreVarCube.get (calleeName);
    calleePostVarCube = (MvSet) mapFuncToPostVarCube.get (calleeName);


    System.err.println("_callerPredicates for " + callerName);
    System.err.println(pm.getFuncPreds (callerName));



    int mapIndex;
    
    mapIndex = ( (Integer) mapFuncToPrePostMapIndex.get (callerName)).intValue ();
    callerPreToPostMap = arrayFuncPreToPostMap[mapIndex];
    callerPostToPreMap = arrayFuncPostToPreMap[mapIndex];


    mapIndex = 
      ( (Integer)  mapFuncToPrePostMapIndex.get (calleeName)).intValue ();
    calleePreToPostMap = arrayFuncPreToPostMap[mapIndex];
    calleePostToPreMap = arrayFuncPostToPreMap[mapIndex];



    PStmt fExit = fpro.getCall ().getFunctionDef ().getReturnSelector ();
    // -- if the function is only called once, do inline
    if (!(fExit instanceof ReturnSelectorPStmt))
      {
	handleAssign (fpro,
		      callerPreVarCube,
		      calleePostVarCube,
		      calleePreToPostMap,
		      calleePostToPreMap);
      }
    
    else
      {
	CallEdge callEdge = doFunctionCallEdge (fpro,
						callerPreVarCube,
						callerPostVarCube,
						calleePreVarCube,
						calleePostVarCube,
						callerPreToPostMap,
						callerPostToPreMap,
						calleePreToPostMap,
						calleePostToPreMap);
	edges.add (callEdge);
	edges.add (callEdge.getFunSummary ());
      }    

  }*/


  /**
   * compile the edges for a function call
   * 
   *
   * @param fpro a <code>FunctionCallPrologue</code> value
   * @param _preVarCube a <code>MvSet</code> value
   * @param _postVarCube a <code>MvSet</code> value
   * @param _preToPostMap an <code>int[]</code> value
   * @param _postToPreMap an <code>int[]</code> value
   * @return a <code>CallEdge</code> value
   */
  /*protected CallEdge doFunctionCallEdge (FunctionCallPrologue fpro,
					 MvSet _callerPreVarCube,
					 MvSet _callerPostVarCube,
					 MvSet _calleePreVarCube,
					 MvSet _calleePostVarCube,
					 int[] _callerPreToPostMap,
					 int[] _callerPostToPreMap,
					 int[] _calleePreToPostMap,
					 int[] _calleePostToPreMap)
  {


    // -- relation corresponding to assigning arguments
    MvRelation callAsmt = doPrllAssign (fpro,
					_callerPreVarCube,
					_calleePostVarCube,
					_calleePreToPostMap,
					_calleePostToPreMap);


    PFunctionDef caller = fpro.getParent ();
    PFunctionDef callee = fpro.getCall ().getFunctionDef ();
    
    

    // XXX this are really globals for the call!
    // -- get a cube of all variables in the scope of both caller and callee
    // -- note that these are just globals if caller == callee
    MvSet callGlobalsCube = 
      mvSetFactory.buildCube
      (pProgram.getPredTable ().getVariableIds (inScope (callee, caller), 0));
    
    MvSetMvRelation.doAssert = false;
    
    // -- create call-summary-relation edge. See CallEdge for a reason
    MvRelation callSumReln = toMvRelation (callAsmt.toMvSet (),
					   callGlobalsCube,
					   _calleePostVarCube,
					   _calleePreToPostMap,
					   _calleePostToPreMap);
    
    MvSetMvRelation.doAssert = true;
    
    // -- the first statement of the callee
    PStmt funEntry = fpro.getDest ();
    // -- the last statement of the callee before return
    PStmt funExit = fpro.getCall ().getFunctionDef ().getReturnSelector ();
    // -- return of the call in the caller
    PStmt callRet = fpro.getNext ();

    SummaryEdge bodySumEdge
      = new SummaryEdge (toNodeId (funEntry),
			 toNodeId (funExit),
			 toMvRelation (mvF,
				       _calleePreVarCube,
				       _calleePostVarCube,
				       _calleePreToPostMap,
				       _calleePostToPreMap));




    
    // XXX callGlobalsCube does not seem to be used in FunSumEdge!
    FunctionSummaryEdge funSumEdge
      = new FunctionSummaryEdge (toNodeId (fpro),
				 toNodeId (callRet),
				 toMvRelation (mvF,
					       _callerPreVarCube,
					       _callerPostVarCube,
					       _callerPreToPostMap,
					       _callerPostToPreMap),
				 caller == callee,
				 callGlobalsCube);

    // -- package everything into a call edge
    return new CallEdge (toNodeId (fpro), 
			 toNodeId (funEntry),
			 callAsmt,
			 funSumEdge,
			 bodySumEdge,
			 doReturnEdge (funExit, callRet),
			 doSumReturnEdge (funExit, callRet),
			 callSumReln);
  }

*/
  
}



