package edu.toronto.cs.yasm.reachability;

import edu.toronto.cs.yasm.pcfa.*;
import edu.toronto.cs.mvset.*;

import java.util.*;

/**
 * Describe class ReachabilityChecker here.
 *
 *
 * Created: Wed Jun  8 15:23:10 2005
 *
 * @author <a href="mailto:arie@eon.cs">Arie Gurfinkel</a>
 * @version 1.0
 */
public class ReachabilityChecker 
{

  boolean DEBUG = false;


  ReachabilityStructure model;
  
  Collection workList;
  Collection sumWorkList;
  
  /**
   * NodeSet for reachability phase of the analysis
   *
   */
  NodeSet nodes;
  
  /**
   * NodeSet for summary computation
   *
   */
  NodeSet sumNodes;
  
  
  /**
   * Creates a new <code>ReachabilityChecker</code> instance.
   *
   */
  public ReachabilityChecker (ReachabilityStructure _model) 
  {
    workList = new LinkedList ();
    sumWorkList = new LinkedList ();
    model = _model;
  }


  public ReachabilityStructure getModel ()
  {
    return model;
  }
  

  public void doReachablity ()
  {
    // -- set nodes and sumNodes 
    nodes = model.getNodeSet ().createNext ();
    sumNodes = model.getSumNodeSet ().createNext ();
    
    // -- create initial worklist
    for (Iterator it = model.getNodeSet ().nodes ().iterator (); 
	 it.hasNext ();)
      workList.addAll 
	(model.getEdgeSet ().edgesTo (((Node) it.next ()).getId ()));
    

    do
      {
	// -- do a step in reachability computation
	doReachabilityStep ();
	// -- check if there is more work to do and 
	// -- if so, update nodes and sumNodes
	if (!workList.isEmpty ())
	  {
	    model.setNodeSet (nodes);
	    nodes = nodes.createNext ();
	  }

	if (!sumWorkList.isEmpty ())
	  {
	    model.setSumNodeSet (sumNodes);
	    sumNodes = sumNodes.createNext ();
	  }
      }
    while (! (workList.isEmpty () && sumWorkList.isEmpty ()));

    model.setNodeSet (nodes);
    model.setSumNodeSet (sumNodes);
  }
  
          
  public void doReachabilityStep ()
  {
    // -- store old worklists
    Collection oldWorkList = workList;
    Collection oldSumWorkList = sumWorkList;
    
    // -- create new ones
    workList = new LinkedList ();
    sumWorkList = new LinkedList ();
    
    // -- process work list
    for (Iterator it = oldWorkList.iterator (); it.hasNext (); )
      {
	Edge edge = (Edge) it.next ();
	// -- on a return edge we start a new summary computation
	if (edge instanceof ReturnEdge)
	  {
	    // -- extract a relation from the edge
	    MvRelation reln = ((ReturnEdge)edge).getRelation ();
	    // -- propogate dst labeling over this relation
	    MvSet v = reln.bwdImage (nodes.get (edge.getDst ()).getValue ());
	    // -- convert the set into a subset of the identity relation
	    v = setToReln (v, ((ReturnEdge)edge).getSrcIdReln ());
	    
	    // -- update sumNodes and add edge to summary worklist
	    Node exitNode = sumNodes.get (edge.getSrc ());
	    if (exitNode != null)
	      v = v.or (exitNode.getValue ());
	    
	    // -- check if anything has changed
	    if (exitNode == null || ! v.equals (exitNode.getValue ()))
	      {
		// -- update summary nodes 
		sumNodes.add (edge.getSrc (), v);
		// -- add edges to exitNode to sumWorkList
		sumWorkList.addAll 
		  (model.getEdgeSet ().edgesTo (edge.getSrc ()));
	      }	    
	  }
	else
	  // -- for any other edge we just compute the bwdImage 
	  bwdImage (edge, model.getNodeSet (), nodes, workList);
      }


    if (DEBUG)
      {    
	System.err.println ("\n *************************** ");
	System.err.println (" *** Summary Computation *** \n");
      }
	

    
    // -- process summary work list
    for (Iterator it = oldSumWorkList.iterator (); it.hasNext (); )
      {
	Edge edge = (Edge) it.next ();
	// -- for return edge we start a new summary computation 
	// -- just like before
	if (edge instanceof ReturnEdge)
	  {
	    // -- extract a relation from the edge
	    MvRelation reln = ((ReturnEdge)edge).getRelation ();
	    // -- propogate dst labeling over this relation
	    MvSet v = reln.bwdImage 
	      (sumNodes.get (edge.getDst ()).getValue ());

	    // -- convert the set into a subset of the identity relation
	    v = setToReln (extractPreStates (v), 
			   ((ReturnEdge)edge).getSrcIdReln ());

	    Node exitNode = sumNodes.get (edge.getSrc ());
	    if (exitNode != null)
	      {
		v = v.or (exitNode.getValue ());
		
	      }
	    
	    if (exitNode == null || ! v.equals (exitNode.getValue ()))
	      {
		sumNodes.add (edge.getSrc (), v);
		sumWorkList.addAll 
		  (model.getEdgeSet ().edgesTo (edge.getSrc ()));
	      }
	  }
	// -- reaching a call edge indicates that a summary computation
	// -- has converged and we should instantiate new summary edges
	else if (edge instanceof CallEdge)
	  {
	    // -- time to update summaries

	    // -- extract current labeling of function entry node
	    MvSet summary = sumNodes.get (edge.getDst ()).getValue ();

	    // -- convert a summary to a relation (i.e. expression over 
	    // -- pre- and post- variables 
	    MvSet sumReln = summaryToRelation (summary);

	    
	    // 	    System.out.println ("Got a new sumReln: ");
	    // 	    System.out.println (new Node (new Node.NodeId (0),
	    // 					  sumReln));
	    // 	    System.out.println ("Done with new summary");
	    
	    // -- update the body summary edge
	    // -- and get a new function summary
	    SummaryEdge funSummary = 
	      ((CallEdge) edge).updateBodySummary (sumReln,
						   model.getPostToAuxMap (),
						   model.getAuxToPostMap ());
	    
	    // -- add function summary to the worklist
	    sumWorkList.add (funSummary);
	    workList.add (funSummary);

	  }
	else
	  bwdImage (edge, model.getSumNodeSet (), sumNodes, sumWorkList);
      }
    

    if (DEBUG)
      {    
	System.err.println ("\n *** End of Summary Computation *** ");
	System.err.println (" ********************************** ");
      }

    
  }
  

  private MvSet summaryToRelation (MvSet v)
  {
    return v.renameArgs (model.getAuxToPostMap ());
  }
  
  private MvSet extractPreStates (MvSet v)
  {
    return v.existAbstract (model.getAuxCube ());
  }
  
  private MvSet setToReln (MvSet set, MvSet idReln)
  {
    //     return set.renameArgs (model.getPreToAuxMap ()).and 
    //       (idReln.renameArgs (model.getPostToAuxMap ()));


    //return idReln.renameArgs (model.getPostToAuxMap ());
    return idReln.and (set).renameArgs (model.getPostToAuxMap ());

  }
  
  private void bwdImage (Edge edge, NodeSet oldNodes, 
			 NodeSet nodes, Collection workList)
  {
    System.err.println ("bwdImage over edge: ( " + edge.getClass () 
			+ " ) " + edge);

    if (DEBUG)
      {    
	System.err.println (" *** before computation *** ");

	System.err.println ("look at Node : " + edge.getSrc ());
	System.err.println (nodes.get (edge.getSrc ()));
      }
	

    Node.NodeId[] allDst;

    allDst = edge.getAllDst ();

    if (DEBUG)
      {
	for (int i = 0; i < allDst.length; i++)
	  {
	    System.err.println ("look at Node : " + allDst [i]);
	    System.err.println (nodes.get (allDst [i]));
	  }
      }
	


    if (edge.bwdImage (oldNodes, nodes, true))
      {
	workList.addAll (model.getEdgeSet ().edgesTo (edge.getSrc ()));	
      }


    if (DEBUG)
      {    

	System.err.println (" *** after computation *** ");
	    
	System.err.println ("look at Node : " + edge.getSrc ());
	System.err.println (nodes.get (edge.getSrc ()));
      }

    
  }
  

}
