package edu.toronto.cs.yasm.refiner;

import edu.toronto.cs.tp.*;
import edu.toronto.cs.expr.*;

import java.util.IdentityHashMap;
import java.util.Map;
import java.util.List;
import java.util.LinkedList;
import java.util.BitSet;
import java.util.ArrayList;

import static edu.toronto.cs.expr.ComparisonOp.*;
import static edu.toronto.cs.expr.NumericOp.*;

import static edu.toronto.cs.cparser.CILTransformerUtil.*;

/**
 * A {@link edu.toronto.cs.yasm.refiner.TPCBJPredicateAbstractor} that
 * implements some optimizations for numeric expressions.
 *
 * Optimization 1:
 *    recognize the x <= y and y <= x cannot be simultaneously false
 *
 * Optimization 2:
 *    recognize that if x <= y is false, then x+1 <= y cannot be true
 *
 * @author <a href="mailto:tomhart@cs.toronto.edu">Tom Hart</a>
 * @version 1.0
 */
public class NumericTPCBJPredicateAbstractor extends TPCBJPredicateAbstractor
{

  /* maps x <= y to y <= x */
  Map <Expr,Expr> sisterMap;

  /* given p = (x <= y), let INCS(p) = {x (+ 1)* <= y}, ordered by the 
   * number of +1's (increasing)
   *
   * incsMap maps x+1 <= y to x <= y, if x <= y is in our set of 
   * predicates
   *
   * revIncsMap is incsMap, but with the direction reversed
   * we use revIncsMap and nextPick to (try to) make sure that, once we start 
   * picking variables from INCS(p), we keep picking from INCS(p) until 
   * we're done with it
   */
  Map <Expr, Expr> incsMap;
  Map <Expr, Expr> revIncsMap;
  Expr nextPick;

  private void init_sisterMap (List<Expr> el)
  {
    sisterMap = new IdentityHashMap<Expr, Expr> ();

    for (Expr e : el) {
      if (isLeq (e)) {
        Expr sis = leq (e.arg (1), e.arg (0));
        if (el.contains (sis)) sisterMap.put (e, sis);
      }
    }
  }

  /* If e is of the form "x+1", return x. Otherwise, return null.
   */
  private Expr getPlusOne (Expr e)
  {
    if (e.op () == PLUS)
      if (e.arg (1) == e.getFactory ().intExpr (1))
        return e.arg (0);
    return null;
  }

  private void init_incsMap (List<Expr> el)
  {
    incsMap = new IdentityHashMap<Expr, Expr> ();
    revIncsMap = new IdentityHashMap<Expr, Expr> ();
    nextPick = null;

    for (Expr e : el) {
      if (isLeq (e)) {
        Expr lhs = e.arg (0);
        Expr base = getPlusOne (lhs);
        if (base != null) {
          Expr parent = leq (base, e.arg(1));
          if (el.contains (parent)) {
            incsMap.put (e, parent);
            revIncsMap.put (parent, e);
          }
        }
      }
    }
  }

  public NumericTPCBJPredicateAbstractor (List<Expr> _srcPredicates, 
                                          Expr _targetPredicate, 
                                          ITheoremProver _tp)
  {
    super (_srcPredicates, _targetPredicate, _tp);

    init_sisterMap (_srcPredicates);
    init_incsMap (_srcPredicates);
 }

  public BitSet checkConsistency (CSPVariable<Expr> curVar)
  {
    Expr e = curVar.getExpr ();

    //System.out.println ("checking consistency of " + e);

    // Make sure that we don't have x <= y and y <= x both false
    if (curVar.getCurVal () == 0) {
      Expr sis = sisterMap.get (e);
      if (sis != null) {
        CSPVariable<Expr> sisVar = exprToVarMap.get (sis);
        assert sisVar != null;

        if (sisVar.getCurVal () == 0) {
          BitSet b = new BitSet ();
          b.set (0);
          b.set (curVar.getCurLevel ());
          return b;
        }
      }
    }

    // Make sure that we don't have !(x <= y) and x+1 <= y
    // Since INCS(p) is stored in increasing order, we just have to 
    // check the previous inc
    if (curVar.getCurVal () == 0) {
      Expr parent = incsMap.get (e);
      if (parent != null) {
        CSPVariable<Expr> parentVar = exprToVarMap.get (parent);
        assert parentVar != null;

        if (parentVar.getCurVal () == 1) {
          BitSet b = new BitSet ();
          b.set (0);
          b.set (curVar.getCurLevel ());
          return b;
        }
      }
    }

    return super.checkConsistency (curVar);
  }
 
  /* Make sure that for any p, we always go through INCS(p) in
   * increasing order.
   */
  public CSPVariable<Expr> pickNextVariable ()
  {
    if (nextPick != null) {
      // We've started along an INCS(p) chain --- stay on it.
      CSPVariable<Expr> ret = exprToVarMap.get (nextPick);
      nextPick = revIncsMap.get (nextPick);
      return ret;
    }

    CSPVariable<Expr> candidate = super.pickNextVariable ();
    Expr e = candidate.getExpr ();
    
    // We want to pick the elements of INCS(p) in increasing order;
    // hence, stop if we get to the tail (ie. e has no parent), or if
    // the next variable is assigned (ie. e's parent's value isn't -1)
    while (true) {
      Expr parent = incsMap.get (e);
      if (parent == null) break; // reached the tail
      //System.out.println ("parent of " + e + " is " + parent);
      CSPVariable<Expr> parentVar = exprToVarMap.get (parent);
      assert parentVar != null;
      if (parentVar.getCurVal () != -1) break; // parent is assigned
      e = parent;
    }

    // set nextPick, possibly to a null value
    nextPick = revIncsMap.get (e);

    return exprToVarMap.get (e);
  }

  public static void testcase1 () throws Exception 
  {
    ExprFactory fac = new ExprFactoryImpl ();
    SymbolTable st = new SymbolTable ();

    List<Expr> assumptions = new LinkedList<Expr> ();
    List<Expr> preds = new LinkedList<Expr> ();

    Expr intT = fac.op (SimpleTypeOp.INT);
    Expr intArrayT = CompoundTypeOp.arrayT (intT, intT);

    Expr src = fac.var ("src");
    st.put (src, intArrayT);
    Expr dest = fac.var ("dest");
    st.put (dest, intArrayT);
    Expr src_nullpos = fac.var ("src_nullpos");
    st.put (src_nullpos, intT);
    Expr dest_nullpos = fac.var ("dest_nullpos");
    st.put (dest_nullpos, intT);
    Expr j = fac.var ("j");
    st.put (j, intT);
    Expr i = fac.var ("i");
    st.put (i, intT);
    Expr j_s = fac.var ("j_VERISEC_ij2_line14_0_s");
    st.put (j_s, intT);
    Expr i_s = fac.var ("i_VERISEC_ij2_line14_0_s");
    st.put (i_s, intT);

    assumptions.add (getExpr (fac, "src[src_nullpos] == 0"));
    assumptions.add (getExpr (fac, "src_nullpos >= 0"));
    assumptions.add (getExpr (fac, "dest[dest_nullpos] == 0"));
    assumptions.add (getExpr (fac, "dest_nullpos >= 0"));

    CTheoremProver ctp = 
      new CTheoremProver (fac,  new CLCVCLTheoremProver (fac, true), 
                          st, assumptions);

    preds.add (getExpr (fac, "j <= 99"));
    preds.add (getExpr (fac, "j == dest_nullpos"));
    preds.add (getExpr (fac, "i <= src_nullpos"));
    preds.add (getExpr (fac, "src_nullpos <= i"));
    preds.add (getExpr (fac, "src_nullpos <= 99"));
    preds.add (getExpr (fac, "src[i] == 0"));
    preds.add (getExpr (fac, "(j - j_VERISEC_ij2_line14_0_s) <= (i - i_VERISEC_ij2_line14_0_s)"));
    preds.add (getExpr (fac, "(src_nullpos - i_VERISEC_ij2_line14_0_s) <= (99 - j_VERISEC_ij2_line14_0_s)"));

    Expr target = getExpr (fac, "((0 - j_VERISEC_ij2_line14_0_s) <= (0 - i_VERISEC_ij2_line14_0_s))");
    Expr target2 = getExpr (fac, "i_VERISEC_ij2_line14_0_s <= j_VERISEC_ij2_line14_0_s");

    NumericTPCBJPredicateAbstractor cbj =
      new NumericTPCBJPredicateAbstractor (preds, target2, ctp);

    cbj.recSearch (1);

    System.out.println ("TExpr = " + cbj.getTExpr ());
    System.out.println ("FExpr = " + cbj.getFExpr ());
  }

  public static void test2Str () throws Exception
  {
    ExprFactory fac = new ExprFactoryImpl ();
    Expr i = fac.var ("i");
    Expr j = fac.var ("j");
    Expr i_s = fac.var ("i_s");
    Expr j_s = fac.var ("j_s");
    Expr M = fac.var ("M");
    Expr N = fac.var ("N");

    Expr intT = fac.op (SimpleTypeOp.INT);
    SymbolTable st = new SymbolTable ();
    st.put (i, intT);
    st.put (j, intT);
    st.put (i_s, intT);
    st.put (j_s, intT);
    st.put (M, intT);
    st.put (N, intT);

    List<Expr> preds = new LinkedList<Expr> ();
    preds.add (getExpr (fac, "i <= M"));
    preds.add (getExpr (fac, "M <= i"));
    preds.add (getExpr (fac, "j - j_s <= i - i_s"));
    preds.add (getExpr (fac, "M - i_s + j_s <= N"));

    Expr target = getExpr (fac, "j < N + 1");

    ITheoremProver tp =
      new CTheoremProver (fac, new CLCVCLTheoremProver (fac, true), st);

    NumericTPCBJPredicateAbstractor cbj =
      new NumericTPCBJPredicateAbstractor (preds, target, tp);

    cbj.recSearch (1);

    System.out.println ("true approx: " + cbj.getTExpr ());
    System.out.println ("false approx: " + cbj.getFExpr ());
  }


  public static void main (String[] args)
  {
    try {
      testcase1 ();
      test2Str ();
    } catch (Exception e) {
      System.out.println ("error");
      e.printStackTrace ();
    }

  }

}
