package edu.toronto.cs.yasm.refiner;

import edu.toronto.cs.tp.*;
import edu.toronto.cs.expr.*;

import java.util.IdentityHashMap;
import java.util.Map;
import java.util.List;
import java.util.LinkedList;
import java.util.BitSet;


/**
 * An implementation of CBJ-based predicate abstraction strategy that
 * uses {@link edu.toronto.cs.tp.ITheoremProver} interface to decided
 * validity of implications.
 *
 * @author <a href="mailto:arie@kanga.fm.sandbox">Arie Gurfinkel</a>
 * @version 1.0
 */
public class TPCBJPredicateAbstractor extends CBJPredicateAbstractor<Expr>
{
  // -- our theorem prover
  ITheoremProver tp;

  // -- mapping between expressions and variables of a CSP
  Map<Expr,CSPVariable<Expr>> exprToVarMap;

  // -- an expression that implies the target predicate
  Expr tExpr;
  // -- an expression that implies the negation of the target predicate
  Expr fExpr;

  int tryNum = 0;

  /**
   * Constructs an instance of {@code TPCBJPredicateAbstractor}.
   *
   * @param _srcPredicates a <code>List<Expr></code> a list of
   * expressions used for the approximation
   * @param _targetPredicate a target expression being approximated
   * @param _tp an <code>ITheoremProver</code> used to decide
   * implications
   */
  public TPCBJPredicateAbstractor (List _srcPredicates, 
				   Expr _targetPredicate, 
				   ITheoremProver _tp)
  {
    super (_srcPredicates, _targetPredicate);
    tp = _tp;

//    for (int i = 0; i < variables.length; i++)
//      System.err.println ("Variable: " + i + " is " + variables [i]);
    
//    System.err.println ("target: " + target);
    tExpr = tp.exprFactory ().falseExpr ();
    fExpr = tp.exprFactory ().falseExpr ();

    // -- this has to be an identity map since we assume that 
    // -- Expr are cannonical
    exprToVarMap = new IdentityHashMap<Expr,CSPVariable<Expr>> ();
    for (int i = 0; i < variables.length; i++)
      exprToVarMap.put (variables [i].getExpr (), variables [i]);
    
  }

  
  public BitSet checkConsistency (CSPVariable<Expr> curVar)
  {
    //System.out.println ("try number " + (tryNum++));
//    System.err.println ("in check consistency of " + curVar);
    BitSet set = new BitSet ();
    set.set (0);

    //if (curVar != variables [variables.length - 1]) return set;
    //if (assignedVariables != variables.length - 1) return set;
    if (!(oneVariables.isEmpty () && twoVariables.isEmpty ())) {
      //System.out.println ("done");
      return set;
    }

    int currentStack = tp.stackLevel ();
    tp.push ();


    // -- query 
    //System.err.println ("\tquery: " + target);
    Boolean result = tp.checkTruth (target);

    Expr solution = tp.exprFactory ().trueExpr ();
    
    if (result != null)
      {
	
	List<Expr> assumptions = tp.getAssumptionsUsed ();
	
	
	for (int i = 0; i < assumptions.size (); i++)
	  {
	    // -- obtain the assumption
	    Expr expr = assumptions.get (i);
	    
	    // debug print
//	    System.err.println ("\t" + i + " " + expr);

	    // -- add it to the solution we are constructing
	    solution = 
	      tp.exprFactory ().op (BoolOp.AND).binApply (solution, expr);

	    // -- remove negation if it is present
	    if (expr.op () == BoolOp.NOT)
	      expr = expr.arg (0);
	    
	    CSPVariable<Expr> var = exprToVarMap.get (expr);

	    assert var != null : 
	      "Could not find a variable for " + expr;
	    
	    set.set (var.getCurLevel ());

	  }

	newSolutionEvent (result, solution);
	if (result == Boolean.TRUE)
	  tExpr = tp.exprFactory ().op (BoolOp.OR).binApply (tExpr, solution);
	else 
	  fExpr = tp.exprFactory ().op (BoolOp.OR).binApply (fExpr, solution);
      }

    

    tp.popto (currentStack);
//    System.err.println ("Done with checkConsistency of " + curVar);
    //System.out.println ("Done");
    return set;
  }

  protected void newSolutionEvent (Boolean result, Expr solution)
  {
    
  }
  

  void assign (CSPVariable<Expr> var, int val, int level)
  {
    super.assign (var, val, level);

    Expr assertion = var.getExpr ();
    if (var.getCurVal () == 0)
      assertion = tp.exprFactory ().op (BoolOp.NOT).unaryApply (assertion);
    
    
    var.setCurScope (tp.stackLevel ());
    tp.push ();
    tp.assertFormula (assertion);

    
  }
  void undo (CSPVariable<Expr> var)
  {	    
    tp.popto (var.getCurScope ());
    super.undo (var);
  }
  


  public Expr getTExpr ()
  {
    return (new SimplifyingRewriter ()).rewrite (tExpr);
  }
  public Expr getFExpr ()
  {
    return (new SimplifyingRewriter ()).rewrite (fExpr);
  }

  public static Expr[] testCLCVCLTheoremProver (ExprFactory fac)
  {
	  return testTheoremProver (new CLCVCLTheoremProver (fac));
  }

  public static Expr[] testJCVCLTheoremProver (ExprFactory fac)
  {
	  return testTheoremProver (new JCVCLTheoremProver (fac));
  }

  public static Expr[] testTheoremProver (ITheoremProver tp)
  {
    // -- a very simple tester method
	  ExprFactory exprFac = tp.exprFactory ();

	  List<Expr> preds = new LinkedList<Expr> ();

	  Expr x = exprFac.var ("x");
	  Expr y = exprFac.var ("y");

	  Expr intType = exprFac.op (SimpleTypeOp.INT);
	  tp.declare (x, intType);
	  tp.declare (y, intType);

	  // x == 3
	  preds.add 
		  (exprFac.op (ComparisonOp.EQ).binApply (x, exprFac.intExpr (3)));

	  // x > 0
	  preds.add 
		  (exprFac.op (ComparisonOp.GT).binApply (x, exprFac.intExpr (0)));

	  // y > 0
	  preds.add
		  (exprFac.op (ComparisonOp.GT).binApply (y, exprFac.intExpr (0)));

	  //2
	  Expr x_plus_y = exprFac.op (NumericOp.PLUS).binApply (x, y);

	  // x + y > 0
	  preds.add
		  (exprFac.op (ComparisonOp.GT).binApply (x_plus_y, exprFac.intExpr (0)));



	  // -- (x + y) + ((x + y) + y)   > 0
	  Expr target = exprFac.op (ComparisonOp.GT).
		  binApply 
		  (exprFac.op (NumericOp.PLUS).binApply 
		   (x_plus_y, exprFac.op (NumericOp.PLUS).binApply (x_plus_y, y)),
		   exprFac.intExpr (0));

	  TPCBJPredicateAbstractor cbj = 
		  new TPCBJPredicateAbstractor (preds, target, tp);

	  // edu.toronto.cs.util.StopWatch sw = new edu.toronto.cs.util.StopWatch ();
	  cbj.recSearch (1);

	  return new Expr[] {cbj.getTExpr (), cbj.getFExpr ()};
  }

  public static void main (String[] args)
  {

    ExprFactory exprFac = new ExprFactoryImpl ();

    Expr[] JCVCLResults = testJCVCLTheoremProver (exprFac);
    Expr[] CLCVCLResults = testCLCVCLTheoremProver (exprFac);

    assert (JCVCLResults.length == CLCVCLResults.length);

    for (int i = 0; i < JCVCLResults.length; i++)
    {
	    assert (JCVCLResults [i].equals (CLCVCLResults [i]));
	    System.out.println ("JCVCL says:  " + JCVCLResults [i]);
	    System.out.println ("CLCVCL says: " + CLCVCLResults [i]);
    }

    /*
    edu.toronto.cs.util.StopWatch sw = new edu.toronto.cs.util.StopWatch ();
    cbj.recSearch (1);
    System.err.println ("Done in: " + sw);

    System.err.println ("True set is: " + cbj.getTExpr ());
    System.err.println ("False set is: " + cbj.getFExpr ());
    */
  }
}
