package edu.toronto.cs.yasm.wp;

import edu.toronto.cs.expr.*;
import java.util.*;

/**
 * A weakest pre-condition computer for statements expressed in the
 * language of {@code edu.toronto.cs.expr} expression
 * package. Currently, only supports weakest pre-condition of
 * assignments via substitution.
 *
 * @author <a href="mailto:arie@kanga.fm.sandbox">Arie Gurfinkel</a>
 * @version 1.0
 */
public class ExprWPComputer extends BaseWPComputer implements WPComputer
{
  // -- <Expr,Expr> map that maps variables to their values after an
  // -- assignment
  Map varMap;


  // -- identity WP computer (maps Expr e to e)
  /**
   * An identity {@code WPComputer} that always maps {@code Expr e} to
   * {@code e}
   *
   */
  public static final WPComputer ID =
    new
    BaseWPComputer ()
    {
      public Expr computeWP (Expr expr)
      {
        return expr;
      }
    };

  /**
   * returns true if all children are non-null, false otherwise.
   *
   * Note: replace with {@code recKnownOp (Expr e)} when {@code null}
   * ops are eliminated
   *
   * @param e an <code>Expr</code> value
   * @return a <code>boolean</code> value
   */
  private boolean recNonNull (Expr e)
  {
    if (e == null) return false;

    for (Iterator it = e.args ().iterator (); it.hasNext ();)
      if (!recNonNull ((Expr) it.next ())) return false;
    return true;
  }

  /**
   * Creates a new <code>ExprWPComputer</code> instance.
   *
   * @param stmt a program statement represented by an <code>Expr</code> value
   */
  public ExprWPComputer (Expr stmt)
  {
    varMap = new HashMap ();

    // XXX sadly, stmt.arg (0) can be null if its the non-existent variable
    // which stores the return value of a function call
    if (stmt.op () == CILAssignOp.ASSIGN
	|| CStmtOp.isModifyExpr (stmt))
    {
      if (stmt.arity () > 0 && ArrayOp.isArraySel (stmt.arg (0)))
      {
        // Update the array; mapping is (A, A[x |-> e])
        varMap.put (ArrayOp.getArray(stmt.arg(0)), 
            ArrayOp.upd(ArrayOp.getArray(stmt.arg(0)), 
              ArrayOp.getIndex(stmt.arg(0)),
              stmt.arg(1)));
      }
      else
      {
          varMap.put (stmt.arg (0), stmt.arg (1));
      }
    }
    else
    {
      throw new IllegalArgumentException ("WP of " + stmt + 
          " is not yet supported");
    }
  }

  // -- multiple parallel assignments; PRE: lhs.arity () == rhs.arity ()
  
  /**
   * Creates a new <code>ExprWPComputer</code> instance for a parallel
   * assignment. A parallel assignment is represented by a pair of
   * lists with sematic meaning:
   <br/>
   {@code
     lhs[0] = rhs[0]
     ...
     lhs[n] = rhs[n]}
   <br/>
   *
   * @param lhs a list of variables being assigned
   * @param rhs a list of values assigned
   */
  public ExprWPComputer (Expr lhs, Expr rhs)
  {
    /* XXX suppress this for now; VARARGS workaround	  
       assert (lhs.op ().arity () == rhs.op ().arity ()) :
       lhs.op ().arity () + "!=" + rhs.op ().arity ();
       assert (lhs.arity () == rhs.arity ()) :
       "lhs " + lhs + "\nrhs " + rhs + "\narity mismatch: " + 
       lhs.arity () + "!=" + rhs.arity ();
       */  

    varMap = new HashMap ();

    for (int i = 0; i < lhs.arity (); i++)
      if (ArrayOp.isArraySel (lhs.arg (i)))
        varMap.put (ArrayOp.getArray (lhs.arg (i)), 
            ArrayOp.upd(ArrayOp.getArray (lhs.arg (i)), 
              ArrayOp.getIndex (lhs.arg (i)),
              rhs.arg (i)));
      else
        varMap.put (lhs.arg (i), rhs.arg (i));
  }

  // A SubstGuard which prevents substitution of & (address-of) Expr's
  // XXX where should this go?
  protected static final SubstGuard addrGuard = 
    new SubstGuard ()
    {
      public boolean okSubst (Expr e)
      {
        // XXX Generalize to work with &A[x]
        // XXX May require an interface change. :(
        return e.op () != CILIndirectionOp.ADDR_OF;
      }
    };

  // Rewrites "pseudo boolean": (x > 5) = 0 is rewritten as !(x > 5).  A zero
  // value on the RHS is treated as FALSE and a non-zero value is treated as
  // TRUE. Note that inequality comparison to a non-zero value is unhandled,
  // since (x > 5) != 1 has no specific boolean value under this
  // interpretation.
  private static final ExprRewriter pseudoBoolean =
    new ExprRewriter ()
    {
      public Expr rewrite (Expr e)
      {
        // base case
        if (e.arity () == 0)
          return e;

        // special recursive case
        if (e.arity () == 2)
        {
          if (e.arg (0).op () instanceof BoolOp
              || e.arg (0).op () instanceof ComparisonOp)
          {
            if (RationalOp.isInt (e.arg (1)))
            {
              if (RationalOp.getN (e.arg (1)) == 0)
              {
                if (e.op () == ComparisonOp.EQ)
                  return BoolOp.not (rewrite (e.arg (0)));
                else if (e.op () == ComparisonOp.NEQ)
                  return rewrite (e.arg (0));
                else
                  throw new RuntimeException
                    ("Can't convert pseudo-boolean expression: " + e);
              }
              else
              {
                if (e.op () == ComparisonOp.EQ)
                  return rewrite (e.arg (0));
                else
                  throw new RuntimeException
                    ("Can't convert pseudo-boolean expression: " + e);
              }
            }
          }
        }

        // recursive case
        List<Expr> rwArgs = new ArrayList<Expr> (e.arity ());
        for (Object arg : e.args ())
          rwArgs.add (rewrite ((Expr) arg));

        return e.getFactory ().op (e.op ()).naryApply (rwArgs);
      }
    };

  // Removes double negations: !!(x > 5) is rewritten as (x > 5)
  private static final ExprRewriter doubleNeg =
    new ExprRewriter ()
    {
      public Expr rewrite (Expr e)
      {
        // base case
        if (e.arity () == 0)
          return e;

        // special recursive case
        if (BoolOp.isNot (e) && BoolOp.isNot (e.arg (0)))
          return rewrite (e.arg (0).arg (0));
        
        // recursive case
        List<Expr> rwArgs = new ArrayList<Expr> (e.arity ());
        for (Object arg : e.args ())
          rwArgs.add (rewrite ((Expr) arg));

        return e.getFactory ().op (e.op ()).naryApply (rwArgs);
      }
    };

  // Removes *&, so *&x is rewritten as x.
  private static final ExprRewriter starAddr =
    new ExprRewriter ()
    {
      public Expr rewrite (Expr e)
      {
        // base case
        if (e.arity () == 0)
          return e;

        // special recursive case
        if (CILIndirectionOp.isDeref (e)
            && CILIndirectionOp.isAddrOf (e.arg (0)))
          return rewrite (e.arg (0).arg (0));
        
        // recursive case
        List<Expr> rwArgs = new ArrayList<Expr> (e.arity ());
        for (Object arg : e.args ())
          rwArgs.add (rewrite ((Expr) arg));

        return e.getFactory ().op (e.op ()).naryApply (rwArgs);
      }
    };

  /**
   * Computes the weakest precondition for this statement relative to
   * a predicate phi.
   *
   * @param expr a predicate phi
   * @return an Expr representing phi[x/e], where x is the lhs of an
   * assignment, and e is the rhs
   */
  public Expr computeWP (Expr expr)
  {
    Expr result = expr.guardedSubst (varMap, addrGuard);
    result = pseudoBoolean.rewrite (result);
    result = doubleNeg.rewrite (result);
    result = starAddr.rewrite (result);
    return result;
  }
}

