(* Copyright (C) 2007 Tom Hart.
 * 
 * Traverse a program, and, for each (loop, modified index) pair, try
 * to guess which abstraction protects the modified index. 
 *
 * Now requires that CFG information be computed. Invoke as:
 *  $ cilly --domakeCFG --doscanloops --save-temps --noPrintLn \
 *      --printCilAsIs file.c
 *)

(* Notes on the RD module.
 * 
 * Call computeRDs f on a fundec f to get reaching definitions for a
 * function.
 * 
 * Given a statement s, call RD.getRDs to get a (_,_,IOS.t IH.t)
 * option. The IOS.t IH.t is the important part. It maps variable ids
 * to the ids of definitions of the variable that reach the statement
 * s.
 * 
 * See lines 324-327 of reachingdefs.ml.
 * 
 * RD.defIdStmtHash maps a defID to the statement in which it takes
 * place --- see line 8 of reachingdefs.ml. But, you don't use this
 * directly, but do lookups via RD.getDefIdStmt.  *)


open Pretty
open Cil

module E = Errormsg
module UD = Usedef
module VS = UD.VS (* Sets of varinfos *)  
module RD = Reachingdefs
module IH = Inthash
module D = Dominators

(* A hash table indexed by SIDs of statements. Requires that CFG information 
 * has already been computed. 
 *
 * NB: SIDs are only unique within a function. *)
module SH = Hashtbl.Make (
  struct
    type t = stmt
    let hash (s: stmt) = s.sid
    let equal s1 s2 = s1.sid == s2.sid
  end
)

module IS = Set.Make (struct
  type t = int
  let compare i1 i2 = i1 - i2
end)

module SS = Set.Make (struct 
  type t = stmt
  let compare s1 s2 = s1.sid - s2.sid
end)

module VM = Map.Make (struct 
  type t = varinfo
  let compare v1 v2 = v1.vid - v2.vid
end)

module STM = Map.Make (String)
module STS = Set.Make (String)

type direction_t = Dir_up | Dir_down | Dir_unknown

type ptr_type_t = Ptr_safe | Ptr_seq | Ptr_wild

(* so that we just need one visitor wrapper for each of our utility
   visitors, and not getXXXStmt, getXXXFunction, etc. *)
type scope = ScB of block | ScF of fundec | ScS of stmt

(* For a pointer dereference or array index, say if:
   - the access is safe, not needing any analysis 
   (ie. A[0], *p where p has no arithmetic, etc.)
   - the access is (potentially) analyzable
   - the access is something we can't analyze 
   (non-linear arithmetic, multi-dimensional arrays, etc.)
*)
type anal_t = Anal_safe | Anal_analyzable | Anal_weird

type bounder = {
  var : varinfo;
  bound : exp option;
  array : varinfo option;
  mutable direction : direction_t;
}

module BS = Set.Make (struct 
  type t = bounder
  let compare b1 b2 = b1.var.vid - b2.var.vid
end)

type apeType_t = 
    AT_1ex of exp
    | AT_1str of varinfo option
    | AT_2ex of direction_t * varinfo * exp * direction_t
    | AT_2str of varinfo * varinfo option
        
(* APE = Abstraction Pattern Entry *)
type ape = {
  ape_func: fundec;
  ape_loopLabel: string;
  ape_var: varinfo;
  ape_type: apeType_t;
}

(* ******************* LITTLE UTILITY FUNCTIONS ************************* *)

let isPotentialIterator v =
  match unrollType v.vtype with (* unrollType gets rid of TNamed *)
      TInt _ | TPtr _ | TEnum _ -> true
    | _ -> false

let isChar_type t = 
  match t with
      TInt (ik,_) -> begin
        match ik with
            IChar | ISChar | IUChar -> true
          | _ -> false
      end
    | _ -> false

let isChar e = 
  isChar_type (typeOf e)

let isChar_var v =
  isChar_type (unrollType v.vtype)

let isInt_var v =
  match unrollType v.vtype with
      TInt _ -> true
    | _ -> false

let isPtr_var v =
  match unrollType v.vtype with
      TPtr _ -> true
    | _ -> false

let isArray_var v =
  match unrollType v.vtype with
      TArray _ -> true
    | _ -> false

let isComparison (b: binop) =
  match b with
      Lt | Gt | Le | Ge | Eq | Ne | BAnd | BXor | BOr | LAnd | LOr -> true
    | _ -> false

let isArith (b: binop) =
  if (isComparison b) then false else true

let sanitize_filename_abs (s: string) =
  let end_s =
    let r_prefix = Str.regexp ".*/" in
      Str.global_replace r_prefix "" s
  in
  let r_doti = Str.regexp "\.i" in
  let r_dotc = Str.regexp "\.c" in
  let r_dotslash = Str.regexp "\./" in
  let r_minus = Str.regexp "\-" in
    Str.global_replace r_doti ""
      (Str.global_replace r_dotc ""
         (Str.global_replace r_dotslash "" 
             end_s
         )
      )

let sanitize_filename (s: string) =
  let r_minus = Str.regexp "\-" in
    (Str.global_replace r_minus "_"
        (sanitize_filename_abs s)
    )

(* OPERATIONS ON BOUNDERS *)

let bounder_of_varinfo (v: varinfo) =
  {var = v; bound = None; array = None; direction = Dir_unknown;}

let bounder_of_varinfo_eo (v: varinfo) (e: exp option) =
  let newBound =
    match e with
        Some e' -> e
      | None -> Some zero
  in
    {var = v; bound = newBound; array = None; direction = Dir_unknown;}

let bounder_of_varinfo_ao (v: varinfo) (ao: varinfo option) =
  {var = v; bound = None; array = ao; direction = Dir_unknown;}

let bounder_of_varinfo_array (v: varinfo) (a: varinfo) =
  bounder_of_varinfo_ao v (Some a)

let bs_of_vs (vs: VS.t) =
  let bs = ref BS.empty in
    VS.iter
      (fun v -> bs := BS.add (bounder_of_varinfo v) !bs)
      vs;
    !bs

let vars_in_bs (bs: BS.t) =
  let vs = ref VS.empty in
    BS.iter (fun b -> vs := VS.add b.var !vs) bs;
    !vs

let bs_has_var (v: varinfo) (bs: BS.t) =
  BS.exists (fun b -> b.var.vid = v.vid) bs

let bs_vs_inter (bs: BS.t) (vs: VS.t) =
  BS.filter (fun b -> VS.exists (fun v -> v.vid == b.var.vid) vs) bs

(* can throw Not_found *)
let bs_find_var (v: varinfo) (bs: BS.t) =
  let candidates = BS.filter (fun b -> b.var.vid == v.vid) bs in
  let _ = if (BS.cardinal candidates > 1) then E.warn "multiple candidates" in
  BS.choose candidates

let getVarString (b: bounder) =
  b.var.vname

let string_of_bound (b: exp) =
  (sprint 80 (d_exp () b))

let getBoundString (b: bounder) =
  match b.bound with
      Some e -> string_of_bound e
    | None -> "XXX-UNKNOWN-XXX"

let getArrayString_var_array (var: varinfo) (array: varinfo option) =
  match array with
      None -> if (isPtr_var var) then "base(" ^ var.vname ^ ")" else ""
    | Some v -> v.vname

let getArrayString (b: bounder) =
  getArrayString_var_array b.var b.array

let string_of_direction_t (d: direction_t) =
  match d with
      Dir_up -> "up"
    | Dir_down -> "down"
    | Dir_unknown -> "unknown"

let getDirectionString (b: bounder) =
  string_of_direction_t b.direction

let hasDirection (b: bounder) =
  match b.direction with
      Dir_up | Dir_down -> true
    | Dir_unknown -> false 

let getBoundOrBust (b:bounder) =
  match b.bound with 
      Some e -> e
    | None -> failwith "INVALID BOUNDER -- NO BOUND!"

(* OPERATIONS ON APES *)


let mkApe (f: fundec) (label: string) (var: varinfo) (apt: apeType_t) =
  { ape_func = f;
    ape_loopLabel = label;
    ape_var = var;
    ape_type = apt;
  }

let ape_1ex (f: fundec) (label: string) (var: varinfo) (b: exp) =
  mkApe f label var (AT_1ex b)

let ape_1str 
    (f: fundec) (label: string) (var: varinfo) (array: varinfo option) =
  mkApe f label var (AT_1str array)

let ape_2ex (f: fundec) (label: string) (var: varinfo) (varDir: direction_t) 
    (leader: varinfo) (bound: exp) (leaderDir: direction_t) =
  mkApe f label var (AT_2ex (varDir, leader, bound, leaderDir))

let ape_2str (f: fundec) (label: string) (var: varinfo)
    (leader: varinfo) (array: varinfo option) =
  mkApe f label var (AT_2str (leader, array))

let string_of_ape (a: ape) =
  let part2 =
    match a.ape_type with
        AT_1ex bound -> 
          "<numvars='1'><type='ex'><bound='" 
          ^ string_of_bound bound
          ^ "'>"
      | AT_1str array -> 
          "<numvars='1'><type='str'><array='" 
          ^ getArrayString_var_array a.ape_var array ^ "'>"
      | AT_2ex (varDir, leader, bound, leaderDir) -> 
          "<numvars='2'><type='ex'><varDirection='" 
          ^ string_of_direction_t varDir ^ "'><leader='"
          ^ leader.vname ^ "'><bound='"
          ^ string_of_bound bound ^ "'><leaderDirection='"
          ^ string_of_direction_t leaderDir ^ "'>"
      | AT_2str (leader, array) -> 
          "<numvars='2'><type='str'><leader='"
          ^ leader.vname ^ "'><array='"
          ^ getArrayString_var_array leader array ^ "'>"
  in
    "<function='" ^ a.ape_func.svar.vname 
    ^ "'><loop='" ^ a.ape_loopLabel 
    ^ "'><var=\'" ^ a.ape_var.vname ^ "\'>" 
    ^ part2



let getVerisecLabel (s: stmt) =
  let hasVerisec (l: label) =
    match l with 
        Label (str,_,_) ->
          let r = Str.regexp "VERISEC.*" in
            Str.string_match r str 0
      | _ -> false
  in
    try
      let l = List.find hasVerisec s.labels in
        match l with 
            Label (str,_,_) -> str
          | _ -> E.s (E.bug "logic error")
    with
        Not_found -> failwith "loop not labelled"
          

let getAPEs (al: ape list) (s: stmt) =
  List.filter (fun a -> a.ape_loopLabel = (getVerisecLabel s)) al

let has2vars (al: ape list) =
  let is2var (a: ape) =
    match a.ape_type with 
        AT_2ex _ | AT_2str _ -> true
      | _ -> false
  in
    List.exists (fun a -> is2var a) al


(* ******************* OUTPUT FUNCTIONS ********************************* *)

let printVar (v:varinfo) = 
  ignore (printf "%s,  " v.vname)

let printVS (name: string) (vs : VS.t) = 
  ignore (printf "%s = { " name);
  VS.iter (fun v -> printVar v) vs;
  ignore (printf "}\n")

let string_of_bounder (b: bounder) = 
  let varString =
    "<var=\'" ^ b.var.vname ^ "\'> "
  in
  let expString =
    match b.bound with
        None -> ""
      | Some e -> "<bound=\'" ^ getBoundString b ^  "\'> "
  in
  let arrayString =
    match b.array with
        None -> ""
      | Some v -> "<array=\'" ^ v.vname ^  "\'> "
  in
  let dirString =
    "<direction=\'" ^ getDirectionString b ^ "\'> "
  in
    varString ^ expString ^ arrayString ^ dirString
    
let printBS (name: string) (bs : BS.t) =
  ignore (printf "%s = { " name);
  BS.iter (fun b -> ignore (printf "%s " (string_of_bounder b))) bs;
  ignore (printf "}\n")


let string_of_ptr_type_t (ptt : ptr_type_t) =
  match ptt with
      Ptr_safe -> "SAFE" | Ptr_seq -> "SEQ" | Ptr_wild -> "WILD"

let string_of_anal a =
  match a with Anal_safe -> "SAFE" 
    | Anal_analyzable -> "ANALYZABLE" 
    | Anal_weird -> "UNANALYZABLE"

(* ******************* HELPER VISITORS, ETC. **************************** *)

let isScalar (lv: lval) : bool = 
  match lv with (Var v, NoOffset) -> true | _ -> false

(* return true if an offset involves array indexing *)
let rec hasIndexing (off: offset) : bool = 
  match off with
      NoOffset -> false
    | Field (_, off') -> hasIndexing off'
    | Index _ -> true

let isUnOpLinear o = 
  match o with Neg -> true | _ -> false

let isBinOpLinear b =
  match b with
      PlusA | PlusPI | IndexPI | MinusA | MinusPI | MinusPP -> true
    | _ -> false


(* return true iff exp is linear arithmetic of scalar variables *)
let rec isExpLinear e =
  match e with
      Const _ -> true
    | Lval lv -> isScalar lv
    | CastE (t,e') -> isExpLinear e'
    | UnOp (o,e',t) -> isUnOpLinear o && isExpLinear e'
    | BinOp (o,e1,e2,t) -> isBinOpLinear o && isExpLinear e1 && isExpLinear e2
    | _ -> false

let isUnidimensionalArray lv =
  match lv with
      Var v, Index (e, off) -> not (hasIndexing off) && (isExpLinear e)
    | Mem e, off -> not (hasIndexing off) && (isExpLinear e)
    | _ -> false

let isExpUnidimensionalArray e =
  match e with Lval lv -> isUnidimensionalArray lv | _ -> false

let isBitmasking e =
  match e with
      BinOp (o,e1,e2,t) -> (match o with BAnd -> true | _ -> false)
    | _ -> false

let isExpStartOfLv e = match e with StartOf _ -> true | _ -> false

(* can assign an iterator any combination of constants, scalar
   variables, function call results and linear arithmetic *)
let isInstrValid (i: instr) : bool =
  match i with 
      Set (_, e, _) -> isExpStartOfLv e or isExpLinear e
    | Call _ -> true (* be optimistic --- fcn may be analyzable *)
    | _ -> false     (* inline assembly definitely bad *)

(* Return true if instr sets lhs to the start of some object *)
let isInstrStartOfLv (i: instr) : bool =
  match i with
      Set (_, e, _) -> isExpStartOfLv e
    | _ -> false


let visitScope (vis: cilVisitor) (sc: scope) = 
  match sc with
      ScB b -> ignore (visitCilBlock vis b)
    | ScF f -> ignore (visitCilFunction vis f)
    | ScS s -> ignore (visitCilStmt vis s)

(* Collect all loops in an AST *)
class loopCollectorVisitor (ss: SS.t ref) = object(self)
  inherit nopCilVisitor

  method vstmt (s: stmt) : stmt visitAction =
    let _ = match s.skind with 
        Loop _ -> ss := SS.add s !ss
      | _ -> ()
    in
      DoChildren

end

let getLoops (f: fundec) : SS.t = 
  let ss = ref SS.empty in
    visitCilFunction (new loopCollectorVisitor ss) f;
    !ss

(* Collect all assignments that define v, an <instr,stmt> pairs *)
class varDefCollectorVisitor (v: varinfo) (isl: (instr * stmt) list ref) = 
object(self)
  inherit nopCilVisitor

  method vstmt (s: stmt) : stmt visitAction =
    let rec proc_il il =
      match il with
          hd :: tl -> 
            let uses, defs = UD.computeUseDefInstr hd in
              if (VS.mem v defs) then isl := (hd,s) :: !isl;
              proc_il tl
        | _ -> ()
    in
    let _ = match s.skind with
        Instr il -> proc_il il | _ -> ()
    in
      DoChildren

end

(* Get all instructions in a statement which define a variable, plus
   the statements containing each instruction. *)
let getDefs (v: varinfo) (sc: scope) : (instr * stmt) list =
  let ret = ref [] in
  let vis = new varDefCollectorVisitor v ret in
  let _ = visitScope vis sc in
    !ret

(* This does the following:
    - noffVS gets things outside of [] and *()
    - offBS gets things under *one* level of [] and *()
    - anything put in noffVS or offBS must obey filter

   - combine this with the usedef module to get *other* vars

   - note that the bounder stuff is really ghetto -- it only works
   with explicit array indexing, otherwise just says "base(ptr)"
*)
class separateCollectionVisitor 
  (noffVS: VS.t ref) (offBS: BS.t ref) (filter: varinfo -> bool) = object(self) 
    inherit nopCilVisitor
      
    method vlval lv =
      let getExpVars e = VS.filter filter (UD.computeUseExp e) in
      let rec proc_off off base = 
        if not (hasIndexing off) then ()
        else begin
          match off with
              Index (e, off') -> 
                if isConstant e then proc_off off' base else begin
                  if not (hasIndexing off') then 
                    VS.iter 
                      (fun v' -> 
                        offBS := 
                          BS.add (bounder_of_varinfo_array v' base) !offBS) 
                      (getExpVars e)
                end
            | _ -> ()
        end
      in
      let _ =
        match lv with
            (* make no attempt to do the right thing if we have 
               ( *p )[i], where p is a safe pointer *)
            Mem e, NoOffset ->
              if isExpLinear e then 
                offBS := BS.union !offBS (bs_of_vs (getExpVars e))
          | Var v, off -> 
              if filter v then noffVS := VS.add v !noffVS;
              proc_off off v
          |  _ -> ()
      in
        SkipChildren

end

let useBreakdownExp (e: exp) (f: varinfo -> bool) : VS.t * BS.t * VS.t =
  let normal = ref VS.empty in
  let off = ref BS.empty in
  let vis = new separateCollectionVisitor normal off f in
  let _ = ignore (visitCilExpr vis e) in
  let offVS = vars_in_bs !off in
  let uses = UD.computeUseExp e in 
    !normal, !off, VS.diff (VS.diff uses !normal) offVS


(* ripped off from usedef.ml. grabs:
   - variables that do not appear in offsets
   - variables that do appear in offsets
   - anything "weird"

   - f is used by the visitor to decide if variables should be tracked
   in the special sets, or left in the "weird" set
   - ig is "instr guard" ---> only visit instructions that match this 
   predicate
*)
let useBreakdownStmt (s: stmt) (f: varinfo -> bool) (ig: instr -> bool) : 
    VS.t * BS.t * VS.t =
  let normal = ref VS.empty in
  let off = ref BS.empty in
  let vis = new separateCollectionVisitor normal off f in
  let ve e = ignore (visitCilExpr vis e) in
  let _ =
    match s.skind with
        Return (None, _) -> ()
      | Return (Some e, _) -> ve e
      | If (e, _, _, _) -> ve e
      | Break _ | Goto _ | Continue _ -> ()
      | Loop (_, _, _, _) -> ()
      | Switch (e, _, _, _) -> ve e
      | Instr il -> 
          List.iter (fun i -> if ig i then ignore (visitCilInstr vis i)) il
      | TryExcept _ | TryFinally _ -> ()
      | Block _ -> ()
  in
  let offVS = vars_in_bs !off in
  let uses,_ = UD.computeUseDefStmtKind s.skind in 
    !normal, !off, VS.diff (VS.diff uses !normal) offVS

(* Given a bounder and a loop, tell if the bounders iterator goes up,
   down, or is unknown. *)
let fillInDirection (b: bounder) (s: stmt) =
  let var = b.var in
  let exp_has_var e = VS.mem var (UD.computeUseExp e) in
  let rec strip_casts e = 
    match e with
        CastE (t,e') -> strip_casts e'
      | _ -> e
  in
  let defList, _ = List.split (getDefs var (ScS s)) in
  let instr_is i func =
    match i with 
        Set (lv, e, _) -> begin
          match (strip_casts e) with
              BinOp (b,e1,e2,_) -> 
                (func b) && (exp_has_var e1 || exp_has_var e2)
            | _ -> false
        end
      | _ -> false
  in
  let instr_is_plus i =
    let opIsPlus o =
      match o with PlusA | PlusPI | IndexPI -> true | _ -> false
    in
      instr_is i opIsPlus
  in
  let instr_is_minus i =
    let opIsMinus o =
      match o with  MinusA | MinusPI | MinusPP -> true | _ -> false
    in
      instr_is i opIsMinus
  in
  let has_pluses = List.exists instr_is_plus defList in
  let has_minuses = List.exists instr_is_minus defList in
    if (has_pluses && (not has_minuses)) then b.direction <- Dir_up
    else if (has_minuses && (not has_pluses)) then b.direction <- Dir_down

(* Stupid bloody hack..... :p *)
let getDirection_var (v: varinfo) (s: stmt) = 
  let fakeBounder = bounder_of_varinfo v in
    fillInDirection fakeBounder s;
    fakeBounder.direction
      
let getDirectionString_var (v: varinfo) (s: stmt) =
  let fakeBounder = bounder_of_varinfo v in
    fillInDirection fakeBounder s;
    getDirectionString fakeBounder

let hasDirection_var (v: varinfo) (s: stmt) =
  let fakeBounder = bounder_of_varinfo v in
    fillInDirection fakeBounder s;
    hasDirection fakeBounder

(* Gives each loop in a program a unique label. *)
class loopLabellerVisitor (f: file)
  = object(self)
    inherit nopCilVisitor

    val mutable counter : int = 0

    method vstmt (s: stmt) : stmt visitAction =
      let filename = sanitize_filename (f.fileName) in
      let _ =
        match s.skind with
            Loop (_,loc,_,_) -> 
              printf "processing %a\n" d_stmt s;
              s.labels <- s.labels @ 
                [(Label ("VERISEC_" ^ filename ^ 
                            "_line" ^ (string_of_int loc.line) ^
                            "_" ^ (string_of_int counter), 
                        !currentLoc, false))];
              counter <- counter + 1
          | _ -> ()
      in
        DoChildren

end

(* Tell if a statement is contained within a given AST *)
class hasStmtVisitor (needle: stmt) (br: bool ref) 
  = object(self)
  inherit nopCilVisitor

  method vstmt (s: stmt) : stmt visitAction =
    if (s.sid == needle.sid) then 
      (br := true; SkipChildren)
    else 
      DoChildren

end

(* Tell if a statement is contained within a given scope (AST). *)
let hasStmt (sc: scope) (needle: stmt) : bool =
  let br = ref false in
  let vis = new hasStmtVisitor needle br in
  let _ = visitScope vis sc in
    !br

(* Get the gotos and returns out of a given loop *)
class exitCollectorVisitor (loop: stmt) (ss : SS.t ref) = object(self)
  inherit nopCilVisitor

  method vstmt (s: stmt) : stmt visitAction =
    match s.skind with
        Return _ -> (ss := SS.add s !ss; SkipChildren)
      | Goto (sr, l) -> begin
          if (not (hasStmt (ScS loop) !sr)) then begin
            ss := SS.add s !ss;
          end;
          DoChildren
        end
      | _ -> DoChildren

end

let getExits (s: stmt) : SS.t =
  let ss = ref SS.empty in
    visitCilStmt (new exitCollectorVisitor s ss) s;
    !ss

(* Get exit conditionals of a set of loop exits 

   XXX We're using an approximation of control dependence based on
   dominators rather than postdominators.
*)
class exitConditionalCollector 
  (ih: stmt option IH.t) (breaks: SS.t) (condsOfBreak: stmt SH.t)
  = object(self)
    inherit nopCilVisitor

    method vstmt (s: stmt) : stmt visitAction =
      let blockDominates (b: block) (break: stmt) =
        try let first = List.hd b.bstmts in D.dominates ih first break
        with Failure "hd" -> false
      in
        match s.skind with 
            If (e,tb,eb,l) -> begin
              SS.iter 
                (fun break -> 
                  if (blockDominates tb break || blockDominates eb break)
                  then SH.add condsOfBreak break s
                )
                breaks;
              DoChildren
            end
          | _ -> DoChildren
              
  end

(* Get a loop's exits and the associated exit conditionals *)
let getLoopExitConditionals 
    (loopBreakMap: stmt SH.t) (breakCondsMap: stmt SH.t)
    (loop: stmt) (ih: stmt option IH.t) = 
  let ss = getExits loop in
  let vis = (new exitConditionalCollector ih ss breakCondsMap) in
    SS.iter (fun ex -> SH.add loopBreakMap loop ex) ss;
    ignore (visitCilStmt vis loop)


(* Writes the list of abstractions to a file *)
let output_abstractions (al: ape list) (filename: string) = 
  let out_chan = open_out (filename ^ ".abs") in
  let _ =
    print_endline ("about to write to " ^ filename ^ ".abs .....")
  in
    List.iter
      (fun a ->
        output_string out_chan ((*"<file='" ^ filename ^ "'>" ^*) 
          string_of_ape a ^ "\n")
      )
      al;
    flush out_chan

(* ****************************** FLOW STUFF ******************************* *)

(* Map a definition ID which we know is correct to its corresponding
   statement. Crash the program if we pass in an invalid definition
   ID. *)
let stmt_of_def (defid: int) : stmt =
  let so = RD.getDefIdStmt defid in
    match so with
        Some (st) -> st
      | None -> failwith ("WOO Invalid def id: " ^ (string_of_int defid))

(* Given a definition id, get the associated definition. The RD module
   doesn't seem to give us a map to the exact instruction, so return a
   list of possible candidates. List should normally have size 1, or
   close to 1. *)
let instrs_of_def (defid: int) : (instr * stmt) list =
  let stm = stmt_of_def defid in
  let rhso = RD.getSimpRhs defid in
  let isDef inst =
    match rhso with
        Some r -> begin match r with
            (* multiple assignments could have the same rhs... be conservative *)
            (* For some reason, checking e = e' can caust
               out-of-memory errors; observed when e was "dp.dname" *)
            RD.RDExp e -> (match inst with Set (_,e',_) -> Util.equals e e' | _ -> false)
          | RD.RDCall i -> Util.equals inst i
        end
      | _ -> false
  in
    match stm.skind with
        Instr il -> 
          let il' = List.find_all (fun i -> isDef i) il in
          let ret = ref [] in
            List.iter (fun i -> ret := (i,stm) :: !ret) il';
            !ret
      | _ -> failwith "not an il"

(* Inverse of instrs_of_def. Unfortunately expensive. *)
(* XXX if necessary, could construct a map from sids to def ids in the statement *)
let instr_to_defIds (i: instr) : int list =
  let (all_defs,_) = List.split (IH.tolist RD.ReachingDef.defIdStmtHash) in
    List.filter 
      (fun did ->
        let (candidates,_) = List.split (instrs_of_def did) in
          List.mem i candidates) 
      all_defs

(* Given a var used in a stmt/instruction/whatever, and the iosh
   associated with the stmt/instruction/whatever, get the reaching
   definitions relevant to that variable, represented as (instruction,
   stmt containing said instruction) pairs *)
let getVarDepInstrStmts (rhsVar: varinfo) (iosh: RD.IOS.t IH.t) : (instr * stmt) list =
  let ret = ref [] in
    try
      let ios = IH.find iosh rhsVar.vid in
        RD.IOS.iter 
          (fun io -> match io with
              Some i -> ret := !ret @ (instrs_of_def i)
            | None -> ()
          )
          ios;
        !ret
    with Not_found -> []

(* Given a variable in a statement and the RD information for that
   statement, get the statements which define the variable. *)
(* is v an lhs or rhs variable????
 * 
 * I'm almost sure v is a rhs var. RDs are computed on a
 * statement. Therefore, we care about what reaches INTO the statement;
 * ie. the rhs vars.
 *)
let getVarDepStmts (rhsVar: varinfo) (iosh: RD.IOS.t IH.t) : SS.t =
  let ss = ref SS.empty in
    try
      let ios = IH.find iosh rhsVar.vid in
        RD.IOS.iter 
          (fun io -> match io with
              Some i -> ss := SS.add (stmt_of_def i) !ss
            | None -> ()
          )
          ios;
        !ss
    with Not_found -> SS.empty

(* Given a definition and a loop, tell if the definition is in scope of the given 
   loop.
*)
let def_in_loop (defId: int) (loop: stmt) =
  let stmo = RD.getDefIdStmt defId in
    match stmo with
        None -> E.s (E.bug "no stmt for def id %d\n" defId)
      | Some stm -> hasStmt (ScS loop) stm

(* Given a definition and a loop, tell if the definition is dominated
   by the given loop.
*)
let def_dominated_by_loop (defId: int) (loop: stmt) (ih: stmt option IH.t) =
  let stmo = RD.getDefIdStmt defId in
    match stmo with
        None -> E.s (E.bug "no stmt for def id %d\n" defId)
      | Some stm -> D.dominates ih loop stm


(* given an <instr,stmt> pair, return the definition ids of all
   reaching definitions through all vars *)
let getReachingDefIds ((i, s) : (instr * stmt)) : IS.t = 
  let ret = ref IS.empty in
  let proc_ildl ildl = 
    List.iter
      (fun (i',(_,x,iosh)) -> 
        begin
          let uses, _ = UD.computeUseDefInstr i' in
            VS.iter 
              (fun usedVar ->
                try
                  let ios = IH.find iosh usedVar.vid in
                    RD.IOS.iter 
                      (fun io -> match io with
                          Some defid -> ret := IS.add defid !ret
                        | None -> ()
                      )
                      ios
                with 
                    Not_found -> ()
              )
              uses
        end
      )
      ildl;
  in
  let _ =
    match RD.getRDs s.sid with
        None -> ()
      | Some (_,x,iosh) -> begin
          match s.skind with
              Instr il -> begin
                (* Standard stuff stolen from CIL extensions *)
                let rd_dat_lst = RD.instrRDs il s.sid ((),x,iosh) false in
                let il_dat_lst = List.combine il rd_dat_lst in
                let filteredList = List.filter (fun (i',_) -> Util.equals i i') il_dat_lst in
                  proc_ildl filteredList
              end
            | _ -> E.s (E.bug "not an instr list: %a\n" d_stmt s)
        end
  in
    !ret

let dummy_iosh = IH.create 1

let getInstrIosh ((i, s) : (instr * stmt)) : RD.IOS.t IH.t = 
  match RD.getRDs s.sid with
      None -> dummy_iosh
    | Some (_,x,iosh) -> begin
        match s.skind with
            Instr il -> begin
              (* Standard stuff stolen from CIL extensions *)
              let rd_dat_lst = RD.instrRDs il s.sid ((),x,iosh) false in
              let il_dat_lst = List.combine il rd_dat_lst in
              let filteredList = List.filter (fun (i',_) -> Util.equals i i') il_dat_lst in
                if List.length filteredList = 1 then begin
                  let _,tuples = List.split filteredList in
                  let (_,_,ret_iosh) = List.hd tuples in
                    ret_iosh
                end
                else
                  E.s (E.bug "List.length filteredList = %d\n" (List.length filteredList))
            end
          | _ -> E.s (E.bug "not an instr list: %a\n" d_stmt s)
      end

let printIS name is =
  printf "%s = {" name;
  IS.iter (fun i -> ignore (printf "%d, " i)) is;
  printf "}\n"; ()

(* assumption: loop dominates "dominated", and v is a loop constant
   tell whether v is defined between the end of the loop, and the
   dominated statement *)
let defined_between (loop: stmt) (dominated: stmt) (v: varinfo) : bool = 
  let visited = ref SS.empty in
  let br = ref false in
  let rec visit (s: stmt) =
    if not (SS.mem s !visited) && not !br && not (s.sid == loop.sid) 
    then begin
      visited := SS.add s !visited;
      let defd,_ = UD.computeUseDefStmtKind s.skind in
        if VS.mem v defd then br := true
        else List.iter visit s.preds
    end
  in
    visit dominated;
    !br

let derivation_set
    (iterator: varinfo)
    (init: IS.t) (loop: stmt) (func: fundec) (filter: instr -> bool) (ih: stmt option IH.t) =
  let (all_defs,_) = List.split (IH.tolist RD.ReachingDef.defIdStmtHash) in
  let rds v iosh : IS.t =
    let ret_rds = ref IS.empty in
    let _ =
      try
        let ios = IH.find iosh v.vid in
          RD.IOS.iter
            (fun io -> match io with
                Some defid -> ret_rds := IS.add defid !ret_rds
              | None -> ()
            )
            ios
      with
          Not_found -> ()
    in
      !ret_rds
  in
  let inits = 
    match RD.getRDs loop.sid with
        None -> IS.empty
      | Some (_,_,iosh) -> rds iterator iosh
  in
  let _ =
    ignore (printf "iterator = %s, " iterator.vname); 
    printIS "inits" inits;
  in
  let uses, defd = UD.computeDeepUseDefStmtKind loop.skind in
  let loopConstants = VS.diff uses defd in
  let f (old : IS.t) : IS.t =
    let ret = ref IS.empty in
    let _ = ret := IS.union old !ret in
    let inNew (d : int) : bool =
      let iss = instrs_of_def d in
        match iss with
            hd :: tl -> 
              List.for_all
                (fun (i,s) ->
                  (
                    let uses,_ = UD.computeUseDefInstr i in
                    let iosh = getInstrIosh (i,s) in
                    let defIds = getReachingDefIds (i,s) in
                      (
                        if hasStmt (ScS loop) s then begin
                          VS.for_all
                            (fun v -> (VS.mem v loopConstants) or
                              (IS.for_all 
                                  (fun r -> (IS.mem r old) or (IS.mem r inits))
                                  (rds v iosh)
                              )
                            )
                            uses
                        end
                        else begin
                          if D.dominates ih loop s then begin
                            let restrict = VS.filter 
                              (fun v -> not (defined_between loop s v)) 
                              loopConstants
                            in
                              VS.for_all
                                (fun v -> (VS.mem v restrict) or
                                  (IS.for_all 
                                      (fun r -> (IS.mem r old) or (IS.mem r inits))
                                      (rds v iosh)
                                  )
                                )
                                uses
                          end
                          else
                            false
                        end
                      )
                      &&
                        (* depends on the previous step *)
                        IS.exists (fun did -> IS.mem did old) defIds
                      &&
                        (* so we can restrict the derivation set to
                           linear expressions, etc. *)
                        filter (i)
                  )
                )
                iss
          | [] -> false (* any def with no instrs shouldn't be in derivation set *)
    in
      List.iter (fun d -> if (inNew d) then ret := IS.add d !ret) all_defs;
      !ret
  in
  let rec fixF (cur: IS.t) =
    let next = f cur in
      if (IS.equal cur next) then cur
      else fixF next
  in
    fixF init

let derivation_set_var 
    (iterator: varinfo) (loop: stmt) (func: fundec) (filter: instr -> bool) (ih: stmt option IH.t) =
  let (defs,_) = List.split (getDefs iterator (ScS loop)) in
  let init = ref IS.empty in
  let _ =
    List.iter (fun i -> 
      let dl = instr_to_defIds i in
        List.iter (fun defid -> init := IS.add defid !init) dl
    )
      defs in
  let ds = derivation_set iterator !init loop func filter ih in
    printf "derivation set of %s = {" iterator.vname;
    IS.iter (fun d -> ignore (printf "%d, " d)) ds; printf "}\n";
    ds


(* Return a map from each of a loop's iterators to the iterators' derivation sets *)
let derivation_sets_of_iterators 
    (iterators: VS.t) (loop: stmt) 
    (func: fundec) (filter: instr -> bool) 
    (ih: stmt option IH.t) : IS.t IH.t =
  let ht = IH.create (VS.cardinal iterators) in
    VS.iter 
      (fun v -> IH.add ht v.vid (derivation_set_var v loop func filter ih)) 
      iterators;
    ht

(* Given an expression which comes from a conditional, get the sides of the 
   conditional.
*)
let rec getSides (e: exp) : exp * exp option =
  let hasRhs (lhs,rhs: exp * exp option) =
    match rhs with
        Some exp -> true
      | None -> false
  in
    match e with
        BinOp (b,lhs,rhs,t) -> (lhs, Some rhs)
      | CastE (_,e') | UnOp (_,e',_) -> begin
          let sides = getSides e' in
            if (hasRhs sides) then sides else (e', None)
        end
      | _ -> (e, None)          

(* Given a comparison function which has v on one side, and for which
   both sides are linear expressions, move things around so that v is
   on one side, and everything else is on the other. Return None if v
   isn't present, or if something non-linear stops us from doing the
   rewriting. *)
let mangle (v: varinfo) (e: exp) : exp option =
  let rec moveTo (vside: exp) (otherside: exp) =
    ignore (printf "vside = %a \t otherside = %a\n" d_exp vside d_exp otherside);
    match vside with
        Lval (Var v', NoOffset) -> if v'.vid == v.vid then Some otherside else None
      | BinOp (b,e1,e2,t) -> begin
          let e1used = UD.computeUseExp e1 in
          let oldt = typeOf otherside in
            if VS.mem v e1used then begin
              print_endline "woot";
              match b with
                  PlusA | PlusPI -> moveTo e1 (BinOp (MinusA, otherside, e2, oldt))
                | MinusA | MinusPI -> moveTo e1 (BinOp (PlusA, otherside, e2, oldt))
                | MinusPP -> moveTo e1 (BinOp (PlusPI, e2, otherside, t))
                | _ -> None
            end else begin
              match b with
                  (* the minuses may not even be legitimate expressions, but we 
                     just use them for bounders, so it's probably okay *)
                  PlusA | PlusPI -> moveTo e2 (BinOp (MinusA, otherside, e1, oldt))
                | MinusA | MinusPI -> moveTo e2 (BinOp (MinusA, e1, otherside, oldt))
                | MinusPP -> moveTo e2 (BinOp (MinusA, e1, otherside, t))
                | _ -> None
            end 
        end
      | _ -> None (* give up *)
  in
  let lhs, rhso = getSides e in
  let _ = 
    match rhso with
        Some rhs -> ignore (printf "v = %s \t lhs = %a \t rhs = %a\n" v.vname d_exp lhs d_exp rhs)
      | None -> ignore (printf "v = %s \tlhs = %a \t rhs = NONE\n" v.vname d_exp lhs)
  in
  let lused = UD.computeUseExp lhs in
    if VS.mem v lused then begin
      match rhso with
          Some rhs -> moveTo (stripCasts lhs) (stripCasts rhs)
        | None -> moveTo (stripCasts lhs) zero
    end
    else begin
      match rhso with
          None -> None
        | Some rhs -> begin
            let rused = UD.computeUseExp rhs in
              if VS.mem v rused then moveTo (stripCasts rhs) (stripCasts lhs)
              else None
          end
    end

let printIOS name ios =
  printf "%s = {" name;
  RD.IOS.iter
    (fun io -> 
      match io with 
          Some i -> 
            ignore (printf "%d (" i); 
            List.iter (fun (ins,s) -> ignore (printf "%a, " d_instr ins)) (instrs_of_def i);
            ignore (printf "), ")
        | None -> ignore (printf "NONE, "))
    ios;
  print_endline "}"

(* This is more "derived from an iterator", and restricted to statements within the loop. *)
let depOnIterator v iter dh (ino: instr option) s loop =
  let ds = try IH.find dh iter.vid with Not_found -> failwith "dh broken" in
  let proc_iosh iosh =
    try 
      let ios = IH.find iosh v.vid in       
        (* rds are all either in derivation set, or outside loop *)
        (* NOTE: "None" means you can get here without the
           variable being defined in this function; for example,
           if the variable is a parameter of this function. So "None"
           is okay.
        *)
        RD.IOS.for_all
          (fun io -> match io with
              Some defid -> 
                IS.mem defid ds 
                  (* this latter part we allow for statements within the loop *)
                or not (hasStmt (ScS loop) (stmt_of_def defid))
            | None -> true
          )
          ios
        &&
          (* has an rd in the derivation set *)
          RD.IOS.exists
          (fun io -> match io with
              Some defid -> IS.mem defid ds
            | None -> false
          )
          ios
    with Not_found -> false
  in
    match RD.getRDs s.sid with 
        None -> false
      | Some (_,x,iosh) -> begin
          match s.skind with
              If _ -> proc_iosh iosh
            | Instr il -> begin
                match ino with
                    Some i -> begin
                      let rd_dat_lst = RD.instrRDs il s.sid ((),x,iosh) false in
                      let il_dat_lst = List.combine il rd_dat_lst in
                      let filteredList = List.filter (fun (i',_) -> Util.equals i i') il_dat_lst in
                      let _,ioshs = List.split filteredList in
                        List.exists (fun (_,_,iosh') -> proc_iosh iosh') ioshs
                    end
                  | None -> proc_iosh iosh
              end
            | _ -> failwith "stmt kind not supported"
        end
            
let getIex 
    (iters: VS.t) (dh: IS.t IH.t) (cond: stmt) (loop: stmt) (iex: BS.t ref) = 
  match cond.skind with
      If (e,_,_,_) -> begin
        let noff,off,weird = useBreakdownExp e isPotentialIterator in
          VS.iter
            (fun v -> 
              VS.iter 
                (fun iter ->
                  if depOnIterator v iter dh None cond loop then begin
                    match mangle v e with 
                        Some e' -> iex := BS.add (bounder_of_varinfo_eo v (Some e')) !iex
                      | _ -> print_endline "mangling failed!"
                  end
                )
                iters
            )
            noff
      end
    | _ -> failwith "non-loop passed to getIex"
        
let rec getArray ((i,s) : instr * stmt) (istr: BS.t ref) (stop: bool ref) 
    (iter: varinfo) (dhex: IS.t IH.t) (loop: stmt) : unit =
  let noff, off, weird = useBreakdownStmt s isPotentialIterator (fun i' -> Util.equals i i') in
    if not !stop then begin
      try
        let found = 
          BS.choose 
            (BS.filter
                (fun b -> depOnIterator b.var iter dhex (Some i) s loop)
                off
            ) in
          istr := BS.add found !istr
      with
          Not_found -> begin
            (* recurse *)
            let rds = getReachingDefIds (i,s) in
              IS.iter
                (fun defid -> 
                  List.iter 
                    (fun (i',s') -> getArray (i',s') istr stop iter dhex loop)
                    (instrs_of_def defid)
                )
                rds
          end
    end

let getIstr 
    (iters: VS.t) (dhex: IS.t IH.t) (dhstr: IS.t IH.t) (cond: stmt) (loop: stmt) (istr: BS.t ref) = 
  match cond.skind with
      If (e,_,_,_) -> begin
        let noff,off,weird = useBreakdownExp e isPotentialIterator in
          (* "if (c) break", where c = A[i] and i is an iterator *)
          VS.iter
            (fun v -> 
              VS.iter 
                (fun iter ->
                  if (depOnIterator v iter dhstr None cond loop) 
                    && not (depOnIterator v iter dhex None cond loop)
                  then begin
                    (* v is dependent on i using array indexing and
                       linear expressions, but not using linear
                       expressions alone --- find the array *)
                    match RD.getRDs cond.sid with
                        None -> failwith "dependence without reaching definitions"
                      | Some (_,_,iosh) -> begin
                          try
                            let ios = IH.find iosh v.vid in
                              RD.IOS.iter
                                (fun io -> match io with
                                    Some defid -> 
                                      let stop = ref false in
                                          List.iter 
                                            (fun (i,s) -> getArray (i,s) istr stop iter dhex loop) 
                                            (instrs_of_def defid)
                                  | None -> ()
                                )
                                ios
                          with
                              Not_found -> failwith "dependence without reaching definitions of v"
                        end
                  end
                )
                iters
            )
            noff;
          (* "if (A[i]) break", where i is an iterator *)
          BS.iter
            (fun b ->
              VS.iter
                (fun iter ->
                  if depOnIterator b.var iter dhex None cond loop then begin
                    (* we alredy have the array -- woot! *)
                    istr := BS.add (bounder_of_varinfo_ao iter b.array) !istr
                  end
                )
                iters
            )
            off
      end
    | _ -> failwith "non-loop passed to getIstr"
        

(* given a statement which is a list of instructions, and a variable,
   return all rhs vars in all definitions of this variable in the
   statement*)
let getRhsVars (v: varinfo) (s: stmt) : VS.t =
  let vs = ref VS.empty in
  let _ =
    let proc_instr i =
      let uses, defd = UD.computeUseDefInstr i in
        if (VS.mem v defd) then vs := VS.union !vs uses
    in
      match s.skind with
          Instr il -> List.iter proc_instr il
        | _ -> E.s (E.bug "non-Instr statement unexpected on var %s\n********" v.vname)
  in
    !vs
      
      
(* This is a beast.
 * 
 * This function traverses use-def chains at the granularity of
 * individual instructions. We start with an initial instruction,
 * examine the variables it uses, and traverse the use-def chains for
 * these. When traversing a list of instructions and their associated
 * RD data, we only process an instruction if it defines a variable
 * through which we reached the instruction in the use-def chain.
 *)
let getTransitiveRDs ((root_i,root_s) : (instr * stmt)) : (instr * stmt) list =
  let acc = ref [] in
  let rec inner (v: varinfo) ((i,s) : (instr * stmt)) =
    let proc_ildl ildl = 
      List.iter
        (fun (i',(_,x,iosh)) -> begin
          let uses, defd = UD.computeUseDefInstr i' in
            (* Only examine instructions which define the variable through which we 
               reached them. *)
            if (VS.mem v defd) then begin
              VS.iter 
                (fun usedVar ->
                  List.iter 
                    (fun (new_i,new_s) ->
                      if not (List.mem (new_i,new_s) !acc) then begin
                        acc := (new_i,new_s) :: !acc;
                        inner usedVar (new_i,new_s)
                      end
                    )
                    (* Get reaching definitions for usedVar, as
                       <instr, containing stmt> pairs *)
                    (getVarDepInstrStmts usedVar iosh)
                ) 
                uses
            end
        end
        )
        ildl;
    in
      if (not (List.mem (i,s) !acc)) then begin
        match RD.getRDs s.sid with
            None -> ()
          | Some (_,x,iosh) -> begin
              match s.skind with
                  Instr il -> begin
                    (* Standard stuff stolen from CIL extensions *)
                    let rd_dat_lst = RD.instrRDs il s.sid ((),x,iosh) false in
                    let il_dat_lst = List.combine il rd_dat_lst in
                      proc_ildl il_dat_lst
                  end
                | _ -> ()
            end
      end
  in
  let uses, _ = UD.computeUseDefInstr root_i in
    VS.iter (fun v -> inner v (root_i,root_s)) uses;
    !acc

let isSelfDependent ((i,s) : (instr * stmt)) =
  let rds_star = getTransitiveRDs (i,s) in
    List.mem (i,s) rds_star

let isRhsLoopConst (v: varinfo) (s: stmt) (consts: VS.t) =
  let rhsVars = getRhsVars v s in
    VS.for_all (fun v' -> VS.mem v' consts) rhsVars

(* An iterator for a loop L is a variable v which
   (1) has a self-dependent definition, and 
   (2) is not dominated by any statement s in L which assigns an L-constant to v
*)
let isLoopIterator (v: varinfo) (loop: stmt) (ih: stmt option IH.t) =
  let hasDominatorInSet (def: stmt) (defSet: (instr * stmt) list) =
    List.exists (fun (i,s) -> D.dominates ih s def) defSet 
  in
  let uses, defd = UD.computeDeepUseDefStmtKind loop.skind in
  let loopConstants = VS.diff uses defd in
  let defs = getDefs v (ScS loop) in
  let selfDepDefs = List.filter (fun (i,s) -> isSelfDependent (i,s)) defs in
  let loopConstDefs = List.filter (fun (i,s) -> isRhsLoopConst v s loopConstants) defs in
  let iteratingDefs = List.filter 
    (fun (i,s) -> not (hasDominatorInSet s loopConstDefs))
    selfDepDefs
  in
    List.length iteratingDefs > 0

let getLoopIterators (loop: stmt) (ih: stmt option IH.t) =
  let (_,defd) = UD.computeDeepUseDefStmtKind loop.skind in
    VS.filter (fun v -> isLoopIterator v loop ih) defd


let isDefValid (d: int) : bool =
  let (il,_) = List.split (instrs_of_def d) in
    List.for_all (fun i -> isInstrValid i) il

(* Given a pointer and a function, say (approximately) whether this
   pointer is SAFE, SEQ, or WILD. (Using CCured terms.) *)
let getPtrType (v: varinfo) (f: fundec) : ptr_type_t =
  let (vDefs,_) = List.split (getDefs v (ScF f)) in
    if (List.length vDefs = 0) then Ptr_safe
    else begin
      if (List.for_all isInstrStartOfLv vDefs) then Ptr_safe
      else begin
        if (List.for_all isInstrValid vDefs) then Ptr_seq
        else Ptr_wild
      end
    end

let getPtrTypes (f: fundec) : ptr_type_t IH.t =
  let all_ptrs = List.filter (fun v -> isPtr_var v) (f.sformals @ f.slocals) in
  let ret = IH.create (List.length all_ptrs) in
    List.iter (fun p -> IH.add ret p.vid (getPtrType p f)) all_ptrs;
    ret

let getIexIstr (checks: SS.t) (iterators: VS.t) (loop: stmt) (dhex: IS.t IH.t) (dhstr: IS.t IH.t)
    : BS.t * BS.t =
  let i_ex2 = ref BS.empty in
  let i_str2 = ref BS.empty in
    SS.iter
      (fun s -> getIex iterators dhex s loop i_ex2)
      checks;
    SS.iter
      (fun s -> getIstr iterators dhex dhstr s loop i_str2)
      checks;
    BS.iter (fun b -> fillInDirection b loop) !i_ex2;
    BS.iter (fun b -> fillInDirection b loop) !i_str2;
    (!i_ex2,!i_str2)

(* Get all loop exits and the corresponding exit conditionals
   (computed approximately), put all exit conditionals in a list, and
   turn this list into a set. *)
let getChecks (breakTable: stmt SH.t) (condsOfBreak: stmt SH.t) 
    (ih: stmt option IH.t) (loop: stmt) : SS.t =
  let _ = getLoopExitConditionals breakTable condsOfBreak loop ih in
  let breaks = SH.find_all breakTable loop in
  let rec inner (sl: stmt list) =
    match sl with 
        hd :: tl -> (SH.find_all condsOfBreak hd) @ (inner tl)
      | [] -> []
  in
  let checkslist = inner breaks in
  let ss = ref SS.empty in
    List.iter (fun s -> ss := SS.add s !ss) checkslist;
    !ss

let all_defs_are_cool (v: varinfo) (loop: stmt) =
  let (defs,_) = List.split (getDefs v (ScS loop)) in
    List.for_all (fun i -> isInstrValid i) defs

let analyze_loop   
    (fd: fundec)
    (ih: stmt option IH.t)
    (breakTable: stmt SH.t) 
    (condsOfBreak: stmt SH.t) 
    (abstractions: ape list ref)
    (s: stmt)
    =
  match s.skind with 
      Loop(b,loc,s1,s2) -> begin
        let _ =
          ignore (printf "processing loop on line %d with label %s\n" 
                     loc.line (getVerisecLabel s))
        in
        let loopIterators = getLoopIterators s ih in
        let coolIterators = VS.filter (fun v -> all_defs_are_cool v s) loopIterators in
        let _ = printVS "loopIterators" loopIterators; printVS "coolIterators" coolIterators in
        let fex,fstr = 
          (fun i -> match i with
              Set (_,e,_) -> isExpLinear e
            | _ -> false
          ),
          (fun i -> let ok e = (isExpLinear e) or (isExpUnidimensionalArray e) or (isBitmasking e) in
                      match i with
                          Set (_,e,_) -> ok e
                        | Call (_,_,el,_) -> List.exists ok el
                        | _ -> false
          )
        in
        let derivationHashEx = derivation_sets_of_iterators coolIterators s fd fex ih in
        let derivationHashStr = derivation_sets_of_iterators coolIterators s fd fstr ih in
        let checks = getChecks breakTable condsOfBreak ih s in
        let (i_ex, i_str) = getIexIstr checks coolIterators s derivationHashEx derivationHashStr in
        let verisecLabel = getVerisecLabel s in
        let suggest_abstraction (v: varinfo) =
          let add_abstraction (a: ape) = 
            abstractions := !abstractions @ [a]
          in
          let _ = ignore (printf "for var %s...  " v.vname) in
          let try_1ex () =
            try
              let b = bs_find_var v i_ex in
                add_abstraction 
                  (ape_1ex fd verisecLabel v (getBoundOrBust b));
                ignore (printf "use single-var explicit\n");
                true
            with
                Not_found -> false
          in
          let try_1str () =
            try 
              let b = bs_find_var v i_str in
                add_abstraction 
                  (ape_1str fd verisecLabel v b.array);
                ignore (printf "use single-var string\n");
                true
            with 
                Not_found -> false
          in let try_2ex () =
            if (hasDirection_var v s) then begin
              let i_ex' = BS.filter hasDirection i_ex in
                BS.iter (fun b -> begin
                  add_abstraction 
                    (ape_2ex fd verisecLabel v 
                        (getDirection_var v s)
                        b.var
                        (getBoundOrBust b)
                        b.direction
                    );
                  ignore (printf "use two-var explicit\n")
                end)
                  i_ex';
                not (BS.is_empty i_ex')
            end
            else false
          in let try_2str () =
            BS.iter (fun b -> begin
              add_abstraction 
                (ape_2str 
                    fd verisecLabel v
                    b.var
                    b.array
                );
              ignore (printf "use two-var string\n")
            end)
              i_str
          in
               if (not (try_1ex ()) && not (try_1str ())) then
                 if (not (try_2ex ())) then 
                   try_2str ()
        in
          (* Suggest an abstraction for each variable which appears
             in an offset expression, changes within the loop's
             body, and is one of this function's iterators. *)
          VS.iter suggest_abstraction loopIterators
      end
    | _ -> failwith "not a loop"

class iteratorScannerVisitor (abstractions: ape list ref) (fi: file)
  = object(self)
    inherit nopCilVisitor

    method vfunc (f: fundec) : fundec visitAction =
      let breakTable: stmt SH.t = SH.create 100 in
      let condsOfBreak: stmt SH.t = SH.create 100 in
      let ih = D.computeIDom ~doCFG:false f in
      let _ = RD.computeRDs f in
      let _ = 
        (* Give each loop a label we can use to refer to it when 
           writing abstraction file. For some reason, if we do this earlier,
           it gets borked by makeCFG.
        *)
        ignore (visitCilFunction (new loopLabellerVisitor fi) f);
      in
      let ptrh = getPtrTypes f in
      let all_ptrs = List.filter (fun v -> isPtr_var v) (f.sformals @ f.slocals) in
      let _ =
        try List.iter (fun p -> ignore (printf "%s : %s\n" p.vname (string_of_ptr_type_t (IH.find ptrh p.vid)))) all_ptrs
        with Not_found -> failwith "wtf"
      in
      let loops = getLoops f in
        (* Do the main analysis *)
        SS.iter 
          (fun loop -> 
            analyze_loop
              f
              ih
              breakTable 
              condsOfBreak 
              abstractions
              loop
          )
          loops;
        SkipChildren

  end

(* *************************** INSTRUMENTER ******************************* *)

(* model assume(e) as "if (!e) { while (1) {} }" *)

(* used to give each assumption a unique label *)
let assCnt = ref 0

let mkLabelledNegAssumeStmt e labelString=
  let infLoop = mkStmt (Loop (mkBlock [], locUnknown, None, None)) in
    infLoop.labels <- 
      Label (labelString, locUnknown, false) :: infLoop.labels;
    mkStmt (If (e, 
               mkBlock [infLoop],
               mkBlock [],
               locUnknown)
    )

let mkNegAssumeStmt e = 
  let labelString = "VERISEC_ASSUMPTION_" ^ (string_of_int !assCnt) in
  let ret = mkLabelledNegAssumeStmt e labelString in
    assCnt := !assCnt + 1;
    ret

let mkLabelledAssumeStmt e labelString = 
  mkLabelledNegAssumeStmt (UnOp (LNot, e, intType)) labelString

let mkAssumeStmt e =
  mkNegAssumeStmt (UnOp (LNot, e, intType))

class instrumenterVisitor (abstractions: ape list)
  = object(self)
    inherit nopCilVisitor

      (* map A -> A_nullpos *)
    val nullposMap = ref VM.empty

      (* map a loop to all it's i_s variables *)
    val loopToShadowsMap : (varinfo list) STM.t ref = ref STM.empty 

      (* map i_s back to i *)
    val shadowsToVarsMap = ref VM.empty

    method vstmt (s: stmt) : stmt visitAction =
        (* add instrumentation to "A[expr1] = expr2" 
           v -> A[]
           off -> expr1
           e -> expr2

           A[expr1] = expr2;
           if (expr1 >= 0) {
             if (expr2 == 0) {
               A_nullpos = NONDET;
               assume (A_nullpos <= expr1);
             } else {
               if (expr1 == A_nullpos) {
                 A_nullpos = NONDET;
               }
             } 
           }
        *)
      let instrument_instr i =
        let instrumentation v off e =
          try 
            (* make a branch, simplifying if cond is a constant *)
            let makeBranch (cond: exp) (tb: block) (eb: block) : block =
              if (isZero cond) then
                eb
              else begin
                if (isConstant cond) then tb
                else 
                  mkBlock [ mkStmt (If (cond,tb,eb,locUnknown)) ]
              end
            in
            let e_eq_zero = constFoldBinOp false Eq e zero intType in
            let a_nullpos_lval = var (VM.find v !nullposMap) in
            let a_nullpos = Lval a_nullpos_lval in
            let idx =
              match off with
                  Index (e, NoOffset) -> e
                | _ -> failwith "unexpected offset, cannot instrument"
            in
            let idx_geq_zero = constFoldBinOp false Le zero idx intType in
            let a_nullpos_leq_idx = BinOp (Le, a_nullpos, idx, intType) in
            let nondet = mkString "NONDET" in
            let havok_a_nullpos =
              mkStmtOneInstr (Set (a_nullpos_lval, nondet, locUnknown)) in
            let if_idx_eq_a_nullpos_havok = 
              (* use two <='s instead of one == *)
              let thenBlock2 = 
                mkBlock [ havok_a_nullpos; 
                          mkNegAssumeStmt a_nullpos_leq_idx;
                ]
              in
              let idx_leq_a_nullpos = BinOp (Le, idx, a_nullpos, intType) in
                mkStmt (If (idx_leq_a_nullpos, 
                           mkBlock [ mkStmt (If (a_nullpos_leq_idx,
                                                thenBlock2, 
                                                mkBlock [],
                                                locUnknown)) ],
                           mkBlock [],
                           locUnknown))
            in
            let mainBlock = 
              let set_a_nullpos = mkBlock [ havok_a_nullpos;
                                            mkAssumeStmt a_nullpos_leq_idx ] in
              let elseBlock = mkBlock [ if_idx_eq_a_nullpos_havok ] in
                makeBranch e_eq_zero set_a_nullpos elseBlock
            in
              mkStmt (Block (makeBranch idx_geq_zero mainBlock (mkBlock [])))
          with
              Not_found -> mkEmptyStmt ()
        in
        let wrap (sl: stmt list) : stmt =
          mkStmt (Block (mkBlock sl))
        in
          match i with 
              Set(lv, e, l) -> begin
                match lv with
                    (Var(v), off) -> 
                      wrap (compactStmts [mkStmtOneInstr i ; instrumentation v off e])
                  | _ -> failwith "unreachable match case"
              end
            | Call(Some lv, f, args, l) -> begin
                failwith "not implemented"
              end
            | _ -> failwith "unreachable match case"
      in
      let instrWritesToArray i =
        match i with 
            Set(lv, e, l) -> begin
              match lv with
                  (Var v, _) -> isArray_var v
                | _ -> false
            end
          | Call(Some lv, f, args, l) -> begin
              match lv with
                  (Var v, _) -> isArray_var v
                | _ -> false
            end
          | _ -> false
      in
      let newStmts = ref [] in
      let curIL = ref [] in
      let proc_instr i =
        if (instrWritesToArray i) then begin
          newStmts := !newStmts @ [mkStmt (Instr !curIL)];
          curIL := [];
          newStmts := !newStmts @ [instrument_instr i]
        end
        else
          curIL := !curIL @ [i]
      in
        (* condInd, condRhs, and a_nullpos get the dummy value zero;
           isTestOnAi will set them iff it returns true *)
      let condRhs = ref zero in
      let condInd = ref zero in
      let a_nullpos = ref zero in
      let isTestOnAi e =
        match e with
            BinOp (Eq, e', rhs, _) -> begin
              match stripCasts e' with
                  Lval (Var v, Index (ind, NoOffset)) -> begin
                    try 
                      condRhs := rhs;
                      condInd := ind;
                      a_nullpos := 
                        Lval ((Var (VM.find v !nullposMap)), NoOffset);
                      true
                    with
                        Not_found -> false
                  end
                | _ -> false
            end
          | _ -> false
      in
              (*let _ =
                let ts = 
                  match s.skind with
                    Instr _-> "Instr"
                    | Block _ -> "Block"
                    | _ -> "OTHER"
                in

                printf "instrumenting %a\nof type %s" d_stmt s ts;
              in*)
        match s.skind with
            Block b ->
              (* so i++; j++ doesn't become {i++;}{j++;} *)
              b.bstmts <- compactStmts b.bstmts;
              DoChildren
          | Instr il -> 
              (* quick hack not to ugly up the .cil.c file too much; not 
                 necessary for correctness *)
              if il == [] then DoChildren
              else begin
              let _ =
                List.iter proc_instr il;
                newStmts := !newStmts @ [mkStmt (Instr !curIL)];
              in
              let replacement = 
                mkStmt (Block (mkBlock (compactStmts !newStmts)) );
              in
                replacement.labels <- s.labels;
                ChangeTo replacement
            end
          | If (e, thenBlock, elseBlock, loc) -> begin
              (* check if e is of the form A[expr1]==expr2 *)
              if (isTestOnAi e) then begin
                let a_nullpos_leq_i = 
                  BinOp (Le, !a_nullpos, !condInd, intType) in
                let rhs_eq_zero = BinOp (Eq, !condRhs, zero, intType) in
                let newStmt = 
                  (* again, trivial simplifications straight-up *)
                  if (isZero !condRhs) then
                    mkAssumeStmt a_nullpos_leq_i
                  else
                    if (isConstant !condRhs) then
                      mkEmptyStmt () (* non-zero constant *)
                    else
                      mkStmt (If (rhs_eq_zero,
                                  mkBlock [ mkAssumeStmt a_nullpos_leq_i ],
                                  mkBlock [],
                                  locUnknown)
                             )
                in
                  thenBlock.bstmts <- [newStmt] @ thenBlock.bstmts;
                  DoChildren
              end else
                DoChildren
            end
          | Loop _ -> begin
              let lab = getVerisecLabel s in
              let myShadows =
                try
                  STM.find lab !loopToShadowsMap
                with
                    Not_found -> []
              in
                (* create the i_s = i instructions *)
              let il = ref [] in
              let _ =
                List.iter 
                  (fun v -> begin
                    let shadow_lv = (Var v, NoOffset) in
                    let orig_exp = try
                        Lval (var (VM.find v !shadowsToVarsMap))
                      with
                          Not_found -> failwith "ERROR: no shadow"
                    in
                      il := !il @ [Set (shadow_lv, orig_exp, locUnknown)]
                  end)
                  myShadows;
              in
              let newStmt = mkStmt 
                (Block (
                  mkBlock 
                    [ mkStmt (Instr !il); 
                      mkLabelledNegAssumeStmt zero (lab ^ "_PRECONDITION1");
                      mkLabelledNegAssumeStmt zero (lab ^ "_PRECONDITION2");
                      mkLabelledNegAssumeStmt zero (lab ^ "_PRECONDITION3");
                      mkLabelledNegAssumeStmt zero (lab ^ "_PRECONDITION4");
                      mkLabelledNegAssumeStmt zero (lab ^ "_PRECONDITION5");
                      s ]))
              in
                ChangeDoChildrenPost (s, (fun s' -> newStmt))
            end
          | _ -> DoChildren
              
    (* Create all the _nullpos and _s variables we may need *)
    method vfunc (func: fundec) : fundec visitAction =
      let var_exists (s: string) =
        try
          List.find (fun v -> (String.compare v.vname s) == 0) func.slocals;
          true
        with
            Not_found -> false
      in
      let addShadowToLoop (l: string) (v: varinfo) =
        try
          let oldList = STM.find l !loopToShadowsMap in
            loopToShadowsMap := STM.add l (oldList @ [v]) !loopToShadowsMap
        with
            Not_found -> loopToShadowsMap := STM.add l [v] !loopToShadowsMap
      in
      let register_shadow 
          (shadowName: string) (original: varinfo) 
          (t: typ) (loopLabel: string)
          =
        if (not (var_exists shadowName)) then begin
          let shadow = makeLocalVar func shadowName t in
            shadowsToVarsMap := 
              VM.add shadow original !shadowsToVarsMap;
            addShadowToLoop loopLabel shadow;
        end
      in
      let _ =
        (* create the _nullpos vars *)
        let arrays = ref [] in
          print_endline ("visiting " ^ func.svar.vname);
          List.iter 
            (fun v -> begin 
               if (isArray_var v) then
                 arrays := !arrays @ [v]
             end) 
            func.slocals;
          List.iter
            (fun v -> nullposMap := 
               VM.add 
                 v 
                 (makeLocalVar func (v.vname ^ "_nullpos") intType)
                 !nullposMap
            )
            !arrays;
      in
        (* create the _s vars *)
        List.iter
          (fun a -> begin
             match a.ape_type with
                 AT_2ex (_, leader, _, _) | AT_2str (leader, _) -> begin
                  let newName v =
                    v.vname ^ "_" ^ a.ape_loopLabel ^ "_s"
                  in
                    register_shadow (newName a.ape_var) a.ape_var 
                      a.ape_var.vtype a.ape_loopLabel;
                    register_shadow 
                      (newName leader) leader leader.vtype a.ape_loopLabel
                end
              | _ -> ()
          end
          )
          abstractions;
        DoChildren


end


(* A memory expression is 
   - safe if it's a linear expression of a single SAFE pointer
   - analyzable if it's a linear expression of a single SEQ pointer
   - unanalyzable otherwise
*)
let memExpAnal (me: exp) (ptrMap: ptr_type_t IH.t) : anal_t =
  let isSeq v =
    if isPtr_var v then
      try match IH.find ptrMap v.vid with Ptr_seq -> true | _ -> false
      with Not_found -> E.s (E.error "lookup failure on %s\n" v.vname)
    else true
  in
  let isSafe v =
    if isPtr_var v then
      try match IH.find ptrMap v.vid with Ptr_safe -> true | _ -> false
      with Not_found -> E.s (E.error "lookup failure on %s\n" v.vname)
    else true
  in
  let uses = UD.computeUseExp me in
    if (VS.cardinal uses > 1) or not (isExpLinear me) then Anal_weird
    else if VS.for_all isSafe uses then Anal_safe
    else if VS.for_all isSeq uses then Anal_analyzable
    else Anal_weird
      
(* An offset is:
   - safe if it has no indexing, or is unidimensional and indexed by a constant
   - analyzable if it is unidimensional and the index is a linear expression
   - unanalyzable otherwise
*)
let rec offsetAnal (off: offset) : anal_t =
  if not (hasIndexing off) then Anal_safe
  else begin
    match off with
        Index (e, off') -> 
          if isConstant e then offsetAnal off' else begin
            if isExpLinear e && not (hasIndexing off') 
            then Anal_analyzable 
            else Anal_weird
          end
      | _ -> Anal_weird
  end

let lvalAnal (lv: lval) (ptrMap: ptr_type_t IH.t) =
  match lv with
      Mem e, off -> begin
        let a1 = memExpAnal e ptrMap in
        let a2 = offsetAnal off in
          match a1 with
              Anal_safe -> a2
            | Anal_analyzable -> begin
                match a2 with
                    Anal_safe -> Anal_analyzable
                  | _ -> Anal_weird
              end
            | _ -> Anal_weird
      end
    | Var v, off -> if isScalar lv then Anal_safe else offsetAnal off
    | _ -> Anal_weird

(* something about one thing in the lval's "offset" expression being
   derived from a loop iterator, and all the rest of the things being
   loop constants that aren't changed between the loop and s 

   furthermore, derived in a koshur manner
*)

(* want the following property: every variable in an index which is
   dependent on an iterator is derived from an iterator with a pattern 

   dependencies on statements which are iterating....

*)

(* iter is an iterator for loop, lval is a potentially-analyzable lval in s *)
let isProtected 
    (lv:lval) (s:stmt) (loop:stmt) (iter:varinfo) (ih: stmt option IH.t) : bool =
  if not (hasStmt (ScS loop) s) && not (D.dominates ih loop s) then false
  else begin
    false
  end

class lvalClassifier = object(self)
  inherit nopCilVisitor

  val mutable pmo = None

  method vfunc f =
    pmo <- Some (getPtrTypes f);
    DoChildren

  method vlval lv =
    let pm = match pmo with Some (map) -> map | None -> failwith "no map!" in
    let tp = lvalAnal lv pm in
    let _ =
      match tp with
          Anal_safe | _ -> ignore (printf "lval %a is %s\n" d_lval lv (string_of_anal tp))
    in
      SkipChildren

end


(* *************************** MAIN PROGRAM ******************************* *)

(* breakTable: A map from a loop to the statements that break out of it. *)
(* condsOfBreak: A map from breaks to conditionals *)
let scanloops_main (f: file) = 
  let abstractions = ref [] in
  let isv = new iteratorScannerVisitor abstractions f in
  let _ =
    
    (* Do the main analysis *)
    visitCilFileSameGlobals (isv :> cilVisitor) f;

    (* Write the .abs file for this file *)
    output_abstractions !abstractions (sanitize_filename_abs f.fileName);

    (* Add the instrumentation we need for our predicates to make sense *)
    visitCilFileSameGlobals (new instrumenterVisitor !abstractions) f

    (*;visitCilFileSameGlobals (new lvalClassifier) f*)
  in
    ()

let feature : featureDescr = 
  { fd_name = "scanloops";
    fd_enabled = ref false;
    fd_description = "stub for scanning loops";
    fd_extraopt = [];
    fd_doit = scanloops_main; 
    fd_post_check = true;
  } 



